/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.action.impl;

import java.util.HashMap;

import magma.agent.action.IAction;
import magma.agent.connection.IServerConnection;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.meta.agent.IBodyPartConfiguration;
import magma.agent.meta.agent.IHingeJointConfiguration;
import magma.agent.meta.agent.IJointConfiguration;
import magma.agent.meta.agent.IUniversalJointConfiguration;

/**
 * This class generates server messages from Effector objects.
 */
public class Action implements IAction
{

	private final HashMap<String, UniversalEffector> universalEffectors;

	private final HashMap<String, HingeEffector> hingeEffectors;

	private final BeamEffector beam;

	private final InitEffector init;

	private final SayEffector say;

	private final String actionScene;

	/** the connection to the server */
	private final IServerConnection connection;

	/**
	 * Initializes all effectors that occur in the simulation
	 * 
	 * @param connection module for communicating with the server
	 * @param actionScene string for the robot type to use
	 * @param teamname name of the team this player plays in
	 * @param number player number within the team
	 */
	public Action(IServerConnection connection, IAgentMetaModel metaModel,
			String teamname, int number)
	{
		this.connection = connection;
		this.actionScene = metaModel.getSceneString();
		universalEffectors = new HashMap<String, UniversalEffector>();
		hingeEffectors = new HashMap<String, HingeEffector>();
		init = new InitEffector(teamname, number);
		beam = new BeamEffector();
		say = new SayEffector();

		initJointEffectors(metaModel);
	}

	/**
	 * Initializes all HingeJoint- and UniversalJoint-Effectors based on the
	 * given meta-model
	 * 
	 * @param metaModel - the meta model of the agent
	 */
	private void initJointEffectors(IAgentMetaModel metaModel)
	{
		for (IBodyPartConfiguration config : metaModel.getBodyPartConfiguration()) {
			IJointConfiguration jointConfig = config.getJointConfiguration();

			// create Joint Effectors
			if (jointConfig instanceof IHingeJointConfiguration) {
				hingeEffectors.put(jointConfig.getEffectorName(),
						new HingeEffector(jointConfig.getEffectorName()));
			} else if (jointConfig instanceof IUniversalJointConfiguration) {
				universalEffectors.put(jointConfig.getEffectorName(),
						new UniversalEffector(jointConfig.getEffectorName()));
			} else {
				// Further Joints go here
			}
		}
	}

	@Override
	public void setUniversalEffectorSpeed(String name, float speed1, float speed2)
	{
		universalEffectors.get(name).setAxis(speed1, speed2);
	}

	@Override
	public void setHingeEffectorSpeed(String name, float speed)
	{
		hingeEffectors.get(name).setAxisSpeed(speed);
	}

	@Override
	public void sendSceneString()
	{
		String message = "(scene " + actionScene + ")";
		connection.sendMessage(message);
	}

	@Override
	public void sendInitString()
	{
		String message = init.getMessageString();
		connection.sendMessage(message);
	}

	@Override
	public void sendBeamString(float x, float y, float rotation)
	{
		beam.setBeam(x, y, rotation);
		connection.sendMessage(beam.getMessageString());
	}

	@Override
	public void sayMessage(String message)
	{
		say.setMessage(message);
	}

	@Override
	public String sendMotorString()
	{
		String message = getMotorString();
		connection.sendMessage(message);
		return message;
	}

	/**
	 * Returns the string necessary to send to the server to trigger joint motors
	 * 
	 * @return the string necessary to send to the server to trigger joint motors
	 */
	private String getMotorString()
	{
		StringBuffer buffer = new StringBuffer(200);
		for (UniversalEffector u : universalEffectors.values()) {
			buffer.append(u.getMessageString());
		}

		for (HingeEffector h : hingeEffectors.values()) {
			buffer.append(h.getMessageString());
		}

		buffer.append(say.getMessageString());

		return buffer.toString();
	}
}
