/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel;

import javax.vecmath.Vector3d;
import javax.vecmath.Vector3f;

import magma.agent.perception.IPerception;
import magma.util.observer.IObserver;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Provides read only access to the agent model including universal joints,
 * hinge joints, force resistance and gyro rate sensors
 * 
 * @author Stefan Glaser, Klaus Dorer
 */
public interface IAgentModel
{
	/**
	 * Called to trigger a update of the AgentModel based on the given perception
	 * object.
	 * 
	 * @param perception the Perception
	 */
	public void update(IPerception perception);

	/**
	 * This method is called from the AgentRuntime in order to reflect the
	 * target-state of the agent model into effector-actions, which can then be
	 * send to the server.
	 */
	public void reflectTargetStateToAction();

	/**
	 * Returns a specified universal joint of the agent
	 * @param name the name of the joint as specified in
	 *        IServerConfigFilesConstants.UNIVERSALJOINT_NAMES
	 * @return read only version of the universal joint specified by name
	 */
	IUniversalJoint getUJ(String name);

	/**
	 * Returns read only access to a specified hinge joint in the sensed body
	 * model of the agent
	 * @param name the name of the joint as specified in
	 *        IServerConfigFilesConstants.HINGEJOINT_NAMES
	 * @return read only version of the hinge joint specified by name
	 */
	IHingeJointR getHJ(String name);

	/**
	 * Returns read only access to a specified hinge joint in the expected body
	 * model of the agent
	 * @param name the name of the joint as specified in
	 *        IServerConfigFilesConstants.HINGEJOINT_NAMES
	 * @return read only version of the hinge joint specified by name
	 */
	IHingeJointR getHJExpected(String name);

	/**
	 * Returns a specified hinge joint of the agent
	 * @param name the name of the joint as specified in
	 *        IServerConfigFilesConstants.HINGEJOINT_NAMES
	 * @return read write version of the hinge joint specified by name
	 */
	IHingeJoint getWriteableHJ(String name);

	/**
	 * Returns a specified force resistance sensor of the agent
	 * @param name the name of the sensor as specified in
	 *        IServerConfigFilesConstants.FORCERESISTANCE_NAMES
	 * @return read only version of the force sensor specified by name
	 */
	IForceResistance getForceResistance(String name);

	/**
	 * Retrieve accelerometer values
	 * 
	 * @param name the name of the sensor
	 * @return Accelerations
	 */
	IAccelerometer getAccelerometer(String name);

	/**
	 * Retrieves the specified gyro sensor
	 * 
	 * @param name the name of the sensor as specified in
	 *        IServerConfigFilesConstants.GYRORATE_NAMES
	 * @return read only version of the gyro sensor specified by name
	 */
	IGyroRate getGyroRate(String name);

	/**
	 * Get the Center of mass
	 * 
	 * @return Center of Mass
	 */
	Vector3d getCenterOfMass();

	/**
	 * Set the message to say in that cycle.
	 * 
	 * @param message
	 */
	public void sayMessage(String message);

	/**
	 * Set the beam position of the agent.
	 * 
	 * @param position x, y, rot (in meter, resp. degrees)
	 */
	public void beamToPosition(Vector3f position);

	/**
	 * @see magma.util.observer.IPublishSubscribe#attach(magma.util.observer.
	 *      IObserver)
	 * 
	 * @param newObserver Observer object to add to the list
	 */
	void attach(IObserver<IAgentModel> newObserver);

	/**
	 * @see magma.util.observer.IPublishSubscribe#detach(magma.util.observer.IObserver)
	 * 
	 * @param oldObserver Observer object to delete from the list
	 * @return True if successfully detached, false if not
	 */
	boolean detach(IObserver<IAgentModel> oldObserver);

	/**
	 * Calculate the rectangle within the ball is kickable by both legs
	 * 
	 * @return Calculated rectangle
	 */
	public Vector3D[] getBothLegKickRectangle();

	/**
	 * Calculate the rectangle within the ball is kickable by the left leg
	 * 
	 * @return Calculated rectangle
	 */
	public Vector3D[] getLeftLegKickRectangle();

	/**
	 * Calculate the rectangle within the ball is kickable by the right leg
	 * 
	 * @return Calculated rectangle
	 */
	public Vector3D[] getRightLegKickRectangle();

	/**
	 * Returns the root body part of the sensed body model.
	 * 
	 * @return the root body part
	 */
	public IBodyPart getTorso();

	/**
	 * @param name the name of the body part to retrieve
	 * @return the body part with the specified name
	 */
	public IBodyPart getBodyPart(String name);

	/**
	 * Returns the future body model, which is a copy of the sensed/expected body
	 * model, on which the actions for the current cycle are performed.
	 * 
	 * @return the future body model
	 */
	public IBodyModel getFutureBodyModel();

	/**
	 * Called to copy back joint speed values from the passed body model to this
	 * agents body model. The passed model is usually a copy that has been
	 * created earlier from this object so that the joint model matches.
	 * @param source the body model to take the values from
	 */
	void updateJointsSpeed(IBodyModel source);

	/**
	 * Returns the body part (of the sensed body model), which contains the
	 * camera.
	 * 
	 * @return the body part containing the camera, or null, if no body part
	 *         contains a camera
	 */
	IBodyPart getBodyPartContainingCamera();
}