/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.agentmodel;

import java.util.Collection;
import java.util.Map;

import javax.vecmath.Matrix3d;
import javax.vecmath.Matrix4d;
import javax.vecmath.Vector3d;

import magma.agent.action.IAction;
import magma.agent.perception.IPerception;

/**
 * GEneric BodyPart interface
 * 
 * @author Ingo Schindler
 */
public interface IBodyPart
{
	/**
	 * get name of the bodypart
	 * @return name of the BodyPart
	 */
	String getName();

	/**
	 * Get parent BodyPart
	 * 
	 * @return Parent BodyPart
	 */
	IBodyPart getParent();

	/**
	 * Position of the center of this BodyPart relative to the torso. Torso is
	 * Vector3d(0,0,0)
	 * 
	 * @return Position of the center of this BodyPart relative to the torso
	 */
	Vector3d getPosition();

	/**
	 * The complete orientation matrix from torso to the last Joint in this row.
	 * E.g. Torso -> Shoulder -> UpperArm -> Elbow -> LowerArm. Then this is the
	 * orientation of the LowerArm.
	 * 
	 * @return orientation Matrix
	 */
	Matrix3d getOrientation();

	/**
	 * The complete transformation matrix from torso to the last Joint in this
	 * row. E.g. Torso -> Shoulder -> UpperArm -> Elbow -> LowerArm. Then this is
	 * the position and orientation of the LowerArm-joint.
	 * 
	 * @return Transformation Matrix
	 */
	Matrix4d getTransformation();

	/**
	 * Mass of this BodyPart
	 * 
	 * @return Mass
	 */
	float getMass();

	/**
	 * @return the mass of this body part and all of its children (in kg)
	 */
	float getWholeMass();

	/**
	 * Get geometry of this BodyPart
	 * 
	 * @return Geometry
	 */
	Vector3d getGeometry();

	/**
	 * Informs the action component to create action commands from this body
	 * model
	 * @param action the action component that creates the action protocol
	 */
	void generateJointActions(IAction action);

	/**
	 * Updates the joint values in the body model from perception
	 * @param perception the new perception we made
	 */
	void updateFromPerception(IPerception perception);

	/**
	 * @return the number of child body parts
	 */
	int getNoOfChildren();

	/**
	 * A collection of the child body parts to this body part.
	 * 
	 * @return a collection of child body parts
	 */
	Collection<IBodyPart> getChildren();

	/**
	 * @param name name of the child body part
	 * @return the child body part, null if not existing
	 */
	IBodyPart getChild(String name);

	/**
	 * @return the joint to which this body part is attached.
	 */
	IJoint getJoint();

	/**
	 * @param name the name of the part to retrieve
	 * @return the specified body part, null if not existing
	 */
	IBodyPart getBodyPart(String name);

	/**
	 * @return the center of mass of this body part and all parts connected to it
	 */
	Vector3d getCenterOfMass();

	/**
	 * Searches the body part hierarchy for a sensor with passed name
	 * @param name the name of the sensor
	 * @return the sensor with passed name, null if not existing
	 */
	ISensor getSensorDeep(String name);

	/**
	 * Returns the sensor instance of the requested type, if this body part
	 * contains such a sensor type.
	 * 
	 * @param <T> - the type if the sensor
	 * @param type - the type of the sensor
	 * @return the sensor instance of the requested type, if existing
	 */
	public <T extends ISensor> T getSensor(Class<T> type);

	/**
	 * Deep update of joint's desired speed from the passed body part. The passed
	 * part has to match to this part and is usually created as copy from this
	 * sometimes before.
	 * @param part the body part from which to take the update values
	 */
	void updateJointsSpeed(IBodyPart part);

	/**
	 * @param globalOrientation matrix representing the global orientation of the
	 *        torso
	 * @return the pitch angle (deg) of this body part in the global coordinate
	 *         system
	 */
	double getGlobalPitch(Matrix4d globalOrientation);

	/**
	 * @param globalOrientation matrix representing the global orientation of the
	 *        torso
	 * @return the roll angle (deg) of this body part in the global coordinate
	 *         system
	 */
	double getGlobalRoll(Matrix4d globalOrientation);

	public Matrix4d getInverseAnchor();

	public Vector3d[] getEdges();

	/**
	 * @param result map to add all sensors deep for caching
	 */
	void collectSensors(Map<String, ISensor> result);

	/**
	 * Retrieve the jacobian matrix, having this IBodyPart as the end of the
	 * chain. The jacobian is created with respect to the given endEffectorPos in
	 * the local system of this body.
	 * 
	 * @param endEffectorPos - the position of the end effector of the
	 *        configuration in the local system of this IBodyPart
	 * @return the jacobian matrix
	 */
	public double[][] getJacobian(Vector3d endEffectorPos);
}
