/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

/**
 * SimSpark "GyroRate" sensor interface. Provides read-only access only.<br>
 * The GyroRate Sensor represents the orientation change of an body in the
 * 3-dimensional space, respective to the global (an other) coordinate system.
 * This orientation change is represented through three unit-vectors along the
 * three system-axis, stored in a {@link Matrix3d} object.
 * 
 * <p>
 * The rotation matrix setup is as follows:
 * <table>
 * <tr align="center">
 * <td><code>m<sub>xx</sub></code></td>
 * <td><code>m<sub>x0</sub><br>(vx)</code></td>
 * <td><code>m<sub>x1</sub><br>(vy)</code></td>
 * <td><code>m<sub>x2</sub><br>(vz)</code></td>
 * </tr>
 * <tr align="center">
 * <td><code>m<sub>0x</sub>(x)</code></td>
 * <td><b>m<sub>00</sub></b></td>
 * <td><b>m<sub>01</sub></b></td>
 * <td><b>m<sub>02</sub></b></td>
 * </tr>
 * <tr align="center">
 * <td><code>m<sub>1x</sub>(y)</code></td>
 * <td><b>m<sub>10</sub></b></td>
 * <td><b>m<sub>11</sub></b></td>
 * <td><b>m<sub>12<sub></b></td>
 * </tr>
 * <tr align="center">
 * <td><code>m<sub>2x</sub>(z)</code></td>
 * <td><b>m<sub>20</sub></b></td>
 * <td><b>m<sub>21</sub></b></td>
 * <td><b>m<sub>22</sub></b></td>
 * </tr>
 * </table>
 * The three unit-vectors <code>(vx)</code>, <code>(vy)</code>,
 * <code>(vz)</code> are distributed along the columns of the matrix.<br>
 * </p>
 * 
 * <p>
 * <strong>Note:</strong> Relative to the Nao-robot (center of it's torso), the
 * x-axis points to the right arm, while the y-axis is pointing forward and the
 * z-axis go's straight up through it's head.<br>
 * </p>
 * 
 * @author Stefan Glaser, Srinivasa Ragavan
 */
public interface IGyroRate extends ISensor
{
	/**
	 * Retrieve gyro values <br>
	 * where &lt;x, y, z&gt; is a vector containing the angle-change around the
	 * corresponding axis of the previous cycle in degrees per second <br>
	 * 
	 * @return The change in orientation of the body with respect to the global
	 *         coordinate system of the previous cycle
	 */
	public Vector3d getGyro();

	/**
	 * Retrieve the current change in orientation
	 * 
	 * @return The current change in orientation as matrix representing a
	 *         3D-rotation
	 */
	public Matrix3d getOrientationChange();
}