/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel.impl;

import java.io.Serializable;

import javax.vecmath.Vector3d;
import javax.vecmath.Vector3f;

import magma.agent.action.IAction;
import magma.agent.agentmodel.IAgentModel;
import magma.agent.agentmodel.IBodyModel;
import magma.agent.agentmodel.IBodyPart;
import magma.agent.agentmodel.IHingeJoint;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.perception.IPerception;
import magma.util.geometry.Geometry;
import magma.util.observer.IObserver;
import magma.util.observer.IPublishSubscribe;
import magma.util.observer.Subject;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Implementation of the AgentModel. Used to represent all the information the
 * agent has about itself.
 * 
 * @author Stefan Glaser, Ingo Schindler
 */

public class AgentModel implements IAgentModel, Serializable
{
	private static final long serialVersionUID = 1L;

	/** observers of agent model */
	private final transient IPublishSubscribe<IAgentModel> observer;

	/** action processing of the agent */
	protected final transient IAction action;

	// TODO: Think about how to remove the following (robot specific constants)
	// from the agent model
	/** length of the kickable area */
	private static final double MAX_X_KICK = 0.18;

	/** width of the kickable area */
	private static final double MAX_Y_KICK = 0.09;

	/** length of the kickable area both legs */
	private static final double MAX_X_BOTH = 0.18;

	/** widht of the kickable area both legs */
	private static final double MAX_Y_BOTH = 0.03;

	/** the body part model of this robot as sensed by last perception */
	protected final BodyModel bodyModelSensed;

	/** the body part model of this robot as we expect it after next sense */
	protected BodyModel bodyModelExpected;

	/** the body part model of this robot to define the future */
	protected BodyModel bodyModelFuture;

	protected final String nameOfCameraBodyPart;

	/**
	 * initializes all known Sensors like: UniversalJoints, HingeJoints,
	 * ForceResistances and GyroRates
	 */
	public AgentModel(IAction action, IAgentMetaModel metaModel)
	{
		this.action = action;
		bodyModelSensed = new BodyModel(metaModel);
		bodyModelExpected = bodyModelSensed;
		bodyModelFuture = new BodyModel(bodyModelSensed);
		nameOfCameraBodyPart = metaModel.getNameOfCameraBodyPart();
		observer = new Subject<IAgentModel>();
	}

	@Override
	public void update(IPerception perception)
	{
		// update all Sensors
		bodyModelSensed.updateFromPerception(perception);
		// checkForDiscrepancies();
		bodyModelExpected = bodyModelFuture;
		bodyModelExpected.updateFromPerception(perception);
		bodyModelFuture = new BodyModel(bodyModelSensed);
		// bodyModelFuture = new BodyModel(bodyModelExpected);

		// now inform observers about changes
		observer.onStateChange(this);
	}

	@Override
	public void attach(IObserver<IAgentModel> newObserver)
	{
		observer.attach(newObserver);
	}

	@Override
	public boolean detach(IObserver<IAgentModel> oldObserver)
	{
		return observer.detach(oldObserver);
	}

	@Override
	public UniversalJoint getUJ(String name)
	{
		return (UniversalJoint) bodyModelSensed.getJoint(name);
	}

	@Override
	public HingeJoint getHJ(String name)
	{
		return (HingeJoint) bodyModelSensed.getJoint(name);
	}

	@Override
	public HingeJoint getHJExpected(String name)
	{
		return (HingeJoint) bodyModelExpected.getJoint(name);
	}

	@Override
	public IHingeJoint getWriteableHJ(String name)
	{
		return (IHingeJoint) bodyModelFuture.getSensor(name);
	}

	@Override
	public IBodyPart getBodyPart(String name)
	{
		// TODO: check which model to use
		// Better would be I think to move this method to BodyModel and out of
		// here
		// so that the user decides.
		return bodyModelSensed.getBodyPart(name);
	}

	@Override
	public IBodyPart getTorso()
	{
		return bodyModelSensed.getTorso();
	}

	@Override
	public Vector3d getCenterOfMass()
	{
		// TODO: check which model to use
		// Better would be I think to move this method to BodyModel and out of
		// here
		// so that the user decides.
		return bodyModelSensed.getCenterOfMass();
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof AgentModel)) {
			return false;
		}
		AgentModel other = (AgentModel) o;
		return bodyModelSensed.equals(other.bodyModelSensed);
	}

	@Override
	public Vector3D[] getBothLegKickRectangle()
	{
		// points are relative to player coordinates
		double[][] points = { { 0, MAX_Y_BOTH }, { MAX_X_BOTH, MAX_Y_BOTH },
				{ MAX_X_BOTH, -MAX_Y_BOTH }, { 0, -MAX_Y_BOTH } };
		return Geometry.getPolygon(points);
	}

	@Override
	public Vector3D[] getLeftLegKickRectangle()
	{
		// points are relative to player coordinates
		double[][] points = { { 0, MAX_Y_KICK }, { MAX_X_KICK, MAX_Y_KICK },
				{ MAX_X_KICK, 0 }, { 0, 0 } };
		return Geometry.getPolygon(points);
	}

	/**
	 * Calculate the rectangle within the ball is kickable by the right leg
	 * 
	 * @return Calculated rectangle
	 */
	@Override
	public Vector3D[] getRightLegKickRectangle()
	{
		// points are relative to player coordinates
		double[][] points = { { 0, 0 }, { MAX_X_KICK, 0 },
				{ MAX_X_KICK, -MAX_Y_KICK }, { 0, -MAX_Y_KICK } };
		return Geometry.getPolygon(points);
	}

	@Override
	public int hashCode()
	{
		return bodyModelSensed.hashCode();
	}

	@Override
	public void beamToPosition(Vector3f position)
	{
		bodyModelFuture.beamToPosition(position);
	}

	@Override
	public void sayMessage(String message)
	{
		bodyModelFuture.sayMessage(message);
	}

	@Override
	public ForceResistance getForceResistance(String name)
	{
		// TODO: check which model to use
		// Better would be I think to move this method to BodyModel and out of
		// here
		// so that the user decides.
		return (ForceResistance) bodyModelSensed.getSensor(name);
	}

	@Override
	public GyroRate getGyroRate(String name)
	{
		// TODO: check which model to use
		// Better would be I think to move this method to BodyModel and out of
		// here
		// so that the user decides.
		return (GyroRate) bodyModelSensed.getSensor(name);
	}

	@Override
	public Accelerometer getAccelerometer(String name)
	{
		// TODO: check which model to use
		// Better would be I think to move this method to BodyModel and out of
		// here
		// so that the user decides.
		return (Accelerometer) bodyModelSensed.getSensor(name);
	}

	@Override
	public String toString()
	{
		return bodyModelSensed.toString();
	}

	@Override
	public void reflectTargetStateToAction()
	{
		bodyModelFuture.generateTargetStateToAction(action);
	}

	@Override
	public void updateJointsSpeed(IBodyModel source)
	{
		bodyModelFuture.updateJointsSpeed(source);
	}

	@Override
	public IBodyModel getFutureBodyModel()
	{
		return bodyModelFuture;
	}

	@Override
	public IBodyPart getBodyPartContainingCamera()
	{
		if (nameOfCameraBodyPart != null) {
			return getBodyPart(nameOfCameraBodyPart);
		}

		return null;
	}
}
