/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.agentmodel.impl;

import java.io.Serializable;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;
import javax.vecmath.Vector3f;

import magma.agent.action.IAction;
import magma.agent.agentmodel.IBodyModel;
import magma.agent.agentmodel.IBodyPart;
import magma.agent.agentmodel.IJoint;
import magma.agent.agentmodel.ISensor;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.meta.agent.IBodyPartConfiguration;
import magma.agent.perception.IPerception;
import magma.util.geometry.MatrixUtil;

/**
 * Represents the body with its body parts and joints
 * @author dorer
 */
class BodyModel implements Serializable, IBodyModel
{
	/**  */
	private static final long serialVersionUID = 1L;

	/** link to the parent body part */
	private final IBodyPart torso;

	/** the next say message to the server */
	protected String sayMessage;

	/** the next beam position */
	protected Vector3f beamPosition;

	/** a cache for all sensors including joints for faster access */
	private Map<String, ISensor> sensorCache;

	/**
	 * Constructor to create the body model from a passed meta model
	 * @param metaModel the meta model containing body part information
	 */
	BodyModel(IAgentMetaModel metaModel)
	{

		Map<String, IBodyPart> bodyParts = createBodyParts(metaModel);
		torso = connectBodyParts(metaModel, bodyParts);
		sensorCache = collectAllSensors();
		sayMessage = null;
		beamPosition = null;
	}

	/**
	 * Copy constructor
	 * @param source the source body model to copy
	 */
	BodyModel(BodyModel source)
	{
		torso = new BodyPart((BodyPart) source.torso, null);
		sayMessage = source.sayMessage;
		if (source.beamPosition == null) {
			beamPosition = null;
		} else {
			beamPosition = new Vector3f(source.beamPosition);
		}
		sensorCache = collectAllSensors();
	}

	private Map<String, IBodyPart> createBodyParts(IAgentMetaModel metaModel)
	{
		Map<String, IBodyPart> bodyParts = new HashMap<String, IBodyPart>();

		for (IBodyPartConfiguration config : metaModel.getBodyPartConfiguration()) {
			BodyPart newPart = new BodyPart(config);
			bodyParts.put(config.getName(), newPart);
		}
		return bodyParts;
	}

	/**
	 * @param metaModel
	 * @param bodyParts
	 */
	private BodyPart connectBodyParts(IAgentMetaModel metaModel,
			Map<String, IBodyPart> bodyParts)
	{
		BodyPart result = null;
		for (IBodyPartConfiguration config : metaModel.getBodyPartConfiguration()) {
			BodyPart parent = (BodyPart) bodyParts.get(config.getParent());
			BodyPart currentPart = (BodyPart) bodyParts.get(config.getName());
			currentPart.setParent(parent);
			if (parent == null) {
				// we have what we call torso
				result = currentPart;
			}
		}
		return result;
	}

	/**
	 * @return
	 */
	private Map<String, ISensor> collectAllSensors()
	{
		Map<String, ISensor> result = new HashMap<String, ISensor>(40);
		torso.collectSensors(result);
		return result;
	}

	/**
	 * Creates the action content for sending to the server
	 */
	void generateTargetStateToAction(IAction action)
	{
		generateBeamAction(action);
		generateJointActions(action);
		generateSayAction(action);
	}

	/**
	 * Informs the action component to create action commands from this body
	 * model
	 * @param action the action component that creates the action protocol
	 */
	void generateJointActions(IAction action)
	{
		torso.generateJointActions(action);
	}

	/**
	 * Called to forward the beam action.
	 */
	void generateBeamAction(IAction action)
	{
		if (beamPosition != null) {
			action.sendBeamString(beamPosition.x, beamPosition.y, beamPosition.z);
		}
	}

	void beamToPosition(Vector3f position)
	{
		beamPosition = position;
	}

	/**
	 * Called to forward the say message.
	 */
	void generateSayAction(IAction action)
	{
		if (sayMessage != null) {
			action.sayMessage(sayMessage);
		}
	}

	void sayMessage(String message)
	{
		sayMessage = message;
	}

	/**
	 * Updates the joint values in the body model from perception
	 * @param perception the new perception we made
	 */
	void updateFromPerception(IPerception perception)
	{
		// reset actions
		sayMessage = null;
		beamPosition = null;

		torso.updateFromPerception(perception);
	}

	/**
	 * Called to copy back joint speed values from the passed model to this
	 * model. The passed model is usually a copy that has been created earlier
	 * from this object so that the joint model matches.
	 * @param source the body model to take the values from
	 */
	public void updateJointsSpeed(IBodyModel source)
	{
		torso.updateJointsSpeed(source.getTorso());
	}

	@Override
	public ISensor getSensor(String name)
	{
		return sensorCache.get(name);
		// return torso.getSensorDeep(name);
	}

	@Override
	public IJoint getJoint(String name)
	{
		return (IJoint) getSensor(name);
	}

	@Override
	public IBodyPart getTorso()
	{
		return torso;
	}

	@Override
	public IBodyPart getBodyPart(String name)
	{
		return torso.getBodyPart(name);
	}

	@Override
	public Vector3d getCenterOfMass()
	{
		return torso.getCenterOfMass();
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof BodyModel)) {
			return false;
		}
		BodyModel other = (BodyModel) o;
		return torso.equals(other.torso);
	}

	@Override
	public void setHingeJointPosition(String name, double value)
	{
		IJoint joint = getJoint(name);
		((HingeJoint) joint).performAxisPosition(value);
	}

	@Override
	public void adjustHingeJointPosition(String name, double delta)
	{
		IJoint joint = getJoint(name);
		((HingeJoint) joint).adjustAxisPosition(delta);
	}

	@Override
	public void performInitialPose()
	{
		LinkedList<IBodyPart> bodyPartsToInitialize = new LinkedList<IBodyPart>();
		Collection<IBodyPart> children;
		IBodyPart currentBodyPart = torso;

		while (currentBodyPart != null) {
			if (currentBodyPart.getJoint() != null) {
				currentBodyPart.getJoint().performInitialPosition();
			}

			// Add all children to list of body part to initialize and poll the
			// next body part to initialize
			children = currentBodyPart.getChildren();
			if (children != null) {
				bodyPartsToInitialize.addAll(children);
			}
			currentBodyPart = bodyPartsToInitialize.pollFirst();
		}
	}

	@Override
	public void resetAllMovements()
	{
		LinkedList<IBodyPart> bodyPartsToInitialize = new LinkedList<IBodyPart>();
		Collection<IBodyPart> children;
		IBodyPart currentBodyPart = torso;

		while (currentBodyPart != null) {
			if (currentBodyPart.getJoint() != null) {
				currentBodyPart.getJoint().resetMovement();
			}

			// Add all children to list of body part to initialize and poll the
			// next body part to initialize
			children = currentBodyPart.getChildren();
			if (children != null) {
				bodyPartsToInitialize.addAll(children);
			}
			currentBodyPart = bodyPartsToInitialize.pollFirst();
		}
	}

	@Override
	public Vector3d[] getEdges(IBodyPart part)
	{
		Matrix3d orientation = new Matrix3d();
		part.getTransformation().get(orientation);

		Vector3d pos = part.getPosition();

		Vector3d[] edges = part.getEdges();

		Vector3d[] edgesReturn = new Vector3d[edges.length];

		for (int i = 0; i < edges.length; i++) {
			edgesReturn[i] = new Vector3d(pos);

			edgesReturn[i].add(MatrixUtil.mulMatrixVector(orientation, edges[i]));
		}

		return edgesReturn;
	}
}
