/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.agentmodel.impl;

import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import javax.vecmath.Matrix3d;
import javax.vecmath.Matrix4d;
import javax.vecmath.Vector3d;

import magma.agent.action.IAction;
import magma.agent.agentmodel.IBodyPart;
import magma.agent.agentmodel.IHingeJoint;
import magma.agent.agentmodel.IJoint;
import magma.agent.agentmodel.ISensor;
import magma.agent.meta.agent.IBodyPartConfiguration;
import magma.agent.meta.agent.IHingeJointConfiguration;
import magma.agent.meta.agent.IJointConfiguration;
import magma.agent.meta.agent.ISensorConfiguration;
import magma.agent.meta.agent.IUniversalJointConfiguration;
import magma.agent.perception.IPerception;
import magma.util.geometry.MatrixUtil;

/**
 * @author stilnox
 * 
 */
class BodyPart implements IBodyPart, Serializable
{
	private static final long serialVersionUID = -7487102876064697313L;

	/** the name of this body part with which it can be retrieved */
	private final String name;

	/** reference to the body part to which this is attached, null if no parent */
	private BodyPart parent;

	/** reference to the body parts that are attached to this */
	private final Map<String, BodyPart> children;

	/**
	 * the joint that is used to rotate this bodypart (redundant reference from
	 * sensors)
	 */
	private IJoint joint;

	/** Sensors attached to this body part including joint sensor if existing. */
	private Map<String, ISensor> sensors;

	/**
	 * translation from this bodypart's center to its parent's center if
	 * joint-angle == 0
	 */
	private final Vector3d translation;

	/**
	 * the position (in local coordinates) where this podypart in connected to
	 * the parent bodypart.
	 */
	private final Vector3d anchor;

	/** the inverse matrix of anchor */
	private final Matrix4d inverseAnchor;

	@Override
	public Matrix4d getInverseAnchor()
	{
		return inverseAnchor;
	}

	/** A Vector that represents the distance(-Vector) from one Joint to the next */
	private Vector3d distanceToParentJoint;

	/** the mass of this body part in kg */
	private final float mass;

	/** Dimensions of the BodyPart. For now each BodyPart is represented by a Box */
	private final Vector3d geometry;

	/**
	 * Constructor for this bodypart
	 * 
	 * @param name name of the bodypart
	 * @param parent reference to the parent bodypart
	 * @param joint joint that rotates this bodypart
	 * @param translation translation to the parent bodypart (center to center)
	 * @param anchor position where this bodypart rotates around (position of the
	 *        joint relative to this bodyparts center)
	 * @param mass mass of this bodypart
	 * @param geometry the dimensions of this bodypart
	 */
	public BodyPart(String name, BodyPart parent, IJoint joint,
			Vector3d translation, Vector3d anchor, float mass, Vector3d geometry)
	{
		this.name = name;
		this.joint = joint;
		sensors = new HashMap<String, ISensor>(3);
		if (joint != null) {
			sensors.put(joint.getName(), joint);
		}
		this.translation = translation;

		this.anchor = anchor;
		this.inverseAnchor = MatrixUtil.createInverseTranslationMatrix(anchor);

		this.mass = mass;
		this.geometry = geometry;

		children = new HashMap<String, BodyPart>(3);
		setParent(parent);
	}

	/**
	 * Constructor to create body part from IBodyPartConfiguration
	 * @param config the configuration
	 */
	public BodyPart(IBodyPartConfiguration config)
	{
		this(config.getName(), null, null, config.getTranslation(), config
				.getJointAnchor(), config.getMass(), config.getGeometry());

		sensors = new HashMap<String, ISensor>(3);
		IJointConfiguration jointConfig = config.getJointConfiguration();

		if (jointConfig != null) {
			IJoint newJoint = null;
			if (jointConfig instanceof IHingeJointConfiguration) {
				// If the joint configuration is a HingeJoint-Type, add a new
				// HingeJoint to the sensors list
				newJoint = new HingeJoint((IHingeJointConfiguration) jointConfig);
			} else if (jointConfig instanceof IUniversalJointConfiguration) {
				// If the joint configuration is a UniversalJoint-Type, add a new
				// UniversalJoint to the sensors list
				newJoint = new UniversalJoint(
						(IUniversalJointConfiguration) jointConfig);
			} else {
				// further joint types
			}
			if (newJoint != null) {
				joint = newJoint;
				sensors.put(jointConfig.getName(), newJoint);
			}
		}
		ISensorConfiguration sensorConfig;

		sensorConfig = config.getGyroRateConfiguration();
		if (sensorConfig != null) {
			// If the body part has a gyro-rate perceptor attached, create
			// one
			sensors.put(
					sensorConfig.getName(),
					new GyroRate(sensorConfig.getName(), sensorConfig
							.getPerceptorName()));
		}

		sensorConfig = config.getAccelerometerConfiguration();
		if (sensorConfig != null) {
			// If the body part has a accelerometer attached, create one
			sensors.put(
					sensorConfig.getName(),
					new Accelerometer(sensorConfig.getName(), sensorConfig
							.getPerceptorName()));
		}

		sensorConfig = config.getForceResistanceConfiguration();
		if (sensorConfig != null) {
			// If the body part has a force-resistance perceptor attached,
			// create one
			sensors.put(
					sensorConfig.getName(),
					new ForceResistance(sensorConfig.getName(), sensorConfig
							.getPerceptorName()));
		}

	}

	/**
	 * Copy constructor
	 * @param source the object from which to copy values from
	 * @param parent the parent of this object (!= parent of other)
	 */
	public BodyPart(BodyPart source, BodyPart parent)
	{
		// TODO kdo: check if this is really non-mutable
		// or should we just make a deep copy since it is cheap for matrices?
		// non deep copy for non changing parts
		this.name = source.name;
		this.parent = parent;
		this.translation = source.translation;
		this.anchor = source.anchor;
		this.inverseAnchor = source.inverseAnchor;
		this.mass = source.mass;
		this.geometry = source.geometry;
		setDistanceToParentJoint();

		// deep copy of the mutable parts and children
		sensors = new HashMap<String, ISensor>();
		for (ISensor sensor : source.sensors.values()) {
			ISensor newSensor = sensor.copy();
			sensors.put(newSensor.getName(), newSensor);
		}
		if (source.joint != null) {
			// we have already copied joint through sensors
			this.joint = (IJoint) sensors.get(source.joint.getName());
		} else {
			joint = null;
		}

		// we have to be sure that body parts are strictly hierarchical
		children = new HashMap<String, BodyPart>();
		for (BodyPart child : source.children.values()) {
			BodyPart copy = new BodyPart(child, this);
			addChild(copy);
		}
	}

	@Override
	public String getName()
	{
		return name;
	}

	@Override
	public IBodyPart getParent()
	{
		return parent;
	}

	@Override
	public Vector3d getPosition()
	{
		if (parent != null) {

			Matrix4d position = new Matrix4d(getTransformation());

			position.mul(inverseAnchor);

			return new Vector3d(position.m03, position.m13, position.m23);
		}

		return new Vector3d(0, 0, 0);
	}

	@Override
	public Matrix3d getOrientation()
	{
		Matrix3d orientation;

		if (parent != null) {
			orientation = parent.getOrientation();
			if (joint != null) {
				orientation.mul(joint.getRotationMatrix());
			}
		} else {
			orientation = new Matrix3d();
			orientation.setIdentity();
		}

		return orientation;
	}

	@Override
	public Matrix4d getTransformation()
	{
		// TODO kdo: we might want to calculate this information once
		// that would require that the object is immutable and created each time
		// we have updates on the joints. In that case it should be calculated top
		// down starting with the root parent.
		Matrix4d transformation;
		Matrix4d jointTransform;

		if (parent != null) {
			transformation = parent.getTransformation();

			if (joint != null) {
				jointTransform = new Matrix4d(joint.getRotationMatrix(),
						distanceToParentJoint, 1);
			} else {
				jointTransform = MatrixUtil
						.createTranslationMatrix(distanceToParentJoint);
			}
			transformation.mul(jointTransform);
		} else {
			transformation = new Matrix4d();
			transformation.setIdentity();
		}

		return transformation;
	}

	/**
	 * Set Parent BodyPart
	 * @param iBodyPart
	 */
	public void setParent(BodyPart parent)
	{
		this.parent = parent;

		if (parent != null) {
			parent.addChild(this);
		}

		setDistanceToParentJoint();
	}

	private void setDistanceToParentJoint()
	{
		if (parent == null) {
			distanceToParentJoint = new Vector3d();
		} else {
			distanceToParentJoint = new Vector3d(//
					-parent.anchor.x + translation.x + anchor.x,//
					-parent.anchor.y + translation.y + anchor.y,//
					-parent.anchor.z + translation.z + anchor.z);
		}
	}

	@Override
	public float getMass()
	{
		return mass;
	}

	@Override
	public Vector3d getGeometry()
	{
		return geometry;
	}

	/**
	 * Adds the passed body part to the map of child body parts
	 * @param bodyPart the child body part to add
	 */
	private void addChild(BodyPart bodyPart)
	{
		children.put(bodyPart.getName(), bodyPart);
	}

	/**
	 * @return the number of child body parts
	 */
	@Override
	public int getNoOfChildren()
	{
		return children.size();
	}

	/**
	 * @param name name of the child body part
	 * @return the child body part, null if not existing
	 */
	@Override
	public BodyPart getChild(String name)
	{
		return children.get(name);
	}

	@Override
	public Collection<IBodyPart> getChildren()
	{
		return Collections.<IBodyPart> unmodifiableCollection(children.values());
	}

	/**
	 * @return a map of all hinge joints in this model
	 */
	@Override
	public ISensor getSensorDeep(String sensorName)
	{
		if (sensors.containsKey(sensorName)) {
			return sensors.get(sensorName);
		}
		for (BodyPart child : children.values()) {
			ISensor result = child.getSensorDeep(sensorName);
			if (result != null) {
				return result;
			}
		}
		return null;
	}

	@Override
	public void collectSensors(Map<String, ISensor> result)
	{
		result.putAll(sensors);
		for (BodyPart child : children.values()) {
			child.collectSensors(result);
		}
	}

	@Override
	public <T extends ISensor> T getSensor(Class<T> type)
	{
		for (ISensor s : sensors.values()) {
			if (type.isInstance(s)) {
				return type.cast(s);
			}
		}

		return null;
	}

	/**
	 * Informs the action component to create action commands from this body
	 * model
	 * @param action the action component that creates the action protocol
	 */
	@Override
	public void generateJointActions(IAction action)
	{
		if (joint != null) {
			joint.generateJointAction(action);
		}
		for (BodyPart child : children.values()) {
			child.generateJointActions(action);
		}
	}

	/**
	 * Updates the joint values in the body model from perception
	 * @param perception the new perception we made
	 */
	@Override
	public void updateFromPerception(IPerception perception)
	{
		for (ISensor sensor : sensors.values()) {
			// this also updates joint
			sensor.updateFromPerception(perception);
		}

		for (BodyPart child : children.values()) {
			child.updateFromPerception(perception);
		}
	}

	@Override
	public IJoint getJoint()
	{
		return joint;
	}

	@Override
	public String toString()
	{
		return name;
	}

	@Override
	public BodyPart getBodyPart(String name)
	{
		if (name.equals(this.name)) {
			return this;
		}
		for (BodyPart child : children.values()) {
			BodyPart result = child.getBodyPart(name);
			if (result != null) {
				return result;
			}
		}
		return null;
	}

	@Override
	public Vector3d getCenterOfMass()
	{
		Vector3d coM = new Vector3d();
		Vector3d pos = new Vector3d(getPosition());
		pos.scale(mass);
		coM.add(pos);
		for (BodyPart child : children.values()) {
			coM.add(child.getCenterOfMass());
		}

		if (parent == null) {
			coM.scale(1.0f / getWholeMass());
		}
		return coM;
	}

	@Override
	public float getWholeMass()
	{
		float wholeMass = mass;
		for (BodyPart child : children.values()) {
			wholeMass += child.getWholeMass();
		}
		return wholeMass;
	}

	@Override
	public boolean equals(Object obj)
	{
		if (!(obj instanceof BodyPart)) {
			return false;
		}
		BodyPart other = (BodyPart) obj;
		if (!name.equals(other.name) || !translation.equals(other.translation)
				|| !anchor.equals(other.anchor) || !geometry.equals(other.geometry)
				|| !children.equals(other.children)
				|| !sensors.equals(other.sensors)) {
			return false;
		}
		if (parent == null && other.parent != null) {
			// we do not check the parent for equality to avoid endless loop
			return false;
		}
		return !(joint == null && other.joint != null);

	}

	@Override
	public void updateJointsSpeed(IBodyPart part)
	{
		if (joint != null) {
			joint.updateJointPositionFromJoint(part.getJoint());
		}
		assert children.size() == part.getNoOfChildren() : "updating joint speeds from invalid source model: children: "
				+ children.size()
				+ " other: "
				+ part.getNoOfChildren()
				+ " this part: " + this;

		for (IBodyPart child : children.values()) {
			IBodyPart otherChild = part.getChild(child.getName());
			assert otherChild != null : "updating joint speeds from invalid source model: child is null";
			child.updateJointsSpeed(otherChild);
		}
	}

	@Override
	public double getGlobalPitch(Matrix4d globalOrientation)
	{
		Matrix4d orientation = getTransformation();
		// transform local into global orientation
		orientation.mul(globalOrientation);
		return Math.toDegrees(Math.asin(orientation.m12));
	}

	@Override
	public double getGlobalRoll(Matrix4d globalOrientation)
	{
		Matrix4d orientation = getTransformation();
		// transform local into global orientation
		orientation.mul(globalOrientation);
		return Math.toDegrees(Math.asin(orientation.m02));
	}

	@Override
	public Vector3d[] getEdges()
	{
		Vector3d geometry = new Vector3d(this.geometry);

		geometry.scale(0.5);

		Vector3d[] edge = new Vector3d[8];

		edge[0] = new Vector3d(-geometry.getX(), geometry.getY(),
				-geometry.getZ());
		edge[1] = new Vector3d(geometry.getX(), geometry.getY(), -geometry.getZ());
		edge[3] = new Vector3d(-geometry.getX(), -geometry.getY(),
				-geometry.getZ());
		edge[2] = new Vector3d(geometry.getX(), -geometry.getY(),
				-geometry.getZ());

		edge[4] = new Vector3d(geometry.getX(), geometry.getY(), geometry.getZ());
		edge[5] = new Vector3d(geometry.getX(), -geometry.getY(), geometry.getZ());
		edge[7] = new Vector3d(-geometry.getX(), geometry.getY(), geometry.getZ());
		edge[6] = new Vector3d(-geometry.getX(), -geometry.getY(),
				geometry.getZ());

		return edge;
	}

	@Override
	public double[][] getJacobian(Vector3d endEffectorPos)
	{
		double[][] jacobian = null;

		// Transform end effector position
		Matrix4d trans = new Matrix4d(getTransformation());
		trans.mul(inverseAnchor);
		Vector3d globalEndEffectorPos = MatrixUtil.mulMatrixVector(trans,
				endEffectorPos);

		// Collect all body parts up to the torso (without the torso)
		LinkedList<IBodyPart> parts = new LinkedList<IBodyPart>();
		IBodyPart currentBodyPart = this;

		while (currentBodyPart.getParent() != null) {
			parts.addFirst(currentBodyPart);
			currentBodyPart = currentBodyPart.getParent();
		}

		// Create the jacobian matrix
		jacobian = new double[3][parts.size()];

		for (int i = 0; i < parts.size(); i++) {
			addJacobianColumn(jacobian, i, globalEndEffectorPos,
					((IHingeJoint) parts.get(i).getJoint()).getJointAxis(), parts
							.get(i).getTransformation());
		}

		return jacobian;
	}

	/**
	 * Add a column to the jacobian matrix.
	 * 
	 * @param jacobian - jacobian matrix
	 * @param col - column to add
	 * @param endEffectorPos - position of the end effector (in root-body system)
	 * @param jointAxis - rotation axis of the joint
	 * @param jointTransformation - transformation of the joint
	 */
	private void addJacobianColumn(double[][] jacobian, int col,
			Vector3d endEffectorPos, Vector3d jointAxis,
			Matrix4d jointTransformation)
	{
		Vector3d axis = new Vector3d(jointAxis);
		Vector3d n = new Vector3d(endEffectorPos);
		Vector3d jointPos = new Vector3d();
		jointTransformation.get(jointPos);

		n.sub(jointPos);

		jointTransformation.transform(axis);
		axis.cross(axis, n);

		jacobian[0][col] = axis.x;
		jacobian[1][col] = axis.y;
		jacobian[2][col] = axis.z;
	}
}
