/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel.impl;

import javax.vecmath.Vector3d;

import magma.agent.agentmodel.IForceResistance;
import magma.agent.agentmodel.ISensor;
import magma.agent.perception.IForceResistancePerceptor;
import magma.agent.perception.IPerception;
import magma.util.FuzzyCompare;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Implementation of the SimSpark "ForceRestistance" sensor
 * 
 * @author Klaus Dorer
 */
public class ForceResistance extends Sensor implements IForceResistance
{
	private static final long serialVersionUID = 7431711977473627245L;

	/** point where the force acts */
	private Vector3d forceOrigin;

	/** the force itself */
	private Vector3d force;

	/**
	 * Instantiates a new ForceResistance sensor
	 * 
	 * @param name Sensor name
	 */
	public ForceResistance(String name, String perceptorName)
	{
		super(name, perceptorName);
		forceOrigin = new Vector3d();
		force = new Vector3d();
	}

	/**
	 * Copy constructor
	 * @param source the object to copy from
	 */
	private ForceResistance(ForceResistance source)
	{
		super(source.name, source.perceptorName);
		forceOrigin = new Vector3d(source.forceOrigin);
		force = new Vector3d(source.force);
	}

	@Override
	public Vector3d getForceOrigin()
	{
		return forceOrigin;
	}

	/**
	 * Set the force origin
	 * 
	 * @param forceOrigin Force origin point
	 */
	void setForceOrigin(Vector3d forceOrigin)
	{
		this.forceOrigin = forceOrigin;
	}

	@Override
	public Vector3d getForce()
	{
		return force;
	}

	/**
	 * Set the force vector
	 * 
	 * @param force Force vector
	 */
	void setForce(Vector3d force)
	{
		this.force = force;
	}

	/**
	 * Updates this ForceResistances from perception
	 * @param perception the result from server message parsing
	 */
	@Override
	public void updateFromPerception(IPerception perception)
	{
		IForceResistancePerceptor frPerceptor = perception
				.getForceResistancePerceptor(getPerceptorName());

		if (frPerceptor == null) {
			setForce(new Vector3d(0, 0, 0));
			return;
		}

		// Fetch new Force Values
		Vector3D newForce = frPerceptor.getForce();
		Vector3D newForceOrigin = frPerceptor.getForceOrigin();

		setForce(new Vector3d(newForce.getX(), newForce.getY(), newForce.getZ()));
		setForceOrigin(new Vector3d(newForceOrigin.getX(), newForceOrigin.getY(),
				newForceOrigin.getZ()));
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof ForceResistance)) {
			return false;
		}
		ForceResistance other = (ForceResistance) o;
		if (!super.equals(other)) {
			return false;
		}
		if (!FuzzyCompare.eq(force, other.force, 0.00001)) {
			return false;
		}
		return FuzzyCompare.eq(forceOrigin, other.forceOrigin, 0.00001f);
	}

	@Override
	public ISensor copy()
	{
		return new ForceResistance(this);
	}
}
