/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel.impl;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

import magma.agent.agentmodel.IGyroRate;
import magma.agent.agentmodel.ISensor;
import magma.agent.perception.IPerception;
import magma.util.FuzzyCompare;
import magma.util.geometry.MatrixUtil;

import org.apache.commons.math.geometry.Vector3D;

/**
 * SimSpark "GyroRate" sensor implementation.
 * 
 * @author Stefan Glaser, Srinivasa Ragavan
 */
public class GyroRate extends Sensor implements IGyroRate
{
	private static final long serialVersionUID = 6859542740368696037L;

	/** handle for logging */
	private static final Logger logger = Logger.getLogger("");

	/** angular change to previous orientation */
	private Vector3d gyro;

	/**
	 * Instantiates a new GyroRate sensor and initializes all values to their
	 * default
	 * 
	 * @param name sensor name
	 */
	public GyroRate(String name, String perceptorName)
	{
		super(name, perceptorName);
		gyro = new Vector3d();
	}

	/**
	 * Copy constructor
	 * @param source the object to copy from
	 */
	public GyroRate(GyroRate source)
	{
		super(source.name, source.perceptorName);
		gyro = new Vector3d(source.gyro);
	}

	@Override
	public Vector3d getGyro()
	{
		return gyro;
	}

	/**
	 * Set 3-dimensional gyro values
	 * 
	 * @param gyro Gyro values
	 */
	public void setGyro(Vector3d gyro)
	{
		// Save current rotation-vector
		this.gyro = gyro;

		logger.log(Level.FINER, "gyro update: ({0}, {1}, {2})", new Object[] {
				gyro.x, gyro.y, gyro.z });
	}

	@Override
	public Matrix3d getOrientationChange()
	{
		double x, y, z;
		Matrix3d m3d = new Matrix3d();

		// Extract the angles of the given rotation-speeds
		x = Math.toRadians(gyro.getX() * 0.02);
		y = Math.toRadians(gyro.getY() * 0.02);
		z = Math.toRadians(gyro.getZ() * 0.02);

		// Rotate the identity matrix
		m3d.setIdentity();
		MatrixUtil.rotateMatrix(m3d, x, y, z);

		return m3d;
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof GyroRate)) {
			return false;
		}
		GyroRate other = (GyroRate) o;
		if (!super.equals(other)) {
			return false;
		}
		return FuzzyCompare.eq(gyro, other.gyro, 0.00001f);
	}

	/**
	 * Updates this Gyro from perception
	 * @param perception the result from server message parsing
	 */
	@Override
	public void updateFromPerception(IPerception perception)
	{
		Vector3D rate = perception.getGyroRatePerceptor(getPerceptorName())
				.getGyro();
		setGyro(new Vector3d(rate.getX(), rate.getY(), rate.getZ()));
	}

	@Override
	public ISensor copy()
	{
		return new GyroRate(this);
	}
}
