/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel.impl;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

import magma.agent.action.IAction;
import magma.agent.agentmodel.IHingeJoint;
import magma.agent.agentmodel.IJoint;
import magma.agent.agentmodel.ISensor;
import magma.agent.meta.agent.IHingeJointConfiguration;
import magma.agent.perception.IPerception;
import magma.util.FuzzyCompare;
import magma.util.geometry.Geometry;

/**
 * SimSpark "HingeJoint" sensor implementation
 * 
 * @author Stefan Glaser, Klaus Dorer
 * 
 */
public class HingeJoint extends Joint implements IHingeJoint
{
	private static final long serialVersionUID = -9142499983257109635L;

	/** current angle of the joint */
	private float ax;

	/** the original angle of the source body model */
	private float originalAx;

	/** a vector describing the rotation axis of the joint */
	private final Vector3d jointAxis;

	/** minimum axis position */
	private final float min_ax;

	/** maximum axis position */
	private final float max_ax;

	/** maximum axis position */
	private final float max_speed;

	/** the rotation matrix to this joint */
	private transient Matrix3d rotationMatrix;

	/**
	 * Instantiates a new HingeJoint sensor
	 * 
	 * @param name Sensor name
	 * @param effectorName Effector name linked to this sensor
	 * @param min Minimum sensor value
	 * @param max Maximum sensor value
	 */
	private HingeJoint(String name, String perceptorName, String effectorName,
			Vector3d jointAxis, float min, float max, float max_speed)
	{
		super(name, perceptorName, effectorName);

		this.ax = 0;
		this.originalAx = 0;
		this.jointAxis = jointAxis;
		this.min_ax = min;
		this.max_ax = max;
		this.max_speed = max_speed;
	}

	public HingeJoint(IHingeJointConfiguration jointConfig)
	{
		this(jointConfig.getName(), jointConfig.getPerceptorName(), jointConfig
				.getEffectorName(), jointConfig.getJointAxis(), jointConfig
				.getMinAngle(), jointConfig.getMaxAngle(), jointConfig
				.getMaxSpeed());
	}

	/**
	 * Copy constructor
	 * @param source the object to copy from
	 */
	private HingeJoint(HingeJoint source)
	{
		super(source);

		this.ax = source.ax;
		// we do not copy the original value but use current as new original
		this.originalAx = ax;
		// we do not expect jointAxis to change. Anyhow, as soon as immutable
		// Vector3ds are used, we need no copy of jointAxis
		this.jointAxis = source.jointAxis;
		this.min_ax = source.min_ax;
		this.max_ax = source.max_ax;
		this.max_speed = source.max_speed;
	}

	@Override
	public float getAxis()
	{
		return ax;
	}

	@Override
	public float getMin()
	{
		return min_ax;
	}

	@Override
	public float getMax()
	{
		return max_ax;
	}

	/**
	 * Called during the update process of the agent model, in order to update
	 * the axis position with the currently perceived joint axis angle.
	 * 
	 * @param ax axis angle
	 */
	public void updateAxisPosition(float ax)
	{
		this.ax = ax;

		// set next expected axis position
		this.originalAx = ax;

		// Reset rotation matrix
		rotationMatrix = null;
	}

	@Override
	public void performAxisSpeed(float speed)
	{
		ax = originalAx + Geometry.limitAbs(speed, max_speed);

		// Reset rotation matrix
		rotationMatrix = null;
	}

	@Override
	public void adjustAxisPosition(double delta)
	{
		performAxisPosition(ax + delta, max_speed);
	}

	@Override
	public float performAxisPosition(double position)
	{
		return performAxisPosition(position, max_speed);
	}

	@Override
	public float performAxisPosition(double position, float maxSpeed)
	{
		float axisSpeed = (float) (position - originalAx) / (max_speed / 2);
		// float axisSpeed = (float) (position - originalAx) / 2.0f;
		// float axisSpeed = (float) (position - originalAx);

		performAxisSpeed(Geometry.limitAbs(axisSpeed, maxSpeed));
		return axisSpeed;
	}

	@Override
	public void performInitialPosition()
	{
		performAxisPosition(0);
	}

	@Override
	public void resetMovement()
	{
		performAxisSpeed(0);
	}

	@Override
	public float getNextAxisSpeed()
	{
		return ax - originalAx;
	}

	@Override
	public Matrix3d getRotationMatrix()
	{
		if (rotationMatrix == null) {
			rotationMatrix = createRotationMatrix(ax, jointAxis);
		}

		return rotationMatrix;
	}

	@Override
	public Matrix3d getInverseRotationMatrix()
	{
		Matrix3d m3d = getRotationMatrix();
		m3d.transpose();

		return m3d;
	}

	@Override
	public Vector3d getJointAxis()
	{
		return jointAxis;
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof HingeJoint)) {
			return false;
		}
		HingeJoint other = (HingeJoint) o;
		if (!super.equals(other)) {
			return false;
		}
		if (!FuzzyCompare.eq(ax, other.ax, 0.00001f)) {
			return false;
		}
		if (!FuzzyCompare.eq(min_ax, other.min_ax, 0.00001f)) {
			return false;
		}
		return FuzzyCompare.eq(max_ax, other.max_ax, 0.00001f);
	}

	@Override
	public ISensor copy()
	{
		return new HingeJoint(this);
	}

	@Override
	public void generateJointAction(IAction action)
	{
		action.setHingeEffectorSpeed(getEffectorName(), getNextAxisSpeed());
	}

	@Override
	public void updateFromPerception(IPerception perception)
	{
		float axis = perception.getHingeJointPerceptor(getPerceptorName())
				.getAxis();
		float delta = ax - originalAx;
		originalAx = axis;
		ax = originalAx + delta;

		// Reset rotation matrix
		rotationMatrix = null;
	}

	@Override
	public void updateJointPositionFromJoint(IJoint joint)
	{
		ax = ((HingeJoint) joint).ax;

		// Reset rotation matrix
		rotationMatrix = null;
	}

	/**
	 * @return the max_speed
	 */
	@Override
	public float getMaxSpeed()
	{
		return max_speed;
	}
}
