/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.agentmodel.impl;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

import magma.agent.action.IAction;
import magma.agent.agentmodel.IJoint;
import magma.agent.agentmodel.ISensor;
import magma.agent.agentmodel.IUniversalJoint;
import magma.agent.meta.agent.IUniversalJointConfiguration;
import magma.agent.perception.IPerception;
import magma.agent.perception.IUniversalJointPerceptor;
import magma.util.FuzzyCompare;

/**
 * SimSpark "UniversalJoint" sensor implementation
 * 
 * @author Stefan Glaser, Klaus Dorer
 * 
 */
public class UniversalJoint extends Joint implements IUniversalJoint
{
	private static final long serialVersionUID = 3124679237337000038L;

	/** angle of first axis */
	private float ax1;

	/** angle of second axis */
	private float ax2;

	/** expected angle of joint1 after the next cycle */
	private float expected_ax1;

	/** expected angle of joint2 after the next cycle */
	private float expected_ax2;

	/** next axis speed to perform for joint1 */
	private float next_axis1_speed;

	/** next axis speed to perform for joint2 */
	private float next_axis2_speed;

	/** a vector describing the rotation around axis1 of the joint */
	private final Vector3d jointAxis1;

	/** minimum axis position for joint1 */
	private final float min_ax1;

	/** maximum axis position for joint1 */
	private final float max_ax1;

	/** maximum axis position for joint1 */
	private final float max_speed1;

	/** a vector describing the rotation around axis2 of the joint */
	private final Vector3d jointAxis2;

	/** minimum axis position for joint2 */
	private final float min_ax2;

	/** maximum axis position for joint2 */
	private final float max_ax2;

	/** maximum axis position for joint2 */
	private final float max_speed2;

	/**
	 * Instantiates a new UniversalJoint sensor
	 * 
	 * @param name - sensor name
	 * @param perceptorName - perceptor name
	 * @param effectorName - effector name
	 * @param minAxis1 - minimum possible angle of axis 1 (in degrees)
	 * @param maxAxis1 - maximum possible angle of axis 1 (in degrees)
	 * @param maxSpeed1 - maximum possible speed of axis 1 (in degrees/cycle)
	 * @param minAxis2 - minimum possible angle of axis 2 (in degrees)
	 * @param maxAxis2 - maximum possible angle of axis 2 (in degrees)
	 * @param maxSpeed2 - maximum possible speed of axis 2 (in degrees/cycle)
	 */
	private UniversalJoint(String name, String perceptorName,
			String effectorName, Vector3d jointAxis1, float minAxis1,
			float maxAxis1, float maxSpeed1, Vector3d jointAxis2, float minAxis2,
			float maxAxis2, float maxSpeed2)
	{
		super(name, perceptorName, effectorName);

		this.ax1 = 0;
		this.ax2 = 0;
		this.expected_ax1 = 0;
		this.expected_ax2 = 0;
		this.next_axis1_speed = 0;
		this.next_axis2_speed = 0;
		this.jointAxis1 = jointAxis1;
		this.min_ax1 = minAxis1;
		this.max_ax1 = maxAxis1;
		this.max_speed1 = maxSpeed1;
		this.jointAxis2 = jointAxis2;
		this.min_ax2 = minAxis2;
		this.max_ax2 = maxAxis2;
		this.max_speed2 = maxSpeed2;
	}

	/**
	 * Instantiates a new UniversalJoint sensor
	 * 
	 * @param jointConfig - the joint meta model
	 */
	public UniversalJoint(IUniversalJointConfiguration jointConfig)
	{
		this(jointConfig.getName(), jointConfig.getPerceptorName(), jointConfig
				.getEffectorName(), jointConfig.getJoint1Axis(), jointConfig
				.getJoint1MinAngle(), jointConfig.getJoint1MaxAngle(), jointConfig
				.getJoint1MaxSpeed(), jointConfig.getJoint2Axis(), jointConfig
				.getJoint2MinAngle(), jointConfig.getJoint2MaxAngle(), jointConfig
				.getJoint2MaxSpeed());
	}

	/**
	 * Copy constructor
	 * @param source to copy from
	 */
	private UniversalJoint(UniversalJoint source)
	{
		super(source);

		this.ax1 = source.ax1;
		this.ax2 = source.ax2;
		this.expected_ax1 = source.expected_ax1;
		this.expected_ax2 = source.expected_ax2;
		this.next_axis1_speed = source.next_axis1_speed;
		this.next_axis2_speed = source.next_axis2_speed;
		this.jointAxis1 = source.jointAxis1;
		this.min_ax1 = source.min_ax1;
		this.max_ax1 = source.max_ax1;
		this.max_speed1 = source.max_speed1;
		this.jointAxis2 = source.jointAxis2;
		this.min_ax2 = source.min_ax2;
		this.max_ax2 = source.max_ax2;
		this.max_speed2 = source.max_speed2;
	}

	@Override
	public float getAxis1()
	{
		return ax1;
	}

	/**
	 * Called during the update process of the agent model, in order to update
	 * the axis1 position with the currently perceived joint axis1 angle.
	 * 
	 * @param ax1 axis1 angle
	 */
	void updateAx1(float ax1)
	{
		this.ax1 = ax1;

		// set next expected axis1 position
		this.expected_ax1 = ax1 + next_axis1_speed;

		// reset next axis1 speed
		this.next_axis1_speed = 0;
	}

	@Override
	public float getAxis2()
	{
		return ax2;
	}

	/**
	 * Called during the update process of the agent model, in order to update
	 * the axis2 position with the currently perceived joint axis2 angle.
	 * 
	 * @param ax2 axis2 angle
	 */
	void updateAx2(float ax2)
	{
		this.ax2 = ax2;

		// set next expected axis2 position
		this.expected_ax2 = ax2 + next_axis2_speed;

		// reset next axis2 speed
		this.next_axis2_speed = 0;
	}

	@Override
	public float getMinAxis1()
	{
		return min_ax1;
	}

	@Override
	public float getMinAxis2()
	{
		return min_ax2;
	}

	@Override
	public float getMaxAxis1()
	{
		return max_ax1;
	}

	@Override
	public float getMaxAxis2()
	{
		return max_ax2;
	}

	@Override
	public float getNextAxis1Speed()
	{
		return max_speed1;
	}

	@Override
	public float getNextAxis2Speed()
	{
		return max_speed2;
	}

	@Override
	public void performAxis1Speed(float speed)
	{
		next_axis1_speed = speed;
	}

	@Override
	public void performAxis2Speed(float speed)
	{
		next_axis2_speed = speed;
	}

	@Override
	public void performAxis1Position(float position)
	{
		performAxis1Position(position, max_speed1);
	}

	@Override
	public void performAxis2Position(float position)
	{
		performAxis2Position(position, max_speed2);
	}

	@Override
	public void performAxis1Position(float position, float maxSpeed)
	{
		float axisSpeed = (position - ax1) / (max_speed1 / 2);

		if (axisSpeed > maxSpeed)
			axisSpeed = maxSpeed;
		if (axisSpeed < -maxSpeed)
			axisSpeed = -maxSpeed;

		performAxis1Speed(axisSpeed);
	}

	@Override
	public void performAxis2Position(float position, float maxSpeed)
	{
		float axisSpeed = (position - ax2) / (max_speed2 / 2);

		if (axisSpeed > maxSpeed)
			axisSpeed = maxSpeed;
		if (axisSpeed < -maxSpeed)
			axisSpeed = -maxSpeed;

		performAxis2Speed(axisSpeed);
	}

	@Override
	public void performInitialPosition()
	{
		performAxis1Position(0);
		performAxis2Position(0);
	}

	@Override
	public void resetMovement()
	{
		performAxis1Speed(0);
		performAxis2Speed(0);
	}

	@Override
	public float getExpectedAxis1Position()
	{
		return expected_ax1;
	}

	@Override
	public float getExpectedAxis2Position()
	{
		return expected_ax2;
	}

	@Override
	public Matrix3d getRotationMatrix()
	{
		Matrix3d res = createInverseRotationMatrix(ax2, jointAxis2);
		res.mul(createInverseRotationMatrix(ax1, jointAxis1));

		return res;
	}

	@Override
	public Matrix3d getInverseRotationMatrix()
	{
		Matrix3d res = createInverseRotationMatrix(ax1, jointAxis1);
		res.mul(createInverseRotationMatrix(ax2, jointAxis2));

		return res;
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof UniversalJoint)) {
			return false;
		}
		UniversalJoint other = (UniversalJoint) o;
		if (!super.equals(other)) {
			return false;
		}
		if (!FuzzyCompare.eq(ax1, other.ax1, 0.00001f)) {
			return false;
		}
		return FuzzyCompare.eq(ax2, other.ax2, 0.00001f);
	}

	@Override
	public ISensor copy()
	{
		return new UniversalJoint(this);
	}

	@Override
	public void generateJointAction(IAction action)
	{
		action.setUniversalEffectorSpeed(getEffectorName(), getNextAxis1Speed(),
				getNextAxis2Speed());
	}

	@Override
	public void updateFromPerception(IPerception perception)
	{
		IUniversalJointPerceptor ujPerceptor = perception
				.getUniversalJointPerceptor(getPerceptorName());

		updateAx1(ujPerceptor.getAxis1());
		updateAx2(ujPerceptor.getAxis2());
	}

	@Override
	public void updateJointPositionFromJoint(IJoint joint)
	{
		next_axis1_speed = ((UniversalJoint) joint).next_axis1_speed;
		next_axis2_speed = ((UniversalJoint) joint).next_axis2_speed;
	}

	@Override
	public int hashCode()
	{
		return name.hashCode() + (int) (ax1 * ax2);
	}
}
