/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.basic;

import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehavior;
import magma.agent.worldmodel.IGlobalMap;

/**
 * Abstract base class for all "simple" behaviors.
 * 
 * @author Klaus Dorer
 */
public abstract class Behavior implements IBehavior
{
	/** handle for logging */
	protected static Logger logger = Logger.getLogger("");

	/** the name of this behavior */
	protected String name;

	/** link to the world model */
	protected IGlobalMap worldModel;

	/** link to the agent model */
	protected IAgentModel agentModel;

	/** the intensity with which to perform the behavior */
	private float intensity;

	/** number of times this behavior was performed since creation */
	protected int performs;

	/** number of times this behavior was performed successfully */
	private int successfulPerforms;

	/** number of times this behavior was performed not successfully */
	private int unsuccessfulPerforms;

	/** the simulation time at which the this behavior started last time */
	private float simulationStart;

	/** the average time needed for successful execution */
	private float successfulAverageTime;

	/** the average time needed for successful execution */
	private float unsuccessfulAverageTime;

	/**
	 * Default constructor for Behaviors
	 * 
	 * @param name Behavior name
	 * @param worldModel World model
	 * @param agentModel Agent model
	 * @param coordinator Coordinator
	 */
	public Behavior(String name, IGlobalMap worldModel, IAgentModel agentModel)
	{
		this.name = name;
		this.worldModel = worldModel;
		this.agentModel = agentModel;
		intensity = 1.0f;
		performs = 0;
		successfulPerforms = 0;
		unsuccessfulPerforms = 0;
		simulationStart = 0;
		successfulAverageTime = 0;
		unsuccessfulAverageTime = 0;
	}

	/**
	 * Copy constructor
	 * 
	 * @param source Source object to copy instance variables from
	 */
	public Behavior(Behavior source)
	{
		this.name = source.name;
		this.worldModel = source.worldModel;
		this.agentModel = source.agentModel;
		this.intensity = source.intensity;
		this.performs = source.performs;
		this.successfulPerforms = source.successfulPerforms;
		this.unsuccessfulPerforms = source.unsuccessfulPerforms;
		this.simulationStart = source.simulationStart;
		this.successfulAverageTime = source.successfulAverageTime;
		this.unsuccessfulAverageTime = source.unsuccessfulAverageTime;
	}

	/**
	 * Non-argument Constructor
	 */
	public Behavior()
	{
	}

	public int getPerforms()
	{
		return performs;
	}

	/**
	 * @return the successfulPerforms
	 */
	public int getSuccessfulPerforms()
	{
		return successfulPerforms;
	}

	/**
	 * @return the unsuccessfulPerforms
	 */
	public int getUnsuccessfulPerforms()
	{
		return unsuccessfulPerforms;
	}

	public final String getName()
	{
		return name;
	}

	public void init()
	{
		statisticsStart();

		logger.log(Level.FINE, "init behavior: {0}", name);
	}

	public boolean isFinished()
	{
		statisticsEnd(true);
		return true;
	}

	/**
	 * Set behavior name
	 * 
	 * @param name New name
	 */
	public void setName(String name)
	{
		this.name = name;
	}

	public void abort()
	{
		statisticsEnd(false);
		init();
	}

	public IBehavior switchTo(IBehavior newBehavior)
	{
		if (isFinished()) {
			onLeavingBehavior();
			return newBehavior;
		}
		return this;
	}

	/**
	 * Called when this behavior is left to perform another behavior
	 */
	protected void onLeavingBehavior()
	{
		init();
	}

	/**
	 * Called when this behavior starts to collect some statistics
	 */
	private void statisticsStart()
	{
		performs++;
		if (worldModel != null) {
			simulationStart = worldModel.getGlobalTime();
		}
	}

	/**
	 * Called when this behavior starts to collect some statistics
	 */
	protected void statisticsEnd(boolean successful)
	{
		float deltaTime = 0;
		if (worldModel != null) {
			deltaTime = worldModel.getGlobalTime() - simulationStart;
		}
		if (successful) {
			successfulAverageTime = runningAverage(successfulAverageTime,
					successfulPerforms, deltaTime);
			successfulPerforms++;
			// System.out.println("Successful: " + name + " count: " + performs
			// + " successful: " + successfulPerforms + " time: " + deltaTime);
		} else {
			unsuccessfulAverageTime = runningAverage(unsuccessfulAverageTime,
					unsuccessfulPerforms, deltaTime);
			unsuccessfulPerforms++;
			// System.out.println("Unsuccessful: " + name + " count: " + performs
			// + " unsuccessful: " + unsuccessfulPerforms + " time: "
			// + deltaTime);
		}
	}

	private float runningAverage(float oldAverage, int oldCount, float newValue)
	{
		return (oldCount * oldAverage + newValue) / (oldCount + 1);
	}

	/**
	 * @return the default is to return the behavior itself
	 */
	public IBehavior getBasicBehavior()
	{
		return this;
	}

	@Override
	public String toString()
	{
		return getName();
	}

	/**
	 * @return the intensity with which to perform the behavior
	 */
	public float getIntensity()
	{
		return intensity;
	}

	/**
	 * @param intensity the intensity with which to perform the behavior. Values
	 *        are free to use, but are typically in the range of [0..1]
	 */
	public void setIntensity(float intensity)
	{
		this.intensity = intensity;
	}
}
