/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.complex;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.IKick;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;

import org.apache.commons.math.geometry.Vector3D;

/**
 * This Behavior is the basic behavior for the closest player to the ball. It
 * navigates the robot to the preferred shoot position, dribbles the ball, or
 * triggers a specific shoot, in a appropriate situation.
 * 
 * @author Stefan Glaser
 * 
 */
public class Attack extends ComplexBehavior
{
	private static final float BALL_KEEP_AWAY_DISTANCE = 0.3f;

	// handle for logging
	private static final Logger logger = Logger.getLogger("");

	/** length of the dribble area */
	private static final double MAX_X_DRIBBLE = 0.35;

	/** width of the dribble area */
	private static final double MAX_Y_DRIBBLE = 0.11;

	/** assumed maximum distance the ball would move while dribbling */
	private static final double MAX_DRIBBLE_KICK_DISTANCE = 3;

	private final Vector3D[] dribbleablePoly;

	private Map<String, IKick> availableKicks;

	private static final String DEFAULT_KICK = IBehaviorConstants.SHOOT_STRAIGHT;

	/**
	 * Default constructor
	 * 
	 * @param worldModel Reference to the world model
	 * @param agentModel Reference to the agent model
	 * @param behaviors List of applicable behaviors
	 */
	public Attack(IGlobalMap worldModel, IAgentModel agentModel,
			Map<String, IBehavior> behaviors)
	{
		super(IBehaviorConstants.ATTACK, worldModel, agentModel, behaviors);

		dribbleablePoly = Geometry.getPolygon(new double[][] {
				{ 0, MAX_Y_DRIBBLE }, { MAX_X_DRIBBLE, MAX_Y_DRIBBLE },
				{ MAX_X_DRIBBLE, -MAX_Y_DRIBBLE }, { 0, -MAX_Y_DRIBBLE } });

		availableKicks = new HashMap<String, IKick>();
		IKick avail;
		avail = (IKick) behaviors.get(IBehaviorConstants.SHOOT_STRAIGHT);
		availableKicks.put(avail.getName(), avail);
		avail = (IKick) behaviors.get(IBehaviorConstants.SHOOT_BACK);
		availableKicks.put(avail.getName(), avail);
		avail = (IKick) behaviors.get(IBehaviorConstants.SHOOT_SIDE_LEFT);
		availableKicks.put(avail.getName(), avail);
		avail = (IKick) behaviors.get(IBehaviorConstants.SHOOT_SIDE_RIGHT);
		availableKicks.put(avail.getName(), avail);
		avail = (IKick) behaviors.get(IBehaviorConstants.SHOOT_BACK_PULL);
		availableKicks.put(avail.getName(), avail);
	}

	@Override
	public IBehavior decideNextBasicBehavior()
	{
		float minRunSpeed = 15;

		// 1. --------------------------------------------------
		// Check if we should run over the ball into the goal
		double[] shootPos = shouldWeRunIntoGoal();
		if (shootPos != null) {
			return runToPosition(shootPos, 100);
		}

		// 2. --------------------------------------------------
		// We don't want to shoot, so maybe we want to dribble
		shootPos = shouldWeDribble();
		if (shootPos != null) {
			return runToPosition(shootPos, 100);
		}

		// 3. --------------------------------------------------
		// Check kick options
		List<IKick> preferredKicks = getPreferredKicks();
		IKick k;
		for (int i = 0; i < preferredKicks.size(); i++) {
			k = preferredKicks.get(i);
			if (!isPlayerInDirection(k.getKickDirection(), 25, 0.5, true)) {
				return k;
			}
		}

		// 4. --------------------------------------------------
		// If all directions are blocked with other agents but we should kick,
		// take the best direction
		if (preferredKicks.size() > 0) {
			return preferredKicks.get(0);
		}

		// 5. --------------------------------------------------
		// We neither want to shoot, nor to dribble, so get in some better
		// position to the ball
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		IPlayer closestOpponent = worldModel.getOpponentAtBall();
		// Vector3D ballPos = worldModel.getBall().getPosition();
		Vector3D ballPos = worldModel.getBall().getFuturePositions(10)[9];

		double weBallDistance = thisPlayer.getDistanceToXY(ballPos);
		double ballOtherGoalDistance = worldModel.getBall().getDistanceToXY(
				worldModel.getOtherGoalPosition());

		Angle shootAngle;

		if (worldModel.getBall().getDistanceToXY(worldModel.getOwnGoalPosition()) > worldModel
				.fieldHalfLength() / 2.0f) {
			// if ball is not near own goal we kick towards other goal
			shootAngle = worldModel.getBall().getDirectionTo(
					worldModel.getOtherGoalPosition());
		} else {
			// If ball is close to own goal we kick away from own goal
			shootAngle = worldModel.getBall()
					.getDirectionTo(worldModel.getOwnGoalPosition())
					.add(Angle.RAD180);
		}

		if (closestOpponent != null) {
			double closestOpponentBallDistance = closestOpponent
					.getDistanceToXY(ballPos);
			if (closestOpponentBallDistance < 2 || ballPos.getX() < 0) {
				if (ballOtherGoalDistance < worldModel.goalHalfWith() * 2) {
					// If the ball is directly in front of the opponents goal
					shootPos = getFastestShootPosition(shootAngle, ballPos);
				} else if (ballPos.getX() > 0 && weBallDistance < .3
						&& closestOpponentBallDistance > 1) {
					// If we are really close to the ball and in the opponent's half,
					// choose the fastest path
					shootPos = getFastestShootPosition(shootAngle, ballPos);
				} else {
					// By default go behind the ball, if a opponent is close to it
					shootPos = getBehindBallPosition(shootAngle, ballPos);
				}

			} else {
				// No opponent near the ball, so simply run for the fastest possible
				// shoot option
				shootPos = getFastestShootPosition(shootAngle, ballPos);
			}
		} else {
			// No opponent visible
			shootPos = getFastestShootPosition(shootAngle, ballPos);
		}

		// TODO: average on shootPos

		return runToPosition(shootPos, minRunSpeed);
	}

	/**
	 * Returns the parameterized RunToPosition behavior.
	 * 
	 * @param targetPos - the target position to run to
	 * @param minRunSpeed - minimum run speed
	 * @return the RunToPosition behavior
	 */
	private IBehavior runToPosition(double[] targetPos, float minRunSpeed)
	{
		RunToPosition run = (RunToPosition) behaviors
				.get(IBehaviorConstants.RUN_TO_POSITION);
		run.setPosition(new Vector3D(targetPos[0], targetPos[1], 0),
				(float) targetPos[2], minRunSpeed, 50, 100);

		return run;
	}

	/**
	 * Check if we should run over the ball into the goal.
	 * 
	 * @return x,y,rot, if we should run over the ball into the goal, or null
	 *         otherwise
	 */
	private double[] shouldWeRunIntoGoal()
	{
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		Vector3D ballPos = worldModel.getBall().getPosition();
		float goalHalfWidth = worldModel.goalHalfWith() - 0.05f;

		if (ballPos.getX() > (worldModel.fieldHalfLength() - 2)
				&& ballPos.getY() < goalHalfWidth
				&& ballPos.getY() > -goalHalfWidth
				&& thisPlayer.isInsidePolygonXY(ballPos, dribbleablePoly)) {
			double viewDirection = thisPlayer.getHorizontalAngle().degrees();

			// Ball is very close or inside the opponent's goal
			// Therefore just allow dribblings towards the goal
			if (ballPos.getX() > (worldModel.fieldHalfLength() - 0.3)) {
				double limit = 80 - (Math.abs(ballPos.getY()) / goalHalfWidth) * 80;

				if ((ballPos.getY() < 0 && 80 > viewDirection && viewDirection > -limit)
						|| (ballPos.getY() >= 0 && limit > viewDirection && viewDirection > -80)) {
					// 3. Dribble 10cm forwards with full speed
					return getDribblePosition();
				} else {
					return null;
				}
			} else {
				IMoveableObject ball = worldModel.getBall();
				Vector3D otherGoalPos = worldModel.getOtherGoalPosition();

				double ballToOtherUpperPost = ball.getDirectionTo(
						otherGoalPos.add(new Vector3D(0, goalHalfWidth, 0)))
						.degrees();
				double ballToOtherLowerPost = ball.getDirectionTo(
						otherGoalPos.subtract(new Vector3D(0, goalHalfWidth, 0)))
						.degrees();

				if (ballToOtherUpperPost > viewDirection
						&& viewDirection > ballToOtherLowerPost) {
					return getDribblePosition();
				} else {
					return null;
				}
			}
		}

		return null;
	}

	/**
	 * Checks if we are in a position to dribble an as well in the need to. It
	 * prevents bad situations with respect to our goal and the field border.
	 * 
	 * @return the position to dribble to, or null otherwise
	 */
	private double[] shouldWeDribble()
	{
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		Vector3D ballPos = worldModel.getBall().getPosition();

		if (thisPlayer.isInsidePolygonXY(ballPos, dribbleablePoly)) {
			IPlayer closestOpponent = worldModel.getOpponentAtBall();

			if (closestOpponent != null
					&& closestOpponent.getDistanceToXY(ballPos) < 1.5) {
				// Since there is a opponent near the ball and we are in dribbleable
				// distance, we should maybe dribble
				double viewDirection = thisPlayer.getHorizontalAngle().degrees();

				// 1. Check if we would run towards our goal
				if (Math.abs(viewDirection) > 140
						|| (ballPos.getX() < 0 && Math.abs(viewDirection) > 70)) {
					return null;
				}

				// 2. Check if we would run outside the field
				Vector3D resultingBallPos = ballPos.add(new Vector3D(
						MAX_DRIBBLE_KICK_DISTANCE, new Vector3D(Math
								.toRadians(viewDirection), 0)));

				if (resultingBallPos.getX() > worldModel.fieldHalfLength()
						|| resultingBallPos.getX() < -worldModel.fieldHalfLength()
						|| resultingBallPos.getY() > worldModel.fieldHalfWith()
						|| resultingBallPos.getY() < -worldModel.fieldHalfWith()) {
					return null;
				}

				// 3. Dribble 10cm forwards with full speed
				return getDribblePosition();
			}
		}

		return null;
	}

	/**
	 * Calculates target position for dribbling, based on our current position
	 * and orientation.
	 * 
	 * @return a double array containing x, y, rot as target position for
	 *         dribbling (usually 10 cm in front of us)
	 */
	private double[] getDribblePosition()
	{
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		double viewDirection = thisPlayer.getHorizontalAngle().radians();

		Vector3D targetPos = thisPlayer.getPosition().add(
				new Vector3D(0.1, new Vector3D(viewDirection, 0)));
		return new double[] { targetPos.getX(), targetPos.getY(), viewDirection };
	}

	/**
	 * Check which kick is currently possible and as well suitable to the current
	 * situation.
	 * 
	 * @return a list of possible, valid kicks, sorted from best to worst
	 */
	private List<IKick> getPreferredKicks()
	{
		List<IKick> preferredKicks = new ArrayList<IKick>();

		for (IKick currentKick : availableKicks.values()) {
			double util = shouldWeKick(currentKick);
			if (util > 0) {
				preferredKicks.add(currentKick);
				currentKick.setUtility(util);
			}
		}

		Collections.sort(preferredKicks, new KickUtilityComparator());

		return preferredKicks;
	}

	/**
	 * Calculate a position behind the ball to get a more defense like position.
	 * 
	 * @param ballTargetDirection
	 * @return x,y,rot
	 */
	private double[] getBehindBallPosition(Angle ballTargetDirection,
			Vector3D ballPos)
	{
		IKick bestKick = availableKicks.get(DEFAULT_KICK);
		return calculateTargetPositionOrientation(ballTargetDirection, bestKick,
				ballPos);
	}

	/**
	 * Calculate the fastest possible shoot position
	 * 
	 * @param ballTargetDirection
	 * @return x,y,rot
	 */
	private double[] getFastestShootPosition(Angle ballTargetDirection,
			Vector3D ballPos)
	{
		// Check, if the ball is directly in from of the opponent's goal, where
		// choosing the direction to the goal center is a bad choice.
		if (worldModel.getBall().getDistanceToXY(
				worldModel.getOtherGoalPosition()) < worldModel.goalHalfWith()) {
			// We could set the direction just straight towards the goal
			ballTargetDirection = Angle.RAD0;

			// Or we adjust the point of relation to goalHalfWidth behind the goal
			// ballTargetDirection = worldModel.getBall().getDirectionTo(
			// worldModel.getOtherGoalPosition().add(
			// new Vector3D(worldModel.goalHalfWith(), 0, 0)));
		}

		Angle weBallDirection = worldModel.getThisPlayer()
				.getDirectionTo(ballPos);
		double deltaPos = ballTargetDirection.subtract(weBallDirection).degrees();
		double deltaDir = worldModel.getThisPlayer().getBodyDirectionTo(ballPos)
				.degrees();

		IKick bestKick = availableKicks.get(DEFAULT_KICK);

		if (Math.abs(deltaPos) <= 65) {
			// Towards shoot direction
			// if (Math.abs(deltaDir) < 120) {
			// // forward kick
			// } else {
			// // backward kick
			// bestKick = availableKicks.get(IBehaviorConstants.SHOOT_BACK);
			// }

		} else if (deltaPos < -65 && deltaPos >= -115 && Math.abs(deltaDir) < 40) {
			// Shoot direction right of us, so make a left side kick
			bestKick = availableKicks.get(IBehaviorConstants.SHOOT_SIDE_LEFT);

		} else if (deltaPos > 65 && deltaPos <= 115 && Math.abs(deltaDir) < 40) {
			// Shoot direction left of us, so make a right side kick
			bestKick = availableKicks.get(IBehaviorConstants.SHOOT_SIDE_RIGHT);

		} else if ((deltaPos > 115 && deltaPos < 150)
				|| (deltaPos < -115 && deltaPos > -150)) {
			// limit applicability of pull kick to small range
			// default kick
		} else if (Math.abs(deltaDir) < 40) {
			// Shoot direction behind us, so make a back kick pull
			bestKick = availableKicks.get(IBehaviorConstants.SHOOT_BACK_PULL);
		}

		logger.log(Level.FINE, "fastestKick: {0} pos: {1} dir: {2}",
				new Object[] { bestKick.getName(), deltaPos, deltaDir });

		return calculateTargetPositionOrientation(ballTargetDirection, bestKick,
				ballPos);
	}

	/**
	 * Calculate a position and orientation behind the ball relative to the given
	 * targetDirection and the keepAwayDistance.
	 * 
	 * @param targetDirection global direction in which to kick
	 * @param bestKick the kick to use
	 * @return x,y,rot
	 */
	private double[] calculateTargetPositionOrientation(Angle targetDirection,
			IKick bestKick, Vector3D ballPos)
	{
		Angle dir = targetDirection.subtract(bestKick.getKickDirection()).add(
				bestKick.getBallFacingDirection());

		Angle weBallDirection = worldModel.getThisPlayer()
				.getDirectionTo(ballPos);

		double distance = bestKick.getMaxBallDistance();
		double delta = Math.abs(dir.subtract(weBallDirection).degrees());

		if (delta > 20) {
			distance = Geometry.getLinearFuzzyValue(20, 80, true, delta)
					* BALL_KEEP_AWAY_DISTANCE + distance;
		}

		Vector3D targetPosition = ballPos.add(-distance,
				new Vector3D(dir.radians(), 0));

		return new double[] { targetPosition.getX(), targetPosition.getY(),
				dir.radians() };
	}

	/**
	 * Checks if we should use the given kick-behavior, in order to get into the
	 * direction of the center of the oppoenent's goal.
	 * 
	 * @param shoot the kick behavior to check
	 * @return +/-1 <br>
	 *         -1 don't kick <br>
	 *         0 < ret < 0.6 => kick possible (the bigger the more towards the
	 *         opponent's goal center) <br>
	 *         1 will end in goal
	 */
	private double shouldWeKick(IKick shoot)
	{
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		IMoveableObject ball = worldModel.getBall();
		Vector3D ballPos = ball.getPosition();

		// Check if ball is in kickable rectangle
		if (thisPlayer.isInsidePolygonXY(ballPos, shoot.getKickableRectangle())) {

			// 1. --------------------------------------------------
			// First check, if we would end up towards the own goal
			double ballToOwnUpperPost = ball.getDirectionTo(
					worldModel.getOwnGoalPosition().add(
							new Vector3D(0, worldModel.goalHalfWith() * 2, 0)))
					.degrees();
			double ballToOwnLowerPost = ball.getDirectionTo(
					worldModel.getOwnGoalPosition().subtract(
							new Vector3D(0, worldModel.goalHalfWith() * 2, 0)))
					.degrees();
			double globalShootDirection = thisPlayer.getHorizontalAngle()
					.add(shoot.getKickDirection()).degrees();

			if (Math.abs(globalShootDirection) > 140
					|| (ballPos.getX() < -(worldModel.fieldHalfLength() - 0.2) && Math
							.abs(globalShootDirection) > 80)
					|| (ballToOwnUpperPost < globalShootDirection && globalShootDirection < ballToOwnLowerPost)) {
				// Don't shoot in directions less than -140 or more than 140 (which
				// is +/-40 deg towards negative x)
				// If the see the ball very close to or behind the field line, just
				// allow a kick-range of +/-80 degrees towards the opponents goal
				// Also definitively don't shoot in this direction, if it points
				// straight into our goal
				return -1;
			}

			// 2. --------------------------------------------------
			// Then check, if we are in score-able distance to the opponent's goal
			if (ball.getDistanceToXY(worldModel.getOtherGoalPosition()) < shoot
					.getMaxKickDistance() * 1.1) {
				float goalHalfWidth = worldModel.goalHalfWith() - 0.1f;

				// Special case 1
				// Check if we see the ball behind the opponent's goal line - in
				// this case we just allow absolute +/-80 deg shoot direction
				if (ballPos.getX() > worldModel.fieldHalfLength()
						&& ballPos.getY() < goalHalfWidth
						&& ballPos.getY() > -goalHalfWidth) {
					if (Math.abs(globalShootDirection) < 80) {
						return 1;
					} else {
						return -1;
					}
				}

				// Special case 2
				// Check if we see the ball very close in front of the opponent's
				// goal - in this case we adjust the kickable direction range
				// depending on the position of the ball
				if (ballPos.getX() > (worldModel.fieldHalfLength() - 0.1)
						&& ballPos.getY() < goalHalfWidth
						&& ballPos.getY() > -goalHalfWidth) {
					double limit = 80 - (Math.abs(ballPos.getY()) / goalHalfWidth) * 80;
					if (ballPos.getY() < 0 && 80 > globalShootDirection
							&& globalShootDirection > -limit) {
						return 1;
					} else if (limit > globalShootDirection
							&& globalShootDirection > -80) {
						return 1;
					} else {
						return -1;
					}
				}

				// Special Case 3
				// Check if we see the ball outside the field, but not in the
				// opponent's goal area
				if (ballPos.getX() >= worldModel.fieldHalfLength()) {
					// In this case we can just assume, where we really are (in x),
					// so we restrict the absolute kick directions to |kickDir| > 90
					// deg towards the middle of the field
					if ((ballPos.getY() < 0 && globalShootDirection > 90)
							|| (ballPos.getY() > 0 && globalShootDirection < -90)) {
						return 1;
					} else {
						return -1;
					}
				}

				// Normal case
				// Check if ball will end up in goal
				double ballToOtherUpperPost = ball.getDirectionTo(
						worldModel.getOtherGoalPosition().add(
								new Vector3D(0, worldModel.goalHalfWith() - 0.05, 0)))
						.degrees();
				double ballToOtherLowerPost = ball.getDirectionTo(
						worldModel.getOtherGoalPosition().subtract(
								new Vector3D(0, worldModel.goalHalfWith() - 0.05, 0)))
						.degrees();

				if (ballToOtherUpperPost > globalShootDirection
						&& globalShootDirection > ballToOtherLowerPost) {
					return 1;
				} else {
					return -1;
				}
			} else {

				// 3. --------------------------------------------------
				// We are in the "urban open field" and won't hit the opponent's nor
				// the own goal - so check if we would kick the ball outside the
				// field with the current shoot behavior
				Vector3D resultingBallPos = ballPos.add(new Vector3D(shoot
						.getMaxKickDistance(), new Vector3D(Math
						.toRadians(globalShootDirection), 0)));

				if (resultingBallPos.getX() > worldModel.fieldHalfLength()
						|| resultingBallPos.getX() < -worldModel.fieldHalfLength()
						|| resultingBallPos.getY() > worldModel.fieldHalfWith()
						|| resultingBallPos.getY() < -worldModel.fieldHalfWith()) {
					return -1;
				}

				double delta = Math.abs(ball
						.getDirectionTo(worldModel.getOtherGoalPosition())
						.subtract(Math.toRadians(globalShootDirection)).degrees());
				double range = Geometry.getLinearFuzzyValue(0, worldModel
						.fieldHalfLength() / 2, false, ball.getPosition().getX()) * 50 + 50;

				// 4. --------------------------------------------------
				// We are in the opponent half so the range to kick gets smaller
				if (ball.getPosition().getX() > 0 && delta > range) {
					return -1;
				}

				// 5. --------------------------------------------------
				// We are in the "urban open field" so return a utility between 0
				// and 0.6 dependent on the direction (>100 to 0 deg difference) to
				// the center of the opponent's goal
				int maxOffset = 60;
				IPlayer opponentAtBall = worldModel.getOpponentAtBall();
				if (opponentAtBall != null
						&& opponentAtBall.getDistanceToXY(ball) < 1.0) {
					maxOffset += 40;
				}
				return Geometry.getLinearFuzzyValue(0, maxOffset, false, delta) * 0.6;
			}
		}

		// Not in kickable areas
		return -1;
	}

	/**
	 * Checks is there is a player in the specified direction relative to the
	 * ball.
	 * 
	 * @param direction the angle to check (local)
	 * @param range the +/- range which indicates a positive hit
	 * @param maxDistance the maximum distance a player is away from the ball
	 * @param countJustOpponents just count opponent players
	 * @return true, if there was at least one player in the given direction -
	 *         false otherwise
	 */
	private boolean isPlayerInDirection(Angle direction, double range,
			double maxDistance, boolean countJustOpponents)
	{
		IMoveableObject ball = worldModel.getBall();
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		Angle ballToPlayer;
		Angle absDirection;
		List<IPlayer> players;

		if (countJustOpponents) {
			players = worldModel.getOpponentsAtBallList();
		} else {
			players = worldModel.getPlayersAtBallList();
		}

		for (IPlayer p : players) {
			if (p.getDistanceToXY(ball) > maxDistance) {
				return false;
			}

			ballToPlayer = ball.getDirectionTo(p.getPosition());
			absDirection = thisPlayer.getHorizontalAngle().add(direction);
			if (Math.abs(ballToPlayer.subtract(absDirection).degrees()) < range) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Sorts kicks in descending order by utility
	 * @author dorer
	 */
	public static class KickUtilityComparator implements Comparator<IKick>
	{
		@Override
		public int compare(IKick o1, IKick o2)
		{
			if (o1.getUtility() < o2.getUtility()) {
				return 1;
			}

			if (o1.getUtility() > o2.getUtility()) {
				return -1;
			}

			return 0;
		}
	}
}
