/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.complex;

import java.util.Map;
import java.util.logging.Level;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.basic.Behavior;
import magma.agent.worldmodel.IGlobalMap;

/**
 * Base class for all behaviors that use a couple of other behaviors
 * @author Klaus Dorer
 */
public abstract class ComplexBehavior extends Behavior
{
	/** the behaviors that are available */
	protected final Map<String, IBehavior> behaviors;

	/** the behavior that was last selected to perform */
	private IBehavior currentBehavior;

	// the behavior to which we initialize
	private final IBehavior defaultBehavior;

	/**
	 * Default constructor for complex behaviors
	 * 
	 * @param name name of behavior
	 * @param worldModel link to world model
	 * @param agentModel link to agent model
	 * @param behaviors map of all available behaviors
	 * @param defaultBehavior the default behavior to start with
	 */
	public ComplexBehavior(String name, IGlobalMap worldModel,
			IAgentModel agentModel, Map<String, IBehavior> behaviors,
			IBehavior defaultBehavior)
	{
		super(name, worldModel, agentModel);
		this.behaviors = behaviors;
		this.defaultBehavior = defaultBehavior;
		currentBehavior = defaultBehavior;
	}

	/**
	 * Default constructor for complex behaviors
	 * 
	 * @param name name of behavior
	 * @param worldModel link to world model
	 * @param agentModel link to agent model
	 * @param behaviors map of all available behaviors
	 */
	public ComplexBehavior(String name, IGlobalMap worldModel,
			IAgentModel agentModel, Map<String, IBehavior> behaviors)
	{
		this(name, worldModel, agentModel, behaviors, behaviors
				.get(IBehaviorConstants.NONE));
	}

	@Override
	public void perform()
	{
		IBehavior toExecute = decideNextBasicBehavior();

		if (toExecute != getCurrentBehavior()) {
			currentBehavior = getCurrentBehavior().switchTo(toExecute);
			logger.log(Level.INFO, "current: {0} new: {1}", new Object[] {
					getCurrentBehavior().getName(), toExecute.getName() });
		}

		getCurrentBehavior().perform();
	}

	/**
	 * Decide which basic behavior to use next
	 * 
	 * @return Next basic behavior
	 */
	public abstract IBehavior decideNextBasicBehavior();

	@Override
	public IBehavior switchTo(IBehavior newBehavior)
	{
		IBehavior newB = getCurrentBehavior().switchTo(newBehavior);
		if (newB == getCurrentBehavior()) {
			return this;
		}
		onLeavingBehavior();
		return newBehavior;
	}

	@Override
	public void abort()
	{
		getCurrentBehavior().abort();
		super.abort();
	}

	/**
	 * @return the currentBehavior
	 */
	public IBehavior getCurrentBehavior()
	{
		return currentBehavior;
	}

	@Override
	public boolean isFinished()
	{
		return currentBehavior.isFinished();
	}

	@Override
	public String toString()
	{
		return super.toString() + " performing " + currentBehavior.toString();
	}

	@Override
	public void init()
	{
		super.init();
		// we have to initialize the current behavior. Otherwise on the next
		// invocation of this complex behavior the old currentBehavior will be
		// asked if it is finished which does not make sense.
		if (behaviors != null) {
			currentBehavior = defaultBehavior;
		}
	}

	/**
	 * A call to this method has side effects!! DecideNextBasicBehavior sets
	 * values to the behavior. This does typically not matter if it is called
	 * again later for real decision.
	 * @return the default is to return the behavior itself
	 */
	@Override
	public IBehavior getBasicBehavior()
	{
		IBehavior basicBehavior = this;
		while (basicBehavior instanceof ComplexBehavior) {
			basicBehavior = ((ComplexBehavior) basicBehavior)
					.decideNextBasicBehavior();
		}
		return basicBehavior;
	}
}
