/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.complex;

import java.util.Map;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Behavior to get in a suitable scoring position, behind the ball, in line with
 * opponent's goal and ball
 * 
 * @author Mathias Ehret
 */
public class GetInScorePosition extends ComplexBehavior
{
	/**  */
	static final float BEHIND_BALL_DISTANCE = 0.15f;

	private static final float BALL_KEEP_AWAY_DISTANCE = 0.5f;

	/** the minimal speed for running */
	private float minSpeed;

	/**
	 * Default constructor
	 * 
	 * @param worldModel Reference to the world model
	 * @param agentModel Reference to the agent model
	 * @param behaviors List of applicable behaviors
	 */
	public GetInScorePosition(IGlobalMap worldModel, IAgentModel agentModel,
			Map<String, IBehavior> behaviors)
	{
		super(IBehaviorConstants.GET_IN_SCORE_POSITION, worldModel, agentModel,
				behaviors);
		minSpeed = 25;
	}

	/**
	 * Default constructor
	 * 
	 * @param worldModel Reference to the world model
	 * @param agentModel Reference to the agent model
	 * @param behaviors List of applicable behaviors
	 */
	public GetInScorePosition(String name, IGlobalMap worldModel,
			IAgentModel agentModel, Map<String, IBehavior> behaviors,
			float minSpeed)
	{
		super(name, worldModel, agentModel, behaviors);
		this.minSpeed = minSpeed;
	}

	@Override
	public IBehavior decideNextBasicBehavior()
	{
		RunToPosition run = (RunToPosition) behaviors
				.get(IBehaviorConstants.RUN_TO_POSITION);

		double[] position = getPosition();
		Vector3D posVector = new Vector3D(position[0], position[1], 0);
		float angle = (float) position[2];

		// if (worldModel.getThisPlayer().getDistanceToXYZ(posVector) > 0.4) {
		run.setPosition(posVector, angle, minSpeed, 50, 100);
		// } else {
		// // when close we run directly to ball
		// run.setPosition(worldModel.getBall().getPosition(), angle);
		// }
		return run;
	}

	double[] getPosition()
	{
		IMoveableObject ball = worldModel.getBall();
		Vector3D ballPosition = ball.getPosition();
		IThisPlayer thisPlayer = worldModel.getThisPlayer();

		Vector3D otherGoalPosition = worldModel.getOtherGoalPosition();
		Vector3D shootVector = otherGoalPosition.subtract(ballPosition);
		// FEATURE: uncomment this lines to use the dynamic shoot direction
		// provided by the worldModel
		// Vector3D shootVector = new Vector3D(worldModel
		// .getPreferredKickDirection().radians(), 0);

		Vector3D weBall = ballPosition.subtract(thisPlayer.getPosition());
		Vector3D scorePosition = ballPosition.add(-BEHIND_BALL_DISTANCE,
				shootVector.normalize());

		double ballGoalDistance = ball.getDistanceToXY(otherGoalPosition);
		if (ballGoalDistance < worldModel.goalHalfWith()) {
			// we run to ball - some y space into goal direction
			double x = otherGoalPosition.getX();
			if (x < 0) {
				// Playing to the left goal
				scorePosition = new Vector3D(ballPosition.getX()
						+ BEHIND_BALL_DISTANCE, ballPosition.getY(), 0);
				shootVector = new Vector3D(-1.0, 0.0, 0.0);
			} else {
				// Playing to the right goal
				scorePosition = new Vector3D(ballPosition.getX()
						- BEHIND_BALL_DISTANCE, ballPosition.getY(), 0);
				shootVector = new Vector3D(1.0, 0.0, 0.0);
			}

		} else {
			// we run to some point that is opposite of ball-goal direction
			Vector3D shoot2D = new Vector3D(shootVector.getX(),
					shootVector.getY(), 0);
			Vector3D meBall2D = new Vector3D(weBall.getX(), weBall.getY(), 0);
			Angle delta1 = Angle.rad(Vector3D.angle(shoot2D, meBall2D));
			double delta = Math.abs(delta1.degrees());

			if (delta > 20) {
				double distanceFromBall = Geometry.getLinearFuzzyValue(20, 80,
						true, delta) * BALL_KEEP_AWAY_DISTANCE + BEHIND_BALL_DISTANCE;

				scorePosition = ballPosition.add(-distanceFromBall,
						shootVector.normalize());
			}
		}
		return new double[] { scorePosition.getX(), scorePosition.getY(),
				shootVector.getAlpha() };
	}
}
