/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.complex;

import java.util.Map;
import java.util.logging.Level;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.worldmodel.GameState;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Behavior to run to the ball
 * 
 * @author Klaus Dorer
 */
public class MoveWithBall extends ComplexBehavior
{
	/**
	 * Default constructor
	 * 
	 * @param worldModel Reference to the world model
	 * @param agentModel Reference to the agent model
	 * @param behaviors List of applicable behaviors
	 */
	public MoveWithBall(IGlobalMap worldModel, IAgentModel agentModel,
			Map<String, IBehavior> behaviors)
	{
		super(IBehaviorConstants.MOVE_WITH_BALL, worldModel, agentModel,
				behaviors);
	}

	@Override
	public IBehavior decideNextBasicBehavior()
	{
		Angle direction = Angle.RAD0;
		Vector3D newPosition = calculateNewXY();

		// avoid to position too close to the ball
		IMoveableObject ball = worldModel.getBall();
		if (ball.getDistanceToXY(newPosition) < 1.3) {
			double dir = worldModel.getThisPlayer().getDirectionTo(ball).radians();
			newPosition = ball.getPosition().add(-1.3, new Vector3D(dir, 0));
		}

		RunToPosition runToPosition = (RunToPosition) behaviors
				.get(IBehaviorConstants.RUN_TO_POSITION);
		runToPosition.setPosition(newPosition, (float) direction.radians(), 0, 0,
				90);

		return runToPosition;
	}

	/**
	 * Calculates new XY Position
	 * 
	 * @return a vector to the target position
	 */
	private Vector3D calculateNewXY()
	{
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		Vector3D target;

		int id = thisPlayer.getID();
		switch (id) {
		// defender
		case 2:
		case 3:
		case 4:
			target = getDefenderPosition(id);
			break;

		// midfield position
		case 5:
		case 6:
		case 7:
			target = getMidfielderPosition(id);
			break;

		// Wing position
		case 8:
		case 9:
			target = getWingPosition(id);
			break;

		default:
			target = getDefenderPosition(id);
			break;
		}

		return target;
	}

	/**
	 * Check if the opponent scored a goal
	 * 
	 * @return True or false
	 */
	private boolean opponentGoalKick()
	{
		return worldModel.getGameState() == GameState.OPPONENT_GOAL_KICK;
	}

	/**
	 * Calculates a position for the passed player that is in the line of ball
	 * and our own goal
	 * @param id id of this player
	 * @return a position on the line between ball and own goal
	 */
	private Vector3D getDefenderPosition(int id)
	{
		Vector3D ownGoalPos = worldModel.getOwnGoalPosition();
		Vector3D ballPos = worldModel.getBall().getPosition();

		Vector3D homePos = worldModel.getHomePosition();
		double deltaBallGoalX = Math.abs(ownGoalPos.getX() - ballPos.getX());

		// we try to be positioned relative to the ball goal distance x
		double minX = -8.5;
		double maxX = -5.5;
		double targetX = minX + 0.3 * deltaBallGoalX;
		if (id == 4) {

		} else if (id == 3) {
			minX = -8.5;
		} else {
			// id == 2
		}

		targetX = keepLimits(targetX, minX, maxX);

		double targetY = homePos.getY() + ballPos.getY() / 3;
		targetY = Geometry.limitAbs(targetY, 3.4);

		Vector3D target = new Vector3D(targetX, targetY, 0);

		logger.log(Level.FINE,
				"move with ball defender: min: {0} target({1},{2})", new Object[] {
						minX, target.getX(), target.getY() });

		return target;
	}

	/**
	 * @param id
	 * @return
	 */
	private Vector3D getMidfielderPosition(int id)
	{
		Vector3D ballPos = worldModel.getBall().getPosition();
		Vector3D homePos = worldModel.getHomePosition();

		// we try to be positioned relative to the ball goal distance x
		double minX = -6.5;
		double maxX = 4.0;
		double baseX = -6.5;
		double factor = 0.6;
		double deltaX = 0;
		double deltaBallBaseX = ballPos.getX() - baseX;
		if (id == 6) {
			minX = -5.5;
			maxX = 5.5;
			baseX = -5.5;
			factor = 0.7;
			deltaX = 0.0;
		}
		double targetX = baseX + factor * deltaBallBaseX + deltaX;

		targetX = keepLimits(targetX, minX, maxX);

		double targetY = homePos.getY() + ballPos.getY() / 3;
		targetY = Geometry.limitAbs(targetY, 3);

		Vector3D target = new Vector3D(targetX, targetY, 0);

		logger.log(Level.FINE,
				"move with ball midfielder: min: {0} target({1},{2})",
				new Object[] { minX, target.getX(), target.getY() });

		return target;
	}

	/**
	 * @param id
	 * @return
	 */
	private Vector3D getWingPosition(int id)
	{
		Vector3D ballPos = worldModel.getBall().getPosition();
		Vector3D homePos = worldModel.getHomePosition();

		// we try to be positioned relative to the ball goal distance x
		double minX = -1.5;
		double maxX = 8.0;
		double targetX = ballPos.getX() + 2.5;
		targetX = keepLimits(targetX, minX, maxX);

		double targetY = homePos.getY() + ballPos.getY() / 3;
		targetY = Geometry.limitAbs(targetY, 3.4);
		Vector3D target = new Vector3D(targetX, targetY, 0);

		if (opponentGoalKick()) {
			// we should wait for the ball at the corner
			final float cornerX = 7;
			final float cornerY = 3;
			if (id == 8) {
				target = new Vector3D(cornerX, cornerY, 0);
			} else {
				target = new Vector3D(cornerX, -cornerY, 0);
			}
		}

		logger.log(Level.FINE, "move with ball wing: min: {0} target({1},{2})",
				new Object[] { minX, target.getX(), target.getY() });

		return target;
	}

	/**
	 * @param target
	 * @param min
	 * @param max
	 * @return
	 */
	private double keepLimits(double target, double min, double max)
	{
		if (target < min) {
			target = min;
		}

		if (target > max) {
			target = max;
		}
		return target;
	}
}
