/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.complex;

import java.util.Map;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.positioning.model.IStrategyPosition;
import magma.agent.behavior.positioning.xml.IXMLRead;
import magma.agent.behavior.positioning.xml.impl.XMLFactory;
import magma.agent.worldmodel.GameState;
import magma.agent.worldmodel.IGlobalMap;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Class PassivePositioning<br>
 * Extends the class ComplexBehavoir.<br>
 * Is responsible for determining the passive position of a player.
 * 
 * @author Denis Zimmermann
 */
public class PassivePositioning extends ComplexBehavior
{
	private final IStrategyPosition strategy; // strategy

	/**
	 * Constructor PassivePositioning()<br>
	 * Creates a XMLFactory and reads the strategy.
	 */
	public PassivePositioning(IGlobalMap worldModel, IAgentModel agentModel,
			Map<String, IBehavior> behaviors)
	{
		super("passivePositioning", worldModel, agentModel, behaviors);

		IXMLRead xmlFact = new XMLFactory();
		strategy = xmlFact.readXML("log/default_strategy.xml");

	}

	@Override
	public IBehavior decideNextBasicBehavior()
	{
		String playerRole = "Player_" + worldModel.getThisPlayer().getID();

		Vector3D newPosition = null;

		try {
			newPosition = strategy.getPositionData(playerRole,
					determineSituation(), worldModel);

			RunToPosition runToPosition = (RunToPosition) behaviors
					.get(IBehaviorConstants.RUN_TO_POSITION);

			runToPosition.setPosition(newPosition,
					(float) Math.toRadians(newPosition.getZ()), 0, 0, 90);

			return runToPosition;
		} catch (Exception e) {
			e.printStackTrace();
		}

		return null;
	}

	/**
	 * Function determineSituation<br>
	 * Determines the situation with the given playmode, ball-position and the
	 * player at ball
	 * 
	 * @return situation actual situation
	 */
	protected String determineSituation()
	{
		// TODO: Check Side compatibility
		// recieve the playmode from the worldmodel
		GameState gameState = worldModel.getGameState();

		// own player at ball?
		boolean ownPlayer = worldModel.getPlayersAtBallList().get(0).isOwnTeam();

		// get the position of the ball
		Vector3D ballPosition = worldModel.getBall().getPosition();

		switch (gameState) {
		case OWN_KICK_OFF:
			return "own_kick_off";

		case OPPONENT_KICK_OFF:
			return "opponent_kick_off";

		case OWN_CORNER_KICK:
			if (ballPosition.getY() > 0) {
				return "own_corner_kick_above";
			}
			return "own_corner_kick_below";

		case OPPONENT_CORNER_KICK:
			if (ballPosition.getY() > 0) {
				return "opponent_corner_kick_above";
			}
			return "opponent_corner_kick_below";

		case OWN_GOAL_KICK:
			return "own_goal_kick";

		case OPPONENT_GOAL_KICK:
			return "opponent_goal_kick";

		case OWN_OFFSIDE:
			return "own_offside";

		case OPPONENT_OFFSIDE:
			return "opponent_offside";

		case OWN_FREE_KICK:
			if (ballPosition.getY() > 0) {
				return "own_free_kick_above";
			}
			return "own_free_kick_below";

		case OPPONENT_FREE_KICK:
			if (ballPosition.getY() > 0) {
				return "opponent_free_kick_above";
			}
			return "opponent_free_kick_below";

		case OWN_KICK_IN:
			if (ballPosition.getY() > 0) {
				return "own_kick_in_above";
			}
			return "own_kick_in_below";

		case OPPONENT_KICK_IN:
			if (ballPosition.getY() > 0) {
				return "opponent_kick_in_above";
			}
			return "opponent_kick_in_below";

		case PLAY_ON:
			if (ownPlayer) {
				if (worldModel.isInOwnHalf(ballPosition)) {
					return "Attack_own_half";
				}
				return "Attack_opponent_half";
			}

			if (worldModel.isInOwnHalf(ballPosition)) {
				return "Defend_own_half";
			}
			return "Defend_opponent_half";
		}

		return "default";
	}
}
