/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.complex;

import java.util.List;
import java.util.Map;
import java.util.logging.Level;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.IWalk;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IPositionManager;
import magma.agent.worldmodel.IThisPlayer;
import magma.agent.worldmodel.IVisibleObject;
import magma.util.geometry.Angle;
import magma.util.geometry.PositionOrientation;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Implements a behavior which is able to walk/run to a specified target
 * position
 */
public class RunToPosition extends ComplexBehavior
{
	private static final int ANGULAR_PRECISION = 60;

	static final float DISTANCE_PRECISION = 0.15f;

	static final float STOP_RUN_DISTANCE = 0.1f;

	/** distance at which we start turning to desired direction */
	private static final float TURN_TO_DESIRED_DIR_DISTANCE = 0.30f;

	/** distance above which it is always best to turn and run */
	private static final double FASTEST_TO_RUN_DISTANCE = 1.0;

	/** the minimum speed with which to walk forward */
	private double minSpeedForward;

	/** the minimum speed with which to walk forward */
	private double maxSpeedForward;

	/** the minimum speed with which to walk sideward */
	private double minSpeedSideward;

	/**
	 * Constructor
	 * 
	 * @param worldModel Reference to the world model
	 * @param agentModel Reference to the agent model
	 * @param behaviors List of applicable behaviors
	 */
	public RunToPosition(IGlobalMap worldModel, IAgentModel agentModel,
			Map<String, IBehavior> behaviors)
	{
		super(IBehaviorConstants.RUN_TO_POSITION, worldModel, agentModel,
				behaviors);
		minSpeedForward = 0;
		maxSpeedForward = 100;
		minSpeedSideward = 0;
	}

	/**
	 * 
	 */
	@Override
	public IBehavior decideNextBasicBehavior()
	{
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		IPositionManager posManager = thisPlayer.getPositionManager();
		PositionOrientation targetPositionOrientation = posManager
				.getFinalPosition();
		Vector3D finalPosition = targetPositionOrientation.getPosition();
		Angle relDirAtPosition = targetPositionOrientation.getOrientationX()
				.subtract(thisPlayer.getHorizontalAngle());

		PositionOrientation desiredPos = avoidCollision(targetPositionOrientation);

		// clear desired list from any previous collisions avoidances
		posManager.clear();
		posManager.addDesiredPosition(0, targetPositionOrientation);
		double distanceToFinal = thisPlayer.getDistanceToXY(finalPosition);
		double distanceToDesired = distanceToFinal;
		if (desiredPos != targetPositionOrientation) {
			// player is avoiding an obstacle so add new position to the list
			posManager.addDesiredPosition(0, desiredPos);
			distanceToDesired = thisPlayer.getDistanceToXY(desiredPos
					.getPosition());
		}

		// dir is the relative direction in which to face
		// when far we want to face to the destination position
		Vector3D position = desiredPos.getPosition();
		Angle dirToGo = thisPlayer.getBodyDirectionTo(position);
		if (distanceToFinal < TURN_TO_DESIRED_DIR_DISTANCE) {
			// when close to destination we want to face into desired direction
			dirToGo = relDirAtPosition;
		}

		// desired position is in direction of the own goal or the opponents goal
		double forwards_backwards = position.getX()
				- thisPlayer.getPosition().getX();

		// desired position is on the left or right
		double right_left = position.getY() - thisPlayer.getPosition().getY();

		IWalk walk = (IWalk) behaviors.get(IBehaviorConstants.WALK);

		double dirDegree = dirToGo.degrees();
		if (Math.abs(dirDegree) > ANGULAR_PRECISION
				&& distanceToDesired > FASTEST_TO_RUN_DISTANCE) {
			walk.walk(0, 0, dirDegree);
		} else {

			if (distanceToFinal <= FASTEST_TO_RUN_DISTANCE) {
				// adjust dir to allow forwards, backwards, sidewards walk
				if (Math.abs(relDirAtPosition.degrees()) <= 50) {
					// only slide if facing into proper destination direction

					if (dirDegree > 70 && dirDegree < 110) {
						// step sidewise left
						dirDegree = dirDegree - 90;
					} else if (dirDegree < -70 && dirDegree > -110) {
						// step sidewise right
						dirDegree = 90 + dirDegree;
					} else if (dirDegree <= -110) {
						// walk backwards
						dirDegree = 180 + dirDegree;
					} else if (dirDegree >= 110) {
						// walk backwards
						dirDegree = dirDegree - 180;
					}
				}
			}
			if (desiredPos != targetPositionOrientation) {
				if (distanceToDesired <= 0.2) {
					// if we run to an intermediate position and if we are almost
					// there we do not turn
					dirDegree = 0;
					// System.out.println("no turn: " + distanceToDesired);
				}
			}

			walk.globalWalk(forwards_backwards, right_left, dirDegree,
					minSpeedForward, minSpeedSideward, maxSpeedForward);
		}

		return walk;
	}

	/**
	 * Calculates if obstacles are in our way
	 * @param goalPos the destination position to run to
	 * @return old position (goalPos) if there is nothing in way, a new position
	 *         to avoid collisions otherwise
	 */
	PositionOrientation avoidCollision(PositionOrientation goalPos)
	{
		PositionOrientation result = goalPos;
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		double goalPosDistance = thisPlayer
				.getDistanceToXY(goalPos.getPosition());

		List<IVisibleObject> obstacles = worldModel.getObstacles();
		for (IVisibleObject obstacle : obstacles) {

			// check if the player intersects our trajectory
			Vector3D obstaclePos = obstacle.getPosition();
			double distanceToMe = thisPlayer.getDistanceToXY(obstaclePos);
			if (distanceToMe > goalPosDistance) {
				// farer than goal position
				continue;
			}

			// calculate distance to line me -> goalPos
			Vector3D mePos2D = new Vector3D(thisPlayer.getPosition().getX(),
					thisPlayer.getPosition().getY(), 0);
			Vector3D goalPos2D = new Vector3D(goalPos.getPosition().getX(),
					goalPos.getPosition().getY(), 0);
			Vector3D meGoalPos2D = goalPos2D.subtract(mePos2D);

			Vector3D obstaclePos2D = new Vector3D(obstaclePos.getX(),
					obstaclePos.getY(), 0);
			Vector3D mePlayerPos2D = obstaclePos2D.subtract(mePos2D);

			// check if destination position and obstacle position lie in opposite
			// direction
			Angle angle = Angle.rad(Vector3D.angle(meGoalPos2D, mePlayerPos2D));
			logger.log(Level.FINEST, "avoidCollision angle: {0}",
					new Object[] { angle });

			if (Math.abs(angle.degrees()) > 90) {
				// obstacle is not in goal direction but opposite direction
				continue;
			}

			// get a normalized orthogonal (in 2 D space)
			meGoalPos2D = meGoalPos2D.normalize();
			Vector3D norm = new Vector3D(-meGoalPos2D.getY(), meGoalPos2D.getX(),
					0);
			double distanceStartOrigin = Vector3D.dotProduct(mePos2D, norm);

			double distanceToLine = Vector3D.dotProduct(obstaclePos2D, norm);
			distanceToLine = Math.abs(distanceToLine - distanceStartOrigin);

			double criticalDistance = obstacle.getCollisionDistance();
			if (distanceToLine > criticalDistance) {
				continue;
			}

			// we have to run around the obstacle
			// calculate base for intermediate positions
			double factor = distanceToMe / (3.0 * criticalDistance);
			if (factor > 1.0) {
				factor = 1.0;
			}
			Vector3D base = mePos2D.add(factor, obstaclePos2D.subtract(mePos2D));

			Vector3D intermediate;
			Vector3D intermediate1 = base.add(criticalDistance, norm);
			Vector3D intermediate2 = base.subtract(criticalDistance, norm);

			// decide for the shortest path
			double distance1 = thisPlayer.getDistanceToXY(intermediate1);
			double distance2 = thisPlayer.getDistanceToXY(intermediate2);
			if (distance1 < distance2) {
				intermediate = intermediate1;
			} else {
				intermediate = intermediate2;
			}

			// calculate desired angle at intermediate point
			double desiredAngle = goalPos2D.subtract(intermediate).getAlpha();
			result = new PositionOrientation(intermediate, Angle.rad(desiredAngle));

			logger.log(Level.FINE,
					"avoiding obstacle {0} at (x/y): {1}/{2} going to {3}",
					new Object[] { obstacle.getName(), obstaclePos2D.getX(),
							obstaclePos2D.getY(), result });

			// TODO: do not stop in this case but calculate a path taking all
			// players into account using AStar search
			break;
		}
		return result;
	}

	/**
	 * Set new target position
	 * 
	 * @param newPosition New target Position, null, if current position should
	 *        be used
	 * @param rotation Target rotation (radians)
	 * @param minSpeedForward the minimum speed for forward direction
	 * @param minSpeedSideward the minimum speed for sideward direction
	 */
	public void setPosition(Vector3D newPosition, float rotation,
			double minSpeedForward, double minSpeedSideward, double maxSpeedForward)
	{
		this.minSpeedForward = minSpeedForward;
		this.minSpeedSideward = minSpeedSideward;
		this.maxSpeedForward = maxSpeedForward;
		Angle rot = Angle.rad(rotation);
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		if (newPosition == null) {
			newPosition = thisPlayer.getPosition();
		} else {
			newPosition = avoidTeammate(newPosition);
		}
		IPositionManager posManager = thisPlayer.getPositionManager();
		posManager.setDesiredPosition(newPosition, rot, false);
		Vector3D current = thisPlayer.getPosition();

		logger.log(
				Level.FINE,
				"setDesiredPosition (x/y/angle): {0}/{1}/{2} current: {3}/{4}/{5}",
				new Object[] { newPosition.getX(), newPosition.getY(),
						rot.degrees(), current.getX(), current.getY(),
						thisPlayer.getHorizontalAngle() });
	}

	/**
	 * We calculate whether teammates stop us from moving to a position
	 * @param newPosition
	 * @return
	 */
	Vector3D avoidTeammate(Vector3D newPosition)
	{
		List<IPlayer> allTeammates = worldModel.getTeammatesAtBall();
		Vector3D ownPosition = worldModel.getThisPlayer().getPosition();
		Vector3D ownPosNewPos = newPosition.subtract(ownPosition);
		double norm = ownPosNewPos.getNorm();

		for (IPlayer player : allTeammates) {
			if (player.getDistanceToXY(newPosition) < 0.5) {
				logger.log(Level.INFO,
						"avoidTeammate norm: {0} newPosition: {1},{2}", new Object[] {
								norm, newPosition.getX(), newPosition.getY() });
				// If we are near enough already we can just stay in place
				if (norm < 1) {
					return ownPosition;
				}

				return ownPosNewPos.normalize().scalarMultiply(norm - 1);
			}
		}

		// If no teammates are near to the new position just go there
		return newPosition;
	}
}
