/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.motor;

import java.io.BufferedReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

import magma.agent.behavior.basic.BehaviorParameter;
import magma.util.function.Function;
import magma.util.function.PiecewiseLinearFunction;
import magma.util.function.SinFunction;
import magma.util.function.SplineFunction;

/**
 * @author Klaus Dorer
 * 
 */
public class FunctionBehaviorParameter extends BehaviorParameter implements
		Serializable
{
	/** serial ID for serialization */
	private static final long serialVersionUID = 1L;

	/** map containing an entry for all joints involved in the behavior */
	private final Map<String, Function> joints;

	/** the version of the file format we use */
	private int version;

	/** the name of the corresponding behavior */
	private String name;

	/** the number of cycles the behavior takes */
	private float period;

	/**
	 * Creates joint functions from behavior function files
	 * 
	 * @param filename the file to read from
	 * @throws IOException If the file could not be successfully opened and
	 *         parsed
	 */
	public FunctionBehaviorParameter(String filename) throws IOException
	{
		joints = new HashMap<String, Function>();
		String line;
		String[] csvLineContents;

		ClassLoader cl = this.getClass().getClassLoader();
		InputStream is = cl.getResourceAsStream(filename);
		BufferedReader in2 = new BufferedReader(new InputStreamReader(is));
		line = in2.readLine();
		// read the date from the first line
		csvLineContents = line.split(",");
		version = Integer.valueOf(csvLineContents[1]);
		name = csvLineContents[0];
		period = Float.valueOf(csvLineContents[2]);

		// read all data for each joint
		while ((line = in2.readLine()) != null) {
			csvLineContents = line.split(",");
			Function jointFunction = null;
			if (csvLineContents[1].equals(SplineFunction.SPLINE_NAME)) {
				// new joint need to be a spline
				jointFunction = new SplineFunction(csvLineContents);

			} else if (csvLineContents[1].equals(SinFunction.SIN_NAME)) {
				// new joint need to be a sinus
				jointFunction = new SinFunction(csvLineContents, period);

			} else if (csvLineContents[1]
					.equals(PiecewiseLinearFunction.LINEAR_NAME)) {
				// new joint need to be a piecewise linear
				jointFunction = new PiecewiseLinearFunction(csvLineContents);
			}
			joints.put(csvLineContents[0], jointFunction);
		}
	}

	/**
	 * Writes a function motor file
	 * @param filename the name to use as filename
	 */
	@Override
	public void writeToFile(String filename)
	{
		PrintWriter out = null;
		try {
			out = new PrintWriter(new FileWriter(filename));
			out.println(version + "," + name + "," + period);

			// write all data for each joint
			for (String key : joints.keySet()) {
				Function jointFunction = joints.get(key);
				String line = key + "," + jointFunction.toCSVString();
				out.println(line);
			}

		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (out != null) {
				out.close();
			}
		}

	}

	/**
	 * @return a map of functions for each joint that is involved in the behavior
	 */
	public Map<String, Function> getJoints()
	{
		return joints;
	}

	/**
	 * @return the period
	 */
	public float getPeriod()
	{
		return period;
	}

}
