/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.behavior.motor;

import java.io.IOException;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.agentmodel.IBodyModel;
import magma.agent.agentmodel.IBodyPart;
import magma.agent.agentmodel.IForceResistance;
import magma.agent.behavior.basic.BehaviorParameter;
import magma.agent.worldmodel.IGlobalMap;
import magma.robots.nao.INaoConstants;
import magma.util.FuzzyCompare;

/**
 * Decorator of the get up from back motor file behavior that does dynamic
 * adjustment of foot pitch angles due to gyro
 * @author dorer
 */
public class GetUpFromBackBehavior extends AdjustableMotorFileBehavior
{

	/**
	 * @param name
	 * @param worldModel
	 * @param agentModel
	 * @param params
	 * @param filepath
	 * @param slowDown
	 * @param startSlowDown
	 * @throws IOException
	 */
	public GetUpFromBackBehavior(String name, IGlobalMap worldModel,
			IAgentModel agentModel, BehaviorParameter params, String filepath,
			int slowDown, int startSlowDown) throws IOException
	{
		super(name, worldModel, agentModel, params, filepath, slowDown,
				startSlowDown);
	}

	@Override
	public boolean isFinished()
	{
		boolean result = super.isFinished();
		if (result) {
			// check if this was successful
			Matrix3d orientation = worldModel.getThisPlayer().getOrientation();

			double zOrientation = orientation.m22;
			if (zOrientation > 0.9) {
				statisticsEnd(true);
			} else {
				statisticsEnd(false);
			}

		}
		return result;
	}

	/**
	 * Adjusts hip pitch in the last phase of getting up to keep torso straight
	 */
	@Override
	protected void pastPerform(IBodyModel body, int realStep)
	{

		if ((realStep > 21)) {

			IBodyModel bodyModel = agentModel.getFutureBodyModel();

			IBodyPart lFoot = bodyModel.getBodyPart(INaoConstants.LFoot);
			IBodyPart rFoot = bodyModel.getBodyPart(INaoConstants.RFoot);

			Vector3d lFootForceOrigin = ((IForceResistance) lFoot
					.getSensorDeep(INaoConstants.LFootForce)).getForceOrigin();
			Vector3d rFootForceOrigin = ((IForceResistance) rFoot
					.getSensorDeep(INaoConstants.RFootForce)).getForceOrigin();

			double lFootForce = ((IForceResistance) lFoot
					.getSensorDeep(INaoConstants.LFootForce)).getForce().length() / 50;
			double rFootForce = ((IForceResistance) rFoot
					.getSensorDeep(INaoConstants.RFootForce)).getForce().length() / 50;

			double distToTip = 0.06;
			double distToHeel = 0.1;

			if (FuzzyCompare.eq(lFoot.getEdges()[0].getY(),
					lFootForceOrigin.getY(), distToTip)
					&& FuzzyCompare.eq(rFoot.getEdges()[0].getY(),
							rFootForceOrigin.getY(), distToTip)) {
				bodyModel.adjustHingeJointPosition(INaoConstants.LFootPitch,
						-lFootForce);
				bodyModel.adjustHingeJointPosition(INaoConstants.RFootPitch,
						-rFootForce);

			} else if (FuzzyCompare.eq(lFoot.getEdges()[2].getY(),
					lFootForceOrigin.getY(), distToHeel)
					&& FuzzyCompare.eq(rFoot.getEdges()[2].getY(),
							rFootForceOrigin.getY(), distToHeel)) {
				bodyModel.adjustHingeJointPosition(INaoConstants.LFootPitch,
						lFootForce);
				bodyModel.adjustHingeJointPosition(INaoConstants.RFootPitch,
						rFootForce);

			}
		}
	}
}
