/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.motor;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.agentmodel.IBodyModel;
import magma.agent.agentmodel.IHingeJoint;
import magma.agent.behavior.basic.BehaviorParameter;
import magma.agent.worldmodel.IGlobalMap;
import magma.util.logging.CSVFormatter;

/**
 * Implements a static motor movement behavior, movement data is read from a
 * file
 * 
 * @author Klaus Dorer
 */
public class MotorFileBehavior extends MotorBehavior
{
	/** values of each joint angles (in degrees) for each step */
	private ArrayList<HashMap<String, Float>> steps;

	private int slowDownFactor;

	private int startSlowDownFactor;

	private int currentStep;

	private boolean startModus = true;

	/** the current line in the motor file */
	private int realStep;

	/**
	 * Instantiates and initializes a new Balance behavior object
	 * 
	 * @param name Behavior name
	 * @param worldModel Reference to the world model object
	 * @param agentModel Reference to the agent model object
	 * @param params Behavior parameters
	 * @param filepath Path to movement data file
	 * @param slowDown Slow down factor
	 * @param startSlowDown Startup slowDown factor
	 * @throws IOException If the movement data file could not be successfully
	 *         opened and parsed
	 */
	public MotorFileBehavior(String name, IGlobalMap worldModel,
			IAgentModel agentModel, BehaviorParameter params, String filepath,
			int slowDown, int startSlowDown) throws IOException
	{
		super(name, worldModel, agentModel);

		ReadCSV valuesFromFile = new ReadCSV(filepath);
		this.steps = valuesFromFile.getFormattedCSV();

		this.currentStep = 0;
		this.slowDownFactor = slowDown; // TO PARSE
		this.startSlowDownFactor = startSlowDown; // TO PARSE
		this.finished = false;
		if (LOG) {
			initLogger(Level.INFO);
		}
	}

	@Override
	public void init()
	{
		super.init();
		currentStep = 0;
		startModus = true;
	}

	@Override
	public void perform()
	{
		logger.log(Level.FINEST, "{0}: {1}", new Object[] { getName(),
				currentStep });
		finished = false;

		IBodyModel body = agentModel.getFutureBodyModel();
		if (startModus) {
			performSlowDown(body, startSlowDownFactor);

			currentStep++;
			if (currentStep >= startSlowDownFactor) {
				startModus = false;
				currentStep = slowDownFactor;
			}
		} else {
			performSlowDown(body, slowDownFactor);

			currentStep++;
			if (currentStep >= (steps.size() * slowDownFactor)) {
				finished = true;
				currentStep = 0;
			}
		}
		pastPerform(body, realStep);
	}

	/**
	 * Called after the motor file was translated into joint speeds. The default
	 * implementation does nothing
	 * @param body the body model on which this was done
	 * @param realStep the step within the behavior file
	 */
	protected void pastPerform(IBodyModel body, int realStep)
	{
	}

	private void performSlowDown(IBodyModel body, int slowDown)
	{
		float speedFactor = slowDown - (currentStep % slowDown);
		IHingeJoint hingeJoint;
		float currentAngle;
		float desiredAngle;
		float delta;
		float speed;

		StringBuilder buffer = new StringBuilder();
		realStep = currentStep / slowDown;
		Map<String, Float> desiredAngles = steps.get(realStep);

		// loop through all joints affected
		for (String angleName : desiredAngles.keySet()) {

			// calculate speed
			desiredAngle = desiredAngles.get(angleName);
			hingeJoint = (IHingeJoint) body.getJoint(angleName);

			if (hingeJoint == null) {
				logger.log(Level.SEVERE, "unknown hinge joint: {0}", angleName);
			} else {
				currentAngle = hingeJoint.getAxis();
				delta = (desiredAngle - currentAngle);
				// use this line if calculating on sensed model
				speed = delta / 2.0f;
				// use this line if calculating on expected model
				// speed = delta;

				speed /= speedFactor;

				if (speed > -0.02 && speed < 0.02) {
					speed = 0.0f;
				}

				if (LOG) {
					buffer.append(desiredAngle).append(";");
					buffer.append(currentAngle).append(";");
					buffer.append(speed).append(";");
				}
				// set axis speed
				hingeJoint.performAxisSpeed(speed);
			}
		}
		if (LOG) {
			Logger.getLogger(getName()).log(Level.INFO, buffer.toString());
		}
	}

	/**
	 * Initializes logging functionality
	 * 
	 * @param level Logging level
	 */
	private void initLogger(Level level)
	{
		try {
			Logger loggerCSV = Logger.getLogger(getName());
			FileHandler handler = new FileHandler("log/" + getName() + ".csv");
			handler.setFormatter(new CSVFormatter());
			loggerCSV.addHandler(handler);
			loggerCSV.setLevel(level);

			// write heading
			StringBuilder buffer = new StringBuilder();
			Map<String, Float> namesMap = steps.get(0);

			for (String stepName : namesMap.keySet()) {
				buffer.append(stepName).append(";");
				buffer.append("angleAsIs").append(";");
				buffer.append("speed").append(";");
			}

			loggerCSV.log(Level.INFO, buffer.toString());
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
