/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.behavior.movement;

import java.util.ArrayList;
import java.util.List;

import magma.agent.agentmodel.IAgentModel;

/**
 * One phase of movement containing the information of which single joints to
 * move as MovementSingles
 * @author dorer
 */
public class MovementPhase
{
	/** the list of movements */
	private List<MovementSingle> movementSingles;

	/** the number of cycles this phase should run */
	private final int cycles;

	/** the number of cycles this phase is running */
	private int cyclesPerformed;

	/** name of the phase */
	private final String name;

	/** true if the behavior can be stopped when all joints have reached angle */
	private final boolean skipWhenFinished;

	/**
	 * @param cycles the number of cycles this phase is running
	 */
	public MovementPhase(String name, int cycles)
	{
		this(name, cycles, false);
	}

	/**
	 * @param name the name of this phase
	 * @param cycles the number of cycles this phase is running
	 * @param skipWhenFinished true if this phase is stopped when all joints have
	 *        reached the destination
	 */
	public MovementPhase(String name, int cycles, boolean skipWhenFinished)
	{
		this.name = name;
		this.cycles = cycles;
		this.skipWhenFinished = skipWhenFinished;
		cyclesPerformed = 0;
		movementSingles = new ArrayList<MovementSingle>();
	}

	public MovementPhase add(MovementSingle move)
	{
		movementSingles.add(move);
		return this;
	}

	public MovementPhase add(String jointName, double jointAngle, float speed)
	{
		MovementSingle move = new MovementSingle(jointName, jointAngle, speed);
		movementSingles.add(move);
		return this;
	}

	/**
	 * Resets the state information
	 */
	public void init()
	{
		cyclesPerformed = 0;
	}

	/**
	 * Calls move for all single movements
	 * @param agent reference to the agentmodel
	 * @return true if the phase should continue
	 */
	public boolean perform(IAgentModel agent)
	{
		boolean reached = true;
		for (MovementSingle current : movementSingles) {
			reached = current.move(agent) && reached;
		}

		if (skipWhenFinished && reached) {
			// we have reached the destination, so stop the movement if dynamic
			// System.out.println("stop: " + cyclesPerformed + " cycles: " +
			// cycles);
			cyclesPerformed = cycles;
		}

		cyclesPerformed++;
		if (cyclesPerformed >= cycles) {
			cyclesPerformed = 0;
			return false;
		}
		return true;
	}

	/**
	 * @return the cyclesPerformed
	 */
	public int getCyclesPerformed()
	{
		return cyclesPerformed;
	}

	/**
	 * @return the name
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * @return
	 */
	public MovementPhase getLeftVersion()
	{
		String newName = name.replace("Right", "Left");
		MovementPhase result = new MovementPhase(newName, cycles);
		for (MovementSingle current : movementSingles) {
			MovementSingle newMovement = current.getLeftVersion();
			result.add(newMovement);
		}
		return result;
	}

	@Override
	public String toString()
	{
		return name;
	}

}
