/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.positioning.model.impl;

import java.util.List;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlElementRefs;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

import magma.agent.behavior.positioning.model.IReferenceObject;
import magma.agent.behavior.positioning.model.IRole;
import magma.agent.behavior.positioning.model.IRule;
import magma.agent.worldmodel.IGlobalMap;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Class Role<br>
 * Represents a role with the including rule.
 * 
 * @author Denis Zimmermann
 */
@XmlRootElement(name = "Role")
@XmlType(propOrder = { "roleName", "rule" })
public class Role implements IRole
{
	private String roleName; // Name of the role

	private IRule rule; // rule

	/**
	 * Constructor Role()
	 */
	public Role()
	{
		this.roleName = "default rolename";
		this.rule = new RuleAbsolute();
	}

	/**
	 * Constructor Role()
	 * 
	 * @param roleName Name of the role
	 */
	public Role(String roleName)
	{
		this.roleName = roleName;
		this.rule = new RuleAbsolute();
	}

	/**
	 * Constructor Role()
	 * 
	 * @param roleName Name of the role
	 * @param rule List of rule
	 */
	public Role(String roleName, IRule rule)
	{
		this.roleName = roleName;
		this.rule = rule;
	}

	/**
	 * Function getRoleName()<br>
	 * Returns the name of the role.
	 * 
	 * @return roleName Name of the role
	 */
	@Override
	@XmlElement(name = "Rolename")
	public String getRoleName()
	{
		return roleName;
	}

	/**
	 * Function setRoleName()<br>
	 * Sets the name of the role.
	 * 
	 * @param roleName Name of the role
	 */
	public void setRoleName(String roleName)
	{
		this.roleName = roleName;
	}

	/**
	 * Function getRule()<br>
	 * Returns the rule.
	 * 
	 * @return rule Rule
	 */
	@Override
	@XmlElementRefs({ @XmlElementRef(type = RuleAbsolute.class),
			@XmlElementRef(type = RuleRelative.class),
			@XmlElementRef(type = RuleLineAbsolute.class),
			@XmlElementRef(type = RuleLineRelative.class) })
	public IRule getRule()
	{
		return rule;
	}

	/**
	 * Function setRule()<br>
	 * Sets the rule.
	 * 
	 * @param rule Rule
	 */
	@Override
	public void setRule(IRule rule)
	{
		this.rule = rule;
	}

	/**
	 * Function setRuleName()<br>
	 * Sets the rule name
	 * 
	 * @param ruleName name of the rule
	 */
	@Override
	public void setRuleName(String ruleName)
	{
		rule.setRuleName(ruleName);
	}

	/**
	 * Function setShift()<br>
	 * Sets the shift
	 * 
	 * @param shift to set
	 */
	@Override
	public void setShift(List<Double> shift)
	{
		rule.setShift(shift);
	}

	/**
	 * Function setDefPos()<br>
	 * Sets the default position
	 * 
	 * @param defPos default position to set
	 */
	@Override
	public void setDefPos(List<Double> defPos)
	{
		rule.setDefaultPosition(defPos);
	}

	/**
	 * Function setActionArea()<br>
	 * Sets the action area
	 * 
	 * @param actionArea action area to set
	 */
	@Override
	public void setActionArea(List<Double> actionArea)
	{
		rule.setActionArea(actionArea);
	}

	/**
	 * Function setReferenceObject()<br>
	 * Sets one of the reference objects of the rule from the given role
	 * 
	 * @param referenceObject reference object
	 * @param num reference object one or two
	 */
	@Override
	public void setReferenceObject(int num, IReferenceObject referenceObject)
	{
		rule.setReferenceObject(num, referenceObject);
	}

	/**
	 * Function getPositionData()<br>
	 * Returns the position-data
	 * 
	 * @param worldModel
	 * @return positionData position-data
	 */
	@Override
	public Vector3D getPositionData(IGlobalMap worldModel)
	{
		if (rule != null) {
			return rule.getPosition(worldModel);
		}

		return null;
	}

	/**
	 * Function equals()<br>
	 * Overwrites the equals function.<br>
	 * Compares the names of the given roles.
	 * 
	 * @param obj
	 */
	@Override
	public boolean equals(Object obj)
	{
		return obj != null && (obj instanceof Role)
				&& roleName.equals(((Role) obj).getRoleName());
	}

	@Override
	public int hashCode()
	{
		return roleName.hashCode() + rule.hashCode();
	}
}
