/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.positioning.model.impl;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlList;

import magma.agent.behavior.positioning.model.IReferenceObject;
import magma.agent.behavior.positioning.model.IRule;
import magma.agent.worldmodel.IGlobalMap;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Abstract Class Rule<br>
 * Is the base-class of all concrete rule classes.<br>
 * Implements all common attributes and the corresponding methods.<br>
 * Contains the abstract method getPosition() which is implemented by all
 * concrete rule classes (Strategypattern).
 * 
 * @author Denis Zimmermann
 */
public abstract class Rule implements IRule
{
	private String ruleName; // Name of the rule

	private List<Double> defaultPosition; // default position

	private List<Double> actionArea; // action-area

	private List<Double> shift; // adjustment

	/**
	 * Constructor Rule()
	 */
	Rule()
	{
		this.ruleName = "default rulename";
		this.defaultPosition = new ArrayList<Double>();
		this.actionArea = new ArrayList<Double>();
		this.shift = new ArrayList<Double>();

		initValues();
	}

	/**
	 * Constructor Rule()
	 * 
	 * @param ruleName Name of the rule
	 * @param defaultPosition default position
	 * @param actionArea action-area
	 * @param shift adjustment
	 */
	Rule(String ruleName, List<Double> defaultPosition, List<Double> actionArea,
			List<Double> shift)
	{
		this.ruleName = ruleName;
		this.defaultPosition = defaultPosition;
		this.actionArea = actionArea;
		this.shift = shift;
	}

	private void initValues()
	{
		defaultPosition.add(0.0);
		defaultPosition.add(0.0);

		actionArea.add(-1.0);
		actionArea.add(1.0);
		actionArea.add(2.0);
		actionArea.add(2.0);

		shift.add(0.0);
		shift.add(0.0);
		shift.add(0.0);
	}

	/**
	 * Function getRuleName()<br>
	 * Returns the name of the rule.
	 * 
	 * @return ruleName Name of the rule
	 */
	@Override
	@XmlElement(name = "Rulename")
	public String getRuleName()
	{
		return this.ruleName;
	}

	/**
	 * Function setRuleName()<br>
	 * Sets the name of the rule.
	 * 
	 * @param ruleName Name of the rule
	 */
	@Override
	public void setRuleName(String ruleName)
	{
		this.ruleName = ruleName;
	}

	/**
	 * Function getDefaultPosition()<br>
	 * Returns the default position.
	 * 
	 * @return defaultPosition default position
	 */
	@Override
	@XmlList
	@XmlElement(name = "Defaultposition")
	public List<Double> getDefaultPosition()
	{
		return this.defaultPosition;
	}

	/**
	 * Function setDefaultPosition()<br>
	 * Sets the default position.
	 * 
	 * @param defaultPosition default position
	 */
	@Override
	public void setDefaultPosition(List<Double> defaultPosition)
	{
		this.defaultPosition = defaultPosition;
	}

	/**
	 * Function getActionArea()<br>
	 * Returns the action-area.
	 * 
	 * @return actionArea action-area
	 */
	@Override
	@XmlList
	@XmlElement(name = "Actionarea")
	public List<Double> getActionArea()
	{
		return this.actionArea;
	}

	/**
	 * Function setActionArea()<br>
	 * Sets the action-area.
	 * 
	 * @param actionArea action-area
	 */
	@Override
	public void setActionArea(List<Double> actionArea)
	{
		this.actionArea = actionArea;
	}

	/**
	 * Function getShift()<br>
	 * Returns the adjustment.
	 * 
	 * @return shift adjustment
	 */
	@Override
	@XmlList
	@XmlElement(name = "Shift")
	public List<Double> getShift()
	{
		return this.shift;
	}

	/**
	 * Function setShift()<br>
	 * Sets the adjustment.
	 * 
	 * @param shift adjustment
	 */
	@Override
	public void setShift(List<Double> shift)
	{
		this.shift = shift;
	}

	/**
	 * Function positionInArea()<br>
	 * Checks whether the given position is within the action-area.<br>
	 * If not, the position is adjusted to the needed border value of the
	 * action-area.
	 * 
	 * @param position Calculated position
	 * @return checkedPosition Adjusted position
	 */
	@Override
	public Vector3D positionInArea(Vector3D position)
	{
		if (!this.actionArea.isEmpty()) {
			List<Double> actionArea = this.getActionArea();

			Double x = position.getX();
			Double y = position.getY();

			if (position.getX() < actionArea.get(0)) {
				x = actionArea.get(0);
			} else if (position.getX() > (actionArea.get(0) + actionArea.get(2))) {
				x = (actionArea.get(0) + actionArea.get(2));
			}

			if (position.getY() > actionArea.get(1)) {
				y = actionArea.get(1);
			} else if (position.getY() < (actionArea.get(1) - actionArea.get(3))) {
				y = (actionArea.get(1) - actionArea.get(3));
			}

			return new Vector3D(x, y, position.getZ());
		}

		return new Vector3D(position.getX(), position.getY(), position.getZ());
	}

	/**
	 * Function getPosition()<br>
	 * Returns the position-data.<br>
	 * Is overwritten by the concrete child classes.
	 * 
	 * @return position Position
	 */
	@Override
	public abstract Vector3D getPosition(IGlobalMap worldModel);

	/**
	 * Function getRuleType()<br>
	 * Returns the type of the rule.<br>
	 * Is overwritten by the concrete child classes.
	 * 
	 * @return ruleType type of the rule
	 */
	@Override
	public abstract String getRuleType();

	/**
	 * Function getReferenceObjects().<br>
	 * Returns the reference objects of the concrete rule, whether they are
	 * existent.<br>
	 * Is overwritten by the concrete child classes.
	 * 
	 * @return reference objects
	 */
	@Override
	public abstract List<IReferenceObject> getReferenceObjects();

	/**
	 * Function setReferenceObject()<br>
	 * Sets one of the reference objects of the rule
	 * 
	 * @param referenceObject reference object
	 * @param num reference object one or two
	 */
	@Override
	public abstract void setReferenceObject(int num,
			IReferenceObject referenceObject);

}
