/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.positioning.model.impl;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlRootElement;

import magma.agent.behavior.positioning.model.IReferenceObject;
import magma.agent.worldmodel.IGlobalMap;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Class RuleAbsolute<br>
 * Extends the abstract class Rule.<br>
 * Represents a rule for absolute positioning.<br>
 * Overwrites the abstract method getPosition() (Strategypattern).
 * 
 * @author Denis Zimmermann
 */
@XmlRootElement(name = "RuleAbsolute")
public class RuleAbsolute extends Rule
{
	/**
	 * Constructor RuleAbsolute()
	 */
	public RuleAbsolute()
	{
		super();
	}

	/**
	 * Constructor RuleAbsolute()
	 * 
	 * @param ruleName Name of the rule
	 * @param defaultPosition default position
	 * @param actionArea action-area
	 * @param shift adjustment
	 */
	public RuleAbsolute(String ruleName, List<Double> defaultPosition,
			List<Double> actionArea, List<Double> shift)
	{
		super(ruleName, defaultPosition, actionArea, shift);
	}

	/**
	 * Function getPosition()<br>
	 * Overwrites the abstract method getPosition() (Strategypattern).<br>
	 * Returns the position-data for absolute positioning.
	 * 
	 * @return position Position
	 */
	@Override
	public Vector3D getPosition(IGlobalMap worldModel)
	{
		Vector3D position;

		if (!getShift().isEmpty()) {
			position = new Vector3D(this.getShift().get(0),
					this.getShift().get(1), this.getShift().get(2));
		} else {
			position = new Vector3D(getDefaultPosition().get(0),
					getDefaultPosition().get(1), 0.0);
		}
		return this.positionInArea(position);
	}

	/**
	 * Function getRuleType()<br>
	 * Returns the type of the rule.<br>
	 * Is overwritten by the concrete child classes.
	 * 
	 * @return ruleType type of the rule
	 */
	@Override
	public String getRuleType()
	{
		return "Absolute";
	}

	/**
	 * Function getReferenceObjects().<br>
	 * Returns the reference objects of the concrete rule, whether they are
	 * existent.<br>
	 * 
	 * @return reference objects
	 */
	@Override
	public List<IReferenceObject> getReferenceObjects()
	{
		return new ArrayList<IReferenceObject>();
	}

	/**
	 * Function setReferenceObject()<br>
	 * Sets one of the reference objects of the rule
	 * 
	 * @param referenceObject reference object
	 * @param num reference object one or two
	 */
	@Override
	public void setReferenceObject(int num, IReferenceObject referenceObject)
	{
	}
}
