/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.positioning.model.impl;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlElementRefs;
import javax.xml.bind.annotation.XmlRootElement;

import magma.agent.behavior.positioning.model.IReferenceObject;
import magma.agent.worldmodel.IGlobalMap;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Class RuleLineRelative<br>
 * Extends the abstract class Rule.<br>
 * Represents a rule for relative positioning on a line between two reference
 * objects.<br>
 * Overwrites the abstract method getPosition() (Strategypattern).
 * 
 * @author Denis Zimmermann
 */
@XmlRootElement(name = "RuleLineRelative")
public class RuleLineRelative extends Rule
{
	private List<IReferenceObject> referenceObjects; // reference objects

	/**
	 * Constructor RuleLineRelative()<br>
	 * No-Argument-Constructor for testing.
	 */
	public RuleLineRelative()
	{
		super();
		this.referenceObjects = new ArrayList<IReferenceObject>();
	}

	/**
	 * Constructor RuleLineRelative()
	 * 
	 * @param refObjects reference objects
	 */
	public RuleLineRelative(List<IReferenceObject> refObjects)
	{
		super();
		this.referenceObjects = refObjects;
	}

	/**
	 * Constructor RuleLineRelative()
	 * 
	 * @param ruleName Name of the rule
	 * @param defaultPosition Default position
	 * @param actionArea Action-area
	 * @param shift Adjustment
	 * @param refObjects reference objects
	 */
	public RuleLineRelative(String ruleName, List<Double> defaultPosition,
			List<Double> actionArea, List<Double> shift,
			List<IReferenceObject> refObjects)
	{
		super(ruleName, defaultPosition, actionArea, shift);
		this.referenceObjects = refObjects;
	}

	/**
	 * Function getReferenceObjects()<br>
	 * Returns the reference objects.
	 * 
	 * @return referenceObjects reference objects
	 */
	@Override
	@XmlElementRefs({ @XmlElementRef(type = ReferenceObjectBall.class),
			@XmlElementRef(type = ReferenceObjectLandmark.class),
			@XmlElementRef(type = ReferenceObjectPlayer.class) })
	public List<IReferenceObject> getReferenceObjects()
	{
		return referenceObjects;
	}

	/**
	 * Function setReferenceObj()<br>
	 * Sets the reference objects.
	 * 
	 * @param referenceObjects reference objects
	 */
	public void setReferenceObjects(List<IReferenceObject> referenceObjects)
	{
		this.referenceObjects = referenceObjects;
	}

	/**
	 * Function getPosition()<br>
	 * Overwrites the abstract method getPosition() (Strategypattern).<br>
	 * Returns the position-data for relative positioning on a line between two
	 * reference objects.
	 * 
	 * @param worldModel
	 * @return position Position
	 */
	@Override
	public Vector3D getPosition(IGlobalMap worldModel)
	{
		Vector3D position;

		if (referenceObjects.size() == 2 && !getShift().isEmpty()) {
			Vector3D refCoordOne = this.referenceObjects.get(0).getCoordinates(
					worldModel);
			Vector3D refCoordTwo = this.referenceObjects.get(1).getCoordinates(
					worldModel);

			Double width = (refCoordOne.getX() - refCoordTwo.getX());
			Double height = (refCoordOne.getY() - refCoordTwo.getY());

			Double x = refCoordOne.getX() - (width * this.getShift().get(0));
			Double y = refCoordOne.getY() - (height * this.getShift().get(0));

			position = new Vector3D(x, y, this.getShift().get(1));
		} else {
			position = new Vector3D(getDefaultPosition().get(0),
					getDefaultPosition().get(1), 0.0);
		}

		return this.positionInArea(position);
	}

	/**
	 * Function getRuleType()<br>
	 * Returns the type of the rule.<br>
	 * Is overwritten by the concrete child classes.
	 * 
	 * @return ruleType type of the rule
	 */
	@Override
	public String getRuleType()
	{
		return "LineRelative";
	}

	/**
	 * Function setReferenceObject()<br>
	 * Sets one of the reference objects of the rule
	 * 
	 * @param referenceObject reference object
	 * @param num reference object one or two
	 */
	@Override
	public void setReferenceObject(int num, IReferenceObject referenceObject)
	{
		if (referenceObjects.isEmpty()) {
			referenceObjects.add(referenceObject);
		} else {
			if (num == 1) {
				referenceObjects.set(0, referenceObject);
			} else {
				if (referenceObjects.size() == 2) {
					referenceObjects.set(1, referenceObject);
				} else {
					referenceObjects.add(referenceObject);
				}
			}
		}
	}
}
