/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.positioning.model.impl;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlElementRefs;
import javax.xml.bind.annotation.XmlRootElement;

import magma.agent.behavior.positioning.model.IReferenceObject;
import magma.agent.worldmodel.IGlobalMap;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Class RuleRelative<br>
 * Extends the abstract class Rule.<br>
 * Represents a rule for relative positioning.<br>
 * Overwrites the abstract method getPosition() (Strategypattern).
 * 
 * @author Denis Zimmermann
 */
@XmlRootElement(name = "RuleRelative")
public class RuleRelative extends Rule
{
	private IReferenceObject referenceObj; // Reference object

	/**
	 * Constructor RuleRelative()<br>
	 * No-Argument-Constructor for testing.
	 */
	public RuleRelative()
	{
	}

	/**
	 * Constructor RuleRelative()
	 * 
	 * @param refObj reference object
	 */
	public RuleRelative(IReferenceObject refObj)
	{
		super();
		this.referenceObj = refObj;
	}

	/**
	 * Constructor RuleRelative()
	 * 
	 * @param ruleName Name of the rule
	 * @param defaultPosition default position
	 * @param actionArea action-area
	 * @param shift Adjustment
	 * @param refObj Reference object
	 */
	public RuleRelative(String ruleName, List<Double> defaultPosition,
			List<Double> actionArea, List<Double> shift, IReferenceObject refObj)
	{
		super(ruleName, defaultPosition, actionArea, shift);
		this.referenceObj = refObj;
	}

	/**
	 * Function getReferenceObj()<br>
	 * Returns the name of the reference object.
	 * 
	 * @return referenceObj Reference object
	 */
	@XmlElementRefs({ @XmlElementRef(type = ReferenceObjectBall.class),
			@XmlElementRef(type = ReferenceObjectLandmark.class),
			@XmlElementRef(type = ReferenceObjectPlayer.class) })
	public IReferenceObject getReferenceObj()
	{
		return referenceObj;
	}

	/**
	 * Function setReferenceObject()<br>
	 * Sets one of the reference objects of the rule
	 * 
	 * @param referenceObject reference object
	 * @param num reference object one or two
	 */
	@Override
	public void setReferenceObject(int num, IReferenceObject referenceObject)
	{
		this.referenceObj = referenceObject;
	}

	/**
	 * Function getPosition()<br>
	 * Overwrites the abstract method getPosition() (Strategypattern).<br>
	 * Returns the position-data for relative positioning.
	 * 
	 * @param worldModel
	 * @return position Position
	 */
	@Override
	public Vector3D getPosition(IGlobalMap worldModel)
	{
		Vector3D position;

		if (referenceObj != null && !getShift().isEmpty()) {
			Vector3D refCoord = this.referenceObj.getCoordinates(worldModel);
			position = new Vector3D((refCoord.getX() + this.getShift().get(0)),
					(refCoord.getY() + this.getShift().get(1)), this.getShift().get(
							2));
		} else {
			position = new Vector3D(getDefaultPosition().get(0),
					getDefaultPosition().get(1), 0.0);
		}

		return this.positionInArea(position);
	}

	/**
	 * Function getRuleType()<br>
	 * Returns the type of the rule.<br>
	 * Is overwritten by the concrete child classes.
	 * 
	 * @return ruleType type of the rule
	 */
	@Override
	public String getRuleType()
	{
		return "Relative";
	}

	/**
	 * Function getReferenceObjects().<br>
	 * Returns the reference objects of the concrete rule, whether they are
	 * existent.<br>
	 * Is overwritten by the concrete child classes.
	 * 
	 * @return reference objects
	 */
	@Override
	public List<IReferenceObject> getReferenceObjects()
	{
		List<IReferenceObject> refObj = new ArrayList<IReferenceObject>();
		if (referenceObj != null)
			refObj.add(referenceObj);

		return refObj;
	}
}
