/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.positioning.model.impl;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlElementRefs;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

import magma.agent.behavior.positioning.model.IReferenceObject;
import magma.agent.behavior.positioning.model.IRole;
import magma.agent.behavior.positioning.model.IRule;
import magma.agent.behavior.positioning.model.ISituation;
import magma.agent.worldmodel.IGlobalMap;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Class Situation<br>
 * Represents a situation with its corresponding roles.
 * 
 * @author Denis Zimmermann
 */
@XmlRootElement(name = "Situation")
@XmlType(propOrder = { "situationName", "roles" })
public class Situation implements ISituation
{
	private String situationName; // Name of the situation

	private List<IRole> roles; // List of roles

	/**
	 * Constructor Situation()
	 */
	public Situation()
	{
		this.situationName = "default situationname";
		this.roles = new ArrayList<IRole>();
	}

	/**
	 * Constructor Situation()
	 * 
	 * @param situationName Name of the situation
	 */
	public Situation(String situationName)
	{
		this.situationName = situationName;
		this.roles = new ArrayList<IRole>();
	}

	/**
	 * Constructor Situation()
	 * 
	 * @param situationName Name of the situation
	 * @param roles List of roles
	 */
	public Situation(String situationName, List<IRole> roles)
	{
		this.situationName = situationName;
		this.roles = roles;
	}

	/**
	 * Function getSituationName()<br>
	 * Returns the name of the situation.
	 * 
	 * @return situationName Name of the situation
	 */
	@Override
	@XmlElement(name = "Situationname")
	public String getSituationName()
	{
		return situationName;
	}

	/**
	 * Function setSituationName()<br>
	 * Sets the name of the situation.
	 * 
	 * @param situationName Name of the situation
	 */
	public void setSituationName(String situationName)
	{
		this.situationName = situationName;
	}

	/**
	 * Function getRoles()<br>
	 * Returns the list of roles.
	 * 
	 * @return roles List of roles
	 */
	@Override
	@XmlElementRefs({ @XmlElementRef(type = Role.class) })
	public List<IRole> getRoles()
	{
		return roles;
	}

	/**
	 * Function setRoles()<br>
	 * Sets the list of roles.
	 * 
	 * @param roles List of roles
	 */
	public void setRoles(List<IRole> roles)
	{
		this.roles = roles;
	}

	/**
	 * Function addRole()<br>
	 * Adds a role to the list of roles.
	 * 
	 * @param role New role
	 * @return success or not
	 */
	@Override
	public boolean addRole(IRole role)
	{
		if (!this.roles.contains(role)) {
			this.roles.add(role);
			return true;
		}

		return false;
	}

	/**
	 * Function removeRole()<br>
	 * Removes the given role from the list of roles.
	 * 
	 * @param role Role to remove
	 * @return success or not
	 */
	@Override
	public boolean removeRole(IRole role)
	{
		if (this.roles.contains(role)) {
			this.roles.remove(role);
			return true;
		}

		return false;
	}

	/**
	 * Function setRuleName()<br>
	 * Sets the rule name of the rule from the given role
	 * 
	 * @param roleName Name of the role
	 * @param ruleName name to set
	 */
	@Override
	public void setRuleName(String ruleName, String roleName)
	{
		getRoleByName(roleName).setRuleName(ruleName);
	}

	/**
	 * Function setRuleType()<br>
	 * Sets the rule type of the rule from the given role
	 * 
	 * @param roleName Name of the role
	 * @param rule rule type
	 */
	@Override
	public void setRuleType(IRule rule, String roleName)
	{
		getRoleByName(roleName).setRule(rule);
	}

	/**
	 * Function setShift()<br>
	 * Sets the shift of the rule from the given role
	 * 
	 * @param roleName Name of the role
	 * @param shift shift to set
	 */
	@Override
	public void setShift(List<Double> shift, String roleName)
	{
		getRoleByName(roleName).setShift(shift);
	}

	/**
	 * Function setDefPos()<br>
	 * Sets the default position of the rule from the given role
	 * 
	 * @param roleName Name of the role
	 * @param defPos default position to set
	 */
	@Override
	public void setDefPos(List<Double> defPos, String roleName)
	{
		getRoleByName(roleName).setDefPos(defPos);
	}

	/**
	 * Function setActionArea()<br>
	 * Sets the action area of the rule from the given role
	 * 
	 * @param roleName Name of the role
	 * @param actionArea action area to set
	 */
	@Override
	public void setActionArea(List<Double> actionArea, String roleName)
	{
		getRoleByName(roleName).setActionArea(actionArea);
	}

	/**
	 * Function setReferenceObject()<br>
	 * Sets one of the reference objects of the rule from the given role
	 * 
	 * @param roleName Name of the role
	 * @param referenceObject reference object
	 * @param num reference object one or two
	 */
	@Override
	public void setReferenceObject(int num, IReferenceObject referenceObject,
			String roleName)
	{
		getRoleByName(roleName).setReferenceObject(num, referenceObject);
	}

	/**
	 * Function getRoleByName()<br>
	 * Returns the role with the given name.
	 * 
	 * @param roleName Name of the role
	 * @return situation role with the given name
	 */
	public IRole getRoleByName(String roleName)
	{

		for (IRole role : this.roles) {
			if (role.getRoleName().equals(roleName)) {
				return role;
			}
		}

		return null;
	}

	/**
	 * Function getPositionData()<br>
	 * Returns the position-data.
	 * 
	 * @param worldModel
	 * @param roleName Name of the role
	 * @return positionData position-data
	 */
	@Override
	public Vector3D getPositionData(String roleName, IGlobalMap worldModel)
	{
		IRole role = getRoleByName(roleName);

		if (role != null) {
			return role.getPositionData(worldModel);
		}

		return roles.get(roles.size() - 1).getPositionData(worldModel);
	}

	/**
	 * Function equals()<br>
	 * Overwrites the equals function.<br>
	 * Compares the names of the given situations.
	 * 
	 * @param obj
	 */
	@Override
	public boolean equals(Object obj)
	{
		return obj != null && (obj instanceof Situation)
				&& this.situationName.equals(((Situation) obj).getSituationName());
	}

	@Override
	public int hashCode()
	{
		return situationName.hashCode() + roles.hashCode();
	}
}
