/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.positioning.model.impl;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlElementRefs;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

import magma.agent.behavior.positioning.model.IReferenceObject;
import magma.agent.behavior.positioning.model.IRole;
import magma.agent.behavior.positioning.model.IRule;
import magma.agent.behavior.positioning.model.ISituation;
import magma.agent.behavior.positioning.model.IStrategyDeveloper;
import magma.agent.behavior.positioning.model.IStrategyPosition;
import magma.agent.worldmodel.IGlobalMap;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Class Strategy<br>
 * Represents a strategy with the including roles.<br>
 * Implements the interface IStrategyPosition.
 * 
 * @author Denis Zimmermann
 */
@XmlRootElement(name = "Strategy")
@XmlType(propOrder = { "strategyName", "situations" })
public class Strategy implements IStrategyPosition, IStrategyDeveloper
{
	private String strategyName; // Name of the strategy

	private List<ISituation> situations; // list of situations

	/**
	 * Constructor Strategy()
	 */
	public Strategy()
	{
		this.strategyName = "default strategyname";
		this.situations = new ArrayList<ISituation>();
	}

	/**
	 * Constructor Strategy()
	 * 
	 * @param strategyName Name of the strategy
	 * @param situations List of situations
	 */
	public Strategy(String strategyName, List<ISituation> situations)
	{
		this.strategyName = strategyName;
		this.situations = situations;
	}

	/**
	 * Function getStrategyName()<br>
	 * Returns the name of the strategy.
	 * 
	 * @return strategyName Name of the strategy
	 */
	@Override
	@XmlElement(name = "Strategyname")
	public String getStrategyName()
	{
		return strategyName;
	}

	/**
	 * Function setStrategyName()<br>
	 * Sets the name of the strategy.
	 * 
	 * @param strategyName Name of the strategy
	 */
	@Override
	public void setStrategyName(String strategyName)
	{
		this.strategyName = strategyName;
	}

	/**
	 * Function getSituations()<br>
	 * Returns the list of situations.
	 * 
	 * @return situations List of situations
	 */
	@Override
	@XmlElementRefs({ @XmlElementRef(type = Situation.class) })
	public List<ISituation> getSituations()
	{
		return situations;
	}

	/**
	 * Function setSituations()<br>
	 * Sets the list of situations.
	 * 
	 * @param situations List of situations
	 */
	public void setSituations(List<ISituation> situations)
	{
		this.situations = situations;
	}

	/**
	 * Function addSituation()<br>
	 * Adds a new situation to the list of situations.
	 * 
	 * @param situation New situation
	 * @return success or not
	 */
	@Override
	public boolean addSituation(ISituation situation)
	{
		if (!this.situations.contains(situation)) {
			this.situations.add(situation);
			return true;
		}

		return false;
	}

	/**
	 * Function removeSituation()<br>
	 * Removes the given situation from the list of situations.
	 * 
	 * @param situation situation to remove
	 * @return success or not
	 */
	@Override
	public boolean removeSituation(ISituation situation)
	{
		if (this.situations.contains(situation)) {
			this.situations.remove(situation);
			return true;
		}

		return false;
	}

	/**
	 * Function addRole()<br>
	 * Adds a new role to the given situation.
	 * 
	 * @param situation situation
	 * @param role role to add
	 * @return success or not
	 */
	@Override
	public boolean addRole(String situationName, IRole role)
	{
		return getSituationByName(situationName).addRole(role);
	}

	/**
	 * Function removeRole()<br>
	 * Removes the given role from the given situation.
	 * 
	 * @param situation situation
	 * @param role role to remove
	 * @return success or not
	 */
	@Override
	public boolean removeRole(String situationName, IRole role)
	{
		return getSituationByName(situationName).removeRole(role);
	}

	/**
	 * Function setRuleName()<br>
	 * Sets the name of the rule from the given role an situation
	 * 
	 * @param situationName situation name
	 * @param roleName role name
	 * @param ruleName rule name to set
	 */
	@Override
	public void setRuleName(String ruleName, String situationName,
			String roleName)
	{
		getSituationByName(situationName).setRuleName(ruleName, roleName);
	}

	/**
	 * Function setRuleType()<br>
	 * Sets the type of the rule from the given role an situation
	 * 
	 * @param situationName situation name
	 * @param roleName role name
	 * @param rule rule type to set
	 */
	@Override
	public void setRuleType(IRule rule, String situationName, String roleName)
	{
		getSituationByName(situationName).setRuleType(rule, roleName);
	}

	/**
	 * Function setShift()<br>
	 * Sets the shift of the rule from the given role an situation
	 * 
	 * @param situationName situation name
	 * @param roleName role name
	 * @param shift shift to set
	 */
	@Override
	public void setShift(List<Double> shift, String situationName,
			String roleName)
	{
		getSituationByName(situationName).setShift(shift, roleName);
	}

	/**
	 * Function setDefPos()<br>
	 * Sets the default position of the rule from the given role an situation
	 * 
	 * @param situationName situation name
	 * @param roleName role name
	 * @param defPos default position to set
	 */
	@Override
	public void setDefPos(List<Double> defPos, String situationName,
			String roleName)
	{
		getSituationByName(situationName).setDefPos(defPos, roleName);
	}

	/**
	 * Function setActionArea()<br>
	 * Sets the action area of the rule from the given role an situation
	 * 
	 * @param situationName situation name
	 * @param roleName role name
	 * @param actionArea action area to set
	 */
	@Override
	public void setActionArea(List<Double> actionArea, String situationName,
			String roleName)
	{
		getSituationByName(situationName).setActionArea(actionArea, roleName);
	}

	/**
	 * Function setReferenceObject()<br>
	 * Sets one of the reference objects of the rule from the given role an
	 * situation
	 * 
	 * @param situationName situation name
	 * @param roleName role name
	 * @param num reference object one or two
	 * @param referenceObject reference object
	 */
	@Override
	public void setReferenceObject(int num, IReferenceObject referenceObject,
			String situationName, String roleName)
	{
		getSituationByName(situationName).setReferenceObject(num,
				referenceObject, roleName);
	}

	/**
	 * Function getSituationByName()<br>
	 * Returns the situation with the given name.
	 * 
	 * @param situationName Name of the situation
	 * @return situation Situation with the given name
	 */
	public ISituation getSituationByName(String situationName)
	{

		for (ISituation situation : this.situations) {
			if (situation.getSituationName().equals(situationName)) {
				return situation;
			}
		}

		return null;
	}

	/**
	 * Function getPositionData()<br>
	 * Returns the position-data dependent on the given role- and situation name.
	 * 
	 * @param roleName Name of the role
	 * @param situationName Name of the situation
	 * @param worldModel
	 * @return positionData position-data dependent on the given role- and
	 *         situation name
	 */
	@Override
	public Vector3D getPositionData(String roleName, String situationName,
			IGlobalMap worldModel)
	{
		ISituation situation = getSituationByName(situationName);

		if (situation != null) {
			return situation.getPositionData(roleName, worldModel);
		}

		return situations.get(0).getPositionData(roleName, worldModel);
	}
}
