/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.positioning.xml.impl;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;

import magma.agent.behavior.positioning.model.impl.Strategy;
import magma.agent.behavior.positioning.xml.IXMLWrite;

/**
 * Class XMLFactory<br>
 * Implements the interface IXMLWrite and because of that the methods readXML()
 * and createXML().<br>
 * 
 * @author Denis Zimmermann
 */
public class XMLFactory implements IXMLWrite
{
	private Marshaller mar; // Marshaller to create a XML-File

	private Unmarshaller unMar; // Unmarshaller to read a XML-File

	/**
	 * Constructor XMLFactory<br>
	 * Creates JAXBContext, Marshaller and Unmarshaller.
	 */
	public XMLFactory()
	{
		try {
			JAXBContext context = JAXBContext.newInstance(Strategy.class);
			this.mar = context.createMarshaller();
			this.unMar = context.createUnmarshaller();

			mar.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
		} catch (JAXBException e) {
			e.printStackTrace();
		}

	}

	/**
	 * Function createXML()<br>
	 * Creates a strategy in form of a XML-file out of the given objects, which
	 * represent a strategy.
	 * 
	 * @param strategie objects, which represent a strategy
	 * @param dest destination for the XML-file
	 */
	@Override
	public boolean createXML(Strategy strategie, String dest)
	{
		Writer w;

		try {
			w = new FileWriter(dest);
			mar.marshal(strategie, w);
			w.close();
			return true;
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		} catch (JAXBException e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * Function readXML()<br>
	 * Reads a strategy in form of a XML-file and creates the appropriate
	 * objects.
	 * 
	 * @param dest location of the XML-file
	 * @return Strategy appropriate objects
	 */
	@Override
	public Strategy readXML(String dest)
	{
		Strategy strategy = null;

		try {
			strategy = (Strategy) unMar.unmarshal(new FileReader(dest));
		} catch (FileNotFoundException e) {
			// e.printStackTrace();
		} catch (JAXBException e) {
			// e.printStackTrace();
		}

		return strategy;
	}

	/**
	 * Function deleteXML()<br>
	 * Deletes a XML-file which has the given destination.
	 * 
	 * @param dest destination of the XML-file
	 * @return success whether the deletion was successful or not
	 */
	@Override
	public boolean deleteXML(String dest)
	{
		File file = new File(dest);

		if (file.exists()) {
			file.delete();
			return true;
		}
		return false;
	}
}
