/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.behavior.supportPoint;

import java.io.IOException;
import java.util.Map;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.motor.FunctionBehaviorParameter;
import magma.agent.worldmodel.IGlobalMap;
import magma.util.function.Function;

/**
 * Behavior representing motor commands based on a mathematical function
 * 
 * @author Klaus Dorer
 */
public class FunctionBehavior extends BasicMotorBehavior
{
	/**
	 * Default constructor for motor behaviors
	 * @param name identifier for this behavior
	 * @param model reference to the world model information
	 * @param agentModel reference to the agent's own information
	 * @param filename the name of a file containing the parameters for this
	 *        behavior
	 * @param slowDown factor of the speed with which the behavior is performed.
	 *        2 means half the speed.
	 * @param startSlowDown the speed to approach the first value
	 */
	public FunctionBehavior(String name, IGlobalMap model,
			IAgentModel agentModel, String filename, float slowDown,
			float startSlowDown)
	{
		super(name, model, agentModel, filename, slowDown, startSlowDown);

		FunctionBehaviorParameter parameters;
		try {
			parameters = new FunctionBehaviorParameter(filename);
			extractMotorManagers(parameters);
			period = parameters.getPeriod();

		} catch (IOException e) {
			e.printStackTrace();
			System.exit(-1);
		}
	}

	/**
	 * Default constructor for motor behaviors
	 * @param name identifier for this behavior
	 * @param coordinator handle for action sending to server
	 * @param model reference to the world model information
	 * @param agentModel reference to the agent's own information
	 * @param params parameters to control the behavior
	 * @param slowDown factor of the speed with which the behavior is performed.
	 *        2 means half the speed.
	 * @param startSlowDown the speed to approach the first value
	 */
	public FunctionBehavior(String name, IGlobalMap model,
			IAgentModel agentModel, FunctionBehaviorParameter params,
			float slowDown, float startSlowDown)
	{
		super(name, model, agentModel, BasicMotorBehavior.UNNAMED_FILE_TXT,
				slowDown, startSlowDown);
		extractMotorManagers(params);
		period = params.getPeriod();
	}

	/**
	 * Copy constructor
	 * 
	 * @param source Source object to copy state from
	 */
	private FunctionBehavior(FunctionBehavior source)
	{
		super(source);
	}

	/**
	 * Extract motor managers from behavior parameters
	 * 
	 * @param parameters Function behavior parameters
	 */
	private void extractMotorManagers(FunctionBehaviorParameter parameters)
	{
		Map<String, Function> joints = parameters.getJoints();

		for (String jointName : joints.keySet()) {
			MotorManager manager = new MotorManager(agentModel, jointName,
					joints.get(jointName));
			motorManagers.put(jointName, manager);
		}
	}

	@Override
	public BasicMotorBehavior copy()
	{
		return new FunctionBehavior(this);
	}
}
