/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.belief.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import magma.agent.IMagmaConstants;
import magma.agent.agentmodel.IAgentModel;
import magma.agent.belief.IBeliefConstants;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Geometry;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Checks if any teammate is closer to the ball
 * 
 * @author Klaus Dorer
 */
public class TeammateCloserToBall extends Belief
{

	/**
	 * Constructor
	 * 
	 * @param agent Reference to the agent model object
	 * @param worldModel Reference to the world model object
	 */
	public TeammateCloserToBall(IAgentModel agent, IGlobalMap worldModel)
	{
		super(IBeliefConstants.TEAMMATE_CLOSER_TO_BALL, agent, worldModel);
	}

	@Override
	public float getTruthValue()
	{
		IThisPlayer thisPlayer = worldModel.getThisPlayer();
		IPlayer playerAtBall = getClosestPlayer();
		if (playerAtBall == thisPlayer) {
			return 0.0f;
		}
		return 1.0f;
	}

	IPlayer getClosestPlayer()
	{
		List<IPlayer> playersAtBall = worldModel.getTeammatesAtBall();
		List<IPlayer> newSort = new ArrayList<IPlayer>(playersAtBall);
		// we add ourselves to have the same calculation for all
		newSort.add(worldModel.getThisPlayer());

		Collections.sort(newSort, new DistanceToBallComparator());
		return newSort.get(0);
	}

	/**
	 * For player at ball calculation
	 * @author Klaus Dorer
	 */
	private class DistanceToBallComparator implements Comparator<IPlayer>
	{
		@Override
		public int compare(IPlayer player1, IPlayer player2)
		{
			IMoveableObject ball = worldModel.getBall();
			Vector3D ballPosition = ball.getPosition();
			Vector3D otherGoalPos = worldModel.getOtherGoalPosition();
			double distance1 = player1.getDistanceToXY(ballPosition);
			double distance2 = player2.getDistanceToXY(ballPosition);

			// check if opponent player is between player and ball
			IPlayer opponentAtBall = worldModel.getOpponentAtBall();
			if (opponentAtBall != null) {
				distance1 = checkBetweenBallPenalty(player1, opponentAtBall,
						ballPosition, distance1);
				distance2 = checkBetweenBallPenalty(player2, opponentAtBall,
						ballPosition, distance2);

			}

			distance1 = checkLayingPenalty(player1, ballPosition, distance1);
			distance2 = checkLayingPenalty(player2, ballPosition, distance2);

			distance1 = checkBackwardRunning(player1, ball, otherGoalPos,
					distance1);
			distance2 = checkBackwardRunning(player2, ball, otherGoalPos,
					distance2);

			distance1 = checkMyselfBonus(player1, distance1);

			if (distance1 < distance2) {
				return -1;
			} else if (distance1 > distance2) {
				return 1;
			} else {
				return 0;
			}
		}

		/**
		 * @param player1
		 * @param ballPosition
		 * @param otherGoalPos
		 * @return
		 */
		private double checkBackwardRunning(IPlayer player, IMoveableObject ball,
				Vector3D otherGoalPos, double distance)
		{
			final double DISTANCE_PENALTY_BEHINDBALL = 2.0;
			double distance1 = player.getDistanceToXY(otherGoalPos);
			double distance2 = ball.getDistanceToXY(otherGoalPos);
			double delta = distance2 - distance1;
			if (distance1 < distance2) {
				return distance + Geometry.getLinearFuzzyValue(0, 1, true, delta)
						* DISTANCE_PENALTY_BEHINDBALL;
			}
			return distance;
		}

		/**
		 * @param player1
		 * @param ballPosition
		 * @param distance1
		 * @return
		 */
		private double checkLayingPenalty(IPlayer player, Vector3D ballPosition,
				double distance)
		{
			if (player.isLaying()) {
				distance += IMagmaConstants.DISTANCE_PENALTY_LYING;
			}
			return distance;
		}

		private double checkMyselfBonus(IPlayer player, double distance)
		{
			final double MYSELF_BONUS = 0.1;

			if (player.getID() == worldModel.getThisPlayer().getID()) {
				return distance - MYSELF_BONUS;
			}
			return distance;
		}

		private double checkBetweenBallPenalty(IPlayer player1,
				IPlayer opponentAtBall, Vector3D ballPosition, double distance1)
		{
			final double DISTANCE_PENALTY_BEHIND_OPPONENT = 2.0;

			double xOpponent = opponentAtBall.getPosition().getX();
			double xPlayer = player1.getPosition().getX();
			double xBall = ballPosition.getX();

			if (xBall < xOpponent && xOpponent < xPlayer) {
				distance1 += DISTANCE_PENALTY_BEHIND_OPPONENT;
			}
			return distance1;
		}
	}

}
