/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.decision.simple;

import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.belief.IBelief;
import magma.agent.belief.IBeliefConstants;
import magma.agent.decision.IDecisionMaker;

/**
 * Base class for all decision makers
 * @author Klaus Dorer
 */
public abstract class DecisionMakerBase implements IDecisionMaker
{
	/** handle for logging */
	protected static final Logger logger = Logger.getLogger("");

	/** the behavior that is currently selected */
	protected IBehavior currentBehavior;

	/** the believes that are available */
	protected final Map<String, IBelief> believes;

	/** the behaviors that are available */
	protected final Map<String, IBehavior> behaviors;

	/** the number of decisions taken */
	protected int numberOfDecisions;

	/**
	 * Constructor
	 * 
	 * @param believes List of believes
	 * @param behaviors List of applicable behaviors
	 */
	public DecisionMakerBase(Map<String, IBelief> believes,
			Map<String, IBehavior> behaviors)
	{
		this.believes = believes;
		this.behaviors = behaviors;
		numberOfDecisions = 0;
		if (behaviors != null) {
			currentBehavior = behaviors.get(IBehaviorConstants.NONE);
		}
	}

	/**
	 * Retrieve a specific belief
	 * 
	 * @param name Belief name
	 * @return The specified believe, null if no matching belief was found
	 */
	@Override
	public IBelief getBelieve(String name)
	{
		return believes.get(name);
	}

	/**
	 * Retrieve a list of all beliefs
	 * 
	 * @return Belief list
	 */
	public Map<String, IBelief> getBelieves()
	{
		return believes;
	}

	/**
	 * Retrieve a specific behavior
	 * 
	 * @param name Behavior name
	 * @return The specified behavior, null if no matching behavior was found
	 */
	@Override
	public IBehavior getBehavior(String name)
	{
		return behaviors.get(name);
	}

	/**
	 * Retrieve a list of all behaviors
	 * 
	 * @return Behavior list
	 */
	public Map<String, IBehavior> getBehaviors()
	{
		return behaviors;
	}

	/**
	 * Retrieve the number of times decide() was called
	 * 
	 * @return Call count
	 */
	public int getNumberOfDecisions()
	{
		return numberOfDecisions;
	}

	@Override
	public boolean decide()
	{
		numberOfDecisions++;

		IBehavior toExecute;

		if (numberOfDecisions > 10) {
			toExecute = decideSoccer();

		} else {
			toExecute = behaviors.get(IBehaviorConstants.NONE);
		}

		if (toExecute != currentBehavior) {

			if (lyingOnGroundAndNotTryingToGetUp()) {
				currentBehavior.abort();
				currentBehavior = toExecute;
			} else {
				String oldName = currentBehavior.toString();
				currentBehavior = currentBehavior.switchTo(toExecute);
				logger.log(
						Level.INFO,
						"old: {0} wanted: {1} new: {2}",
						new Object[] { oldName, toExecute.toString(),
								currentBehavior.toString() });
			}
		}
		logger.log(Level.INFO, "going to perform: {0}",
				new Object[] { currentBehavior });
		currentBehavior.perform();
		logger.log(Level.INFO, "performed: {0}", new Object[] { currentBehavior });
		return true;
	}

	/**
	 * Decide which behavior to use next
	 * 
	 * @return Next behavior
	 */
	public IBehavior decideSoccer()
	{
		return null;
	}

	/**
	 * Check if the agent is lying on the ground and not trying to get up
	 * 
	 * @return <code>true</code> if the condition is fullfilled
	 */
	private boolean lyingOnGroundAndNotTryingToGetUp()
	{
		boolean lying = believes.get(IBeliefConstants.LAY_ON_BACK)
				.getTruthValue() > 0.5
				|| believes.get(IBeliefConstants.LAY_ON_FRONT).getTruthValue() > 0.5
				|| believes.get(IBeliefConstants.LAY_ON_SIDE).getTruthValue() > 0.5;

		boolean tryToGetUp = currentBehavior == behaviors
				.get(IBehaviorConstants.GET_UP_BACK)
				|| currentBehavior == behaviors
						.get(IBehaviorConstants.GET_UP_FRONT)
				|| currentBehavior == behaviors.get(IBehaviorConstants.KEEP)
				|| currentBehavior == behaviors
						.get(IBehaviorConstants.MOVE_ARM_TO_FALL_BACK);

		return lying && !tryToGetUp;
	}
}
