/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.decision.simple;

import java.util.Map;

import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.belief.IBelief;
import magma.agent.belief.IBeliefConstants;

/**
 * Simple decision maker for primitive field player
 * 
 * @author Klaus Dorer
 */
public class SoccerDecisionMaker extends DecisionMakerBase
{
	/** The version of the soccer server used */
	protected int serverVersion;

	/**
	 * Default constructor
	 * @param believes map of believes we can decide on
	 * @param behaviors map of behaviors we can decide for
	 * @param serverVersion the version of the soccer server used
	 */
	public SoccerDecisionMaker(Map<String, IBelief> believes,
			Map<String, IBehavior> behaviors, int serverVersion)
	{
		super(believes, behaviors);
		this.serverVersion = serverVersion;
	}

	@Override
	public IBehavior decideSoccer()
	{
		IBehavior focusBehavior = behaviors.get("FocusBall");
		focusBehavior.perform();

		IBehavior toExecute;
		IBehavior result = super.decideSoccer();
		if (result != null) {
			return result;
		}

		// dance or cry
		result = winOrLoose();
		if (result != null) {
			return result;
		}

		// beam
		result = beamHome();
		if (result != null) {
			return result;
		}

		// wait till game starts
		result = waitForGameStart();
		if (result != null) {
			return result;
		}

		result = getUp();
		if (result != null) {
			return result;
		}

		// get ready
		result = getReady();
		if (result != null) {
			return result;
		}

		// wait till opponent has restarted the game
		result = waitForOpponentActions();
		if (result != null) {
			return result;
		}

		// kick
		result = doKickBehavior();
		if (result != null) {
			return result;
		}

		// walk
		toExecute = movement();
		if (toExecute != null) {
			return toExecute;
		}

		// stop any running motors and do nothing

		return behaviors.get(IBehaviorConstants.NONE);
	}

	/**
	 * Called to decide if we should beam us to our home position
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior beamHome()
	{
		IBelief beamTime = believes.get(IBeliefConstants.BEAM_TIME);
		if (beamTime.getTruthValue() > 0.5) {
			return behaviors.get(IBehaviorConstants.BEAM_HOME);
		}
		return null;
	}

	protected IBehavior winOrLoose()
	{
		IBelief gameRunning = believes.get(IBeliefConstants.GAME_RUNNING);
		if (gameRunning.getTruthValue() < 0.5) {
			if (believes.get(IBeliefConstants.WIN).getTruthValue() > 0.5)
				return behaviors.get(IBehaviorConstants.WIN);
			if (believes.get(IBeliefConstants.LOOSE).getTruthValue() > 0.5)
				return behaviors.get(IBehaviorConstants.LOOSE);

		}

		return null;
	}

	/**
	 * Called to decide if we should wait for game start or not. Default
	 * implementation waits.
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior waitForGameStart()
	{
		IBelief gameRunning = believes.get(IBeliefConstants.GAME_RUNNING);
		if (gameRunning.getTruthValue() < 0.5) {
			return behaviors.get(IBehaviorConstants.NONE);
		}

		return null;
	}

	/**
	 * Called to decide if we should wait for opponent actions or not. Default
	 * implementation waits only on OpponentKickOff.
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior waitForOpponentActions()
	{
		// IBelief opponentCornerKick =
		// believes.get(IBelief.OPPONENT_CORNER_KICK);
		// if (opponentCornerKick.getTruthValue() > 0.5) {
		// return behaviors.get(IBehavior.NONE);
		// }
		//
		// IBelief opponentFreeKick = believes.get(IBelief.OPPONENT_FREE_KICK);
		// if (opponentFreeKick.getTruthValue() > 0.5) {
		// return behaviors.get(IBehavior.NONE);
		// }
		//
		// IBelief opponentGoalKick = believes.get(IBelief.OPPONENT_GOAL_KICK);
		// if (opponentGoalKick.getTruthValue() > 0.5) {
		// return behaviors.get(IBehavior.NONE);
		// }
		//
		// IBelief opponentKickIn = believes.get(IBelief.OPPONENT_KICK_IN);
		// if (opponentKickIn.getTruthValue() > 0.5) {
		// return behaviors.get(IBehavior.NONE);
		// }

		IBelief opponentKickOff = believes
				.get(IBeliefConstants.OPPONENT_KICK_OFF);
		if (opponentKickOff.getTruthValue() > 0.5) {
			return behaviors.get(IBehaviorConstants.NONE);
		}

		return null;
	}

	/**
	 * Called to decide if we should stand up from laying on ground
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior getUp()
	{
		// get up back
		IBelief onBack = believes.get(IBeliefConstants.LAY_ON_BACK);
		if (onBack.getTruthValue() > 0.5) {
			return behaviors.get(IBehaviorConstants.GET_UP_BACK);
		}

		// get up front
		IBelief onFront = believes.get(IBeliefConstants.LAY_ON_FRONT);
		if (onFront.getTruthValue() > 0.5) {
			return behaviors.get(IBehaviorConstants.GET_UP_FRONT);
		}

		// get up side
		IBelief onSide = believes.get(IBeliefConstants.LAY_ON_SIDE);
		if (onSide.getTruthValue() > 0.5) {
			return behaviors.get(IBehaviorConstants.MOVE_ARM_TO_FALL_BACK);
		}

		return null;
	}

	/**
	 * Called to decide if we should get into ready position
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior getReady()
	{
		IBelief soccerPosition = believes.get(IBeliefConstants.SOCCER_POSITION);
		if (soccerPosition.getTruthValue() < 0.5) {
			return behaviors.get(IBehaviorConstants.GET_READY);
		}
		return null;
	}

	/**
	 * Called to decide if we should kick the ball
	 * @return the behavior to execute, null if this part of decision does not
	 *         trigger a behavior currently
	 */
	protected IBehavior doKickBehavior()
	{
		// IBelief ballKickable = believes.get(IBeliefConstants.BALL_KICKABLE);
		// if (ballKickable.getTruthValue() > 0.5) {
		//
		// IBelief goalDirection = believes
		// .get(IBeliefConstants.FACING_GOAL_KICK_DIRECTION);
		//
		// if (goalDirection.getTruthValue() > 0.5) {
		//
		// // decide if we have time to kick far but takes longer time
		// IBelief opponentNearBall = believes
		// .get(IBeliefConstants.OPPONENT_NEAR_BALL);
		// if (opponentNearBall.getTruthValue() > 0.5) {
		// return behaviors.get(IBehaviorConstants.FAST_SHOOT);
		// }
		//
		// return behaviors.get(IBehaviorConstants.SHOOT_TO_GOAL);
		// }
		// }
		return null;
	}

	/**
	 * Called to decide if movement is necessary
	 * @return walk behavior if no team mate is closer to ball
	 */
	protected IBehavior movement()
	{
		IBelief teammateBetter = believes
				.get(IBeliefConstants.TEAMMATE_CLOSER_TO_BALL);
		if (teammateBetter.getTruthValue() < 0.5) {
			// we want to go to the ball
			IBelief opponentNearBall = believes
					.get(IBeliefConstants.OPPONENT_NEAR_BALL);
			if (opponentNearBall.getTruthValue() > 0.5) {
				// return behaviors.get(IBehaviorConstants.DRIBBLE_BALL);
				return behaviors.get(IBehaviorConstants.ATTACK);
			}

			// if close to the goal, we dribble into the goal
			IBelief scorable = believes
					.get(IBeliefConstants.BALL_IN_SCORABLE_AREA);
			if (scorable.getTruthValue() > 0.9) {
				// return behaviors.get(IBehaviorConstants.DRIBBLE_BALL);
				return behaviors.get(IBehaviorConstants.ATTACK);
			}

			// run to the ball
			// return behaviors.get(IBehaviorConstants.GET_IN_SCORE_POSITION);
			return behaviors.get(IBehaviorConstants.ATTACK);
		}

		// TODO Feature strategy part activation
		// return behaviors.get(IBehaviorConstants.PASSIVE_POSITIONING);

		return behaviors.get(IBehaviorConstants.MOVE_WITH_BALL);
	}

	/**
	 * @return the serverVersion
	 */
	public int getServerVersion()
	{
		return serverVersion;
	}

}
