/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.general;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.Map;

import magma.agent.action.IAction;
import magma.agent.action.impl.Action;
import magma.agent.agentmodel.IAgentModel;
import magma.agent.agentmodel.impl.AgentModel;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.basic.BeamBehavior;
import magma.agent.behavior.basic.BeamHomeBehavior;
import magma.agent.behavior.basic.BehaviorParameter;
import magma.agent.behavior.basic.NoneBehavior;
import magma.agent.behavior.basic.SayPositionsBehavior;
import magma.agent.behavior.motor.MotorFileBehavior;
import magma.agent.behavior.motor.ReadCSV;
import magma.agent.behavior.supportPoint.FunctionBehavior;
import magma.agent.belief.IBelief;
import magma.agent.belief.IBeliefConstants;
import magma.agent.belief.impl.BallEndsUpInOwnGoal;
import magma.agent.belief.impl.BallInCriticalArea;
import magma.agent.belief.impl.BallKickable;
import magma.agent.belief.impl.BeamTime;
import magma.agent.belief.impl.FacingAdvantageousKickDirection;
import magma.agent.belief.impl.FacingGoalKickDirection;
import magma.agent.belief.impl.FacingKickDirection;
import magma.agent.belief.impl.GameRunning;
import magma.agent.belief.impl.GameStateBelief;
import magma.agent.belief.impl.LayOnBack;
import magma.agent.belief.impl.LayOnFront;
import magma.agent.belief.impl.LayOnSide;
import magma.agent.belief.impl.Lose;
import magma.agent.belief.impl.OpponentNearBall;
import magma.agent.belief.impl.PlayDefense;
import magma.agent.belief.impl.PlayOffense;
import magma.agent.belief.impl.TeammateCloserToBall;
import magma.agent.belief.impl.Win;
import magma.agent.connection.IServerConnection;
import magma.agent.connection.impl.ServerConnection;
import magma.agent.decision.IDecisionMaker;
import magma.agent.decision.simple.DecisionMakerBase;
import magma.agent.decision.simple.DoNothingDecisionMaker;
import magma.agent.decision.simple.GoalieDecisionMaker;
import magma.agent.decision.simple.PenaltyDecisionMaker;
import magma.agent.decision.simple.SimpleDecisionMaker;
import magma.agent.decision.simple.SoccerPlayerDecisionMaker;
import magma.agent.decision.simple.TrainingDecisionMaker;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.meta.server.IRCServerMetaModel;
import magma.agent.meta.server.impl.RCServerMetaModelV62;
import magma.agent.meta.server.impl.RCServerMetaModelV63;
import magma.agent.meta.server.impl.RCServerMetaModelV64;
import magma.agent.perception.IPerception;
import magma.agent.perception.impl.Perception;
import magma.agent.worldmodel.GameState;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.impl.GlobalMap;
import magma.agent.worldmodel.localizer.ILocalizer;
import magma.agent.worldmodel.localizer.impl.CompositeLocalizer;
import magma.agent.worldmodel.localizer.impl.LocalizerFieldNormal;
import magma.agent.worldmodel.localizer.impl.LocalizerGyro;

/**
 * Factory class for the components of our RoboCup player
 * 
 * @author Klaus Dorer
 */
public abstract class ComponentFactory
{
	/**
	 * Create a ServerConnection
	 * 
	 * @param host Host name/IP
	 * @param port Server port
	 * @return New Server connection
	 */
	public IServerConnection createConnection(String host, int port)
	{
		return new ServerConnection(host, port);
	}

	/**
	 * Create a agent meta model
	 * 
	 * @return the agent meta model
	 */
	public abstract IAgentMetaModel createAgentMetaModel();

	/**
	 * Create a new rc-server meta model based on serverVersion.
	 * 
	 * @param serverVerion - the version of the meta model
	 * @return the rc-server meta model
	 */
	public IRCServerMetaModel createRCServerMetaModel(int serverVerion)
	{
		if (serverVerion < 62)
			serverVerion = 62;

		switch (serverVerion) {
		case 62:
			return new RCServerMetaModelV62();
		case 63:
			return new RCServerMetaModelV63();
		case 64:
		default:
			return new RCServerMetaModelV64();

		}
	}

	/**
	 * Create a Perception object
	 * 
	 * @return New Perception
	 */
	public IPerception createPerception(IAgentMetaModel metaModel)
	{
		return new Perception(metaModel);
	}

	/**
	 * Create an new Action
	 * 
	 * @param connection Server connection
	 * @param scene Scene name
	 * @param teamname Team name
	 * @param playerNumber Player number
	 * @return New Action object
	 */
	public IAction createAction(IServerConnection connection,
			IAgentMetaModel metaModel, String teamname, int playerNumber)
	{
		return new Action(connection, metaModel, teamname, playerNumber);
	}

	/**
	 * Create a new AgentModel
	 * 
	 * @return New AgentModel object
	 */
	public IAgentModel createAgentModel(IAction action, IAgentMetaModel metaModel)
	{
		return new AgentModel(action, metaModel);
	}

	/**
	 * Create a new GlobalMap
	 * 
	 * @param agentModel Reference to the AgentModel
	 * @param teamname Team name
	 * @param playerNumber Player number
	 * @param serverVersion Currently used server version
	 * @return New GlobalMap object
	 */
	public IGlobalMap createWorldModel(IAgentModel agentModel,
			IRCServerMetaModel metaModel, String teamname, int playerNumber)
	{
		return new GlobalMap(agentModel, createLocalizer(), metaModel, teamname,
				playerNumber);
	}

	/**
	 * Create a new Localizer
	 * 
	 * @return New Localizer
	 */
	ILocalizer createLocalizer()
	{
		ILocalizer[] localizer = new ILocalizer[2];

		// Note: the order of localizer is important! I should start with the
		// localizer needing the highest number of visible reference points and
		// end with the localizer, which needs the least amount of visible
		// reference points.
		localizer[0] = new LocalizerFieldNormal();
		localizer[1] = new LocalizerGyro();

		return new CompositeLocalizer(localizer);
	}

	/**
	 * Create a new DecisionMaker
	 * 
	 * @param believes Agent believes
	 * @param behaviors Agent behaviors
	 * @param playerNumber Player number
	 * @param serverVersion Server version
	 * @param decisionMakerID 0 (default) if we decide based on player number, 1
	 *        for goalie, 2 for field player, 3 for simple decision maker, 4 for
	 *        Penalty decision maker, 5 for do nothing decision maker
	 * @return new DecisionMaker object
	 */
	public IDecisionMaker createDecisionMaker(Map<String, IBelief> believes,
			Map<String, IBehavior> behaviors, int playerNumber, int serverVersion,
			int decisionMakerID)
	{
		DecisionMakerBase result;

		switch (decisionMakerID) {
		case 1:
			result = new GoalieDecisionMaker(believes, behaviors, serverVersion);
			break;

		case 2:
			result = new SoccerPlayerDecisionMaker(believes, behaviors,
					serverVersion);
			break;

		case 3:
			result = new SimpleDecisionMaker(believes, behaviors);
			break;

		case 4:
			result = new PenaltyDecisionMaker(believes, behaviors, serverVersion);
			break;

		case 5:
			result = new DoNothingDecisionMaker(believes, behaviors);
			break;

		case 6:
			result = new TrainingDecisionMaker(believes, behaviors, serverVersion);
			break;

		default:
			if (playerNumber == 1) {
				result = new GoalieDecisionMaker(believes, behaviors, serverVersion);
			} else {
				result = new SoccerPlayerDecisionMaker(believes, behaviors,
						serverVersion);
			}
		}

		return result;

	}

	/**
	 * Creates all believe objects
	 * 
	 * @param agentModel Agent model
	 * @param worldModel World model
	 * @return A map with all believes that may be used
	 */
	public Map<String, IBelief> createBeliefs(IAgentModel agentModel,
			IGlobalMap worldModel)
	{
		Map<String, IBelief> beliefs = new HashMap<String, IBelief>();

		// General beliefs
		IBelief belief = new GameRunning(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new BeamTime(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new Win(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new Lose(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new PlayOffense(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new PlayDefense(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		// Player beliefs
		belief = new TeammateCloserToBall(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new OpponentNearBall(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		// Kick beliefs
		belief = new BallKickable(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new FacingKickDirection(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new FacingAdvantageousKickDirection(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new FacingGoalKickDirection(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		// Laying beliefs
		belief = new LayOnBack(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new LayOnFront(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new LayOnSide(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		// GameState beliefs
		belief = new GameStateBelief(IBeliefConstants.OWN_CORNER_KICK,
				agentModel, worldModel, GameState.OWN_CORNER_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OWN_FREE_KICK, agentModel,
				worldModel, GameState.OWN_FREE_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OWN_GOAL_KICK, agentModel,
				worldModel, GameState.OWN_GOAL_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OWN_KICK_IN, agentModel,
				worldModel, GameState.OWN_KICK_IN);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OWN_KICK_OFF, agentModel,
				worldModel, GameState.OWN_KICK_OFF);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OPPONENT_CORNER_KICK,
				agentModel, worldModel, GameState.OPPONENT_CORNER_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OPPONENT_FREE_KICK,
				agentModel, worldModel, GameState.OPPONENT_FREE_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OPPONENT_GOAL_KICK,
				agentModel, worldModel, GameState.OPPONENT_GOAL_KICK);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OPPONENT_KICK_IN,
				agentModel, worldModel, GameState.OPPONENT_KICK_IN);
		beliefs.put(belief.getName(), belief);

		belief = new GameStateBelief(IBeliefConstants.OPPONENT_KICK_OFF,
				agentModel, worldModel, GameState.OPPONENT_KICK_OFF);
		beliefs.put(belief.getName(), belief);

		// Goalie beliefs
		belief = new BallInCriticalArea(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new BallEndsUpInOwnGoal(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		// Robot specific beliefs
		createSpecificBeliefs(agentModel, worldModel, beliefs);

		return beliefs;
	}

	/**
	 * Creates all specific belief objects
	 * 
	 * @param agentModel - the agent model
	 * @param worldModel - the world model
	 * @param beliefs - a map for storing the created specific beliefs
	 */
	protected abstract void createSpecificBeliefs(IAgentModel agentModel,
			IGlobalMap worldModel, Map<String, IBelief> beliefs);

	/**
	 * Create all behavior objects that are used during the game
	 * 
	 * @param agentModel Reference to the AgentModel
	 * @param worldModel Reference to the WorldModel
	 * @param coordinator Reference to the Coordinator
	 * @param params Behavior parameterization, null if default should be used
	 * @return A map of available behaviors
	 */
	public Map<String, IBehavior> createBehaviors(IAgentModel agentModel,
			IGlobalMap worldModel, BehaviorParameter params)
	{
		Map<String, IBehavior> behaviors = new HashMap<String, IBehavior>();

		IBehavior behavior = new NoneBehavior();
		behaviors.put(behavior.getName(), behavior);

		behavior = new BeamHomeBehavior(IBehaviorConstants.BEAM_HOME, worldModel,
				agentModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new BeamBehavior(IBehaviorConstants.BEAM_TO_POSITION,
				worldModel, agentModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new SayPositionsBehavior(
				IBehaviorConstants.SAY_POSITIONS_BEHAVIOR, worldModel, agentModel);
		behaviors.put(behavior.getName(), behavior);

		readFileBehaviors(getBehaviorFilesBasePath()
				+ "BehaviorValues/UsedBehaviors/", false, agentModel, worldModel,
				params, behaviors);

		readFileBehaviors(getBehaviorFilesBasePath()
				+ "FunctionBehaviors/UsedBehaviors/", true, agentModel, worldModel,
				params, behaviors);

		createSpecificBehaviors(agentModel, worldModel, params, behaviors);

		return behaviors;
	}

	/**
	 * Return the path to the behavior files base folder. This method is used to
	 * read dynamically all used behaviors of the given path.<br>
	 * The following folder structure is expected:<br>
	 * <i>&lt;path&gt; -> BehaviorValues -> UsedBehaviors</i><br>
	 * for CSV behavior files and:<br>
	 * <i>&lt;path&gt; -> FunctionBehaviors -> UsedBehaviors</i><br>
	 * for function behavior files.
	 * 
	 * @return the path to the behavior files base folder
	 */
	protected abstract String getBehaviorFilesBasePath();

	/**
	 * Create all specific behavior objects.
	 * 
	 * @param agentModel - the agent model
	 * @param worldModel - the world model
	 * @param params - behavior parameterization, null if default should be used
	 * @param behaviors - a map for storing the created specific behaviors
	 */
	protected abstract void createSpecificBehaviors(IAgentModel agentModel,
			IGlobalMap worldModel, BehaviorParameter params,
			Map<String, IBehavior> behaviors);

	/**
	 * Reads either CSV- or function behavior files, creates corresponding
	 * objects and adds them to the map of behaviors.
	 * 
	 * @param agentModel - the agent model
	 * @param worldModel - the world model
	 * @param params - behavior parameterization, null if default should be used
	 * @param behaviors - a map for storing the created behaviors
	 * @throws IOException
	 */
	protected void readFileBehaviors(String pathname, boolean functionBehavior,
			IAgentModel agentModel, IGlobalMap worldModel,
			BehaviorParameter params, Map<String, IBehavior> behaviors)
	{
		try {
			String[] files = ReadCSV.getResourceListing(pathname);

			for (String fileName : files) {
				// to avoid .svn file
				if (fileName.endsWith(".txt")) {
					String filePath = pathname + fileName;
					ReadCSV csvHeaderReader = new ReadCSV(filePath);

					if (functionBehavior) {
						String name = csvHeaderReader.getFunctionBehaviorName();
						behaviors.put(name,
								new FunctionBehavior(name, worldModel, agentModel,
										filePath, csvHeaderReader.getFloatSlowDown(),
										csvHeaderReader.getFloatStartSlowDown()));
					} else {
						String name = csvHeaderReader.getBehaviorValuesName();
						behaviors.put(
								name,
								new MotorFileBehavior(name, worldModel, agentModel,
										params, filePath, csvHeaderReader
												.getIntSlowDown(), csvHeaderReader
												.getIntStartSlowDown()));
					}
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
			System.out.println("Rule file not found" + pathname);
		} catch (URISyntaxException e) {
			e.printStackTrace();
			System.out.println("URI exception when reading: " + pathname);
		}
	}
}
