/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.meta.server.impl;

import java.util.HashMap;
import java.util.Map;

import javax.vecmath.Vector2d;
import javax.vecmath.Vector3d;

public class RCServerMetaModelV64 extends RCServerMetaModel
{
	private static final float BALL_DECAY = 0.94f;

	private static final float FIELD_HALF_LENGTH = 10.5f;

	private static final float FIELD_HALF_WIDTH = 7.0f;

	private static final float GOAL_HALF_WIDTH = 1.05f;

	private static final float GOAL_HEIGHT = 0.8f;

	private static final float PENALTY_HALF_LENGTH = 1.95f;

	private static final float PENALTY_WIDTH = 1.8f;

	private static final float MIDDLE_CIRCLE_RADIUS = 1.8f;

	/** Make sure to keep this and LANDMARK_POSITIONS in same order */
	private static final String[] LANDMARK_NAMES = { "G1L", "G2L", "G1R", "G2R",
			"F1L", "F2L", "F1R", "F2R" };

	/**
	 * The positions of the visible landmarks in order of LANDMARK_NAMES as used
	 * in version 0.6.3 of the server
	 */
	private static final double[][] LANDMARK_POSITIONS = {
			{ -FIELD_HALF_LENGTH, GOAL_HALF_WIDTH, GOAL_HEIGHT },
			{ -FIELD_HALF_LENGTH, -GOAL_HALF_WIDTH, GOAL_HEIGHT }, // left goal
			// posts
			{ FIELD_HALF_LENGTH, GOAL_HALF_WIDTH, GOAL_HEIGHT },
			{ FIELD_HALF_LENGTH, -GOAL_HALF_WIDTH, GOAL_HEIGHT }, // right goal
			// posts
			{ -FIELD_HALF_LENGTH, FIELD_HALF_WIDTH, 0.0 },
			{ -FIELD_HALF_LENGTH, -FIELD_HALF_WIDTH, 0.0 },
			{ FIELD_HALF_LENGTH, FIELD_HALF_WIDTH, 0.0 },
			{ FIELD_HALF_LENGTH, -FIELD_HALF_WIDTH, 0.0 } };

	public RCServerMetaModelV64()
	{
		super(64, BALL_DECAY, new Vector2d(FIELD_HALF_LENGTH * 2,
				FIELD_HALF_WIDTH * 2), new Vector2d(GOAL_HALF_WIDTH * 2,
				GOAL_HEIGHT), new Vector2d(), new Vector2d(PENALTY_HALF_LENGTH * 2,
				PENALTY_WIDTH), MIDDLE_CIRCLE_RADIUS, createLandmarkMap(),
				createFieldLinesMap());
	}

	private static Map<String, Vector3d> createLandmarkMap()
	{
		Map<String, Vector3d> landmarks = new HashMap<String, Vector3d>();

		for (int i = 0; i < LANDMARK_NAMES.length; i++) {
			landmarks.put(LANDMARK_NAMES[i], new Vector3d(
					LANDMARK_POSITIONS[i][0], LANDMARK_POSITIONS[i][1],
					LANDMARK_POSITIONS[i][2]));
		}

		return landmarks;
	}

	private static Map<String, Vector2d[]> createFieldLinesMap()
	{
		Map<String, Vector2d[]> fieldLines = new HashMap<String, Vector2d[]>();

		// Ground lines
		fieldLines.put("UGL", new Vector2d[] { // Upper-ground-line
				new Vector2d(-FIELD_HALF_LENGTH, FIELD_HALF_WIDTH), //
						new Vector2d(FIELD_HALF_LENGTH, FIELD_HALF_WIDTH) });

		fieldLines.put("LGL", new Vector2d[] { // Lower-ground-line
				new Vector2d(-FIELD_HALF_LENGTH, -FIELD_HALF_WIDTH), //
						new Vector2d(FIELD_HALF_LENGTH, -FIELD_HALF_WIDTH) });

		// Side lines
		fieldLines.put("LSL", new Vector2d[] { // Left-Side-Line
				new Vector2d(-FIELD_HALF_LENGTH, FIELD_HALF_WIDTH), //
						new Vector2d(-FIELD_HALF_LENGTH, -FIELD_HALF_WIDTH) });

		fieldLines.put("RSL", new Vector2d[] { // Right-side-line
				new Vector2d(FIELD_HALF_LENGTH, FIELD_HALF_WIDTH), //
						new Vector2d(FIELD_HALF_LENGTH, -FIELD_HALF_WIDTH) });

		// Middle line
		fieldLines.put("ML", new Vector2d[] { // Middle-Line
				new Vector2d(0, FIELD_HALF_WIDTH), //
						new Vector2d(0, -FIELD_HALF_WIDTH) });

		// Left penalty area lines
		fieldLines.put("LPAUL", new Vector2d[] { // Left-penalty-area-upper-line
						new Vector2d(-FIELD_HALF_LENGTH, PENALTY_HALF_LENGTH), //
						new Vector2d(-FIELD_HALF_LENGTH + PENALTY_WIDTH,
								PENALTY_HALF_LENGTH) });

		fieldLines.put("LPALL", new Vector2d[] { // Left-penalty-area-lower-line
						new Vector2d(-FIELD_HALF_LENGTH, -PENALTY_HALF_LENGTH), //
						new Vector2d(-FIELD_HALF_LENGTH + PENALTY_WIDTH,
								-PENALTY_HALF_LENGTH) });

		fieldLines.put("LPAFL", new Vector2d[] { // Left-penalty-area-front-line
						new Vector2d(-FIELD_HALF_LENGTH + PENALTY_WIDTH,
								PENALTY_HALF_LENGTH), //
						new Vector2d(-FIELD_HALF_LENGTH + PENALTY_WIDTH,
								-PENALTY_HALF_LENGTH) });

		// Right penalty area lines
		fieldLines.put("RPAUL", new Vector2d[] { // Right-penalty-area-upper-line
						new Vector2d(FIELD_HALF_LENGTH, PENALTY_HALF_LENGTH), //
						new Vector2d(FIELD_HALF_LENGTH - PENALTY_WIDTH,
								PENALTY_HALF_LENGTH) });

		fieldLines.put("RPALL", new Vector2d[] { // Right-penalty-area-lower-line
						new Vector2d(FIELD_HALF_LENGTH, -PENALTY_HALF_LENGTH), //
						new Vector2d(FIELD_HALF_LENGTH - PENALTY_WIDTH,
								-PENALTY_HALF_LENGTH) });

		fieldLines.put("RPAFL", new Vector2d[] { // Right-penalty-area-front-line
						new Vector2d(FIELD_HALF_LENGTH - PENALTY_WIDTH,
								PENALTY_HALF_LENGTH), //
						new Vector2d(FIELD_HALF_LENGTH - PENALTY_WIDTH,
								-PENALTY_HALF_LENGTH) });

		// Middle circle lines
		double px1 = MIDDLE_CIRCLE_RADIUS;
		double py1 = 0;
		double px2, py2;

		for (int deg = 36; deg <= 360; deg += 36) {
			px2 = Math.cos(Math.toRadians(deg)) * MIDDLE_CIRCLE_RADIUS;
			py2 = Math.sin(Math.toRadians(deg)) * MIDDLE_CIRCLE_RADIUS;

			fieldLines.put("MC_" + (deg - 36) + "-" + deg, new Vector2d[] { //
					new Vector2d(px1, py1), new Vector2d(px2, py2) });

			px1 = px2;
			py1 = py2;
		}

		return fieldLines;
	}
}
