/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.perception.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.IMagmaConstants;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.meta.agent.IBodyPartConfiguration;
import magma.agent.meta.agent.IHingeJointConfiguration;
import magma.agent.meta.agent.IJointConfiguration;
import magma.agent.meta.agent.ISensorConfiguration;
import magma.agent.meta.agent.IUniversalJointConfiguration;
import magma.agent.perception.IAccelerometerPerceptor;
import magma.agent.perception.IAgentStatePerceptor;
import magma.agent.perception.IForceResistancePerceptor;
import magma.agent.perception.IGameStatePerceptor;
import magma.agent.perception.IGyroPerceptor;
import magma.agent.perception.IHearPerceptor;
import magma.agent.perception.IHingeJointPerceptor;
import magma.agent.perception.ILinePerceptor;
import magma.agent.perception.IPerception;
import magma.agent.perception.IPlayerPos;
import magma.agent.perception.ITimerPerceptor;
import magma.agent.perception.IUniversalJointPerceptor;
import magma.agent.perception.IVisibleObjectPerceptor;

/**
 * Represents all data the agent is able to perceive from its environment.
 * Should be updated in every simulation cycle.
 * 
 * @author Klaus Dorer, Simon Raffeiner, Stefan Glaser
 */
public class Perception implements IPerception
{
	/** handle for logging */
	private static final Logger logger = Logger.getLogger("");

	/** If true info is printed to console */
	public static final boolean perceptorInfo = false;

	// Parsing layer
	private final ServerMessageParser parser;

	// Map for all named perceptors
	private final Map<String, Perceptor> perceptors;

	// Map for all accelerometer perceptors
	private final Map<String, AccelerometerPerceptor> accelperceptors;

	// Map for all gyro perceptors
	private final Map<String, GyroPerceptor> gyroperceptors;

	// Map for all force resistance perceptors
	private final Map<String, ForceResistancePerceptor> forceperceptors;

	// Map for all visible objects
	private Map<String, IVisibleObjectPerceptor> visibleObjects;

	// List of visible players
	private List<IPlayerPos> visiblePlayers;

	// List of visible lines
	private List<ILinePerceptor> visibleLines;

	// Time information from the server
	private final TimePerceptor time;

	// Actual Game State
	private final GameStatePerceptor gameState;

	// Actual Agent State
	private final AgentStatePerceptor agentState;

	// Hearing
	private final HearPerceptor hear;

	// Flag - if the current perception contains vision information
	private boolean containsVision;

	/**
	 * initializes all perceptors that appear in the simulation without the
	 * visions of the other players
	 */
	public Perception(IAgentMetaModel metaModel)
	{
		parser = new ServerMessageParser();

		perceptors = new HashMap<String, Perceptor>(40);
		accelperceptors = new HashMap<String, AccelerometerPerceptor>(1);
		gyroperceptors = new HashMap<String, GyroPerceptor>(1);
		forceperceptors = new HashMap<String, ForceResistancePerceptor>(2);

		initPlayers();
		initVisions();
		initPerceptors(metaModel);

		time = new TimePerceptor();
		gameState = new GameStatePerceptor();
		agentState = new AgentStatePerceptor();
		hear = new HearPerceptor();
	}

	/**
	 * Initialize visible players
	 */
	private void initPlayers()
	{
		visiblePlayers = new ArrayList<IPlayerPos>(
				IMagmaConstants.NUMBER_OF_PLAYERS_PER_TEAM * 2);
	}

	/**
	 * Initialize known visible objects
	 */
	private void initVisions()
	{
		visibleObjects = new HashMap<String, IVisibleObjectPerceptor>(20);
		visibleLines = new ArrayList<ILinePerceptor>();
	}

	/**
	 * Initializes all agent perceptors
	 */
	private void initPerceptors(IAgentMetaModel metaModel)
	{
		for (IBodyPartConfiguration config : metaModel.getBodyPartConfiguration()) {
			IJointConfiguration jointConfig = config.getJointConfiguration();
			ISensorConfiguration sensorConfig;

			if (jointConfig != null) {
				if (jointConfig instanceof IHingeJointConfiguration) {
					// If the joint configuration is a HingeJoint-Type, add a new
					// HingeJointPerceptor to the perceptors list
					perceptors.put(jointConfig.getPerceptorName(),
							new HingeJointPerceptor(jointConfig.getPerceptorName()));
				} else if (jointConfig instanceof IUniversalJointConfiguration) {
					// If the joint configuration is a UniversalJoint-Type, add a new
					// UniversalJointPerceptor to the perceptors list
					perceptors
							.put(jointConfig.getPerceptorName(),
									new UniversalJointPerceptor(jointConfig
											.getPerceptorName()));
				} else {
					// further joint types
				}
			}

			sensorConfig = config.getGyroRateConfiguration();
			if (sensorConfig != null) {
				// If the body part has a gyro-rate perceptor attached, create one
				gyroperceptors.put(sensorConfig.getPerceptorName(),
						new GyroPerceptor(sensorConfig.getPerceptorName()));
			}

			sensorConfig = config.getAccelerometerConfiguration();
			if (sensorConfig != null) {
				// If the body part has a accelerometer attached, create one
				accelperceptors.put(sensorConfig.getPerceptorName(),
						new AccelerometerPerceptor(sensorConfig.getPerceptorName()));
			}

			sensorConfig = config.getForceResistanceConfiguration();
			if (sensorConfig != null) {
				// If the body part has a force-resistance perceptor attached,
				// create one
				forceperceptors
						.put(sensorConfig.getPerceptorName(),
								new ForceResistancePerceptor(sensorConfig
										.getPerceptorName()));
			}
		}
	}

	/**
	 * Called when the server message is parsed
	 */
	@Override
	public void update(String message)
	{
		try {
			parse(message);
		} catch (PerceptorConversionException e) {
			// TODO: we should be able to recognize. At least we should explicitly
			// specify which exceptions we ignore on purpose
			logger.log(Level.INFO, "not able to Parse: {0}", e);
		} catch (RuntimeException e) {
			logger.log(Level.SEVERE, "Error parsing message: {0}", message);
			logger.log(Level.SEVERE, "Error parsing message: {0}", e);
			assert false : "Parsing problem";
		}

	}

	protected void parse(String message) throws PerceptorConversionException
	{
		if (message == null) {
			// nothing to do, might happen at disconnection
			return;
		}

		// Clear Force Resistance Perceptors
		forceperceptors.clear();

		// Clear list of visible objects
		containsVision = false;
		visiblePlayers.clear();
		visibleObjects.clear();
		visibleLines.clear();

		hear.init();

		List<Perceptor> rawPerceptors = parser.parseString(message);

		// Process
		for (Perceptor perceptor : rawPerceptors)
			processInputPerceptor(perceptor);
	}

	private void processInputPerceptor(Perceptor perceptor)
	{
		Perceptor target = null;

		// Handle sensor perceptors
		if (perceptor instanceof HingeJointPerceptor
				|| perceptor instanceof UniversalJointPerceptor
				|| perceptor instanceof TouchPerceptor)
			target = perceptors.get(perceptor.getName());

		if (perceptor instanceof ForceResistancePerceptor)
			forceperceptors.put(perceptor.getName(),
					(ForceResistancePerceptor) perceptor);

		if (perceptor instanceof VisionPerceptor) {
			containsVision = true;
			processVisibleObjects((VisionPerceptor) perceptor);
			return;
		}

		if (perceptor instanceof AccelerometerPerceptor)
			target = accelperceptors.get(perceptor.getName());

		if (perceptor instanceof GyroPerceptor)
			target = gyroperceptors.get(perceptor.getName());

		if (perceptor instanceof GameStatePerceptor) {
			gameState.assign(perceptor);
			return;
		}

		if (perceptor instanceof AgentStatePerceptor) {
			agentState.assign(perceptor);
			return;
		}

		if (perceptor instanceof HearPerceptor) {
			hear.assign(perceptor);
			return;
		}

		if (perceptor instanceof TimePerceptor) {
			time.assign(perceptor);
			return;
		}

		if (target != null) {
			target.assign(perceptor);
		} else if (Perception.perceptorInfo)
			System.out.println("Got unknown perceptor \"" + perceptor.getName()
					+ "\"");
	}

	/**
	 * Updates visible objects (Flags, Goals, Players, ...) from the vision
	 * 
	 * @param perceptor Vision perceptor
	 */
	private void processVisibleObjects(VisionPerceptor perceptor)
	{
		for (IVisibleObjectPerceptor object : perceptor.getObjects()) {
			if (object instanceof PlayerPos) {
				visiblePlayers.add((PlayerPos) object);
			} else if (object instanceof LinePerceptor) {
				visibleLines.add((LinePerceptor) object);
			} else {
				visibleObjects.put(object.getName(), object);
			}
		}
	}

	@Override
	public IUniversalJointPerceptor getUniversalJointPerceptor(String name)
	{
		return (IUniversalJointPerceptor) perceptors.get(name);
	}

	@Override
	public IHingeJointPerceptor getHingeJointPerceptor(String name)
	{
		return (IHingeJointPerceptor) perceptors.get(name);
	}

	@Override
	public IAccelerometerPerceptor getAccelerationPerceptor(String name)
	{
		return accelperceptors.get(name);
	}

	@Override
	public IAgentStatePerceptor getAgentState()
	{
		return agentState;
	}

	@Override
	public IGyroPerceptor getGyroRatePerceptor(String name)
	{
		return gyroperceptors.get(name);
	}

	@Override
	public IForceResistancePerceptor getForceResistancePerceptor(String name)
	{
		return forceperceptors.get(name);
	}

	@Override
	public IVisibleObjectPerceptor getObject(String name)
	{
		return visibleObjects.get(name);
	}

	@Override
	public ITimerPerceptor getTime()
	{
		return time;
	}

	@Override
	public IGameStatePerceptor getGameState()
	{
		return gameState;
	}

	@Override
	public IHearPerceptor getHearPerceptor()
	{
		return hear;
	}

	@Override
	public List<IPlayerPos> getVisiblePlayers()
	{
		return visiblePlayers;
	}

	@Override
	public List<ILinePerceptor> getVisibleLines()
	{
		return visibleLines;
	}

	@Override
	public boolean containsVision()
	{
		return containsVision;
	}
}