/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.perception.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.IMagmaConstants;
import magma.agent.perception.IServerMessageParser;
import magma.agent.perception.IVisibleObjectPerceptor;
import magma.util.observer.Subject;
import magma.util.symboltreeparser.IllegalSymbolInputException;
import magma.util.symboltreeparser.SymbolNode;
import magma.util.symboltreeparser.SymbolTreeParser;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Parses incoming server messages, converts them into single Perceptor messages
 * and delivers them to all observers
 * 
 * @author Simon Raffeiner
 * 
 */
public class ServerMessageParser extends Subject<List<Perceptor>> implements
		IServerMessageParser
{
	// handle for logging
	protected static final Logger logger = Logger.getLogger("");

	// Internal tree parser
	private final SymbolTreeParser treeparser;

	/**
	 * Default constructor
	 */
	public ServerMessageParser()
	{
		treeparser = new SymbolTreeParser();
	}

	@Override
	public void update(String message)
	{
		// Just parse the message and ignore all errors
		try {
			parse(message);
		} catch (Exception e) {
			logger.log(Level.SEVERE,
					"Parsing problem message: " + message + "{0}", e);
		}
	}

	/**
	 * Parse server messages, convert symbol tree nodes into Perception objects
	 * and inform the observers for every found object
	 * 
	 * @param message Input message
	 * @return List of perceptors
	 */
	public List<Perceptor> parse(String message)
	{
		if (message.length() == 0)
			return new ArrayList<Perceptor>();

		try {
			List<Perceptor> list = parseString(message);

			if (list != null) {
				onStateChange(list);
			}
			return list;

		} catch (Exception e) {
			logger.log(Level.SEVERE,
					"Could not parse message: " + message + "{0}", e);
		}
		return null;
	}

	/**
	 * Parse a string into a list of Perceptor objects
	 * 
	 * @param message Message string
	 * @return List of parsed Perceptor objects
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	public List<Perceptor> parseString(String message)
			throws PerceptorConversionException
	{
		SymbolNode root;

		// Generate the symbol tree
		try {
			root = treeparser.parse(message);
		} catch (IllegalSymbolInputException e) {
			throw new PerceptorConversionException("Parsing error: "
					+ e.getMessage());
		}

		ArrayList<Perceptor> list = new ArrayList<Perceptor>(root.children.length);

		// Parse all top-level nodes into messages
		for (int i = 0; i < root.children.length; i++) {
			Object node = root.children[i];

			/*
			 * The tree may contain top-level leaves, ATM they are not specified
			 * and therefore not parsed
			 */
			if (node instanceof SymbolNode) {
				Perceptor perceptor = parseNode((SymbolNode) node);

				if (perceptor != null) {
					list.add(perceptor);
				}
			}
		}

		return list;
	}

	/**
	 * Parse a symbol tree node into a Perceptor object (if possible)
	 * 
	 * @param node Symbol tree node
	 * @return Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private Perceptor parseNode(SymbolNode node)
			throws PerceptorConversionException
	{
		/* Check message type */
		Object child = node.children[0];

		if (child == null || !(child instanceof String))
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());

		String type = (String) child;

		/* Hinge Joint */
		if (type.equals("HJ"))
			return parseHingeJoint(node);

		/* Time perceptor */
		if (type.equals("time"))
			return parseTime(node);

		/* GameState perceptor */
		if (type.equals("GS"))
			return parseGameState(node);

		/* Gyro */
		if (type.equals("GYR"))
			return parseGyro(node);

		/* Accelerometer */
		if (type.equals("ACC"))
			return parseAccelerometer(node);

		/* Vision perceptor */
		if (type.equals("See"))
			return parseVision(node);

		/* Touch perceptor */
		if (type.equals("TCH"))
			return parseTouch(node);

		/* AgentState perceptor */
		if (type.equals("AgentState"))
			return parseAgentState(node);

		/* Hear perceptor */
		if (type.equals("hear"))
			return parseHear(node);

		/* Force resistance perceptor */
		if (type.equals("FRP"))
			return parseForceResistance(node);

		/* Universal Joint */
		if (type.equals("UJ"))
			return parseUniversalJoint(node);

		/* Unknown node? */
		logger.log(Level.FINEST, "Got an unknown node: " + node);

		return null;
	}

	/**
	 * Parse a symbol tree node into a Gyro Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Gyro Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private GyroPerceptor parseGyro(SymbolNode node)
			throws PerceptorConversionException
	{
		if (!(node.children[1] instanceof SymbolNode)
				|| !(node.children[2] instanceof SymbolNode))
			throw new PerceptorConversionException("Malformed Message: "
					+ node.toString());

		try {
			/* Check content */
			SymbolNode nameNode = (SymbolNode) node.children[1];
			SymbolNode rotationNode = (SymbolNode) node.children[2];

			if (!nameNode.children[0].equals("n"))
				throw new PerceptorConversionException("name expected: "
						+ node.toString());

			if (!rotationNode.children[0].equals("rt"))
				throw new PerceptorConversionException("rotation expected: "
						+ node.toString());

			return new GyroPerceptor((String) nameNode.children[1],
					Float.valueOf((String) rotationNode.children[1]),
					Float.valueOf((String) rotationNode.children[2]),
					Float.valueOf((String) rotationNode.children[3]));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Gyro perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Gyro perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private AccelerometerPerceptor parseAccelerometer(SymbolNode node)
			throws PerceptorConversionException
	{
		if (!(node.children[1] instanceof SymbolNode)
				|| !(node.children[2] instanceof SymbolNode))
			throw new PerceptorConversionException("Malformed Message: "
					+ node.toString());

		try {
			/* Check content */
			SymbolNode nameNode = (SymbolNode) node.children[1];
			SymbolNode accelerationNode = (SymbolNode) node.children[2];

			if (!nameNode.children[0].equals("n"))
				throw new PerceptorConversionException("name expected: "
						+ node.toString());

			if (!accelerationNode.children[0].equals("a"))
				throw new PerceptorConversionException("rotation expected: "
						+ node.toString());

			return new AccelerometerPerceptor((String) nameNode.children[1],
					Float.valueOf((String) accelerationNode.children[1]),
					Float.valueOf((String) accelerationNode.children[2]),
					Float.valueOf((String) accelerationNode.children[3]));

		} catch (NumberFormatException e) {
			// seems that the server sometimes sends NAN, in which case we ignore
			// the reading for accelerometer
			logger.log(Level.INFO,
					"Parsing accelerometer problem: " + node.toString() + "{0}", e);
			return null;

		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Hinge Joint Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Hinge Joint Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private HingeJointPerceptor parseHingeJoint(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			// if (!(node.children[1] instanceof SymbolNode)
			// || !(node.children[2] instanceof SymbolNode))
			// throw new PerceptorConversionException("Malformed Message: "
			// + node.toString());

			/* Check content */
			SymbolNode nameNode = (SymbolNode) node.children[1];
			SymbolNode rotationNode = (SymbolNode) node.children[2];

			// if (!((String) nameNode.children[0]).equals("n"))
			// throw new PerceptorConversionException("Malformed Message: "
			// + node.toString() + ": name expected");
			//
			// if (!((String) rotationNode.children[0]).equals("ax"))
			// throw new PerceptorConversionException("Malformed Message: "
			// + node.toString() + ": axis expected");

			return new HingeJointPerceptor((String) nameNode.children[1],
					Float.valueOf((String) rotationNode.children[1]));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Universal Joint Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Universal Joint Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private UniversalJointPerceptor parseUniversalJoint(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			// Sanity checks
			if (!(node.children[1] instanceof SymbolNode)
					|| !(node.children[2] instanceof SymbolNode)
					| !(node.children[3] instanceof SymbolNode))
				throw new PerceptorConversionException("Malformed Message: "
						+ node.toString());

			/* Check content */
			SymbolNode nameNode = (SymbolNode) node.children[1];
			SymbolNode axis1Node = (SymbolNode) node.children[2];
			SymbolNode axis2Node = (SymbolNode) node.children[3];

			if (!nameNode.children[0].equals("n"))
				throw new PerceptorConversionException("name expected: "
						+ node.toString());
			if (!axis1Node.children[0].equals("ax1"))
				throw new PerceptorConversionException("axis expected: "
						+ node.toString());
			if (!axis2Node.children[0].equals("ax2"))
				throw new PerceptorConversionException("axis expected: "
						+ node.toString());

			return new UniversalJointPerceptor((String) nameNode.children[1],
					Float.valueOf((String) axis1Node.children[1]),
					Float.valueOf((String) axis2Node.children[1]));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Force Resistance Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Force Resistance Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private ForceResistancePerceptor parseForceResistance(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			// Sanity checks
			if (!(node.children[1] instanceof SymbolNode)
					|| !(node.children[2] instanceof SymbolNode)
					| !(node.children[3] instanceof SymbolNode))
				throw new PerceptorConversionException("Malformed Message: "
						+ node.toString());

			/* Check content */
			SymbolNode nameNode = (SymbolNode) node.children[1];
			SymbolNode originNode = (SymbolNode) node.children[2];
			SymbolNode forceNode = (SymbolNode) node.children[3];

			if (!nameNode.children[0].equals("n"))
				throw new PerceptorConversionException("name expected: "
						+ node.toString());
			if (!originNode.children[0].equals("c"))
				throw new PerceptorConversionException("origin expected: "
						+ node.toString());
			if (!forceNode.children[0].equals("f"))
				throw new PerceptorConversionException("force expected: "
						+ node.toString());

			return new ForceResistancePerceptor((String) nameNode.children[1],
					Float.valueOf((String) originNode.children[1]),
					Float.valueOf((String) originNode.children[2]),
					Float.valueOf((String) originNode.children[3]),
					Float.valueOf((String) forceNode.children[1]),
					Float.valueOf((String) forceNode.children[2]),
					Float.valueOf((String) forceNode.children[3]));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Touch Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Touch Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private TouchPerceptor parseTouch(SymbolNode node)
			throws PerceptorConversionException
	{
		if (node.children.length != 5 || !node.children[1].equals("n")
				|| !node.children[3].equals("val"))
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());

		try {
			String name = (String) node.children[2];

			boolean value = false;
			if (node.children[4].equals("1"))
				value = true;

			return new TouchPerceptor(name, value);
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree vision node into a list of Perceptor objects
	 * 
	 * @param node Symbol tree vision node
	 * @return List of Perceptor objects
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private VisionPerceptor parseVision(SymbolNode node)
			throws PerceptorConversionException
	{
		// Sanity checks
		if (node.children.length == 0)
			throw new PerceptorConversionException("Malformed Node: "
					+ node.toString());

		try {
			VisionPerceptor vision = new VisionPerceptor("See");

			// Parse visible objects
			for (int i = 1; i < node.children.length; i++) {
				Object subitem = node.children[i];

				if (subitem instanceof SymbolNode) {
					SymbolNode subnode = (SymbolNode) subitem;

					// Player object?
					if (!(subnode.children[0] instanceof String))
						throw new PerceptorConversionException(
								"Malformed Node, empty name: " + node.toString());
					else if (subnode.children[0].equals("P"))
						vision.addObject(parsePlayer(subnode));

					else if (subnode.children[0].equals("L"))
						vision.addObject(parseLine(subnode));

					else
						vision.addObject(parseVisibleObject(subnode));
				}
			}

			return vision;
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * @param subnode
	 * @return
	 */
	private IVisibleObjectPerceptor parseLine(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			assert node.children.length == 3 : "Malformed node";
			assert node.children[1] instanceof SymbolNode : "Malformed node: "
					+ node.toString();
			assert node.children[2] instanceof SymbolNode : "Malformed node: "
					+ node.toString();

			String name = (String) node.children[0];
			Vector3D pol1 = parsePol((SymbolNode) node.children[1]);
			Vector3D pol2 = parsePol((SymbolNode) node.children[2]);

			return new LinePerceptor(name, pol1, pol2);
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Player object
	 * 
	 * @param node Symbol tree node
	 * @return Player object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private PlayerPos parsePlayer(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			/* Parse content */
			String teamName = IMagmaConstants.UNKNOWN_PLAYER_TEAMNAME;
			int id = IMagmaConstants.UNKNOWN_PLAYER_NUMBER;
			Vector3D pol = new Vector3D();
			String bodyPartIdentifier = "";
			Map<String, Vector3D> bodyPartMap = new HashMap<String, Vector3D>();

			for (int i = 1; i < node.children.length; i++) {
				if (!(node.children[i] instanceof SymbolNode)) {
					throw new PerceptorConversionException("Malformed node: "
							+ node.toString());
				}

				SymbolNode param = (SymbolNode) node.children[i];

				if (param.children[0].equals("team")) {
					teamName = (String) param.children[1];
				} else if (param.children[0].equals("id")) {
					id = Integer.parseInt((String) param.children[1]);
				} else {
					// in case of seeing parts of opponent, we have to look into them
					if (!param.children[0].equals("pol")) {
						bodyPartIdentifier = (String) param.children[0];
						param = (SymbolNode) param.children[1];
					}

					if (param.children[0].equals("pol")) {
						pol = parsePol(param);
						bodyPartMap.put(bodyPartIdentifier, pol);
					}
				}
			}

			return new PlayerPos(pol, id, teamName, bodyPartMap);
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Player object
	 * 
	 * @param node Symbol tree node
	 * @return Player object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private VisibleObjectPerceptor parseVisibleObject(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			assert node.children.length == 2 : "Malformed node";
			assert node.children[1] instanceof SymbolNode : "Malformed node: "
					+ node.toString();

			String name = (String) node.children[0];
			Vector3D pol = parsePol((SymbolNode) node.children[1]);

			return new VisibleObjectPerceptor(name, pol);
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse object coordinates
	 * 
	 * @param node SymbolNode to parse
	 * @return Coordinates
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private Vector3D parsePol(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			if (node.children.length != 4)
				throw new PerceptorConversionException("Malformed node: "
						+ node.toString());
			if (!node.children[0].equals("pol"))
				throw new PerceptorConversionException("Expecting a pol object: "
						+ node.toString());

			float val1 = Float.parseFloat((String) node.children[1]);
			float val2 = Float.parseFloat((String) node.children[2]);
			float val3 = Float.parseFloat((String) node.children[3]);

			return new Vector3D(val1, new Vector3D(Math.toRadians(val2),
					Math.toRadians(val3)));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());

		} catch (NumberFormatException e) {
			// seems that the server sometimes sends NAN, in which case we ignore
			// the reading for position
			logger.log(Level.INFO, "Parsing position problem: " + node.toString()
					+ "{0}", e);
		}
		return null;
	}

	/**
	 * Parse a symbol tree node into a Game State Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Game State Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private GameStatePerceptor parseGameState(SymbolNode node)
			throws PerceptorConversionException
	{
		float time = 0.0f;
		String playmode = "", team = "";
		int unum = 0;

		SymbolNode child = null;

		// Evaluate content
		for (int i = 1; i < node.children.length; i++) {
			try {
				child = (SymbolNode) node.children[i];

				String type = (String) child.children[0];

				// Check sub-node type
				if (type.equals("t")) {
					time = Float.parseFloat((String) child.children[1]);
				} else if (type.equals("pm")) {
					playmode = (String) child.children[1];
				} else if (type.equals("unum")) {
					unum = Integer.parseInt((String) child.children[1]);
				} else if (type.equals("team")) {
					team = (String) child.children[1];
				} else
					throw new PerceptorConversionException(
							"Malformed GameState node, unknown sub-node: "
									+ child.toString());
			} catch (Exception e) {
				if (child != null)
					throw new PerceptorConversionException(
							"Malformed GameState node, conversion error: "
									+ child.toString());

				throw new PerceptorConversionException(
						"Malformed GameState node: child was null!");
			}
		}
		return new GameStatePerceptor(time, playmode, team, unum);
	}

	/**
	 * Parse a symbol tree node into an Agent State Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Agent State Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private AgentStatePerceptor parseAgentState(SymbolNode node)
			throws PerceptorConversionException
	{
		try {
			// Sanity checks
			if (!(node.children[1] instanceof SymbolNode))
				throw new PerceptorConversionException("Malformed Message: "
						+ node.toString());
			if (!(node.children[2] instanceof SymbolNode))
				throw new PerceptorConversionException("Malformed Message: "
						+ node.toString());

			/* Check content */
			SymbolNode temperatureNode = (SymbolNode) node.children[1];
			SymbolNode batteryNode = (SymbolNode) node.children[2];

			if (!temperatureNode.children[0].equals("temp"))
				throw new PerceptorConversionException("temperature expected: "
						+ node.toString());
			if (!batteryNode.children[0].equals("battery"))
				throw new PerceptorConversionException("battery level expected: "
						+ node.toString());

			return new AgentStatePerceptor(
					Integer.parseInt((String) temperatureNode.children[1]),
					Integer.parseInt((String) batteryNode.children[1]));
		} catch (IndexOutOfBoundsException e) {
			throw new PerceptorConversionException("Malformed node: "
					+ node.toString());
		}
	}

	/**
	 * Parse a symbol tree node into a Hear Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Hear Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private HearPerceptor parseHear(SymbolNode node)
			throws PerceptorConversionException
	{
		float time;
		String message = "", target;

		logger.log(Level.FINEST, "hear: ({0})", new Object[] { node });

		// Sanity checks
		if (node.children.length < 4)
			throw new PerceptorConversionException("Malformed hear node: "
					+ node.toString());

		try {
			time = Float.parseFloat((String) node.children[1]);
			target = (String) node.children[2];

			// Concatenate following nodes
			for (int i = 3; i < node.children.length; i++) {
				if (i > 3)
					message += " " + node.children[i];
				else
					message += node.children[i];
			}
		} catch (Exception e) {
			throw new PerceptorConversionException(
					"Malformed hear node, conversion error: " + node.toString());
		}

		return new HearPerceptor(time, target, message);
	}

	/**
	 * Parse a symbol tree node into a Time Perceptor object
	 * 
	 * @param node Symbol tree node
	 * @return Time Perceptor object
	 * @throws PerceptorConversionException If the input string contains illegal
	 *         data which cannot be converted
	 */
	private TimePerceptor parseTime(SymbolNode node)
			throws PerceptorConversionException
	{
		// Sanity check
		if (node.children.length != 2
				|| !(node.children[1] instanceof SymbolNode))
			throw new PerceptorConversionException("Malformed time node: "
					+ node.toString());

		// Sub-node sanity check
		SymbolNode timeNode = (SymbolNode) node.children[1];
		if (timeNode.children.length != 2 || timeNode.children[1].equals("now"))
			throw new PerceptorConversionException("Malformed time sub-node: "
					+ timeNode.toString());

		try {
			// Sanity checks
			float time = Float.parseFloat((String) timeNode.children[1]);

			return new TimePerceptor(time);
		} catch (Exception e) {
			throw new PerceptorConversionException(
					"Malformed time node, conversion error: " + node.toString());
		}
	}
}
