/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel;

import java.util.Collection;
import java.util.List;

import magma.agent.perception.IPerception;
import magma.util.geometry.Angle;
import magma.util.observer.IObserver;
import magma.util.observer.ISubscribe;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Represents all environment data available to the agent
 */
public interface IGlobalMap extends ISoccerPitchDescription,
		ISubscribe<IGlobalMap>
{
	/**
	 * Called to trigger a update of the AgentModel based on the given perception
	 * object.
	 * 
	 * @param perception the Perception
	 */
	public void update(IPerception perception);

	/**
	 * Get a reference to the ball object
	 * 
	 * @return The ball object
	 */
	public abstract IMoveableObject getBall();

	/**
	 * Get a list of all visible landmarks
	 * 
	 * @return An unmodifiable collection of the landmarks
	 */
	public abstract Collection<ILandmark> getLandmarks();

	/**
	 * @param name the name of the landmark
	 * @return the landmark specified by name
	 */
	public abstract ILandmark getLandmark(String name);

	/**
	 * Calculate the average error between known and calculated position
	 * 
	 * @return the average error of known position and calculated position,
	 *         Double.MAX_VALUE if we do not see any flags
	 */
	double getLandmarkError();

	/**
	 * Get a list of all visible field lines
	 * 
	 * @return An unmodifiable collection of the field lines
	 */
	public abstract Collection<IFieldLine> getFieldLines();

	/**
	 * @param name the name of the field line
	 * @return the field line specified by name
	 */
	public abstract IFieldLine getFieldLine(String name);

	/**
	 * Get a list of all visible players
	 * 
	 * @return An unmodifiable list of the currently visible players
	 */
	public abstract List<IPlayer> getVisiblePlayers();

	/**
	 * Get a reference to the player object of this player
	 * 
	 * @return This players information of position etc
	 */
	public abstract IThisPlayer getThisPlayer();

	/**
	 * Get the current global time
	 * 
	 * @return the time on the global clock
	 */
	float getGlobalTime();

	/**
	 * Get the running time of the current game
	 * 
	 * @return the time the game is running now
	 */
	float getGameTime();

	/**
	 * Get the current play mode
	 * @return the current mode of the game as String
	 */
	GameState getGameState();

	/**
	 * Get home position for this agent in conjunction with the current game
	 * state
	 * 
	 * @return vector to home position
	 */
	Vector3D getHomePosition();

	/**
	 * Get the number of goals our own team scored
	 * 
	 * @return number of goals scored by us
	 */
	int getGoalsWeScored();

	/**
	 * Get the number of goals the opponent team scored
	 * 
	 * @return number of goals scored by opponent
	 */
	int getGoalsTheyScored();

	/**
	 * Get the direction to our own goal
	 * 
	 * @return direction to the middle of own Goal as Vector3D
	 */
	Vector3D getOwnGoalPosition();

	/**
	 * Get the position of the opponent goal
	 * 
	 * @return position to the middle of opponent Goal as Vector3D
	 */
	Vector3D getOtherGoalPosition();

	/**
	 * Detaches the passed observer to this connection
	 * 
	 * @param observer the observer that is no longer interested in new messages
	 * @return true if the observer was in the list and detached
	 */
	boolean detach(IObserver<IGlobalMap> observer);

	/**
	 * Checks if the specified is in the other half or not
	 * 
	 * @param position position for which to check
	 * @return true if the specified is in the other half
	 */
	boolean isInOtherHalf(Vector3D position);

	/**
	 * Checks if the specified is in the own half or not
	 * 
	 * @param position position for which to check
	 * @return true if the specified is in the own half
	 */
	boolean isInOwnHalf(Vector3D position);

	/**
	 * Get the player which is closest to the ball
	 * 
	 * @return the teammate that is closest to ball, null if we see no teammates
	 */
	public IPlayer getTeammateAtBall();

	/**
	 * Get a list of all players, sorted by distance from the own position
	 * 
	 * @return a list of all players sorted by the distance to me
	 */
	public List<IPlayer> getPlayersAtMeList();

	/**
	 * Get a list of all opponents, sorted by distance from the own position
	 * 
	 * @return a list of all opponents sorted by the distance to me
	 */
	public List<IPlayer> getOpponentsAtMeList();

	/**
	 * Get a list of all opponents, sorted by distance from the ball position
	 * 
	 * @return a list of all opponents sorted by the distance to ball
	 */
	public List<IPlayer> getOpponentsAtBallList();

	/**
	 * Get a list of all teammates, sorted by distance from the ball position
	 * 
	 * @return list of teammates
	 */
	public List<IPlayer> getTeammatesAtBall();

	/**
	 * Get a list of all players, sorted by distance from the ball position
	 * 
	 * @return List of players
	 */
	public abstract List<IPlayer> getPlayersAtBallList();

	/**
	 * @return the obstacles to avoid
	 */
	public List<IVisibleObject> getObstacles();

	/**
	 * Get a list of points around ball in order of priority to kick to.
	 * 
	 * @return List of points
	 */
	public Vector3D getPointToKickBall();

	/**
	 * Method gives a list of Polygons around this player and the ball. Is
	 * @deprecated as it is used only by the Debugger tool.
	 * 
	 * @return List of polygons
	 */
	@Deprecated
	public List<Vector3D[]> getInfluencePolys();

	/**
	 * @return the opponent that is closest to ball
	 */
	public IPlayer getOpponentAtBall();

	/**
	 * @param delta how much closer has the opponent to be than own player
	 * @return true if one of our own team (not including myself!) is closer to
	 *         ball than an opponent player
	 */
	public boolean ownTeammateHasBall(double delta);

	/**
	 * @return the preferred kick direction
	 */
	public Angle getPreferredKickDirection();

	/**
	 * @return the kick direction profile
	 */
	public float[] getKickDirectionProfile();

	/**
	 * @param id the player's number (1-11)
	 * @param ownTeam true if we want the player of our own team
	 * @return the referenced player, null if not in list
	 */
	IPlayer getVisiblePlayer(int id, boolean ownTeam);
}