/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel;

import magma.util.geometry.Angle;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Interface to access visible player information
 * @author Klaus Dorer, Stefan Glaser
 */
public interface IThisPlayer extends IPlayer
{
	/**
	 * Returns the relative angle this player's torso has to the specified
	 * position
	 * @param position the position to which to calculate the body angle
	 * @return the relative angle this player's torso has to the specified
	 *         position
	 */
	Angle getBodyDirectionTo(Vector3D position);

	/**
	 * @return true if this is the goal keeper of our team
	 */
	@Override
	boolean isGoalie();

	/**
	 * Calculate global position based on own position and orientation.
	 * 
	 * @param localPosition - the position to transform, given in the local root
	 *        body system
	 * @return the position, relative to the global coordinate system
	 */
	public Vector3D calculateGlobalPosition(Vector3D localPosition);

	/**
	 * Calculate global position in 2D based on own position and body direction.<br>
	 * This calculation is independent of rotations along the view axis of the
	 * agent. Is is useful for top view calculations.
	 * 
	 * @param localPos - the position to transform, given in the local root body
	 *        system (just xy matters)
	 * @return the position in 2D (top view) relative to the global coordinate
	 *         system
	 */
	public Vector3D calculateGlobal2DPosition(Vector3D localPos);

	/**
	 * Calculate local position based on own position and orientation.
	 * 
	 * @param globalPosition - the position to transform, given in the global
	 *        coordinate system
	 * @return the position, relative to the local root body system
	 */
	public Vector3D calculateLocalPosition(Vector3D globalPosition);

	/**
	 * Checks if the passed 2D xy-point (absolute coordinate system) is inside
	 * the passed polygon (xy, relative coordinate system)
	 * @param absolutePosition 2D vector (x,y) of the point to check in the
	 *        global coordinate system
	 * @param polygon corner points of the polygon (x, y), in this player's
	 *        coordinate system and in clockwise order. The coordinates should
	 *        NOT take neck angles into account, i.e. assume head is straight.
	 * @return true if the point is inside the polygon
	 */
	public boolean isInsidePolygonXY(Vector3D absolutePosition,
			Vector3D[] polygon);

	/**
	 * @param position the position to check
	 * @return true if the specified position is left of this player with respect
	 *         to the body direction
	 */
	public boolean positionIsLeft(Vector3D absolutePosition);

	/**
	 * @param position the position to check
	 * @return true if the specified position is right of this player with
	 *         respect to the body direction
	 */
	public boolean positionIsRight(Vector3D absolutePosition);

	/**
	 * @param position the position to check
	 * @return true if the specified position is behind this player with respect
	 *         to the body direction
	 */
	public boolean positionIsBehind(Vector3D absolutePosition);

	/**
	 * @return the positionManager
	 */
	public IPositionManager getPositionManager();
}