/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel;

import magma.util.geometry.Angle;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Base Interface for all visible objects in the world model
 * @author Klaus Dorer
 */
public interface IVisibleObject
{
	/**
	 * @return the name of this object
	 */
	String getName();

	/**
	 * @return the current position in the global coordinate system
	 */
	Vector3D getPosition();

	/**
	 * @return the previous position in the global coordinate system, null,
	 *         before the first perception update.
	 */
	Vector3D getPreviousPosition();

	/**
	 * Retrieves the local position of the visible object.<br>
	 * This position is not directly or necessarily the same "local" position as
	 * observed through the perception! We provide an unique local coordinate
	 * system for the agent, based on the root body system. All local positions
	 * of the model layer are relative to this local system.<br>
	 * <b>Note:</b> The unique local system is facing the x axis!
	 * 
	 * @return the local position (relative to the global root body system)
	 */
	Vector3D getLocalPosition();

	/**
	 * @return time stamp when this object was last updated by perception
	 */
	float getLastSeenTime();

	/**
	 * Returns the delta of current time passed and the last time this object was
	 * seen
	 * @param currentTime the current global time
	 * @return delta of current time passed and the last time this object was
	 *         seen
	 */
	float getAge(float currentTime);

	// TODO Modify the getDistanceToXY-methods to return the nearest point, not
	// just the first
	/**
	 * @param other the point to which to calculate
	 * @return the distance projection to the x y plane between this object and
	 *         the passed coordinate
	 */
	public double getDistanceToXY(Vector3D other);

	/**
	 * @param other the point to which to calculate
	 * @return the distance projection to the x y plane between this object and
	 *         the passed objects position
	 */
	public double getDistanceToXY(IVisibleObject other);

	/**
	 * Calculates the distance of this visible object to the passed one
	 * @param other the object to which to calculate distance
	 * @return the distance of this visible object to the passed one
	 */
	public double getDistanceToXYZ(IVisibleObject other);

	/**
	 * Calculates the distance of this visible object to the passed position
	 * @param other the position to which to calculate distance
	 * @return the distance of this visible object to the passed position
	 */
	public double getDistanceToXYZ(Vector3D other);

	/**
	 * Calculates the Direction of this visible object to the passed one
	 * @param other the object to which to calculate direction
	 * @return the direction (rad) of this visible object to the passed one
	 */
	public Angle getDirectionTo(IVisibleObject other);

	/**
	 * Calculates the global absolute Direction of this visible object to the
	 * passed Vector3D
	 * @param other the Vector3D to which to calculate direction
	 * @return the global absolute direction (rad) of this visible object to the
	 *         passed Vector
	 */
	public Angle getDirectionTo(Vector3D other);

	/**
	 * @return true if this object was seen last perception
	 */
	public boolean isVisible();

	/**
	 * @return The distance required to move around the object without collision
	 */
	double getCollisionDistance();

	/**
	 * Check if the object lies on the ground. The default version always returns
	 * false. To put this method here and not in IPlayer is a workaround to be
	 * able to use VisibleObjects in IFOCalculator
	 * @return true if it is laying around
	 */
	public boolean isLaying();

}