/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.impl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.vecmath.Matrix3d;
import javax.vecmath.Matrix4d;
import javax.vecmath.Vector2d;
import javax.vecmath.Vector3d;

import magma.agent.IMagmaConstants;
import magma.agent.IServerConfigFilesConstants;
import magma.agent.agentmodel.IAgentModel;
import magma.agent.agentmodel.IBodyPart;
import magma.agent.agentmodel.IGyroRate;
import magma.agent.meta.server.IRCServerMetaModel;
import magma.agent.perception.IHearPerceptor;
import magma.agent.perception.ILinePerceptor;
import magma.agent.perception.IPerception;
import magma.agent.perception.IPlayerPos;
import magma.agent.perception.IVisibleObjectPerceptor;
import magma.agent.worldmodel.GameState;
import magma.agent.worldmodel.IFieldLine;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.ILandmark;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IThisPlayer;
import magma.agent.worldmodel.IVisibleObject;
import magma.agent.worldmodel.PlayMode;
import magma.agent.worldmodel.PlaySide;
import magma.agent.worldmodel.localizer.IEnvironmentModel;
import magma.agent.worldmodel.localizer.ILocalizationLine;
import magma.agent.worldmodel.localizer.ILocalizer;
import magma.agent.worldmodel.localizer.IReferenceLine;
import magma.agent.worldmodel.localizer.IReferencePoint;
import magma.util.SayCoder;
import magma.util.geometry.Angle;
import magma.util.geometry.MatrixUtil;
import magma.util.geometry.PositionOrientation3D;
import magma.util.geometry.positionFilter.PositionFilter;
import magma.util.geometry.positionFilter.PositionFilterBase;
import magma.util.observer.IObserver;
import magma.util.observer.IPublishSubscribe;
import magma.util.observer.Subject;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Container class for all visible objects on the field, for the time and game
 * state
 */
public class GlobalMap implements IGlobalMap, IEnvironmentModel, Serializable
{
	/** the time we keep other players in mind if not seeing them (in seconds) */
	private static final int REMEMBERANCE_TIME = 30;

	// handle for logging
	private static final Logger logger = Logger.getLogger("");

	/** id of this class for serialization */
	private static final long serialVersionUID = 182243462936086603L;

	/** observers of world model */
	private final transient IPublishSubscribe<IGlobalMap> observer;

	private final float fieldHalfLength;

	private final float fieldHalfWidth;

	private final float goalHalfWidth;

	private final float goalHeight;

	/** the localizer that is able to calculate the absolute position */
	private final transient ILocalizer localizer;

	/** the ball of the game */
	private Ball ball;

	/** the landmark visual objects (flags and goal posts) */
	private final Map<String, ILandmark> landmarks;

	/** the reference points used for localization (flags and goal posts) */
	private final transient List<IReferencePoint> referencePoints;

	/** the field line visual objects */
	private final Map<String, IFieldLine> fieldLines;

	/** the reference lines (field lines) used for localization */
	private final transient List<IReferenceLine> referenceLines;

	/**
	 * A map of all known players over time. As unique key, the team name
	 * appended by the player number is used.
	 */
	private Map<String, IPlayer> knownPlayers;

	/** the list of all visible players */
	private List<IPlayer> visiblePlayers;

	/** the time on the global clock (in cycles). Not the game time */
	private float globalTime;

	/** the time the game is running now (cycles since start) */
	private float gameTime;

	/** the time of the last change of the game state */
	private float lastGameStateChangeTime;

	/** The side which our team is playing. */
	// Note: It's important, that this value is initialized with the left side!
	private PlaySide playSide = PlaySide.LEFT_SIDE;

	/** the playmode the game is running now */
	private PlayMode playmode = PlayMode.BEFORE_KICK_OFF;

	/** the game state, the game is running now */
	private GameState gameState = GameState.BEFORE_KICK_OFF;

	/** the game state, the game was running previously */
	private GameState previousGameState;

	/** the global coordinates of the agent */
	ThisPlayer thisPlayer;

	/** goals scored by the opponent */
	private int goalsTheyScored = 0;

	/** goals scored by us */
	private int goalsWeScored = 0;

	/** reference to the agent model needed for gyro localization */
	private IAgentModel agentModel;

	/** The version of the soccer server used */
	protected int serverVersion;

	/** the indexical functional object calculator */
	private transient IFOCalculator ifoCalculator;

	/** list of all teammates sorted by the distance to ball */
	private List<IPlayer> teamMateAtBallList;

	/** list of all opponents sorted by the distance to ball */
	private List<IPlayer> opponentsAtBallList;

	private List<IPlayer> playersAtBallList;

	/** list of all players sorted by the distance to me */
	private List<IPlayer> playersAtMeList;

	/** list of all opponents sorted by the distance to me */
	private List<IPlayer> opponentsAtMeList;

	/** list of all obstacles to avoid sorted by the distance to me */
	private List<IVisibleObject> obstacles;

	private InfluenceMap influenceMap;

	private transient PositionFilterBase posFilter;

	/** The kick direction profiler */
	private transient KickDirectionProfiler kickDirector;

	/**
	 * The name of the other team.<br>
	 * <tt>null</tt> as long we haven't seen any other player.
	 */
	private String otherTeamName;

	/**
	 * The transformation from the camera body system to the root body system of
	 * the current cycle.
	 */
	private transient Matrix4d cameraToRootVision;

	/**
	 * Constructor
	 * 
	 * @param agentModel Reference to the agent model object
	 * @param localizer the module that calculates the agent's global position
	 * @param metaModel the meta model of the rc server
	 * @param teamname Team name
	 * @param playerNumber Player number
	 */
	public GlobalMap(IAgentModel agentModel, ILocalizer localizer,
			IRCServerMetaModel metaModel, String teamname, int playerNumber)
	{
		assert playerNumber < IMagmaConstants.startPositionsOwnKickoff.length : "No startPositionLeft for player id: "
				+ playerNumber;

		this.agentModel = agentModel;
		this.localizer = localizer;
		this.serverVersion = metaModel.getVersion();
		fieldHalfLength = (float) metaModel.getFieldDimensions().x / 2;
		fieldHalfWidth = (float) metaModel.getFieldDimensions().y / 2;
		goalHalfWidth = (float) metaModel.getGoalDimensions().x / 2;
		goalHeight = (float) metaModel.getGoalDimensions().y;

		// create objects
		landmarks = new HashMap<String, ILandmark>();
		fieldLines = new HashMap<String, IFieldLine>();
		referencePoints = new ArrayList<IReferencePoint>();
		referenceLines = new ArrayList<IReferenceLine>();
		createObjects(metaModel);

		thisPlayer = new ThisPlayer(teamname, playerNumber);
		thisPlayer.getPositionManager().setDesiredPosition(getHomePosition(),
				Angle.RAD0, false);
		observer = new Subject<IGlobalMap>();
		// TODO: Think about what will happen, when we switch sides during a game
		// (this problem is basically not present, since the flags can only be
		// mirrored, when they are symmetric to the origin - and therefore the
		// Landmarks we try to avoid obstacle with are just wrong labeled)
		ifoCalculator = new IFOCalculator(this);
		influenceMap = new InfluenceMap(this);
		posFilter = new PositionFilter();
		kickDirector = new KickDirectionProfiler(180, 4.7f);
	}

	/**
	 * Creates all objects. We do not create new objects each time on update, but
	 * update the existing objects
	 */
	private void createObjects(IRCServerMetaModel metaModel)
	{
		ball = new Ball(metaModel.getBallDecay());

		// Landmarks
		Map<String, Vector3d> landmarkConfig = metaModel.getLandmarks();
		if (landmarkConfig != null) {
			Vector3d pos;
			Landmark landmark;

			for (String key : landmarkConfig.keySet()) {
				pos = landmarkConfig.get(key);
				landmark = new Landmark(key, new Vector3D(pos.x, pos.y, pos.z));
				landmarks.put(key, landmark);
				referencePoints.add(landmark);
			}
		}

		// FieldLines
		Map<String, Vector2d[]> fieldLineConfig = metaModel.getFieldLines();
		if (fieldLineConfig != null) {
			Vector2d[] pos2d;
			FieldLine fieldLine;

			for (String key : fieldLineConfig.keySet()) {
				pos2d = fieldLineConfig.get(key);
				fieldLine = new FieldLine(key, //
						new Vector3D(pos2d[0].x, pos2d[0].y, 0),//
						new Vector3D(pos2d[1].x, pos2d[1].y, 0));
				fieldLines.put(key, fieldLine);
				referenceLines.add(fieldLine);
			}
		}

		// Players
		int numberOfPlayersPerTeam = IMagmaConstants.NUMBER_OF_PLAYERS_PER_TEAM;
		knownPlayers = new HashMap<String, IPlayer>(numberOfPlayersPerTeam * 2);
		visiblePlayers = new ArrayList<IPlayer>(numberOfPlayersPerTeam * 2);

		// functional Object list
		teamMateAtBallList = new ArrayList<IPlayer>(0);
		opponentsAtBallList = new ArrayList<IPlayer>(0);
		playersAtBallList = new ArrayList<IPlayer>(0);
		playersAtMeList = new ArrayList<IPlayer>(0);
		opponentsAtMeList = new ArrayList<IPlayer>(0);
		obstacles = new ArrayList<IVisibleObject>(0);
	}

	/**
	 * Called once perception is finished parsing a new incoming message
	 * @param perception the object containing the result from parsing
	 */
	@Override
	public void update(IPerception perception)
	{
		// update global time
		globalTime = perception.getTime().getTime();

		// update game time, play side and play mode, as well as scored goals
		processGameState(perception);

		// calculate the new position of the agent
		calculatePosition(perception);

		// update visible objects
		updateBall(perception);
		updateLandmarks(perception);
		updateFieldLines();
		updatePlayers(perception);

		processHear(perception);

		// recalculate kick directions
		if (perception.containsVision()) {
			kickDirector.updateProfile(this);
		}

		logger.log(Level.FINE, "time: {0} gameTime: {1}", new Object[] {
				globalTime, gameTime });

		// now inform observers about changes
		observer.onStateChange(this);
	}

	private void processGameState(IPerception perception)
	{
		// update game time
		gameTime = perception.getGameState().getTime();

		// update playSide
		String playSideString = perception.getGameState().getTeamSide();
		PlaySide newPlaySide = PlaySide.parsePlaySide(playSideString);

		switch (newPlaySide) {
		case LEFT_SIDE:
			if (playSide != PlaySide.LEFT_SIDE) {
				// We switched from right to left side, so remember the new side and
				// mirror the flags
				playSide = PlaySide.LEFT_SIDE;
				mirrorLandmarks();
			}
			break;
		case RIGHT_SIDE:
			if (playSide != PlaySide.RIGHT_SIDE) {
				// We switched from left to right side, so remember the new side and
				// mirror the flags
				playSide = PlaySide.RIGHT_SIDE;
				mirrorLandmarks();
			}
			break;
		case UNKNOWN:
		default:
			// No playSide?!? Maybe I'm the referee?
			break;
		}

		// update playMode
		String playModeString = perception.getGameState().getPlaymode();
		previousGameState = gameState;

		playmode = PlayMode.parsePlayMode(playModeString);
		gameState = GameState.determineGameState(playmode, playSide);

		// If game state has changed, remember the current time
		if (gameState != previousGameState) {
			lastGameStateChangeTime = globalTime;
		}

		if (playmode == PlayMode.NONE) {
			logger.log(Level.SEVERE, "play mode not known: <{0}>",
					new Object[] { playModeString });
		}

		// update scored goals
		updateGoalsScored();
	}

	private void mirrorLandmarks()
	{
		Map<String, ILandmark> newLandmarks = new HashMap<String, ILandmark>();
		Vector3D pos;
		Landmark landmark;

		// Clear list of reference points
		referencePoints.clear();

		for (String key : landmarks.keySet()) {
			pos = landmarks.get(key).getKnownPosition();
			landmark = new Landmark(key, new Vector3D(-pos.getX(), -pos.getY(),
					pos.getZ()));
			newLandmarks.put(key, landmark);
			referencePoints.add(landmark);
		}

		landmarks.clear();
		landmarks.putAll(newLandmarks);
	}

	/**
	 * Update the number of scored goals
	 */
	private void updateGoalsScored()
	{
		if (gameState == GameState.OWN_GOAL
				&& previousGameState != GameState.OWN_GOAL) {
			goalsWeScored++;
		} else if (gameState == GameState.OPPONENT_GOAL
				&& previousGameState != GameState.OPPONENT_GOAL) {
			goalsTheyScored++;
		}
	}

	private void processHear(IPerception perception)
	{
		try {
			IHearPerceptor hear = perception.getHearPerceptor();
			String target = hear.getTarget();

			if ("self".equalsIgnoreCase(target)) {
				// message from ourselves
				return;
			}

			String message = hear.getMessage();
			if (message.length() != 3 * SayCoder.getPositionLength() + 2) {
				// our messages are fixed length
				return;
			}

			int posStart = 1;
			int posLength = SayCoder.getPositionLength();
			int playerID = SayCoder.decodeID(message.substring(0, posStart));
			Vector3D playerPos = SayCoder.decodePosition(message.substring(
					posStart, posStart + posLength));
			posStart += posLength;
			Vector3D ballPos = SayCoder.decodePosition(message.substring(posStart,
					posStart + posLength));

			posStart += posLength;
			Vector3D opponentPos = SayCoder.decodePosition(message.substring(
					posStart, posStart + posLength));

			posStart += posLength;
			String teamCode = message.substring(posStart, posStart + 1);

			if (!SayCoder.encodeName(getTeamname(true)).equals(teamCode)) {
				// message is not from our team
				return;
			}

			if (playerPos != null) {
				upatePlayerPosition(playerID, true, playerPos);
			}

			if (ballPos != null && ball.getAge(globalTime) > 1.0f) {
				// after not seeing the ball for n secs we trust hear
				ball.setPosition(new Vector3D(ballPos.getX(), ballPos.getY(), ball
						.getPosition().getZ()));
				logger.log(Level.FINER, "hear ball: ({0}, {1})", new Object[] {
						ballPos.getX(), ballPos.getY() });
			}

			if (opponentPos != null) {
				upatePlayerPosition(playerID, false, opponentPos);
			}

		} catch (Exception e) {
			// might happen with opponent messages
		}
	}

	/**
	 * Updates the local and global position of the corresponding player instance
	 * according to the given global position.
	 * 
	 * @param playerID the id of the player
	 * @param ownTeam true if the player to update is of the own team
	 * @param playerPos the new global position
	 */
	private void upatePlayerPosition(int playerID, boolean ownTeam,
			Vector3D playerPos)
	{
		String teamName = getTeamname(ownTeam);
		if (teamName == null) {
			// As long we don't know the others team name, we aren't able to create
			// proper player instances
			return;
		}

		Player player = (Player) knownPlayers.get(teamName + playerID);
		if (player == null) {
			// We hear from a player that we have never seen before
			player = new Player(playerID, teamName, ownTeam);
			knownPlayers.put(teamName + playerID, player);
		}

		if (player.getAge(globalTime) < 1.0f) {
			// We have recently seen the player so no update
			return;
		}

		// Update player information
		player.setBodyparts(null);
		Vector3D localPos = thisPlayer.calculateLocalPosition(playerPos);
		player.update(localPos, playerPos, globalTime);
	}

	/**
	 * If ownTeam is true we return the team name of thisPlayer. If false we
	 * return the name of the first seen other player ever. If we haven't seen
	 * any other player yet, this method returns <tt>null</tt>.
	 * @param ownTeam true if from own team, false if from other team
	 * @return the name of the team, if known, null if not
	 */
	public String getTeamname(boolean ownTeam)
	{
		if (ownTeam) {
			return thisPlayer.getTeamname();
		} else {
			return otherTeamName;
		}
	}

	/**
	 * @param perception the object containing the result from parsing
	 * 
	 */
	private void calculatePosition(IPerception perception)
	{
		// Update current camera-to-root-vision transformation, if possible
		IBodyPart cameraBodyPart = agentModel.getBodyPartContainingCamera();
		if (cameraBodyPart != null) {
			cameraToRootVision = cameraBodyPart.getTransformation();
		}

		// Extract visible lines of the current cycle
		List<ILinePerceptor> visibleLines = perception.getVisibleLines();
		List<ILocalizationLine> visibleFieldLines = new ArrayList<ILocalizationLine>(
				visibleLines.size());
		for (ILinePerceptor pl : visibleLines) {
			visibleFieldLines.add(new LocalizationLine(toRootVisionSystem(pl
					.getPosition()), toRootVisionSystem(pl.getPosition2())));
		}

		// Reset reference lines (and implicitly the field lines as well)
		for (IReferenceLine line : referenceLines) {
			line.setVisible(false);
		}

		// Extract visible markers of the current cycle
		for (String name : landmarks.keySet()) {
			Landmark landmark = (Landmark) landmarks.get(name);
			IVisibleObjectPerceptor landmarkSeen = perception.getObject(name);
			if (landmarkSeen != null) {
				// this landmark is currently visible, so update it's local position
				// and store it to the visible flags for localization
				landmark.updateLocalPosition(toRootVisionSystem(landmarkSeen
						.getPosition()));
			} else {
				landmark.setVisible(false);
			}
		}

		// Calculate an estimation of the current orientation, based on the gyro
		// sensor in the root body (if existing)
		Matrix3d orientationEstimation = null;
		IGyroRate gyro = agentModel.getTorso().getSensor(IGyroRate.class);

		// If we have a gyro sensor in the root body, use it to estimate the
		// current orientation
		if (gyro != null) {
			orientationEstimation = new Matrix3d(thisPlayer.getOrientation());
			orientationEstimation.mul(gyro.getOrientationChange());
		}

		// Perform localization
		PositionOrientation3D localizeInfo = null;
		try {
			localizer.assignReferenceLines(this, visibleFieldLines,
					orientationEstimation);
			localizeInfo = localizer.localize(this, visibleFieldLines,
					orientationEstimation);
		} catch (RuntimeException e) {
			logger.log(Level.SEVERE, "Localization problem {0}", e);
		}

		if (localizeInfo != null) {
			thisPlayer
					.update(null,
							posFilter.filterPosition(localizeInfo.getPosition()),
							globalTime);
			thisPlayer.setOrientation(localizeInfo.getOrientation());
		} else {
			thisPlayer.setOrientation(orientationEstimation);
		}
	}

	/**
	 * @param perception the object containing the result from parsing
	 */
	private void updateBall(IPerception perception)
	{
		IVisibleObjectPerceptor ballVision = perception
				.getObject(IPerception.BALL);
		if (ballVision == null) {
			// the ball is currently not visible
			ball.setVisible(false);
		} else {
			Vector3D localPos = toRootVisionSystem(ballVision.getPosition());
			Vector3D globalPos = thisPlayer.calculateGlobalPosition(localPos);

			ball.update(localPos, globalPos, globalTime);
		}
	}

	/**
	 * Updates the local and global position of the visible landmarks.
	 * 
	 * @param perception the object containing the result from parsing
	 */
	private void updateLandmarks(IPerception perception)
	{
		for (String name : landmarks.keySet()) {
			Landmark landmark = (Landmark) landmarks.get(name);
			IVisibleObjectPerceptor landmarkSeen = perception.getObject(name);
			if (landmarkSeen == null) {
				// this landmark is currently not visible
				landmark.setVisible(false);
			} else {
				Vector3D localPos = toRootVisionSystem(landmarkSeen.getPosition());
				Vector3D globalPos = thisPlayer.calculateGlobalPosition(localPos);

				landmark.update(localPos, globalPos, globalTime);
			}
		}
	}

	/**
	 * Updates the global position of the as visible marked field lines.
	 */
	private void updateFieldLines()
	{
		for (IFieldLine fieldLine : fieldLines.values()) {
			if (fieldLine.isVisible()) {
				((FieldLine) fieldLine).updatePositions(thisPlayer
						.calculateGlobalPosition(fieldLine.getLocalPosition1()),
						thisPlayer.calculateGlobalPosition(fieldLine
								.getLocalPosition2()), globalTime);
			}
		}
	}

	/**
	 * Updates the player information according to the new visual perception.
	 * 
	 * @param perception the object containing the result from vision parsing
	 */
	void updatePlayers(IPerception perception)
	{
		// We only act when vision information should be present
		if (!perception.containsVision()) {
			return;
		}

		// Clear visible players
		visiblePlayers.clear();

		// Process player visions
		List<IPlayerPos> playersVision = perception.getVisiblePlayers();
		Player currentPlayer;
		boolean isOwnTeam;

		for (IPlayerPos playerVision : playersVision) {
			isOwnTeam = thisPlayer.getTeamname()
					.equals(playerVision.getTeamname());

			// Don't process myself as a teammate
			if (isOwnTeam && playerVision.getId() == thisPlayer.getID()) {
				continue;
			}

			// Fetch a player instance
			if (playerVision.getTeamname() == null
					|| IMagmaConstants.UNKNOWN_PLAYER_TEAMNAME.equals(playerVision
							.getTeamname())
					|| playerVision.getId() == IMagmaConstants.UNKNOWN_PLAYER_NUMBER) {
				// If the player vision information is not complete, create just a
				// temporary player instance to hold the perceived information and
				// add it to the visiblePlayers array
				// TODO: Instead of simply adding an additional player by default,
				// one can perform some kind of lookup in the known players map and
				// try to match the player to an previously known player near by
				currentPlayer = new Player(playerVision.getId(),
						playerVision.getTeamname(), isOwnTeam);
				visiblePlayers.add(currentPlayer);

				logger.log(
						Level.FINE,
						"Saw player with incomplete information: ({0}, {1})",
						new Object[] { playerVision.getTeamname(),
								playerVision.getId() });
			} else {
				// If the player vision information is complete, fetch or create a
				// permanent instance to him
				currentPlayer = (Player) knownPlayers.get(playerVision
						.getTeamname() + playerVision.getId());
				if (currentPlayer == null) {
					currentPlayer = new Player(playerVision.getId(),
							playerVision.getTeamname(), isOwnTeam);
					knownPlayers.put(
							currentPlayer.getTeamname() + currentPlayer.getID(),
							currentPlayer);

					// Remember the name of the other team once we see the first
					// other player
					if (otherTeamName == null && !isOwnTeam) {
						otherTeamName = playerVision.getTeamname();
					}
				}
			}

			// Update the player information
			updatePlayer(currentPlayer, playerVision);
		}

		// Extract all recently seen players from known players map and mark the
		// others as not visible
		for (IPlayer player : knownPlayers.values()) {
			if (player.getAge(globalTime) < REMEMBERANCE_TIME) {
				visiblePlayers.add(player);
			} else {
				((Player) player).setVisible(false);
			}
		}

		// Recalculate indexical functional objects
		calculateIndexicalFunctionalObjects();
	}

	/**
	 * Updates a player's body parts as well as its local and global position
	 * 
	 * @param player the player to update
	 * @param playerVision the update information
	 */
	private void updatePlayer(Player player, IPlayerPos playerVision)
	{
		Vector3D localPos;
		Vector3D globalPos;

		// Transform and update body parts of player
		Map<String, Vector3D> bodyParts = new HashMap<String, Vector3D>();
		Map<String, Vector3D> allBodyParts = playerVision.getAllBodyParts();
		for (String name : allBodyParts.keySet()) {
			localPos = toRootVisionSystem(allBodyParts.get(name));
			globalPos = thisPlayer.calculateGlobalPosition(localPos);

			bodyParts.put(name, globalPos);
		}
		player.setBodyparts(bodyParts);

		// Transform and update player positions
		localPos = toRootVisionSystem(playerVision.getPosition());
		globalPos = thisPlayer.calculateGlobalPosition(localPos);

		player.update(localPos, globalPos, globalTime);
	}

	/**
	 * @param id the player's number (1-11)
	 * @param ownTeam true if we want the player of our own team
	 * @return the referenced player, null if not in list
	 */
	@Override
	public IPlayer getVisiblePlayer(int id, boolean ownTeam)
	{
		if (id < 1) {
			return null;
		}
		for (int i = 0; i < visiblePlayers.size(); i++) {
			IPlayer player = visiblePlayers.get(i);
			if (id == player.getID() && ownTeam == player.isOwnTeam()) {
				return player;
			}
		}
		return null;
	}

	private Vector3D toRootVisionSystem(Vector3D visionVector)
	{
		if (visionVector == null) {
			return null;
		}

		if (cameraToRootVision == null) {
			// There exists no transformation, so no system conversion
			return visionVector;
		}

		Vector3d vVector = new Vector3d(-visionVector.getY(),
				visionVector.getX(), visionVector.getZ());

		vVector = MatrixUtil.mulMatrixVector(cameraToRootVision, vVector);

		return new Vector3D(vVector.getY(), -vVector.getX(), vVector.getZ());
	}

	/**
	 * Calculates the lists of players sorted by different criteria
	 */
	private void calculateIndexicalFunctionalObjects()
	{
		teamMateAtBallList = ifoCalculator.getTeammatesAtBall(visiblePlayers);
		opponentsAtBallList = ifoCalculator.getOpponentsAtBall(visiblePlayers);
		playersAtBallList = ifoCalculator.getPlayersAtBall(visiblePlayers);
		playersAtMeList = ifoCalculator.getPlayersAtMe(visiblePlayers);
		opponentsAtMeList = ifoCalculator.getOpponentsAtMe(playersAtMeList);
		obstacles = ifoCalculator.getObstacles(visiblePlayers, ball);
	}

	@Override
	public boolean ownTeammateHasBall(double delta)
	{
		IPlayer teammateAtBall = getTeammateAtBall();
		if (teammateAtBall == null) {
			return false;
		}

		IPlayer opponentAtBall = getOpponentAtBall();
		if (opponentAtBall == null) {
			return true;
		}

		Vector3D ballPos = getBall().getPosition();
		return opponentAtBall.getDistanceToXY(ballPos) + delta >= teammateAtBall
				.getDistanceToXY(ballPos);

	}

	// --------------------------------------------------
	// General environment and game information
	@Override
	public float getGlobalTime()
	{
		return globalTime;
	}

	/**
	 * @return the time the game is running now
	 */
	@Override
	public float getGameTime()
	{
		return gameTime;
	}

	@Override
	public int getGoalsTheyScored()
	{
		return goalsTheyScored;
	}

	@Override
	public int getGoalsWeScored()
	{
		return goalsWeScored;
	}

	/**
	 * @return the current mode of the game as String
	 */
	public PlayMode getPlaymode()
	{
		return playmode;
	}

	/**
	 * Get play side
	 * 
	 * @return Play side
	 */
	public PlaySide getPlaySide()
	{
		return playSide;
	}

	@Override
	public GameState getGameState()
	{
		return gameState;
	}

	/**
	 * @return the time when the current game state was entered
	 */
	public float getTimeOfLastGameStateChange()
	{
		return lastGameStateChangeTime;
	}

	@Override
	public Vector3D getHomePosition()
	{
		float[] startPos;

		int id = thisPlayer.getID();
		if (id < 0) {
			return new Vector3D(0, 0, 0);
		}

		if (// gameState == GameState.OWN_GOAL
		gameState == GameState.BEFORE_KICK_OFF
				|| gameState == GameState.OPPONENT_GOAL
				|| gameState == GameState.OWN_KICK_OFF) {
			if (serverVersion <= IServerConfigFilesConstants.SERVER_VERSION_62) {
				startPos = IMagmaConstants.startPositionsOwnKickoff_VERSION62[id];
			} else {
				startPos = IMagmaConstants.startPositionsOwnKickoff[id];
			}
		} else {
			if (serverVersion <= IServerConfigFilesConstants.SERVER_VERSION_62) {
				startPos = IMagmaConstants.startPositionsOtherKickoff_VERSION62[id];
			} else {
				startPos = IMagmaConstants.startPositionsOtherKickoff[id];
			}
		}

		logger.log(Level.FINE, "playmode: {0}", new Object[] { gameState });
		return new Vector3D(startPos[0], startPos[1], startPos[2]);
	}

	// --------------------------------------------------
	// Visual object getters
	@Override
	public Ball getBall()
	{
		return ball;
	}

	@Override
	public Collection<ILandmark> getLandmarks()
	{
		return Collections.unmodifiableCollection(landmarks.values());
	}

	@Override
	public ILandmark getLandmark(String name)
	{
		return landmarks.get(name);
	}

	/**
	 * @return the average error of known position and calculated position,
	 *         Double.MAX_VALUE if we do not see any flags
	 */
	@Override
	public double getLandmarkError()
	{
		double result = 0.0;
		int count = 0;
		for (String name : landmarks.keySet()) {
			Landmark landmark = (Landmark) landmarks.get(name);
			// this landmark is not visible currently
			if (landmark.isVisible()) {
				result += landmark.getPosition()
						.subtract(landmark.getKnownPosition()).getNorm();
				count++;
			}
		}
		if (count < 1) {
			return Double.MAX_VALUE;
		}

		return result / count;
	}

	@Override
	public Collection<IFieldLine> getFieldLines()
	{
		return Collections.unmodifiableCollection(fieldLines.values());
	}

	@Override
	public IFieldLine getFieldLine(String name)
	{
		return fieldLines.get(name);
	}

	@Override
	public List<IPlayer> getVisiblePlayers()
	{
		return Collections.unmodifiableList(visiblePlayers);
	}

	@Override
	public IThisPlayer getThisPlayer()
	{
		return thisPlayer;
	}

	// --------------------------------------------------
	// IFOs
	@Override
	public List<IVisibleObject> getObstacles()
	{
		return obstacles;
	}

	@Override
	public List<IPlayer> getOpponentsAtMeList()
	{
		return opponentsAtMeList;
	}

	@Override
	public List<IPlayer> getPlayersAtMeList()
	{
		return playersAtMeList;
	}

	@Override
	public List<IPlayer> getTeammatesAtBall()
	{
		return teamMateAtBallList;
	}

	@Override
	public IPlayer getTeammateAtBall()
	{
		if (teamMateAtBallList.isEmpty()) {
			return null;
		}
		return teamMateAtBallList.get(0);
	}

	@Override
	public List<IPlayer> getOpponentsAtBallList()
	{
		return opponentsAtBallList;
	}

	@Override
	public IPlayer getOpponentAtBall()
	{
		if (opponentsAtBallList.isEmpty()) {
			return null;
		}
		return opponentsAtBallList.get(0);
	}

	@Override
	public List<IPlayer> getPlayersAtBallList()
	{
		return playersAtBallList;
	}

	// --------------------------------------------------
	// Influence Map
	@Override
	public Vector3D getPointToKickBall()
	{
		return influenceMap.getTargetKickPositionForBall();
	}

	@Override
	public List<Vector3D[]> getInfluencePolys()
	{
		return influenceMap.getInfluenceArea(getBall());
	}

	// --------------------------------------------------
	// Kick direction profile and estimation
	@Override
	public float[] getKickDirectionProfile()
	{
		return kickDirector.getKickProfile();
	}

	@Override
	public Angle getPreferredKickDirection()
	{
		return kickDirector.getBestKickDirection();
	}

	// --------------------------------------------------
	// EnvironmentModel Mothods
	@Override
	public List<IReferenceLine> getReferenceLines()
	{
		return referenceLines;
	}

	@Override
	public List<IReferencePoint> getReferencePoints()
	{
		return referencePoints;
	}

	// --------------------------------------------------
	// Server version and soccer pitch description
	@Override
	public int getServerVersion()
	{
		return serverVersion;
	}

	@Override
	public float fieldHalfLength()
	{
		return fieldHalfLength;
	}

	@Override
	public float fieldHalfWith()
	{
		return fieldHalfWidth;
	}

	@Override
	public float goalHalfWith()
	{
		return goalHalfWidth;
	}

	@Override
	public float goalHeight()
	{
		return goalHeight;
	}

	@Override
	public Vector3D getOwnGoalPosition()
	{
		return new Vector3D(-fieldHalfLength, 0, 0);
	}

	@Override
	public Vector3D getOtherGoalPosition()
	{
		return new Vector3D(fieldHalfLength, 0, 0);
	}

	@Override
	public boolean isInOtherHalf(Vector3D position)
	{
		return position.getX() > 2;
	}

	@Override
	public boolean isInOwnHalf(Vector3D position)
	{
		return !isInOtherHalf(position);
	}

	// --------------------------------------------------
	// Observer methods
	@Override
	public void attach(IObserver<IGlobalMap> newObserver)
	{
		observer.attach(newObserver);
	}

	@Override
	public boolean detach(IObserver<IGlobalMap> oldObserver)
	{
		return observer.detach(oldObserver);
	}

	// --------------------------------------------------
	// Basic object methods
	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof GlobalMap)) {
			return false;
		}
		GlobalMap other = (GlobalMap) o;

		if (!ball.equals(other.ball)) {
			return false;
		}

		if (!ball.equals(other.ball)) {
			return false;
		}

		if (!landmarks.equals(other.landmarks)) {
			return false;
		}

		return visiblePlayers.equals(other.visiblePlayers);
	}
}