/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.worldmodel.impl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IPlayer;
import magma.util.FuzzyCompare;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Class to calculate advantageous positions to kick ball to.
 * 
 * @author Srinivasa Devanatthan
 */
public class InfluenceMap implements Serializable
{
	private static final long serialVersionUID = 829945431767039663L;

	private final IGlobalMap worldModel;

	/**
	 * The radius of the circle around the ball inside which the ball can be
	 * after the kick. Currently the player can kick 2.7m
	 */
	final double maxKickDistance = 2.7d;

	/**
	 * Constructor
	 * 
	 * @param worldModel Reference to world model object
	 */
	public InfluenceMap(IGlobalMap worldModel)
	{
		this.worldModel = worldModel;
	}

	/**
	 * Method returns a list of positions to kick the ball to, in the order of
	 * priority. Only for the ball! For player, use
	 * @return List
	 */
	public Vector3D getTargetKickPositionForBall()
	{
		// sort polygons based on some criteria and give the one with high
		// advantage
		Vector3D[] polyToKickTo = sortPolygons(getInfluenceArea(worldModel
				.getBall()));
		Vector3D kickTo;
		Vector3D ballPos = worldModel.getBall().getPosition();

		double x, y;
		if (polyToKickTo != null) {
			x = (polyToKickTo[1].getX() + polyToKickTo[2].getX()) / 2;
			y = (polyToKickTo[1].getY() + polyToKickTo[2].getY()) / 2;
			if (Math.abs(x - ballPos.getX()) < 0.25
					&& Math.abs(y - ballPos.getY()) < 0.25) {
				kickTo = worldModel.getOtherGoalPosition();
			} else {
				kickTo = new Vector3D(x, y, 0);
			}
		} else {
			kickTo = worldModel.getOtherGoalPosition();
		}
		return kickTo;
	}

	/**
	 * Calculates a semi-circle around the object where it has influence. Reduces
	 * the area if there is an opponent anywhere inside.
	 * 
	 * @param movableObject Reference to the object
	 * @return polygons around the object
	 */
	public List<Vector3D[]> getInfluenceArea(IMoveableObject movableObject)
	{
		Vector3D myPosition = movableObject.getPosition();

		List<Vector3D[]> influencePolygons = calculatePolygonsAroundObject(
				myPosition, maxKickDistance * 2);

		List<IPlayer> opps = this.worldModel.getOpponentsAtBallList();
		for (IPlayer opponent : opps) {

			// if player is farther away than twice our kick distance, we are
			// faster to
			// the ball even after the kick
			if (movableObject.getDistanceToXY(opponent) > maxKickDistance * 2) {
				continue;
			}

			for (Vector3D[] poly : influencePolygons) {
				if (isPlayerInsidePoly(opponent, poly)) {

					// halve the distance to playerPos for the last side of the
					// previous poly and the first side on the next poly.
					Angle left = movableObject.getDirectionTo(poly[1]);
					Angle right = movableObject.getDirectionTo(poly[2]);

					double distance = movableObject.getDistanceToXY(opponent) / 2;

					poly[1] = Geometry.getVectorAtDistanceAndAngle(myPosition,
							distance, left);
					poly[2] = Geometry.getVectorAtDistanceAndAngle(myPosition,
							distance, right);
				}
			}
		}
		return influencePolygons;
	}

	/**
	 * Sorts a list of polygons
	 * 
	 * @param polygons Polygons
	 * @return Sorted list
	 */
	private Vector3D[] sortPolygons(List<Vector3D[]> polygons)
	{
		List<PriorityComparator> sortedPolys = new ArrayList<PriorityComparator>();

		for (Vector3D[] poly : polygons) {
			PriorityComparator p = new PriorityComparator();
			p.setPoly(poly);
			sortedPolys.add(p);
		}

		Collections.sort(sortedPolys);
		// happens when ball is the goal/outside the field
		if (!sortedPolys.isEmpty()) {
			return sortedPolys.get(0).getPoly();
		}
		return null;
	}

	/**
	 * Implements Comparable, to compare two polygons and decide which is better
	 * based on area, angle to goal, distance to goal, team-mate in poly
	 */
	private class PriorityComparator implements Comparable<PriorityComparator>
	{
		Vector3D[] first;

		public void setPoly(Vector3D[] poly)
		{
			this.first = poly;
		}

		public Vector3D[] getPoly()
		{
			return this.first;
		}

		final List<IPlayer> us = worldModel.getTeammatesAtBall();

		@Override
		public int compareTo(PriorityComparator p)
		{
			// TODO another set of weights for offensive and defensive play

			// the weights attributed to each polygon.
			// Order of priority
			// Offensive - Goal, Team, Direction to goal, Somewhere else
			// Defensive - Somewhere else, Team, Direction To Goal, Goal
			// present weights are in the offensive.

			// 1. the polygon with the most area is better as the distance we cover
			// is also large
			double weightForArea = 1.0;

			// 2. compared to area, proximity is more important
			double weightForProximityAndAngleToGoal = 1.2;

			// 3. weight for teamMate > area + (proximity*2)
			double weightForTeamMate = 4.0;

			// 4. weight for goal > teamMate + area (if goal is in a
			// polygon, normally the proximity should also lie there). So the
			// comparison would be goal+proximity against teamMate+area
			double weightForGoalInPolygon = 7.0;

			double weight1 = 0.0;
			double weight2 = 0.0;

			Vector3D[] second = p.getPoly();

			double area1 = Geometry.areaOfTriangle(first);
			double area2 = Geometry.areaOfTriangle(second);
			if (FuzzyCompare.gt(area1, area2, 0.1)) {
				weight1 += weightForArea;
			} else if (FuzzyCompare.lt(area1, area2, 0.1)) {
				weight2 += weightForArea;
			}

			/** weight to the polygon which has our team mate */
			for (IPlayer teamMate : us) {
				if (isPlayerInsidePoly(teamMate, first)) {
					weight1 += weightForTeamMate;
				}
				if (isPlayerInsidePoly(teamMate, second)) {
					weight2 += weightForTeamMate;
				}
			}

			/** weight for goal inside the poly */
			Vector3D goal = worldModel.getOtherGoalPosition();

			if (Geometry.isInsidePolygon(goal, first)) {
				weight1 += weightForGoalInPolygon;
			} else if (Geometry.isInsidePolygon(goal, second)) {
				weight2 += weightForGoalInPolygon;
			}

			Vector3D mid1 = new Vector3D((first[1].getX() + first[2].getX()) / 2,
					(first[1].getY() + first[2].getY()) / 2);

			Vector3D mid2 = new Vector3D(
					(second[1].getX() + second[2].getX()) / 2,
					(second[1].getY() + second[2].getY()) / 2);

			/** add weights for distance to goal */
			double distance1 = mid1.subtract(goal).getNorm();
			double distance2 = mid2.subtract(goal).getNorm();
			if (distance1 < distance2) {
				weight1 += weightForProximityAndAngleToGoal;
			} else if (distance2 < distance1) {
				weight2 += weightForProximityAndAngleToGoal;
			}

			/** add weights by angle to other goal */
			double angle1 = Math.abs(goal.subtract(mid1).getAlpha());
			double angle2 = Math.abs(goal.subtract(mid2).getAlpha());
			if (FuzzyCompare.lt(angle1, angle2, 0.1)) {
				weight1 += weightForProximityAndAngleToGoal;
			} else if (FuzzyCompare.lt(angle2, angle1, 0.1)) {
				weight2 += weightForProximityAndAngleToGoal;
			}

			// the poly with more weight, the better. So we push it to the top of
			// the list.
			if (weight1 < weight2) {
				return 1;
			} else if (weight1 > weight2) {
				return -1;
			} else {
				return 0;
			}
		}
	}

	/**
	 * The semi-circle around the <code>position</code> in the forward direction
	 * (towards opponent's goal) is calculated as a list of polygons, with each
	 * polygon having an internal angle of 20deg and height twice the
	 * <code>distance</code>.
	 * 
	 * @param position Center coordinate
	 * @param distance Distance
	 * @return TODO can be moved to Geometry class. Can wait.
	 */
	private List<Vector3D[]> calculatePolygonsAroundObject(Vector3D position,
			double distance)
	{
		// the max points on left and right give the max angles at which we can
		// kick without kicking towards our goal
		Vector3D pointOnLeftBorder;
		Vector3D pointOnRightBorder;

		pointOnLeftBorder = new Vector3D(position.getX(), 6, 0);
		pointOnRightBorder = new Vector3D(position.getX(), -6, 0);

		Angle maxAngleLeft = Angle.rad(pointOnLeftBorder.subtract(position)
				.getAlpha());
		Angle maxAngleRight = Angle.rad(pointOnRightBorder.subtract(position)
				.getAlpha());

		// influence map for the ball is always the horizontal angel to the
		// opponents goal.
		Angle initialAngle = Angle.rad(worldModel.getOtherGoalPosition()
				.subtract(position).getAlpha());

		// the list of polygons around the object
		List<Vector3D[]> influencePoly = new ArrayList<Vector3D[]>();

		Vector3D[] poly;

		// create polygons on the left
		for (Angle a = initialAngle; !a.isLeftOf(maxAngleLeft);) {

			poly = new Vector3D[3];
			poly[0] = position;
			Vector3D temp = Geometry.getVectorAtDistanceAndAngle(position,
					distance, a);
			// rotating to the left. add 20 degrees.
			a = a.add(Angle.deg(20));
			poly[1] = Geometry.getVectorAtDistanceAndAngle(position, distance, a);
			// this is to keep all polygons points in clockwise direction
			poly[2] = temp;

			influencePoly.add(poly);
		}

		// create polygons on the right
		for (Angle a = initialAngle; !a.isRightOf(maxAngleRight);) {

			poly = new Vector3D[3];
			poly[0] = position;
			poly[1] = Geometry.getVectorAtDistanceAndAngle(position, distance, a);
			// rotating to the right. subtract 20 degrees.
			a = a.subtract(Angle.deg(20));
			poly[2] = Geometry.getVectorAtDistanceAndAngle(position, distance, a);

			influencePoly.add(poly);
		}

		return influencePoly;
	}

	/**
	 * Checks if the circle around the player (with radius as his collision
	 * distance and centre as his position) intersects with the passed poly.
	 * 
	 * @param player Reference to player object
	 * @param poly Polygon
	 * @return True if the condition is met
	 */
	private boolean isPlayerInsidePoly(IPlayer player, Vector3D[] poly)
	{
		Vector3D centre = player.getPosition();

		// if centre is inside poly - intersects
		if (Geometry.isInsidePolygon(centre, poly))
			return true;

		// if the distance to any of the poly points is less than circle radius -
		// intersects
		for (Vector3D p : poly) {
			if (p.subtract(centre).getNorm() < player.getCollisionDistance())
				return true;
		}

		// if the perpendicular distance to any of vertices is less than radius -
		// intersects
		int i1 = poly.length - 1;
		for (int i2 = 0; i2 < poly.length; i2++) {
			Vector3D p0 = poly[i1];
			Vector3D p1 = poly[i2];
			i1 = i2;

			double temp = (p1.getX() - p0.getX());
			if (temp == 0) {
				// happens when points of poly are on same x
				continue;
			}

			double m1 = (p1.getY() - p0.getY()) / temp;
			if (m1 == 0) {
				// happens when points of poly are on same y
				continue;
			}

			double m2 = -1 / m1;

			// x and y at the intersection of the side of the polygon to the
			// perpendicular from the centre of circle
			double x = ((m1 * p1.getX()) - p1.getY() - (m2 * centre.getX()) + centre
					.getY()) / (m1 - m2);

			double y = (m1 * x) - (m1 * p1.getX()) + p1.getY();

			// x and y as a vector
			Vector3D pointAtPerpendicular = new Vector3D(x, y, 0);

			if (pointAtPerpendicular.subtract(centre).getNorm() < player
					.getCollisionDistance())
				return true;
		}
		return false;
	}
}
