/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.impl;

import java.util.ArrayList;
import java.util.List;

import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IPlayer;
import magma.util.geometry.Angle;

import org.apache.commons.math.geometry.Vector3D;

/**
 * The KickDirectionProfiler provides an estimate of a preferred kick direction.
 * It therefore creates a profiles of the <tt>scanRange</tt> which is
 * represented as a histogram in 1 degree steps. The histogram represents a
 * (complex ;)) weighting of players, the opponent goal and (not yet, but maybe
 * sometime) the field diameters. The best kick direction is then determined by
 * the maximum value in the histogram.<br>
 * Known issues:<br>
 * 1. If the ball is directly in front of the opponents goal, but not central,
 * the choice of relating everything to the ball-to-goal-center-direction
 * revenges. But in such a case the agent should decide anyways to run with the
 * ball into the goal...
 * 
 * @author Stefan Glaser
 * 
 */
public class KickDirectionProfiler
{
	// private static final float INVALID_ANGLE = -1000000000;

	/**
	 * The weighting factor for the own player gaussian.
	 */
	public static final float OWN_PLAYER_WEIGHT = 1f;

	/**
	 * The weighting factor for the other player gaussian.
	 */
	public static final float OTHER_PLAYER_WEIGHT = 0.7f;

	/**
	 * The weighting factor for the initial gaussian distribution. (General goal
	 * attractiveness)
	 */
	public static final float INITIAL_GOAL_WEIGHT = 150f;

	/**
	 * This factor is used to give a extra weight the area of the histogram,
	 * which ends up in the goal. This factor is just used, if the center of the
	 * goal is in kickable distance.
	 */
	public static final float GOAL_SCORE_FACTOR = 1000f;

	/**
	 * The factor a histogram-value is multiplied in case of an obstacle in this
	 * direction.
	 */
	public static final float OBSTACLE_FACTOR = 0.1f;

	/**
	 * The factor to define the range of interest (used as factor for the
	 * maxKickDistance to define the maximum range, in which players are taken
	 * into account).
	 */
	public static final float ADDITIONAL_DISTANCE = 1.25f;

	/**
	 * The range in which the environment is scanned. Just values from 0 to 360
	 * are valid - even though 0 is senseless. The scan range is relative to the
	 * direction of the ball to the goal.<br>
	 * E.g., a scan range of 180 will result in a +/-90 degree range around the
	 * goal direction.
	 */
	public final int scanRange;

	/**
	 * The maximum possible distance we are able to kick the ball.
	 */
	public final float maxKickDistance;

	/**
	 * The current scan histogram.
	 */
	private float[] histogram;

	/**
	 * The currently best known direction to kick to.
	 */
	private Angle bestKickDirection;

	/**
	 * An array holding the initial values for the histogram, in order to allow
	 * easy copy instead of busy calculating.
	 */
	private transient final float[] initialHistogram;

	/**
	 * An array holding a discrete overlay function, which defines the kind of
	 * effect an own player has to its neighboring directions in the histogram.
	 * This discrete function is then weighted by several dynamic environment
	 * factors.
	 */
	private transient final float[] ownWeights;

	/**
	 * An array holding a discrete overlay function, which defines the kind of
	 * effect an opponent player has to its neighboring directions in the
	 * histogram. This discrete function is then weighted by several dynamic
	 * environment factors.
	 */
	private transient final float[] otherWeights;

	// just used temporary for histogram calculations
	// (just don't wanted to pass them everywhere...)
	private transient IMoveableObject ball;

	private transient Vector3D goalPos;

	public KickDirectionProfiler(int scanRange, float maxKickDistance)
	{
		this.scanRange = scanRange;
		this.maxKickDistance = maxKickDistance;
		histogram = new float[scanRange];

		// create gaussian weight histograms
		initialHistogram = createGaussian(scanRange, 30, false);
		ownWeights = createGaussian(25, 3, false);
		otherWeights = createGaussian(21, 3, true);
	}

	/**
	 * @return kick profile histogram
	 */
	public float[] getKickProfile()
	{
		return histogram;
	}

	/**
	 * @return best kick direction
	 */
	public Angle getBestKickDirection()
	{
		return bestKickDirection;
	}

	/**
	 * Update histogram to new situation.
	 * 
	 * @param worldModel
	 */
	public void updateProfile(IGlobalMap worldModel)
	{
		// Fetch new values from the worldModel
		goalPos = worldModel.getOtherGoalPosition();
		ball = worldModel.getBall();

		// start new calculation of the kick profile
		// initialize the histogram with a basic attractiveness of the goal
		initializeHistogram();

		// add the weights for the team mates to the histogram
		weightPlayers(worldModel);

		// extra weight the goal range (if the distance to the goal is kickable)
		weightGoal();

		// give directions which point outside the soccer field extra bad values
		clearOutOfSite();

		// some beautification
		normalizeHistogram();

		// Choose maximum histogram intensity as preferred kick direction
		determineBestKickDirection();
	}

	/**
	 * This method initializes the overall histogram with the
	 * initializeHistogram. The initial attractiveness is weighted by
	 * <tt>INITIAL_GOAL_WEIGHT/distanceToGoal</tt>.
	 */
	private void initializeHistogram()
	{
		float distanceToGoal = (float) ball.getDistanceToXY(goalPos);

		for (int i = 0; i < histogram.length; i++) {
			histogram[i] = initialHistogram[i] * INITIAL_GOAL_WEIGHT
					/ distanceToGoal;
		}
	}

	/**
	 * This method adds a dynamic calculated weight to all relevant players to
	 * the overall histogram.
	 */
	private void weightPlayers(IGlobalMap worldModel)
	{
		List<IPlayer> players = fetchRelevantPlayers(worldModel);
		Angle ballToGoalAngle = ball.getDirectionTo(goalPos);
		Angle ballToPlayerAngle;
		int indexAngle;

		// first produce some overlay array (here a gaussian is used) to each
		// player and add it to the histogram
		for (IPlayer p : players) {
			// fetch player to goal angle
			ballToPlayerAngle = ball.getDirectionTo(p);

			// calculate index in histogram based on playerToGoalAngle and
			// ballToGoalAngle
			indexAngle = ((int) ballToPlayerAngle.subtract(ballToGoalAngle)
					.degrees()) + (histogram.length / 2);
			if (indexAngle < 0 || indexAngle >= histogram.length) {
				// if the angle difference results in an invalid index (and is
				// therefore not in the histogram range), continue with next player
				continue;
			} else {
				// if the angle is in the histogram range, choose corresponding
				// weights and add the player to the histogram
				addPlayerToHistogram(p, indexAngle, p.isOwnTeam() ? ownWeights
						: otherWeights);
				addObstacleToHistogram(p, indexAngle);
				// anglesToClear.add(indexAngle);
			}
		}

		// then divide the histogram values to direct obstacles by some factor to
		// reduce attractiveness of such angles
		for (IPlayer p : players) {
			// fetch player to goal angle
			ballToPlayerAngle = ball.getDirectionTo(p);

			// calculate index in histogram based on playerToGoalAngle and
			// ballToGoalAngle
			indexAngle = ((int) ballToPlayerAngle.subtract(ballToGoalAngle)
					.degrees()) + (histogram.length / 2);
			if (indexAngle < 0 || indexAngle >= histogram.length) {
				// if the angle difference results in an invalid index (and is
				// therefore not in the histogram range), continue with next player
				continue;
			} else {
				// if the angle is in the histogram range, clear out it's direct
				// angles
				addObstacleToHistogram(p, indexAngle);
			}
		}
	}

	/**
	 * Fetch all players in maxKickDistance * ADDITIONAL_DISTANCE.
	 * 
	 * @param worldModel the worldModel
	 * @return a list of relevant players
	 */
	private List<IPlayer> fetchRelevantPlayers(IGlobalMap worldModel)
	{
		List<IPlayer> players = new ArrayList<IPlayer>();
		float distance = maxKickDistance * ADDITIONAL_DISTANCE;

		for (IPlayer p : worldModel.getVisiblePlayers()) {
			if (p.getDistanceToXY(ball) < distance) {
				players.add(p);
			}
		}

		return players;
	}

	/**
	 * Weight a specific <tt>player</tt> in the histogram. This method first
	 * calculates a general weight based on the team (own/opponent), the
	 * difference to the goal angle (in case of an own player) and the distance
	 * of the player to the maxKickDistance. This weight is then used to weight
	 * the function overlay (the <tt>weights</tt> array) to a player.
	 * 
	 * @param player the player to weight
	 * @param index the corresponding angle/index in the histogram
	 * @param weights the weights which should be added to the histogram
	 */
	private void addPlayerToHistogram(IPlayer player, int index, float[] weights)
	{
		// calculate overall weight factor
		float weight = OTHER_PLAYER_WEIGHT;

		// if the player is a team mate, we use a different base weight and we
		// also take the so called niceFactor into account.
		if (player.isOwnTeam()) {
			weight = OWN_PLAYER_WEIGHT;
			float distanceToGoal = (float) ball.getDistanceToXY(goalPos);

			// niceFactor [0.4 ... 1] is defined as:
			// 0.4 by the biggest difference of the goal direction and the
			// player direction (+/- 180 deg difference to goal direction)
			// towards 1, the smaller the gap to the goal direction is
			//
			// The niceFactor is also dependent on the distance of the ball to the
			// goal! The values described above are present in a distance to the
			// goal of 3 meter - if the ball has a distance of 6 meters to the
			// goal, the range is reduced to 0.7 to 1. Ergo, the farer the
			// ball is away from the goal, the less interesting it is, to play it
			// towards players in goal direction and the passing range should be
			// wider.
			weight *= 1f - ((Math.abs(index - (histogram.length / 2)) / (180 * 1.75f)) / (distanceToGoal / 3));
		}

		// distanceFactor [0 ... 1] is defined as:
		// 0 for twice the max kick distance as the relative distance to the
		// max kick distance
		// towards 1, the closer a player is located to the max kick distance
		weight *= 1f - ((float) Math.abs(ball.getDistanceToXY(player)
				- maxKickDistance) / (maxKickDistance * 2f));

		// calculate start position for histogram overlay
		int weightsHalfLength = weights.length / 2;
		int start = index - weightsHalfLength;
		if (start < 0) {
			start = 0;
		}

		// calculate end position for histogram overlay
		int end = index + weightsHalfLength;
		if (end > histogram.length) {
			end = histogram.length;
		}

		// update histogram values
		for (int i = start; i < end; i++) {
			histogram[i] += weights[i - index + weightsHalfLength] * weight;
		}
	}

	/**
	 * Pulls down all angles which directly point to a player. The range how many
	 * angle-values are pulled down is 3 for own players and dynamic for an
	 * opponent. The range to an opponent is enlarged / reduced with respect to
	 * it's distance to the ball.
	 * 
	 * @param player
	 * @param index
	 */
	private void addObstacleToHistogram(IPlayer player, int index)
	{
		int width = 3;

		if (!player.isOwnTeam()) {
			// distanceFactor [0 ... 3] is defined as:
			// 0 for 1.5 times the max kick distance as distance to the ball
			// towards 3, the closer a player is located to the ball
			float weight = 3f - ((float) ball.getDistanceToXY(player) / (maxKickDistance * .5f));
			weight *= weight; // take exponential weight
			width *= weight; // adjust the width of obstacle
		}

		// calculate start position
		int halfWidth = width / 2;
		int start = index - halfWidth;
		if (start < 0) {
			start = 0;
		}

		// calculate end position
		int end = index + halfWidth;
		if (end > histogram.length) {
			end = histogram.length;
		}

		// update histogram values
		for (int i = start; i < end; i++) {
			histogram[i] = histogram[i] * OBSTACLE_FACTOR; // INVALID_ANGLE;
		}
	}

	/**
	 * Weight the goal ludicrously high, if it's in kickable distance.
	 */
	private void weightGoal()
	{
		if ((ball.getDistanceToXY(goalPos) - maxKickDistance) < 0) {
			Angle ballToGoalAngle = ball.getDirectionTo(goalPos);
			// TODO: Remove dependency on hard coded goal size
			int a1 = (int) ball
					.getDirectionTo(goalPos.add(new Vector3D(0, -1.05, 0)))
					.subtract(ballToGoalAngle).degrees()
					+ (histogram.length / 2);
			int a2 = (int) ball
					.getDirectionTo(goalPos.add(new Vector3D(0, 1.05, 0)))
					.subtract(ballToGoalAngle).degrees()
					+ (histogram.length / 2);

			// if direction is switched (behind the goal) the indexes have to be
			// switched as well
			if (a2 < a1) {
				int temp = a2;
				a2 = a1;
				a1 = temp;
			}

			a1 = a1 < 0 ? 0 : a1;
			a2 = a2 >= histogram.length ? histogram.length - 1 : a2;

			for (int i = a1; i < a2; i++) {
				histogram[i] *= GOAL_SCORE_FACTOR;
			}
		}
	}

	/**
	 * Check all angles if they result in a position outside the field and cancel
	 * them accordingly.
	 */
	private void clearOutOfSite()
	{

	}

	/**
	 * Normalize the histogram.<br>
	 * This method brings the histogram in the following range: [0 ... 1], where
	 * 0 are not allowed kick directions and 1 is the preferred direction.
	 */
	private void normalizeHistogram()
	{
		int minIndex = -1;
		int maxIndex = 0;

		for (int i = 0; i < histogram.length; i++) {
			if ((minIndex < 0 || histogram[i] < histogram[minIndex])
			// && histogram[i] > INVALID_ANGLE
			) {
				minIndex = i;
			}

			if (histogram[i] > histogram[maxIndex]) {
				maxIndex = i;
			}
		}

		float range = histogram[maxIndex] - histogram[minIndex];
		float minValue = histogram[minIndex];

		for (int i = 0; i < histogram.length; i++) {
			// // Set all bad angles to 0
			// if (histogram[i] <= INVALID_ANGLE) {
			// histogram[i] = 0;
			// continue;
			// }

			// normalize histogram values
			histogram[i] = ((histogram[i] - minValue) / range) + 0.01f;
		}
	}

	/**
	 * Calculate best kick direction based on the maximum value in the histogram.
	 */
	private void determineBestKickDirection()
	{
		int bestIndex = 0;

		for (int i = 0; i < histogram.length; i++) {
			if (histogram[i] > histogram[bestIndex]) {
				bestIndex = i;
			}
		}

		bestKickDirection = ball.getDirectionTo(goalPos).add(
				Math.toRadians(bestIndex - (histogram.length / 2)));
	}

	/**
	 * Creates a float array with length <tt>width</tt>, containing a gaussian
	 * distribution with <tt>sigma</tt>. The resulting values in the array are
	 * normalized to give a sum of 1.<br>
	 * If <tt>negateValues</tt> is set to <tt>true</tt>, all values in the array
	 * are negated.
	 * 
	 * @param width the length of the resulting array
	 * @param sigma for gauss function
	 * @param negateValues flag, if resulting values should be negated
	 * @return a float array containing the discrete gaussian distribution with
	 *         the maximum in the center of the array
	 */
	private float[] createGaussian(int width, float sigma, boolean negateValues)
	{
		float[] res = new float[width];
		int halfWidth = width / 2;
		int evenIndexCorrector = width % 2 == 0 ? 1 : 0;
		int negator = negateValues ? -1 : 1;
		double sum = 0;

		for (int i = 0; i <= halfWidth - evenIndexCorrector; i++) {
			res[halfWidth + i] = res[halfWidth - i - evenIndexCorrector] = negator
					* getGaussian(i, sigma);
			sum += res[halfWidth + i];
		}

		// Normalize gaussian values to the sum of 1
		// since we just counted one half of the array and the gaussian is
		// symmetric, simply double the result
		sum *= 2;
		// if the array is of odd length, the value in the middle has been counted
		// twice - therefore subtract it once again
		if (width % 2 == 1) {
			sum -= res[halfWidth];
		}

		// if we produced negative values, the sum is also negative and will turn
		// the result again positive
		sum = Math.abs(sum);

		// Normalize each value
		for (int i = 0; i <= halfWidth - evenIndexCorrector; i++) {
			res[halfWidth + i] /= sum;
		}

		return res;
	}

	/**
	 * Returns a gaussian value at x.
	 * 
	 * @param x
	 * @param sigma
	 * @return gauss value at x
	 */
	private float getGaussian(int x, float sigma)
	{
		return (float) ((1 / (sigma * Math.sqrt(2 * Math.PI))) * Math
				.exp(-(x * x) / (2 * sigma * sigma)));
	}
}
