/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.impl;

import magma.agent.worldmodel.IMoveableObject;
import magma.util.FuzzyCompare;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Base class for all movable objects on the field
 * 
 * @author Klaus Dorer
 */
public abstract class MovableObject extends VisibleObject implements
		IMoveableObject
{
	/**
	 * Constructor
	 * 
	 * @param name Object name
	 */
	MovableObject(String name)
	{
		super(name);
	}

	private static final long serialVersionUID = 1L;

	/** the speed calculated from previous and current positions */
	private Vector3D speed;

	/**
	 * Returns a list of future positions as predicted by the speed
	 * @param aHowMany the number of future positions that should be returned
	 * @return a list of future positions
	 */
	@Override
	public Vector3D[] getFuturePositions(int aHowMany)
	{
		Vector3D[] futurePositions = new Vector3D[aHowMany];
		futurePositions[0] = getPosition().add(getSpeed());

		for (int i = 1; i < aHowMany; i++) {
			futurePositions[i] = futurePositions[i - 1].add(getSpeed());
		}
		return futurePositions;
	}

	/**
	 * @return the speed of this object
	 */
	@Override
	public Vector3D getSpeed()
	{
		// we don't want null pointers!
		if (speed == null)
			speed = new Vector3D(0.0, 0.0, 0.0);
		return speed;
	}

	@Override
	public void update(Vector3D localPosition, Vector3D globalPosition,
			float time)
	{
		// WARNING: getLastSeenTime() MUST be called before super.update(),
		// because
		// super.update() updates the lastSeenTime with the current time
		float lastSeenTime = getLastSeenTime();
		super.update(localPosition, globalPosition, time);
		calculateSpeed(lastSeenTime, time);
	}

	/**
	 * Calculate the maximum possible speed this object can move at
	 * 
	 * @return Possible speed
	 */
	double getPossibleSpeed()
	{
		return 0.0;
	}

	/**
	 * Calculate the speed this object is moving at
	 * 
	 * @param lastSeenTime Timestamp this object was last seen
	 * @param time Current timestamp
	 */
	void calculateSpeed(float lastSeenTime, float time)
	{
		// need previousPosition, hence speed is calculated after update
		if (getPreviousPosition() != null) {
			Vector3D checkSpeed = new Vector3D();
			checkSpeed = getPosition().subtract(getPreviousPosition());
			if (time - lastSeenTime > 1)
				checkSpeed = new Vector3D(
						checkSpeed.getX() / (time - lastSeenTime), checkSpeed.getY()
								/ (time - lastSeenTime), checkSpeed.getZ()
								/ (time - lastSeenTime));

			// we don't set speed for impossible values
			if (checkSpeed.getNorm() < getPossibleSpeed()) {
				speed = checkSpeed;
			} else if (checkSpeed.getNorm() < 0.1) {
				speed = new Vector3D(0.0, 0.0, 0.0);
			}

		} else {
			speed = new Vector3D(0.0, 0.0, 0.0);
		}
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof MovableObject)) {
			return false;
		}
		MovableObject other = (MovableObject) o;
		if (!super.equals(other)) {
			return false;
		}
		return FuzzyCompare.eq(speed, other.speed, 0.00001f);
	}
}