/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.impl;

import java.util.HashMap;
import java.util.Map;

import javax.vecmath.Matrix3d;

import magma.agent.IMagmaConstants;
import magma.agent.worldmodel.IPlayer;
import magma.util.geometry.Angle;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Represents a player agent on the field
 */
public class Player extends MovableObject implements IPlayer
{
	/** id of this class for serialization */
	private static final long serialVersionUID = 4856779909555166046L;

	/** the player number */
	private final int id;

	/** the team name of this player */
	private final String teamname;

	/** true if this is a player of our own team */
	private final boolean ownTeam;

	/** whether the player is laying */
	private boolean isLaying;

	/** a timestamp for how long we were laying */
	private float layingTimestamp;

	/** maximum speed the player can achieve running forward (m/s) */
	protected float forwardSpeed;

	/** maximum speed the player can turn (deg/s) */
	protected float turnSpeed;

	/** maximum speed the player can step sidewise */
	protected float sideStepSpeed;

	/** maximum speed the player can step backwards */
	private final float backwardSpeed;

	/** the orientation of the player relative to the global coordinate system */
	protected Matrix3d orientation;

	/** time of last horizontal angle measurement */
	private float lastOrientationMeasurement;

	/** list of body parts */
	protected final HashMap<String, Vector3D> bodyParts;

	/**
	 * Constructor
	 * 
	 * @param id The number of the player on the field
	 * @param teamname The name of the team this player plays in
	 * @param ownTeam True if this is a player of our own team
	 */
	public Player(int id, String teamname, boolean ownTeam)
	{
		super(teamname + id);
		this.id = id;
		this.teamname = teamname;
		this.ownTeam = ownTeam;
		this.isLaying = false;
		this.layingTimestamp = 0.0f;
		this.bodyParts = new HashMap<String, Vector3D>();
		orientation = new Matrix3d(1, 0, 0, 0, 1, 0, 0, 0, 1);
		lastOrientationMeasurement = 0.0f;

		// TODO: measure for own players and observe for other players
		forwardSpeed = 0.3f;
		turnSpeed = 20.0f;
		sideStepSpeed = 0.01f;
		backwardSpeed = 0.05f;
	}

	@Override
	public int getID()
	{
		return id;
	}

	@Override
	public String getTeamname()
	{
		return teamname;
	}

	@Override
	public double getPossibleSpeed()
	{
		return 2.0;
	}

	/**
	 * Returns the relative angle (rad) this player's torso has to the specified
	 * position. As long as we do not know other player's heading this method
	 * returns 0.
	 * @param position the position to which to calculate the body angle
	 * @return the relative angle (rad) this player's torso has to the specified
	 *         position
	 */
	public Angle getBodyDirectionTo(Vector3D position)
	{
		if (lastOrientationMeasurement > 0) {
			return getDirectionTo(position).subtract(getHorizontalAngle());
		}

		// if we do not know heading, we assume player is faced to direction
		return Angle.RAD0;
	}

	@Override
	public boolean equals(Object o)
	{
		if (!(o instanceof Player)) {
			return false;
		}

		Player other = (Player) o;
		if (id != other.id && isOwnTeam() == other.isOwnTeam()) {
			return false;
		}

		return teamname.equals(other.teamname);
	}

	@Override
	public boolean isOwnTeam()
	{
		return ownTeam;
	}

	@Override
	public boolean isGoalie()
	{
		return id == 1;
	}

	@Override
	public Angle getHorizontalAngle()
	{
		// calculate view-angle
		Angle horizontalViewAngle = Angle.rad(Math.asin(-orientation.m01));

		if (orientation.m11 < 0) {
			horizontalViewAngle = Angle.RAD180.subtract(horizontalViewAngle);
		}

		return horizontalViewAngle;
	}

	/**
	 * Set player orientation in 3-dimensional space.
	 * 
	 * @param orientation - new orientation
	 */
	public void setOrientation(Matrix3d orientation)
	{
		this.orientation = orientation;
		lastOrientationMeasurement = getLastSeenTime();
	}

	@Override
	public Matrix3d getOrientation()
	{
		return this.orientation;
	}

	/**
	 * Calculates the time a player is expected to take to get from its current
	 * position to the passed position heading in the passed direction running
	 * forward
	 * @param position the destination position (global)
	 * @param directionAtTarget desired body direction (global, rad) at the
	 *        destination position
	 * @return time in seconds we estimate is needed to get to the specified
	 *         position with the specified direction
	 */
	@Override
	public float getTimeToTurnAndRun(Vector3D position, Angle directionAtTarget)
	{
		return getTimeForMoving(position, directionAtTarget, forwardSpeed, 0);
	}

	/**
	 * Calculates the time a player is expected to take to get from its current
	 * position to the passed position heading in the passed direction running
	 * forward
	 * @param position the destination position (global)
	 * @param directionAtTarget desired body direction (global, rad) at the
	 *        destination position
	 * @return time in seconds we estimate is needed to get to the specified
	 *         position with the specified direction
	 */
	@Override
	public float getTimeForBackStep(Vector3D position, Angle directionAtTarget)
	{
		return getTimeForMoving(position, directionAtTarget, backwardSpeed, -180);
	}

	/**
	 * Calculates the time a player is expected to take to get from its current
	 * position to the passed position heading in the passed direction by side
	 * stepping
	 * @param position the destination position (global)
	 * @param directionAtTarget desired body direction (global, rad) at the
	 *        destination position
	 * @return time in seconds we estimate is needed to get to the specified
	 *         position with the specified direction
	 */
	@Override
	public float getTimeForSideStep(Vector3D position, Angle directionAtTarget,
			boolean left)
	{
		if (left) {
			return getTimeForMoving(position, directionAtTarget, sideStepSpeed, 90);
		}

		return getTimeForMoving(position, directionAtTarget, sideStepSpeed, -90);
	}

	/**
	 * Calculates the time a player is expected to take to get from its current
	 * position to the passed position heading in the passed direction
	 * @param position the destination position (global)
	 * @param directionAtTarget desired body direction (global, rad) at the
	 *        destination position
	 * @param speed the average speed we can achieve moving
	 * @param heading the body direction in which to move (forward 0, left 90,
	 *        right -90, backward -180)
	 * @return time in seconds we estimate is needed to get to the specified
	 *         position with the specified direction
	 */
	private float getTimeForMoving(Vector3D position, Angle directionAtTarget,
			float speed, float heading)
	{
		// time for turning into side step direction
		Angle turnAngleNow = Angle.deg(Math.abs(getBodyDirectionTo(position)
				.degrees()) - heading);
		float time = getTimeForTurn(turnAngleNow);

		// time for stepping there
		float distanceToTarget = (float) getDistanceToXY(position);
		time += distanceToTarget / speed;

		// time for turning into desired direction
		Angle turnAngleAtTarget = directionAtTarget.subtract(Angle
				.deg(getDirectionTo(position).degrees() - heading));
		time += getTimeForTurn(turnAngleAtTarget);

		return time;
	}

	/**
	 * @param angleToTurn the angle to turn
	 * @return time it needs (in s) to turn the specified angle
	 */
	private float getTimeForTurn(Angle angleToTurn)
	{
		// TODO: may be we should use a stepwise linear function
		float angle = (float) Math.abs(angleToTurn.degrees());
		return angle / turnSpeed;
	}

	/**
	 * @return the isLaying
	 */
	@Override
	public boolean isLaying()
	{
		return isLaying;
	}

	/**
	 * Sets whether we're laying or not
	 * @param time
	 */
	void setLaying(float time)
	{
		boolean newIsLaying;

		if (bodyParts.size() < 3) {
			isLaying = false;
			layingTimestamp = time;
			return;
		}

		double minZ = Double.MAX_VALUE;
		double maxZ = Double.MIN_VALUE;

		for (Vector3D pos : bodyParts.values()) {
			if (pos.getZ() < minZ) {
				minZ = pos.getZ();
			}
			if (pos.getZ() > maxZ) {
				maxZ = pos.getZ();
			}
		}

		newIsLaying = maxZ - minZ < 0.25;

		if (!newIsLaying && isLaying) {
			if ((time - layingTimestamp) < IMagmaConstants.TIMEDELAY_LAYING) {
				newIsLaying = true;
			}
		} else if (newIsLaying && isLaying) {
			layingTimestamp = time;
		}

		isLaying = newIsLaying;

	}

	/**
	 * Get position of bodyPart
	 * @return the bodyPart
	 */
	public Vector3D getBodyPart(String name)
	{
		return bodyParts.get(name);
	}

	/**
	 * Set visible body parts
	 * 
	 * @param allBodyParts List of visible body parts
	 */
	public void setBodyparts(Map<String, Vector3D> allBodyParts)
	{
		bodyParts.clear();
		if (allBodyParts != null) {
			bodyParts.putAll(allBodyParts);
		}
	}

	@Override
	public double getCollisionDistance()
	{
		if (isOwnTeam()) {
			return 1.0;
		}

		if (isLaying()) {
			return 0.5;
		}

		return 0.3;
	}

	@Override
	public void update(Vector3D localPosition, Vector3D globalPosition,
			float time)
	{
		super.update(localPosition, globalPosition, time);
		setLaying(time);
	}

}