/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.impl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import magma.agent.worldmodel.IPositionManager;
import magma.util.geometry.Angle;
import magma.util.geometry.PositionOrientation;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Manages desired positions of a player
 * 
 * @author Klaus Dorer
 */
public class PositionManager implements IPositionManager, Serializable
{
	/** id of this class for serialization */
	private static final long serialVersionUID = 3307599892830771538L;

	/** positions this player wants to be at */
	private List<PositionOrientation> desiredPositions;

	/** true if the desired Position is set from outside the agent */
	private boolean desiredPositionEnforcedFromOutside;

	/**
	 * Constructor
	 */
	public PositionManager()
	{
		desiredPositions = new ArrayList<PositionOrientation>();
		desiredPositionEnforcedFromOutside = false;
	}

	/**
	 * @return the desiredPositions
	 */
	@Override
	public List<PositionOrientation> getDesiredPositions()
	{
		// creating a new list to avoid changes and concurrent modifications
		return new ArrayList<PositionOrientation>(desiredPositions);
	}

	/**
	 * @return the last position orientation in the list
	 */
	@Override
	public PositionOrientation getFinalPosition()
	{
		if (desiredPositions.isEmpty()) {
			return null;
		}

		return desiredPositions.get(desiredPositions.size() - 1);
	}

	/**
	 * @param desiredPositions the desiredPositions to set
	 */
	public void setDesiredPositions(List<PositionOrientation> desiredPositions)
	{
		this.desiredPositions = desiredPositions;
	}

	/**
	 * Set desired position and rotation
	 * 
	 * @param desiredPosition Desired position
	 * @param desiredRotation Desired rotation
	 * @param enforceFromOutside True if the position is set from outside the
	 *        agent and may not be changed by the agent itself
	 */
	@Override
	public void setDesiredPosition(Vector3D desiredPosition,
			Angle desiredRotation, boolean enforceFromOutside)
	{
		if (desiredPosition != null) {
			PositionOrientation newElement = new PositionOrientation(
					desiredPosition, desiredRotation);
			setDesiredPosition(newElement, enforceFromOutside);
		}
	}

	/**
	 * @param posOrientation the position and orientation of desired position
	 * @param enforceFromOutside true if the position should not be changed
	 */
	@Override
	public void setDesiredPosition(PositionOrientation posOrientation,
			boolean enforceFromOutside)
	{
		if (!desiredPositionEnforcedFromOutside || enforceFromOutside) {
			this.desiredPositions.clear();
			if (posOrientation != null) {
				desiredPositions.add(posOrientation);
			}
		}
		if (enforceFromOutside) {
			desiredPositionEnforcedFromOutside = enforceFromOutside;
		}
	}

	/**
	 * adds an intermediate position (with orientation) to the list of positions
	 * @param index the index where to add the new position
	 * @param pos the position orientation to add
	 */
	@Override
	public void addDesiredPosition(int index, PositionOrientation pos)
	{
		if (pos != null) {
			desiredPositions.add(index, pos);
			if (desiredPositions.size() > 2) {
				System.out.println("Fehler: " + desiredPositions.size());
			}
		}
	}

	/**
	 * Removes all desired positions from the list
	 */
	@Override
	public void clear()
	{
		desiredPositions.clear();
	}
}
