/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.impl;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

import magma.agent.worldmodel.IThisPlayer;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.MatrixUtil;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Container for this agent's information that is unique to the player this
 * agent is representing
 * 
 * @author Klaus Dorer, Stefan Glaser
 */
public class ThisPlayer extends Player implements IThisPlayer
{
	/** id of this class for serialization */
	private static final long serialVersionUID = 3307599892830771538L;

	/** manager for desired positions */
	private final PositionManager positionManager;

	/**
	 * Constructor
	 * 
	 * @param teamname Team name
	 * @param id Player number
	 */
	public ThisPlayer(String teamname, int id)
	{
		super(id, teamname, true);

		positionManager = new PositionManager();
	}

	@Override
	public void update(Vector3D localPosition, Vector3D globalPosition,
			float time)
	{
		// The local position of ourself should never change... therefore we keep
		// it the same.
		super.update(this.localPosition, globalPosition, time);
	}

	@Override
	public Vector3D calculateGlobalPosition(Vector3D localPos)
	{
		Vector3d visionVector = new Vector3d(-localPos.getY(), localPos.getX(),
				localPos.getZ());

		visionVector = MatrixUtil.mulMatrixVector(orientation, visionVector);

		Vector3D res = new Vector3D(visionVector.getY(), -visionVector.getX(),
				visionVector.getZ());

		// create a flag-vector
		return res.add(position);
	}

	@Override
	public Vector3D calculateGlobal2DPosition(Vector3D localPos)
	{
		Vector3d visionVector = new Vector3d(-localPos.getY(), localPos.getX(),
				localPos.getZ());

		Matrix3d zOrientation = new Matrix3d();
		zOrientation.rotZ(getHorizontalAngle().radians());

		visionVector = MatrixUtil.mulMatrixVector(zOrientation, visionVector);

		Vector3D res = new Vector3D(visionVector.getY(), -visionVector.getX(),
				visionVector.getZ());

		return res.add(position);
	}

	@Override
	public Vector3D calculateLocalPosition(Vector3D globalPosition)
	{
		Vector3D localPos = globalPosition.subtract(position);

		Vector3d visionVector = new Vector3d(-localPos.getY(), localPos.getX(),
				localPos.getZ());

		// The inverse of a rotation matrix is it's transpose...
		Matrix3d inverseOrientation = new Matrix3d(orientation);
		inverseOrientation.transpose();

		visionVector = MatrixUtil.mulMatrixVector(inverseOrientation,
				visionVector);

		return new Vector3D(visionVector.getY(), -visionVector.getX(),
				visionVector.getZ());
	}

	/**
	 * Returns the relative angle (rad) this player's torso has to the specified
	 * position
	 * @param position the position to which to calculate the body angle
	 * @return the relative angle (rad) this player's torso has to the specified
	 *         position
	 */
	@Override
	public Angle getBodyDirectionTo(Vector3D position)
	{
		return getDirectionTo(position).subtract(getHorizontalAngle());
	}

	/**
	 * Checks if the passed 2D xy-point (absolute coordinate system) is inside
	 * the passed polygon (xy, relative coordinate system)
	 * @param absolutePosition 2D vector (x,y) of the point to check in the
	 *        global coordinate system
	 * @param polygon corner points of the polygon (x, y), in this player's
	 *        coordinate system and in clockwise order. The coordinates should
	 *        NOT take neck angles into account, i.e. assume head is straight.
	 * @return true if the point is inside the polygon
	 */
	@Override
	public boolean isInsidePolygonXY(Vector3D absolutePosition,
			Vector3D[] polygon)
	{
		Vector3D[] polygonGlobal = new Vector3D[polygon.length];
		for (int i = 0; i < polygon.length; i++) {
			polygonGlobal[i] = calculateGlobalPosition(polygon[i]);
		}

		return Geometry.isInsidePolygon(absolutePosition, polygonGlobal);
	}

	/**
	 * @param position the position to check
	 * @return true if the specified position is left of this player with respect
	 *         to the root body orientation
	 */
	@Override
	public boolean positionIsLeft(Vector3D absolutePosition)
	{
		return calculateLocalPosition(absolutePosition).getY() >= 0;
	}

	/**
	 * @param position the position to check
	 * @return true if the specified position is right of this player with
	 *         respect to the root body orientation
	 */
	@Override
	public boolean positionIsRight(Vector3D absolutePosition)
	{
		return calculateLocalPosition(absolutePosition).getY() <= 0;
	}

	/**
	 * @param position the position to check
	 * @return true if the specified position is behind this player with respect
	 *         to the root body orientation
	 */
	@Override
	public boolean positionIsBehind(Vector3D absolutePosition)
	{
		return calculateLocalPosition(absolutePosition).getX() <= 0;
	}

	/**
	 * @return the positionManager
	 */
	@Override
	public PositionManager getPositionManager()
	{
		return positionManager;
	}
}
