/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.impl;

import java.io.Serializable;
import java.util.logging.Level;
import java.util.logging.Logger;

import magma.agent.worldmodel.IVisibleObject;
import magma.util.FuzzyCompare;
import magma.util.geometry.Angle;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Base class for all visible objects on the field
 */
public abstract class VisibleObject implements IVisibleObject, Serializable
{
	// handle for logging
	protected static final Logger logger = Logger.getLogger("");

	private static final long serialVersionUID = 1L;

	private final String name;

	/** the last perceived local position in the root body system */
	protected Vector3D localPosition;

	/** the current position in the global coordinate system */
	protected Vector3D position;

	/** the previous position in the global coordinate system */
	protected Vector3D previousPosition;

	/** time stamp when this object was last updated by perception */
	protected float lastSeenTime;

	/** true if this object was seen in last perception */
	private boolean visible;

	/**
	 * Default constructor, initializes all fields to zero/null
	 * 
	 * @param name Object name
	 */
	public VisibleObject(String name)
	{
		this.name = name;
		localPosition = new Vector3D();
		position = new Vector3D();
		previousPosition = new Vector3D();
		visible = false;
	}

	@Override
	public String getName()
	{
		return name;
	}

	@Override
	public Vector3D getPosition()
	{
		return position;
	}

	/**
	 * Setting this objects position is usually done when updating the object.
	 * This method is used to set the ball position from hear message only.
	 * 
	 * @param pos New position
	 */
	public void setPosition(Vector3D pos)
	{
		position = pos;
	}

	@Override
	public Vector3D getPreviousPosition()
	{
		return previousPosition;
	}

	@Override
	public float getLastSeenTime()
	{
		return lastSeenTime;
	}

	/**
	 * Updates this object with the latest perception
	 * 
	 * @param localPosition - the local position as observed in the root body
	 *        system
	 * @param globalPosition - the calculated global position of this visible
	 *        object
	 * @param time - time the current absolute time
	 */
	public void update(Vector3D localPosition, Vector3D globalPosition,
			float time)
	{
		this.localPosition = localPosition;

		previousPosition = position;
		position = globalPosition;

		logger.log(
				Level.FINEST,
				"position of: {0} ({1}, {2}, {3})",
				new Object[] { name, position.getX(), position.getY(),
						position.getZ() });

		visible = true;
		lastSeenTime = time;
	}

	/**
	 * @param other the point to which to calculate
	 * @return the distance projection to the x y plane between this object and
	 *         the passed coordinate
	 */
	@Override
	public double getDistanceToXY(Vector3D other)
	{
		Vector3D delta = getPosition().subtract(other);
		return new Vector3D(delta.getX(), delta.getY(), 0).getNorm();
	}

	/**
	 * @param other the point to which to calculate
	 * @return the distance projection to the x y plane between this object and
	 *         the passed coordinate
	 */
	@Override
	public double getDistanceToXY(IVisibleObject other)
	{
		return getDistanceToXY(other.getPosition());
	}

	/**
	 * Calculates the distance of this visible object to the passed one
	 * @param other the object to which to calculate distance
	 * @return the distance of this visible object to the passed one
	 */
	@Override
	public double getDistanceToXYZ(IVisibleObject other)
	{
		return getDistanceToXYZ(other.getPosition());
	}

	/**
	 * Calculates the distance of this visible object to the passed position
	 * @param other the position to which to calculate distance
	 * @return the distance of this visible object to the passed position
	 */
	@Override
	public double getDistanceToXYZ(Vector3D other)
	{
		return getPosition().subtract(other).getNorm();
	}

	/**
	 * Calculates the Direction of this visible object to the passed one
	 * @param other the object to which to calculate direction
	 * @return the direction (rad) of this visible object to the passed one
	 */
	@Override
	public Angle getDirectionTo(IVisibleObject other)
	{
		return getDirectionTo(other.getPosition());
	}

	/**
	 * Calculates the global absolute Direction of this visible object to the
	 * passed Vector3D
	 * @param other the Vector3D to which to calculate direction
	 * @return the global absolute direction (rad) of this visible object to the
	 *         passed Vector
	 */
	@Override
	public Angle getDirectionTo(Vector3D other)
	{
		return Angle.rad(other.subtract(getPosition()).getAlpha());
	}

	@Override
	public Vector3D getLocalPosition()
	{
		return localPosition;
	}

	/**
	 * Set visible state
	 * 
	 * @param state New state
	 */
	public void setVisible(boolean state)
	{
		visible = state;
	}

	/**
	 * @return the visible
	 */
	@Override
	public boolean isVisible()
	{
		return visible;
	}

	@Override
	public float getAge(float currentTime)
	{
		return currentTime - lastSeenTime;
	}

	@Override
	public boolean equals(Object other)
	{
		if (!(other instanceof VisibleObject)) {
			return false;
		}
		VisibleObject otherVis = (VisibleObject) other;

		if (!FuzzyCompare.eq(position, otherVis.position, 0.00001f)) {
			return false;
		}
		if (!FuzzyCompare.eq(previousPosition, otherVis.previousPosition,
				0.00001f)) {
			return false;
		}
		return FuzzyCompare.eq(lastSeenTime, otherVis.lastSeenTime, 0.00001f);
	}

	@Override
	public double getCollisionDistance()
	{
		// default value for all visible objects
		return 0.7;
	}

	@Override
	public boolean isLaying()
	{
		return false;
	}

	@Override
	public int hashCode()
	{
		return name.hashCode() + localPosition.hashCode() + position.hashCode()
				+ previousPosition.hashCode();
	}
}