/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.localizer.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

import magma.agent.worldmodel.localizer.IEnvironmentModel;
import magma.agent.worldmodel.localizer.ILocalizationLine;
import magma.agent.worldmodel.localizer.IReferenceLine;
import magma.agent.worldmodel.localizer.IReferencePoint;
import magma.util.FuzzyCompare;
import magma.util.LocalizerUtil;
import magma.util.geometry.Angle;
import magma.util.geometry.MatrixUtil;
import magma.util.geometry.PositionOrientation3D;

import org.apache.commons.math.geometry.Vector3D;

/**
 * This localization method first calculates a normal to the field ground based
 * on all unlabeled reference lines and reference points on ground level, before
 * it determines the z rotation by relative angle difference of two flags. The
 * resulting orientation is then used together with the reference points to
 * calculate the position.
 * 
 * @author Stefan Glaser
 * 
 */
public class LocalizerFieldNormal extends LocalizerBase
{
	/** handle for logging */
	private static final Logger logger = Logger.getLogger("");

	@Override
	public PositionOrientation3D localize(IEnvironmentModel environment,
			List<ILocalizationLine> lines, Matrix3d estimatedOrientation)
	{
		List<IReferencePoint> refPoints = getVisibleReferencePoints(environment);
		List<IReferenceLine> refLines = getVisibleReferenceLines(environment);

		// We wanna see at least 3 lines and 2 flags to rely on this localization
		// Though, the theoretical minimum amount of lines is 2
		if (lines == null || lines.size() < 3 || refPoints.size() < 2) {
			logger.log(
					Level.FINER,
					"FieldNormal Localization not possible, only have {0} reference points and less than 3 visible lines",
					refPoints.size());
			return null;
		}

		List<Vector3D> bottomPoints = new ArrayList<Vector3D>();
		Vector3D temp;

		// Add fixpoints on ground level to list of bottom-points
		for (IReferencePoint flag : refPoints) {
			if (FuzzyCompare.eq(0, flag.getKnownPosition().getZ(), 0.001)) {
				temp = flag.getLocalPosition();
				bottomPoints.add(new Vector3D(-temp.getY(), temp.getX(), temp
						.getZ()));
			}
		}

		// Add all seen lines to list of bottom-points
		for (ILocalizationLine line : lines) {
			temp = line.getLocalPosition1();
			if (temp != null) {
				bottomPoints.add(new Vector3D(-temp.getY(), temp.getX(), temp
						.getZ()));
			}
			temp = line.getLocalPosition2();
			if (temp != null) {
				bottomPoints.add(new Vector3D(-temp.getY(), temp.getX(), temp
						.getZ()));
			}
		}

		// Calculate field normal
		Vector3d normal = LocalizerUtil.calculatePlaneNormal(bottomPoints);

		// Create rotation matrix based on field normal
		Matrix3d m3d = new Matrix3d();
		double rotX;
		double rotY;

		if (normal.getZ() > 0.998) {
			// North pole => up straight
			rotX = 0;
			rotY = 0;
		} else if (normal.getZ() < -0.998) {
			// South pole => up down
			rotX = Math.PI;
			rotY = 0;
		} else {
			rotX = Math.atan2(-normal.getY(), normal.getZ());
			rotY = Math.asin(normal.getX());
		}

		// Create a rotation matrix for x- and y-rotation based on the field
		// normal
		m3d.setIdentity();
		MatrixUtil.rotateX(m3d, rotX);
		MatrixUtil.rotateY(m3d, rotY);
		MatrixUtil.rotateZ(m3d, 0);
		// Since the field normal is the global z-axis in our system and the
		// rotation is based on it, the rotation is negated - so get it back to
		// our perspective
		m3d.transpose();

		// Calculate z rotation based on normalized local positions of the
		// reference points. Since we still have no clue about our position, we
		// can't just simply use the angle difference to one reference point.
		// Instead we have to use an independent calculation of the angles between
		// two reference points in the local and known system to get the z
		// rotation.
		IReferencePoint flag1;
		IReferencePoint flag2;
		Vector3d flag1Pos;
		Vector3d flag1KnownPos;
		Vector3d flag2Pos;
		Vector3d flag2KnownPos;
		Vector3d vec1to2_3d;
		Vector3d vec1to2Known_3d;
		List<Angle> angles = new ArrayList<Angle>();

		// Instead of just using the reference points to determine the z rotation,
		// the assigned lines also contribute to a more precise result.
		// Theoretically just two unlabeled lines and {at least one assigned line
		// or tow reference points} are required to result in a complete
		// orientation estimation.
		// Therefore: Add reference points for all proper positions of each
		// assigned line to the list of reference points
		if (refLines != null) {
			for (IReferenceLine line : refLines) {
				if (line.getLocalPosition1() != null) {
					refPoints.add(new ReferencePoint(line.getLocalPosition1(), line
							.getAdjustedKnownPosition1()));
				}

				if (line.getLocalPosition2() != null) {
					refPoints.add(new ReferencePoint(line.getLocalPosition2(), line
							.getAdjustedKnownPosition2()));
				}
			}
		}

		for (int i = 0; i < refPoints.size() - 1; i++) {
			flag1 = refPoints.get(i);
			temp = flag1.getKnownPosition();
			flag1KnownPos = new Vector3d(-temp.getY(), temp.getX(), temp.getZ());
			temp = flag1.getLocalPosition();
			flag1Pos = new Vector3d(-temp.getY(), temp.getX(), temp.getZ());
			flag1Pos = MatrixUtil.mulMatrixVector(m3d, flag1Pos);

			for (int j = i + 1; j < refPoints.size(); j++) {
				flag2 = refPoints.get(j);
				temp = flag2.getKnownPosition();
				// Check if the second reference point is at the same known position
				// as the first reference point (can happen, since we also use
				// reference lines). In this case, the selected reference point
				// combination can't provide a rotation estimation.
				if (FuzzyCompare.eq(temp, flag1.getKnownPosition(), 0.01)) {
					continue;
				}

				flag2KnownPos = new Vector3d(-temp.getY(), temp.getX(), temp.getZ());
				temp = flag2.getLocalPosition();
				flag2Pos = new Vector3d(-temp.getY(), temp.getX(), temp.getZ());
				flag2Pos = MatrixUtil.mulMatrixVector(m3d, flag2Pos);

				vec1to2Known_3d = new Vector3d(flag2KnownPos);
				vec1to2Known_3d.sub(flag1KnownPos);

				vec1to2_3d = new Vector3d(flag2Pos);
				vec1to2_3d.sub(flag1Pos);

				Angle knownDirection = Angle.rad(Math.atan2(
						-vec1to2Known_3d.getX(), vec1to2Known_3d.getY()));
				Angle visionDirection = Angle.rad(Math.atan2(-vec1to2_3d.getX(),
						vec1to2_3d.getY()));

				angles.add(knownDirection.subtract(visionDirection));
			}
		}

		// Add the new calculated z rotation to the previous field normal
		// orientation to get the final agent orientation
		Matrix3d orientation = new Matrix3d();
		orientation.rotZ(Angle.average(angles.toArray(new Angle[angles.size()]))
				.radians());
		orientation.mul(m3d);

		// Calculate position based on the calculated orientation and the
		// reference points
		return new PositionOrientation3D(LocalizerUtil.calculatePosition(
				refPoints, orientation), orientation);
	}
}
