/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.localizer.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

import magma.agent.worldmodel.localizer.IEnvironmentModel;
import magma.agent.worldmodel.localizer.ILocalizationLine;
import magma.agent.worldmodel.localizer.IReferenceLine;
import magma.agent.worldmodel.localizer.IReferencePoint;
import magma.util.FuzzyCompare;
import magma.util.LocalizerUtil;
import magma.util.geometry.PositionOrientation3D;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Implements a 3-dimensional localizer based on reference points and the
 * estimated orientation.
 * 
 * @author Stefan Glaser
 */
public class LocalizerGyro extends LocalizerBase
{
	/** handle for logging */
	private static final Logger logger = Logger.getLogger("");

	@Override
	public PositionOrientation3D localize(IEnvironmentModel environment,
			List<ILocalizationLine> lines, Matrix3d estimatedOrientation)
	{
		if (estimatedOrientation == null) {
			// there is no gyro information available
			logger.log(Level.FINER,
					"Gyro Localization not possible, there is no gyro information available!");
			return null;
		}

		// fetch reference points and -lines
		List<IReferencePoint> refPoints = getVisibleReferencePoints(environment);
		List<IReferenceLine> refLines = getVisibleReferenceLines(environment);

		if (refPoints.size() == 0) {
			// there is no visible flag
			logger.log(Level.FINER,
					"Gyro Localization not possible, there is no visible reference point!");
			return null;
		}

		return new PositionOrientation3D(LocalizerUtil.calculatePosition(
				refPoints, refLines, estimatedOrientation), estimatedOrientation);
	}

	// ==================================================
	// Below: experimental localizer implementations
	// ==================================================
	/**
	 * This method searches for line pints near by known flags, to extent the
	 * fix-point base for umeyama orientation estimation.
	 * 
	 * @param flagsList
	 * @param lines
	 * @return
	 */
	@SuppressWarnings("unused")
	private PositionOrientation3D lineSearchingUmeyamaLocalization(
			Collection<IReferencePoint> flagsList, List<ILocalizationLine> lines)
	{
		if (flagsList.size() < 3 || (lines != null && lines.size() < 2)) {
			return null;
		}

		List<Vector3d> visionVec = new ArrayList<Vector3d>();
		List<Vector3d> fixedVec = new ArrayList<Vector3d>();
		Vector3d fixedCOM;
		Vector3d visionCOM;
		Vector3D temp;

		for (IReferencePoint flag : flagsList) {
			// Add fixedPosition
			temp = flag.getKnownPosition();
			fixedVec.add(new Vector3d(-temp.getY(), temp.getX(), temp.getZ()));

			// Add localPosition
			temp = flag.getLocalPosition();
			visionVec.add(new Vector3d(-temp.getY(), temp.getX(), temp.getZ()));
		}

		// Search for lines, ending at fix-points
		for (IReferencePoint flag : flagsList) {
			// Check on corner flag
			if (FuzzyCompare.eq(0, flag.getKnownPosition().getZ(), 0.001)) {
				for (ILocalizationLine line : lines) {
					if (FuzzyCompare.eq(flag.getLocalPosition(),
							line.getLocalPosition1(), 0.5)) {
						// We found a border line, starting/ending at the fix-point
						temp = flag.getKnownPosition();
						fixedVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
								.getZ()));
						temp = line.getLocalPosition1();
						visionVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
								.getZ()));
						// System.out.println("found matching line to corner flag");
					}

					if (FuzzyCompare.eq(flag.getLocalPosition(),
							line.getLocalPosition2(), 0.5)) {
						// We found a border line, starting/ending at the fix-point
						temp = flag.getKnownPosition();
						fixedVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
								.getZ()));
						temp = line.getLocalPosition2();
						visionVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
								.getZ()));
						// System.out.println("found matching line to corner flag");
					}
				}
			}

			// Check on goal post flag
			if (flag.getKnownPosition().getZ() > 0.1) {
				for (ILocalizationLine line : lines) {
					if (FuzzyCompare.eq(flag.getLocalPosition(),
							line.getLocalPosition1(), 1.5)) {
						// We found a border line, starting/ending at the fix-point
						temp = flag.getKnownPosition();

						if (temp.getY() > 0) {
							// Upper goal post
							temp = temp.add(new Vector3D(0, 0.9, -0.8));
						} else {
							// Lower goal post
							temp = temp.add(new Vector3D(0, -0.9, -0.8));
						}
						fixedVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
								.getZ()));

						if (line.getLocalPosition2() != null) {
							if (temp.getX() > 0) {
								temp = temp.add(new Vector3D(-1.8, 0, 0));
							} else {
								temp = temp.add(new Vector3D(1.8, 0, 0));
							}
							fixedVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
									.getZ()));
						}

						temp = line.getLocalPosition1();
						visionVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
								.getZ()));

						if (line.getLocalPosition2() != null) {
							temp = line.getLocalPosition2();
							visionVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
									.getZ()));
						}

						// System.out.println("found matching line to goal post");
					}

					if (FuzzyCompare.eq(flag.getLocalPosition(),
							line.getLocalPosition2(), 1.5)) {
						// We found a border line, starting/ending at the fix-point
						temp = flag.getKnownPosition();

						if (temp.getY() > 0) {
							// Upper goal post
							temp = temp.add(new Vector3D(0, 0.9, -0.8));
						} else {
							// Lower goal post
							temp = temp.add(new Vector3D(0, -0.9, -0.8));
						}
						fixedVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
								.getZ()));

						if (line.getLocalPosition1() != null) {
							if (temp.getX() > 0) {
								temp = temp.add(new Vector3D(-1.8, 0, 0));
							} else {
								temp = temp.add(new Vector3D(1.8, 0, 0));
							}
							fixedVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
									.getZ()));
						}

						temp = line.getLocalPosition2();
						visionVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
								.getZ()));

						if (line.getLocalPosition1() != null) {
							temp = line.getLocalPosition1();
							visionVec.add(new Vector3d(-temp.getY(), temp.getX(), temp
									.getZ()));
						}

						// System.out.println("found matching line to goal post");
					}
				}
			}
		}

		fixedCOM = new Vector3d();
		for (Vector3d vec : fixedVec) {
			fixedCOM.add(vec);
		}
		fixedCOM.scale(1.0 / fixedVec.size());

		visionCOM = new Vector3d();
		for (Vector3d vec : visionVec) {
			visionCOM.add(vec);
		}
		visionCOM.scale(1.0 / visionVec.size());

		for (int i = 0; i < fixedVec.size(); i++) {
			fixedVec.get(i).sub(fixedCOM);
		}

		for (int i = 0; i < visionVec.size(); i++) {
			visionVec.get(i).sub(visionCOM);
		}

		Matrix3d orientation = LocalizerUtil.umeyama(
				visionVec.toArray(new Vector3d[visionVec.size()]),
				fixedVec.toArray(new Vector3d[visionVec.size()]));

		return new PositionOrientation3D(LocalizerUtil.calculatePosition(
				flagsList, orientation), orientation);
	}
}
