/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.worldmodel.localizer.impl;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

import magma.agent.worldmodel.localizer.IEnvironmentModel;
import magma.agent.worldmodel.localizer.ILocalizationLine;
import magma.agent.worldmodel.localizer.IReferenceLine;
import magma.agent.worldmodel.localizer.IReferencePoint;
import magma.util.LocalizerUtil;
import magma.util.geometry.PositionOrientation3D;

import org.apache.commons.math.geometry.Vector3D;

/**
 * This localization method applies the umeyama method on the reference points
 * (and reference lines - not yet implemented) to determine the current
 * orientation in the 3D space. This orientation is then used to transform the
 * local positions of the reference points and lines to calculate the position.
 * 
 * @author Stefan Glaser
 * 
 */
public class LocalizerUmeyama extends LocalizerBase
{
	/** handle for logging */
	private static final Logger logger = Logger.getLogger("");

	@Override
	public PositionOrientation3D localize(IEnvironmentModel environment,
			List<ILocalizationLine> lines, Matrix3d estimatedOrientation)
	{
		List<IReferencePoint> refPoints = getVisibleReferencePoints(environment);
		List<IReferenceLine> refLines = getVisibleReferenceLines(environment);

		// We wanna see at least 3 reference points
		if (refPoints.size() < 3) {
			logger.log(Level.FINER,
					"Umeyama Localization not possible, less than 3 visible reference points");
			return null;
		}

		Vector3d[] visionVecArr = new Vector3d[refPoints.size()];
		Vector3d[] fixedVecArr = new Vector3d[refPoints.size()];
		Vector3d fixedCOM;
		Vector3d visionCOM;
		Vector3D temp;

		// Transform the global and local positions of each reference point to the
		// body-system and store the resulting vectors in the corresponding
		// arrays.
		int i = 0;
		for (IReferencePoint flag : refPoints) {
			// Add fixedPosition
			temp = flag.getKnownPosition();
			fixedVecArr[i] = new Vector3d(-temp.getY(), temp.getX(), temp.getZ());

			// Add localPosition
			temp = flag.getLocalPosition();
			visionVecArr[i] = new Vector3d(-temp.getY(), temp.getX(), temp.getZ());
			i++;
		}

		// Transform the global and local positions of each reference line to the
		// body-system and store the resulting vectors in the corresponding
		// arrays.
		if (refLines != null) {
			for (IReferenceLine line : refLines) {
				temp = line.getLocalPosition1();
				if (temp != null) {
					// Add localPosition
					visionVecArr[i] = new Vector3d(-temp.getY(), temp.getX(),
							temp.getZ());

					// Add fixedPosition
					temp = line.getAdjustedKnownPosition1();
					fixedVecArr[i] = new Vector3d(-temp.getY(), temp.getX(),
							temp.getZ());
					i++;
				}

				temp = line.getLocalPosition2();
				if (temp != null) {
					// Add localPosition
					visionVecArr[i] = new Vector3d(-temp.getY(), temp.getX(),
							temp.getZ());

					// Add fixedPosition
					temp = line.getAdjustedKnownPosition2();
					fixedVecArr[i] = new Vector3d(-temp.getY(), temp.getX(),
							temp.getZ());
					i++;
				}
			}
		}

		// Calculate centroid of fixed points
		fixedCOM = new Vector3d();
		for (Vector3d vec : fixedVecArr) {
			fixedCOM.add(vec);
		}
		fixedCOM.scale(1.0 / fixedVecArr.length);

		// Calculate centroid of vision points
		visionCOM = new Vector3d();
		for (Vector3d vec : visionVecArr) {
			visionCOM.add(vec);
		}
		visionCOM.scale(1.0 / visionVecArr.length);

		// Calculate fixed covariance vectors
		for (i = 0; i < fixedVecArr.length; i++) {
			fixedVecArr[i].sub(fixedCOM);
		}
		// Calculate vision covariance vectors
		for (i = 0; i < visionVecArr.length; i++) {
			visionVecArr[i].sub(visionCOM);
		}

		// Determine orientation by umeyama
		Matrix3d orientation = LocalizerUtil.umeyama(visionVecArr, fixedVecArr);

		// TODO: Here is maybe an open problem! The calculated orientation is with
		// respect to the centroid, not to our own position and is therefore not
		// necessarily the same as our orientation!
		// We maybe have to translate the vision vectors before and after the
		// rotation according to the centroid before position estimation and then
		// perform a second umeyama with all known positions adjusted to the
		// localized position.

		// Determine position and return result
		return new PositionOrientation3D(LocalizerUtil.calculatePosition(
				refPoints, refLines, orientation), orientation);
	}
}
