/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots;

import java.util.logging.Level;

import magma.agent.IMagmaConstants;
import magma.agent.IServerConfigFilesConstants;
import magma.agent.connection.IServerConnection;
import magma.agent.general.AgentRuntime;
import magma.agent.general.ComponentFactory;
import magma.agent.general.PlayerParameters;
import magma.robots.nao.general.NAOComponentFactory;

/**
 * This is the main class of the application used to start the magma Robocup
 * agent.
 * 
 * @author Klaus Dorer
 */
public class RoboCupClient
{
	/**
	 * Instantiates and starts an agent.
	 * 
	 * @param args Command line arguments <br>
	 *        <table>
	 *        <tr>
	 *        <td>arg0:</td>
	 *        <td>TeamName</td>
	 *        <td>Name of the Team</td>
	 *        </tr>
	 *        <tr>
	 *        <td>arg1:</td>
	 *        <td>PlayerID</td>
	 *        <td>PlayerID of the Agent</td>
	 *        </tr>
	 *        <tr>
	 *        <td>arg2:</td>
	 *        <td>Host</td>
	 *        <td>Server IP</td>
	 *        </tr>
	 *        <tr>
	 *        <td>arg3:</td>
	 *        <td>Port</td>
	 *        <td>Server Port</td>
	 *        </tr>
	 *        <tr>
	 *        <td>arg4:</td>
	 *        <td>LogLevel</td>
	 *        <td>Different Logging levels, default: SEVERE<br>
	 *        ('fine','finer','finest','warning','info')</td>
	 *        </tr>
	 *        <tr>
	 *        <td>arg5:</td>
	 *        <td>ServerVersion</td>
	 *        <td>Simspark Server Version<br>
	 *        ('60','61','62','63',...)</td>
	 *        </tr>
	 *        <tr>
	 *        <td>arg6:</td>
	 *        <td>DecisionMakerID</td>
	 *        <td>DecisionMaker of the Agent<br>
	 *        (0 (default) if we decide based on player number, 1 for goalie, 2
	 *        for field player, 3 for simple decision maker, 4 for Penalty
	 *        decision maker, 5 for do nothing decision maker</td>
	 *        </tr>
	 *        </table>
	 */
	public static void main(String[] args)
	{
		String teamname = IMagmaConstants.DEFAULT_TEAMNAME;
		int id = 0;
		String host = IServerConnection.SERVER_IP;
		int port = IServerConnection.SERVER_PORT;
		ComponentFactory componentFactory;

		// TODO: Once we have multiple robot models, add a new start parameter to
		// determine the specific robot model. Up to then, we simply use the
		// component factory of the NAO
		componentFactory = new NAOComponentFactory();

		if (args.length > 0) {
			teamname = args[0];
		}

		if (args.length > 1) {
			id = new Integer(args[1]);
			if (id < 0 || id >= IMagmaConstants.startPositionsOwnKickoff.length) {
				System.out.println("Invalid player ID: " + id + "! Maximum is: "
						+ (IMagmaConstants.startPositionsOwnKickoff.length - 1));
				System.exit(-1);
			}
		}

		if (args.length > 2) {
			host = args[2];
		}

		if (args.length > 3) {
			port = new Integer(args[3]);
			if (port < 0) {
				System.out.println("Invalid port: " + port);
				System.exit(-1);
			}
		}

		Level level = Level.SEVERE;
		if (args.length > 4) {
			String loglevel = args[4];
			if (loglevel.equalsIgnoreCase("fine")) {
				level = Level.FINE;
			} else if (loglevel.equalsIgnoreCase("finer")) {
				level = Level.FINER;
			} else if (loglevel.equalsIgnoreCase("finer")) {
				level = Level.FINER;
			} else if (loglevel.equalsIgnoreCase("finest")) {
				level = Level.FINEST;
			} else if (loglevel.equalsIgnoreCase("warning")) {
				level = Level.WARNING;
			} else if (loglevel.equalsIgnoreCase("info")) {
				level = Level.INFO;
			}
		}

		int serverVersion = IServerConfigFilesConstants.DEFAULT_SERVER_VERSION;
		if (args.length > 5) {
			serverVersion = new Integer(args[5]);
			if (serverVersion < 60) {
				System.out.println("Invalid server version: " + serverVersion);
				System.exit(-1);
			}
		}

		int decisionMakerID = 0;
		if (args.length > 6) {
			decisionMakerID = new Integer(args[6]);
			if (decisionMakerID < 0 || decisionMakerID > 6) {
				System.out.println("Invalid decision maker id: " + decisionMakerID);
				System.exit(-1);
			}
		}

		AgentRuntime client = new AgentRuntime(new PlayerParameters(teamname, id,
				host, port, null, level, serverVersion, componentFactory,
				decisionMakerID));
		client.startClient();
	}
}
