/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
/**
 * 
 */
package magma.robots.nao.behavior.complex;

import java.util.Map;
import java.util.logging.Level;

import javax.vecmath.Vector3d;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.IWalk;
import magma.agent.behavior.basic.BehaviorParameter;
import magma.agent.behavior.complex.ComplexBehavior;
import magma.agent.worldmodel.IGlobalMap;
import magma.robots.nao.behavior.dynamic.walk.Balance;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;
import magma.util.geometry.MatrixUtil;

/**
 * Implements a behavior which performs bi-pedal walking based on the "Balance"
 * behavior.
 * 
 * @author Ingo Schindler
 */
public class Walk extends ComplexBehavior implements IWalk
{
	/**
	 * Constructor
	 * 
	 * @param worldModel Reference to the world model
	 * @param agentModel Reference to the agent model
	 * @param params Behavior parameters
	 * @param behaviors List of available behaviors
	 */
	public Walk(String name, IGlobalMap worldModel, IAgentModel agentModel,
			BehaviorParameter params, Map<String, IBehavior> behaviors)
	{
		super(name, worldModel, agentModel, behaviors, behaviors
				.get(IBehaviorConstants.BALANCE));

		walk(0, 0, 0);
	}

	/**
	 * Constructor
	 * 
	 * @param worldModel Reference to the world model
	 * @param agentModel Reference to the agent model
	 * @param params Behavior parameters
	 * @param behaviors List of available behaviors
	 */
	public Walk(IGlobalMap worldModel, IAgentModel agentModel,
			BehaviorParameter params, Map<String, IBehavior> behaviors)
	{
		this(IBehaviorConstants.WALK, worldModel, agentModel, params, behaviors);
	}

	/**
	 * Set parameters for Walk. It is possible to combine all different
	 * parameters, e.g. forwards and sidesteps.
	 * 
	 * @param forwards_backwards positive = forwards; negative = backwards (-100
	 *        .. 100)
	 * @param step_left_right positive = right; negative = left (-100 .. 100)
	 * @param turn_left_right positive = right; negative = left (-30 .. 30)
	 */
	@Override
	public void walk(double forwards_backwards, double step_left_right,
			double turn_left_right)
	{
		walk(forwards_backwards, step_left_right, turn_left_right, 100);
	}

	/**
	 * Set parameters for Walk. It is possible to combine all different
	 * parameters, e.g. forwards and sidesteps.
	 * 
	 * @param forwards_backwards positive = forwards; negative = backwards (-100
	 *        .. 100)
	 * @param step_left_right positive = right; negative = left (-100 .. 100)
	 * @param turn_left_right positive = right; negative = left (-30 .. 30)
	 */
	@Override
	public void walk(double forwards_backwards, double step_left_right,
			double turn_left_right, double maxSpeedForward)
	{
		logger.log(Level.FINE, "walk (x/y/turn): {0}/{1}/{2}", new Object[] {
				forwards_backwards, step_left_right, turn_left_right });

		Balance balance = (Balance) getCurrentBehavior();
		balance.setMovement(forwards_backwards, step_left_right, turn_left_right,
				false, maxSpeedForward);
	}

	/**
	 * Set parameters to walk in a global direction
	 * 
	 * @param left_right negative = own goal; positive = opponent goal
	 * @param up_down negative = left field side; positive = right field side
	 * @param direction local angle where to go
	 */
	@Override
	public void globalWalk(double left_right, double up_down, double direction,
			double minSpeedForward, double minSpeedSideward, double maxSpeedForward)
	{
		Angle goalDirection = worldModel.getThisPlayer().getHorizontalAngle();

		Vector3d fieldVector = new Vector3d(left_right, up_down, 0);

		Vector3d playerVector = MatrixUtil.mulMatrixVector(
				MatrixUtil.getOrientationMatrix(-(float) goalDirection.degrees()),
				fieldVector);

		double forwardsBackwards = playerVector.getX();
		double speed = (Geometry.getLinearFuzzyValue(0.0, 0.5, true,
				Math.abs(forwardsBackwards)) * 100)
				* Math.signum(forwardsBackwards);

		double stepLeftRight = playerVector.getY();
		double sideSpeed = Geometry.getLinearFuzzyValue(0.0, 0.5, true,
				Math.abs(stepLeftRight))
				* 100 * Math.signum(stepLeftRight);

		logger.log(
				Level.FINE,
				"globalWalk (x/y/direction): {0}/{1}/{2} goalDir: {3} (fb/lr): ({4}/{5})",
				new Object[] { left_right, up_down, direction,
						goalDirection.degrees(), forwardsBackwards, stepLeftRight });

		int turnPreferenceAngle = 25;
		if (Math.abs(forwardsBackwards) > Math.abs(stepLeftRight)) {
			// mainly going forward
			turnPreferenceAngle = 40;
			// sideSpeed = sideSpeed * Math.abs((stepLeftRight /
			// forwardsBackwards));
			sideSpeed = 0;
			// when approaching ball, we should not get below a minimum speed
			if (speed > 0) {
				speed = Math.max(minSpeedForward, speed);
			} else {
				speed = Math.min(-minSpeedForward, speed);
			}
		} else {
			if (sideSpeed > 0) {
				sideSpeed = Math.max(minSpeedSideward, sideSpeed);
				sideSpeed = Math.min(100, sideSpeed);
			} else {
				sideSpeed = Math.min(minSpeedSideward, sideSpeed);
				sideSpeed = Math.max(-100, sideSpeed);
			}
		}

		if (Math.abs(direction) > turnPreferenceAngle) {
			// if we have to turn considerably we have to limit our speed
			// this should have fixed a bug that when running fast it was not
			// possible to run somewhere else
			if (speed > 0) {
				speed = Math.min(10, speed);
			} else {
				speed = Math.max(-10, speed);
			}
			if (sideSpeed > 0) {
				sideSpeed = Math.min(10, sideSpeed);
			} else {
				sideSpeed = Math.max(-10, sideSpeed);
			}
		}

		speed = Geometry.limitAbs(speed, maxSpeedForward);

		walk(speed, sideSpeed, direction, maxSpeedForward);

	}

	@Override
	public IBehavior decideNextBasicBehavior()
	{
		return getCurrentBehavior();
	}
}