/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.robots.nao.behavior.dynamic;

import java.util.Map;

import javax.vecmath.Matrix3d;
import javax.vecmath.Matrix4d;
import javax.vecmath.Vector3d;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.agentmodel.IBodyPart;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.basic.Behavior;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IVisibleObject;
import magma.robots.nao.INaoConstants;
import magma.util.FuzzyCompare;
import magma.util.geometry.MatrixUtil;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Implements a behavior which lets the robot head focus on the ball
 * 
 * @author Ingo Schindler
 */
public class FocusBall extends Behavior
{

	private static final float TIME_TO_LOOK_AROUND = 2;

	private static final float LOOK_ARROUND_DELAY = 8;

	private final IBehavior turnHead;

	private float timeToTurnHead;

	/**
	 * Instantiates and initializes a new Balance behavior object
	 * 
	 * @param worldModel Reference to the world model object
	 * @param agentModel Reference to the agent model object
	 * @param behaviors List of applicable behaviors
	 */
	public FocusBall(IGlobalMap worldModel, IAgentModel agentModel,
			Map<String, IBehavior> behaviors)
	{
		super(IBehaviorConstants.FOCUS_BALL, worldModel, agentModel);
		turnHead = behaviors.get(IBehaviorConstants.TURN_HEAD);
		timeToTurnHead = worldModel.getGlobalTime();
	}

	@Override
	public void perform()
	{
		int visibleLandmarks = 0;
		for (int i = 0; i < worldModel.getLandmarks().size(); i++) {
			float time = ((IVisibleObject) worldModel.getLandmarks().toArray()[i])
					.getAge(worldModel.getGlobalTime());
			if (time < 0.1) {
				visibleLandmarks++;
			}
		}

		Vector3D ballPos = worldModel.getBall().getLocalPosition();

		float timeDelta = worldModel.getGlobalTime() - timeToTurnHead;

		if (worldModel.getBall().getAge(worldModel.getGlobalTime()) > 0.1
				|| visibleLandmarks == 0
				|| (ballPos.getNorm() > 1.5 && FuzzyCompare
						.gte(timeDelta, LOOK_ARROUND_DELAY + TIME_TO_LOOK_AROUND,
								TIME_TO_LOOK_AROUND))) {
			turnHead.perform();
			if (timeDelta > LOOK_ARROUND_DELAY + TIME_TO_LOOK_AROUND)
				timeToTurnHead = worldModel.getGlobalTime();

		} else {
			// Calculate position of ball relative to camera body
			IBodyPart head = agentModel.getBodyPart(INaoConstants.Head);
			Matrix3d headOrientation = head.getOrientation();
			headOrientation.transpose();
			Vector3d headPostition = head.getPosition();
			headPostition.negate();
			Matrix4d headTransform = new Matrix4d(headOrientation, headPostition,
					1);

			Vector3d ballVec = new Vector3d(-ballPos.getY(), ballPos.getX(),
					ballPos.getZ());

			ballVec = MatrixUtil.mulMatrixVector(headTransform, ballVec);
			ballPos = new Vector3D(ballVec.getY(), -ballVec.getX(), ballVec.getZ());

			// Adjust head joints to focus the ball
			agentModel.getWriteableHJ(INaoConstants.NeckYaw).adjustAxisPosition(
					Math.toDegrees(ballPos.getAlpha()));
			agentModel.getWriteableHJ(INaoConstants.NeckPitch).adjustAxisPosition(
					Math.toDegrees(ballPos.getDelta()));
		}
	}
}
