/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.dynamic.walk;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

import magma.agent.behavior.basic.BehaviorParameter;

/**
 * Represents the parameter set necessary for walking
 * 
 * @author Ingo Schindler
 * 
 *         What exactly this parameters are is documented in my Bachelor-Thesis
 *         (Bipedal-Walking in the simulated RoboCup-3D environment)
 */
public class WalkBehaviorParameter extends BehaviorParameter
{
	private static final long serialVersionUID = -2324636896847512436L;

	private double acceleration_forwards = 2;

	private double acceleration_backwards = 2;

	private double acceleration_left_right = 3;

	private double acceleration_turn_left_right = 3;

	private double deceleration_forwards = 3;

	private double deceleration_backwards = 2;

	private double deceleration_left_right = 3;

	private double deceleration_turn_left_right = 3;

	private double hipHeight = 0.03596818;

	/**
	 * k = step
	 */
	private double k = 0;

	// Td < Tm < Tc
	/**
	 * Tc = duration of the whole motion
	 */
	private double Tc = 20;

	/**
	 * Tm = time when the foot should touch the ground again
	 */
	private double Tm = 12;

	/**
	 * Td = middle of the motion
	 */
	private double Td = 4;

	/**
	 * Height of Ground when leaving the Ground
	 */
	private double h_gs = 0;

	/**
	 * Height of Ground when touching the Ground again
	 */
	private double h_ge = 0;

	/**
	 * Ds * 2 = Step-Length
	 */
	private double Ds = 0.036196616;

	/**
	 * Highest Point of the Foot during a step
	 */
	private double Hao = 0.044080388;

	/**
	 * Distance of Hao relative to the Point the foot leaves the Ground
	 */
	private double Lao = 0.03281899;

	/**
	 * Angle of the Ground when leaving the Ground
	 */
	private double qgs = 0;

	/**
	 * Foot-Angle when leaving the Ground
	 */
	private double qb = -2.3566105;

	/**
	 * Foot-Angle when touching the Ground again
	 */
	private double qf = 2.7248816;

	/**
	 * Angle of the Ground when touching the Ground again
	 */
	private double qge = 0;

	/**
	 * Hip to Ankle distance at the start of the single support phase should be a
	 * value 0 < x_ed <= 0.5Ds
	 */
	private double x_ed = -0.1 * Ds;

	/**
	 * Hip to Ankle distance at the end of the single support phase should be a
	 * value 0 < x_sd <= 0.5Ds
	 */
	private double x_sd = 0.1 * Ds;

	/**
	 * Distance Ankle->Heel
	 */
	private double l_ab = 0.05;

	/**
	 * Distance Ankle->Tiptoe
	 */
	private double l_af = 0.11;

	/**
	 * Distance Ankle->Sole
	 */
	private double l_an = 0.015;

	public WalkBehaviorParameter()
	{
		// uncomment to read parameter values from learned file
		// this("walkParameters.txt");
	}

	public WalkBehaviorParameter(String filepath)
	{
		try {

			BufferedReader in = new BufferedReader(new FileReader(filepath));

			String line;

			while ((line = in.readLine()) != null) {
				// Print the content on the console

				if (line.startsWith("Tc ")) {
					setTc(Double.parseDouble(line.substring(3)));
					// System.out.println("Tc: " + getTc());
				} else if (line.startsWith("Td ")) {
					setTm(Double.parseDouble(line.substring(3)));
					// System.out.println("Tm: " + getTd());
				} else if (line.startsWith("Tm ")) {
					setTd(Double.parseDouble(line.substring(3)));
					// System.out.println("Td: " + getTm());
				} else if (line.startsWith("Ds ")) {
					setDs(Double.parseDouble(line.substring(3)));
					// System.out.println("Ds: " + getDs());
				} else if (line.startsWith("Hao ")) {
					setHao(Double.parseDouble(line.substring(4)));
					// System.out.println("Hao: " + getHao());
				} else if (line.startsWith("Lao ")) {
					setLao(Double.parseDouble(line.substring(4)));
					// System.out.println("Lao: " + getLao());
				} else if (line.startsWith("Qb ")) {
					setQb(Double.parseDouble(line.substring(3)));
					// System.out.println("Qb: " + getQb());
				} else if (line.startsWith("Qf ")) {
					setQf(Double.parseDouble(line.substring(3)));
					// System.out.println("Qf: " + getQf());
				} else if (line.startsWith("x_ed ")) {
					setX_ed(Double.parseDouble(line.substring(5)));
					// System.out.println("x_ed: " + getX_ed());
				} else if (line.startsWith("x_sd ")) {
					setX_sd(Double.parseDouble(line.substring(5)));
					// System.out.println("x_sd: " + getX_sd());
				} else if (line.startsWith("HipHeight ")) {
					setHipHeight(Double.parseDouble(line.substring("HipHeight "
							.length())));
					// System.out.println("x_sd: " + getX_sd());
				} else if (line.startsWith("acceleration_forwards ")) {
					setAcceleration_forwards(Double.parseDouble(line
							.substring("acceleration_forwards ".length())));
					// System.out.println("x_sd: " + getX_sd());
				} else if (line.startsWith("acceleration_backwards ")) {
					setAcceleration_backwards(Double.parseDouble(line
							.substring("acceleration_backwards ".length())));

					// System.out.println("x_sd: " + getX_sd());
				} else if (line.startsWith("acceleration_left_right ")) {
					setAcceleration_left_right(Double.parseDouble(line
							.substring("acceleration_left_right ".length())));
					// System.out.println("x_sd: " + getX_sd());
				} else if (line.startsWith("acceleration_turn_left_right ")) {
					setAcceleration_turn_left_right(Double.parseDouble(line
							.substring("acceleration_turn_left_right ".length())));
					// System.out.println("x_sd: " + getX_sd());
				}

			}

		} catch (FileNotFoundException e) {
			// e.printStackTrace();
		} catch (IOException e) {
			// e.printStackTrace();
		}
	}

	public double getDeceleration_forwards()
	{
		return deceleration_forwards;
	}

	public void setDeceleration_forwards(double deceleration_forwards)
	{
		this.deceleration_forwards = deceleration_forwards;
	}

	public double getDeceleration_backwards()
	{
		return deceleration_backwards;
	}

	public void setDeceleration_backwards(double deceleration_backwards)
	{
		this.deceleration_backwards = deceleration_backwards;
	}

	public double getDeceleration_left_right()
	{
		return deceleration_left_right;
	}

	public void setDeceleration_left_right(double deceleration_left_right)
	{
		this.deceleration_left_right = deceleration_left_right;
	}

	public double getDeceleration_turn_left_right()
	{
		return deceleration_turn_left_right;
	}

	public void setDeceleration_turn_left_right(
			double deceleration_turn_left_right)
	{
		this.deceleration_turn_left_right = deceleration_turn_left_right;
	}

	public double getAcceleration_forwards()
	{
		return acceleration_forwards;
	}

	public void setAcceleration_forwards(double acceleration_forwards)
	{
		this.acceleration_forwards = acceleration_forwards;
	}

	public double getAcceleration_backwards()
	{
		return acceleration_backwards;
	}

	public void setAcceleration_backwards(double acceleration_backwards)
	{
		this.acceleration_backwards = acceleration_backwards;
	}

	public double getAcceleration_left_right()
	{
		return acceleration_left_right;
	}

	public void setAcceleration_left_right(double acceleration_left_right)
	{
		this.acceleration_left_right = acceleration_left_right;
	}

	public double getAcceleration_turn_left_right()
	{
		return acceleration_turn_left_right;
	}

	public void setAcceleration_turn_left_right(
			double acceleration_turn_left_right)
	{
		this.acceleration_turn_left_right = acceleration_turn_left_right;
	}

	public double getHipHeight()
	{
		return hipHeight;
	}

	public void setHipHeight(double hipheight)
	{
		this.hipHeight = hipheight;
		this.h_ge = hipheight;
		this.h_gs = hipheight;
	}

	/**
	 * Set the current step value 0 = First Step, 1 = second Step , ...
	 * 
	 * @param k New step value
	 */
	public void setK(double k)
	{
		this.k = k;
	}

	/**
	 * Set the duration of the whole motion
	 * 
	 * @param tc New duration
	 */
	public void setTc(double tc)
	{
		Tc = tc;
	}

	public void setTd(double td)
	{
		Td = td;
	}

	public void setTm(double tm)
	{
		Tm = tm;
	}

	public void setH_gs(double hGs)
	{
		h_gs = hGs;
	}

	public void setH_ge(double hGe)
	{
		h_ge = hGe;
	}

	/**
	 * Set the step length value
	 * 
	 * @param ds New step length value
	 */
	public void setDs(double ds)
	{
		Ds = ds;
	}

	public void setHao(double hao)
	{
		Hao = hao;
	}

	public void setLao(double lao)
	{
		Lao = lao;
	}

	public void setQgs(double qgs)
	{
		this.qgs = qgs;
	}

	public void setQb(double qb)
	{
		this.qb = qb;
	}

	public void setQf(double qf)
	{
		this.qf = qf;
	}

	public void setQge(double qge)
	{
		this.qge = qge;
	}

	public void setX_ed(double xEd)
	{
		x_ed = xEd;
	}

	public void setX_sd(double xSd)
	{
		x_sd = xSd;
	}

	public void setL_ab(double lAb)
	{
		l_ab = lAb;
	}

	public void setL_af(double lAf)
	{
		l_af = lAf;
	}

	public void setL_an(double lAn)
	{
		l_an = lAn;
	}

	public double getK()
	{
		return k;
	}

	public double getTd()
	{
		return Td;
	}

	public double getH_gs()
	{
		return h_gs;
	}

	public double getH_ge()
	{
		return h_ge;
	}

	public double getDs()
	{
		return Ds;
	}

	public double getHao()
	{
		return Hao;
	}

	public double getLao()
	{
		return Lao;
	}

	public double getX_ed()
	{
		return x_ed;
	}

	public double getX_sd()
	{
		return x_sd;
	}

	public double getL_ab()
	{
		return l_ab;
	}

	public double getL_af()
	{
		return l_af;
	}

	public double getL_an()
	{
		return l_an;
	}

	public double getTm()
	{
		return Tm;
	}

	public double getQge()
	{
		return qge;
	}

	public double getQf()
	{
		return qf;
	}

	public double getTc()
	{
		return Tc;
	}

	public double getQb()
	{
		return qb;
	}

	public double getQgs()
	{
		return qgs;
	}
}
