/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.movement;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IKick;
import magma.agent.behavior.movement.Movement;
import magma.agent.behavior.movement.MovementBehavior;
import magma.agent.behavior.movement.MovementPhase;
import magma.agent.worldmodel.IGlobalMap;
import magma.robots.nao.INaoConstants;
import magma.util.geometry.Angle;

import org.apache.commons.math.geometry.Vector3D;

/**
 * @author Stefan Glaser
 * 
 */
public abstract class Kick extends MovementBehavior implements IKick
{
	/** angle in which the shoot will go relative to the body direction */
	private Angle kickDirection;

	/** direction we have to face relative to the ball */
	private Angle ballFacingDirection;

	/** max distance to the ball to hit the ball */
	private double maxBallDistance;

	/** the maximum kick distance this kick is able to kick */
	private double maxKickDistance;

	/** the utility (calculated at runtime) for this kick */
	private double utility;

	/** the area in which the ball has to be to be kickable */
	private final Vector3D[] shootPoly;

	/** get stopped on both legs */
	protected Movement standStill;

	/** move to stand on left leg only */
	protected Movement getOnLeftLeg;

	/** move to stand on right leg only */
	protected Movement getOnRightLeg;

	/**
	 * @param name the name of the kick
	 * @param worldModel reference to information around us
	 * @param agentModel reference to the own information model
	 * @param kickDirection angle in which the shoot will go relative to the body
	 *        direction
	 * @param ballFacingDirection direction we have to face relative to the ball
	 * @param maxBallDistance max distance to the ball to hit the ball
	 * @param maxKickDistance the maximum kick distance this kick is able to kick
	 * @param shootPoly polygon inside which the ball has to be to be kickable
	 * @param initCount the value to start counting with cnt
	 */
	public Kick(String name, IGlobalMap worldModel, IAgentModel agentModel,
			Angle kickDirection, Angle ballFacingDirection,
			double maxBallDistance, double maxKickDistance, Vector3D[] shootPoly)
	{
		super(name, worldModel, agentModel);
		this.kickDirection = kickDirection;
		this.ballFacingDirection = ballFacingDirection;
		this.maxBallDistance = maxBallDistance;
		this.maxKickDistance = maxKickDistance;
		this.shootPoly = shootPoly;

		standStill = new Movement("standStill");
		standStill.add(new MovementPhase("standStill", 50) //
				.add(INaoConstants.LShoulderPitch, -80, 3f) //
				.add(INaoConstants.RShoulderPitch, -80, 3f) //
				.add(INaoConstants.LHipYawPitch, 0, 0.5f) //
				.add(INaoConstants.RHipYawPitch, 0, 0.5f) //
				.add(INaoConstants.LHipRoll, 0, 0.5f) //
				.add(INaoConstants.LFootRoll, 0, 0.5f) //
				.add(INaoConstants.RHipRoll, 0, 0.5f) //
				.add(INaoConstants.RFootRoll, 0, 0.5f) //
				.add(INaoConstants.LHipPitch, 15, 0.5f) //
				.add(INaoConstants.LKneePitch, -30, 1f) //
				.add(INaoConstants.LFootPitch, 15, 0.5f) //
				.add(INaoConstants.RHipPitch, 15, 0.5f) //
				.add(INaoConstants.RKneePitch, -30, 1f) //
				.add(INaoConstants.RFootPitch, 15, 0.5f) //
				);

		getOnLeftLeg = new Movement("getOnLeftLeg");
		getOnLeftLeg.add(new MovementPhase("getOnLeftLeg", 10) //
				.add(INaoConstants.LHipPitch, 20, 2f) //
				.add(INaoConstants.LKneePitch, -40, 4f) //
				.add(INaoConstants.LFootPitch, 20, 2f) //
				.add(INaoConstants.RHipPitch, 0, 2f) //
				.add(INaoConstants.RKneePitch, 0, 4f) //
				.add(INaoConstants.RFootPitch, 0, 2f) //
				);
		getOnLeftLeg.add(new MovementPhase("getOnLeftLegSlowDown", 20) //
				.add(INaoConstants.LHipPitch, 20, 1f) //
				.add(INaoConstants.LKneePitch, -40, 2f) //
				.add(INaoConstants.LFootPitch, 20, 1f) //
				.add(INaoConstants.LHipRoll, 8, 1f) //
				.add(INaoConstants.LFootRoll, 8, 1f) //
				);

		getOnRightLeg = new Movement("getOnRightLeg");
		getOnRightLeg.add(new MovementPhase("getOnRightLeg", 10)
				.add(INaoConstants.RHipPitch, 20, 2f) //
				.add(INaoConstants.RKneePitch, -40, 4f) //
				.add(INaoConstants.RFootPitch, 20, 2f) //
				.add(INaoConstants.LHipPitch, 0, 2f) //
				.add(INaoConstants.LKneePitch, 0, 4f) //
				.add(INaoConstants.LFootPitch, 0, 2f) //
				);
		getOnRightLeg.add(new MovementPhase("getOnRightLegSlowDown", 10) //
				.add(INaoConstants.RHipPitch, 20, 1f) //
				.add(INaoConstants.RKneePitch, -40, 2f) //
				.add(INaoConstants.RFootPitch, 20, 1f) //
				.add(INaoConstants.RFootRoll, -8, 1f) //
				.add(INaoConstants.RHipRoll, -8, 1f) //
				);
	}

	/**
	 * @return the kickable rectangle
	 */
	@Override
	public Vector3D[] getKickableRectangle()
	{
		return shootPoly;
	}

	/**
	 * @return the horizontal direction of the kick (local)
	 */
	@Override
	public Angle getKickDirection()
	{
		return kickDirection;
	}

	/**
	 * @return the horizontal body direction relative to the ball
	 */
	@Override
	public Angle getBallFacingDirection()
	{
		return ballFacingDirection;
	}

	/**
	 * @return the maximum distance to the ball in which we hit
	 */
	@Override
	public double getMaxBallDistance()
	{
		return maxBallDistance;
	}

	/**
	 * @return the maximum kick distance this kick is able to kick
	 */
	@Override
	public double getMaxKickDistance()
	{
		return maxKickDistance;
	}

	/**
	 * @param util
	 */
	@Override
	public void setUtility(double util)
	{
		this.utility = util;
	}

	/**
	 * @return the utility
	 */
	@Override
	public double getUtility()
	{
		return utility;
	}

	@Override
	public void init()
	{
		super.init();
		switchTo(standStill);
	}
}
