/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.movement;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.movement.Movement;
import magma.agent.behavior.movement.MovementPhase;
import magma.agent.worldmodel.IGlobalMap;
import magma.robots.nao.INaoConstants;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Kick ball straight backward. This Behavior first stabilizes and then decides
 * to kick either with the left or the right.
 * 
 * @author Stefan Glaser
 */
public class KickBack extends KickStraight
{
	private static final Angle KICK_DIRECTION = Angle.RAD180;

	private static final Angle BALL_FACING_DIRECTION = Angle.RAD180;

	private static final double BEHIND_BALL_DISTANCE = 0.15;

	private static final double MAX_KICK_DISTANCE = 3;

	/** length of the kickable area */
	private static final double MAX_X_KICK = 0.18;

	/** width of the kickable area */
	private static final double MAX_Y_KICK = 0.09;

	private static final Vector3D[] SHOOT_POLY = Geometry
			.getPolygon(new double[][] { { 0, MAX_Y_KICK }, { 0, -MAX_Y_KICK },
					{ -MAX_X_KICK, -MAX_Y_KICK }, { -MAX_X_KICK, MAX_Y_KICK } });

	/**
	 * Default constructor
	 * 
	 * @param worldModel - the world model
	 * @param agentModel - the agent model
	 */
	public KickBack(IGlobalMap worldModel, IAgentModel agentModel)
	{
		super(IBehaviorConstants.SHOOT_BACK, worldModel, agentModel,
				KICK_DIRECTION, BALL_FACING_DIRECTION, BEHIND_BALL_DISTANCE,
				MAX_KICK_DISTANCE, SHOOT_POLY);

		internalShootPoly = Geometry.getPolygon(new double[][] {
				{ 0, MAX_Y_KICK + 0.04 }, { 0, -MAX_Y_KICK - 0.04 },
				{ -MAX_X_KICK - 0.2, -MAX_Y_KICK - 0.04 },
				{ -MAX_X_KICK - 0.2, MAX_Y_KICK + 0.04 } });

		kickRight = new Movement("kickRightStraight");
		kickRight.add(new MovementPhase("moveLegForwards", 10)
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.RFootRoll, 16, 2f) //
				.add(INaoConstants.RHipPitch, 35, 3f) //
				.add(INaoConstants.RKneePitch, 0, 2f) //
				);
		kickRight.add(new MovementPhase("kickRight", 10)
				.add(INaoConstants.RHipPitch, -45, 7f) //
				.add(INaoConstants.RKneePitch, 0, 2f) //
				.add(INaoConstants.RFootPitch, 15, 2f) //
				.add(INaoConstants.LHipPitch, 35, 1f) //
				);
		kickRight.add(new MovementPhase("moveStraight", 20).add(
				INaoConstants.RHipPitch, 0, 2f) //
				.add(INaoConstants.RFootPitch, 0, 2f) //
				);
		kickRight.add(new MovementPhase("getOnTwoFeet", 20)
				.add(INaoConstants.LFootRoll, 0, 1f) //
				.add(INaoConstants.LHipRoll, 0, 1f) //
				.add(INaoConstants.RFootRoll, 0, 1f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //

				.add(INaoConstants.LHipPitch, 15, 2f) //
				.add(INaoConstants.LKneePitch, -30, 4f) //
				.add(INaoConstants.LFootPitch, 15, 2f) //
				.add(INaoConstants.RHipPitch, 15, 2f) //
				.add(INaoConstants.RKneePitch, -30, 4f) //
				.add(INaoConstants.RFootPitch, 15, 2f) //
				);

		kickLeft = kickRight.getLeftVersion();

		// trigger normal behavior, no fast version
		kickRightFast = getOnLeftLeg;
		kickLeftFast = getOnRightLeg;
	}
}
