/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.movement;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.movement.Movement;
import magma.agent.behavior.movement.MovementPhase;
import magma.agent.worldmodel.IGlobalMap;
import magma.robots.nao.INaoConstants;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Kick ball backward by standing on it and pulling it back before kicking.
 * 
 * @author Stefan Glaser, Klaus Dorer
 */
public class KickBackPull extends KickStraight
{
	private static final Angle KICK_DIRECTION = Angle.RAD180;

	private static final Angle BALL_FACING_DIRECTION = Angle.RAD0;

	private static final double BEHIND_BALL_DISTANCE = 0.12;

	private static final double MAX_KICK_DISTANCE = 3;

	/** length of the kickable area */
	private static final double MAX_X_KICK = 0.18;

	/** width of the kickable area */
	private static final double MAX_Y_KICK = 0.05;

	private static final Vector3D[] SHOOT_POLY = Geometry
			.getPolygon(new double[][] { { 0, MAX_Y_KICK },
					{ MAX_X_KICK, MAX_Y_KICK }, { MAX_X_KICK, -MAX_Y_KICK },
					{ 0, -MAX_Y_KICK } });

	/**
	 * Default constructor
	 * 
	 * @param worldModel - the world model
	 * @param agentModel - the agent model
	 */
	public KickBackPull(IGlobalMap worldModel, IAgentModel agentModel)
	{
		super(IBehaviorConstants.SHOOT_BACK_PULL, worldModel, agentModel,
				KICK_DIRECTION, BALL_FACING_DIRECTION, BEHIND_BALL_DISTANCE,
				MAX_KICK_DISTANCE, SHOOT_POLY);

		internalShootPoly = Geometry.getPolygon(new double[][] {
				{ 0, MAX_Y_KICK + 0.02 }, { MAX_X_KICK + 0.02, MAX_Y_KICK + 0.02 },
				{ MAX_X_KICK + 0.02, -MAX_Y_KICK - 0.02 },
				{ 0, -MAX_Y_KICK - 0.02 } });

		kickRight = new Movement("kickRightStraight");
		kickRight.add(new MovementPhase("moveLegUpwards", 30)
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.RFootRoll, 16, 2f) //
				.add(INaoConstants.RHipPitch, 60, 2f) //
				.add(INaoConstants.RKneePitch, -70, 7f) //
				.add(INaoConstants.RFootPitch, 40, 4f) //
				);
		kickRight.add(new MovementPhase("moveLegForwards", 15)
				.add(INaoConstants.RHipPitch, 25, 2.2f) //
				.add(INaoConstants.RKneePitch, 0, 5f) //
				.add(INaoConstants.RFootPitch, -40, 6f) //
				);
		kickRight.add(new MovementPhase("moveLegOnBall", 7).add(
				INaoConstants.RHipPitch, 18, 1f) //
				);
		kickRight.add(new MovementPhase("pullLegBackwards", 4)
				.add(INaoConstants.RHipPitch, 30, 3f) //
				.add(INaoConstants.RKneePitch, -10, 2f) //
				.add(INaoConstants.RFootPitch, -30, 2f) //
				);
		kickRight.add(new MovementPhase("moveLegForwards", 14)
				.add(INaoConstants.RHipPitch, 35, 4f) //
				.add(INaoConstants.RKneePitch, 0, 1f) //
				.add(INaoConstants.RFootPitch, 0, 7f) //
				);
		kickRight.add(new MovementPhase("moveLegForwards", 10)
				.add(INaoConstants.RHipPitch, -45, 7f) //
				.add(INaoConstants.RKneePitch, 0, 2f) //
				.add(INaoConstants.RFootPitch, 15, 2f) //
				.add(INaoConstants.LHipPitch, 35, 1f) //
				);
		kickRight.add(new MovementPhase("moveLegStraight", 10).add(
				INaoConstants.RHipPitch, 0, 2f) //
				.add(INaoConstants.RFootPitch, 0, 2f) //
				);
		kickRight.add(new MovementPhase("getOnTwoFeet", 10)
				.add(INaoConstants.LFootRoll, 0, 1f) //
				.add(INaoConstants.LHipRoll, 0, 1f) //
				.add(INaoConstants.RFootRoll, 0, 1f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //

				.add(INaoConstants.LHipPitch, 15, 2f) //
				.add(INaoConstants.LKneePitch, -30, 4f) //
				.add(INaoConstants.LFootPitch, 15, 2f) //
				.add(INaoConstants.RHipPitch, 15, 2f) //
				.add(INaoConstants.RKneePitch, -30, 4f) //
				.add(INaoConstants.RFootPitch, 15, 2f) //
				);

		kickLeft = kickRight.getLeftVersion();

		// trigger normal behavior, no fast version
		kickRightFast = getOnLeftLeg;
		kickLeftFast = getOnRightLeg;
	}

	protected void performRightKick()
	{
		// left in to look up how to print contact information
		// IForceResistance contact = agentModel
		// .getForceResistance(INaoConstants.RFootForce);
		// double force = contact.getForce().length();
		// if (Math.abs(force) > 0.01) {
		// System.out.println(cnt + " contact: " + force);
		// }

		// System.out.println("phase: " + phase + " cnt: " + cnt + " hip: "
		// + agentModel.getHJ(INaoConstants.RHipPitch).getAxis() + " knee: "
		// + agentModel.getHJ(INaoConstants.RKneePitch).getAxis() + " foot: "
		// + agentModel.getHJ(INaoConstants.RFootPitch).getAxis());
	}
}
