/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.movement;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.movement.Movement;
import magma.agent.behavior.movement.MovementPhase;
import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IThisPlayer;
import magma.robots.nao.INaoConstants;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Kick ball straight forward. This Behavior first stabilizes and then decides
 * to kick either with the left or the right. It also decides if we should do a
 * fast-kick, if an opponent is close by.
 * 
 * @author Stefan Glaser
 */
public class KickStraight extends Kick
{
	private static final Angle KICK_DIRECTION = Angle.RAD0;

	private static final Angle BALL_FACING_DIRECTION = Angle.RAD0;

	private static final double BEHIND_BALL_DISTANCE = 0.15;

	private static final double MAX_KICK_DISTANCE = 4.7;

	/** length of the kickable area */
	private static final double MAX_X_KICK = 0.18;

	/** width of the kickable area */
	private static final double MAX_Y_KICK = 0.09;

	private static final Vector3D[] SHOOT_POLY = Geometry
			.getPolygon(new double[][] { { 0, MAX_Y_KICK },
					{ MAX_X_KICK, MAX_Y_KICK }, { MAX_X_KICK, -MAX_Y_KICK },
					{ 0, -MAX_Y_KICK } });

	/** used to check if the ball is still kickable */
	Vector3D[] internalShootPoly;

	/** movement to kick with right leg */
	protected Movement kickRight;

	/** movement to fast kick with right leg */
	protected Movement kickRightFast;

	/** movement to kick with left leg */
	protected Movement kickLeft;

	/** movement to fast kick with left leg */
	protected Movement kickLeftFast;

	/**
	 * Default constructor
	 * 
	 * @param worldModel World model
	 * @param agentModel Agent model
	 */
	public KickStraight(String name, IGlobalMap worldModel,
			IAgentModel agentModel, Angle kickDirection,
			Angle ballFacingDirection, double maxBallDistance,
			double maxKickDistance, Vector3D[] shootPoly)
	{
		super(name, worldModel, agentModel, kickDirection, ballFacingDirection,
				maxBallDistance, maxKickDistance, shootPoly);

		internalShootPoly = Geometry
				.getPolygon(new double[][] { { 0, MAX_Y_KICK + 0.04 },
						{ MAX_X_KICK + 0.2, MAX_Y_KICK + 0.04 },
						{ MAX_X_KICK + 0.2, -MAX_Y_KICK - 0.04 },
						{ 0, -MAX_Y_KICK - 0.04 } });

		kickRight = new Movement("kickRightStraight");
		kickRight.add(new MovementPhase("moveLegBackwards", 15) //
				.add(INaoConstants.LHipPitch, 10, 0.5f) //
				.add(INaoConstants.LKneePitch, -20, 1f) //
				.add(INaoConstants.LFootPitch, 10, 0.5f) //
				.add(INaoConstants.RHipPitch, -5, 2f) //
				.add(INaoConstants.RKneePitch, -10, 2f) //
				.add(INaoConstants.RFootPitch, 10, 4f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.RFootRoll, 16, 2f) //
				);
		kickRight.add(new MovementPhase("moveLegBackwards2", 20) //
				.add(INaoConstants.LHipPitch, 20, 0.5f) //
				.add(INaoConstants.LKneePitch, -20, 1f) //
				.add(INaoConstants.LFootPitch, 10, 0.5f) //
				.add(INaoConstants.RHipPitch, 10, 2f) //
				.add(INaoConstants.RKneePitch, -50, 2f) //
				.add(INaoConstants.RFootPitch, 20, 2f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.RFootRoll, 16, 2f) //
				);

		kickRight.add(new MovementPhase("kickRight", 15) //
				.add(INaoConstants.LHipPitch, -10, 7f) //
				.add(INaoConstants.LFootPitch, 20, 5f) //
				.add(INaoConstants.RHipPitch, 35, 7f) //
				.add(INaoConstants.RKneePitch, 0, 7f) //
				.add(INaoConstants.RFootPitch, -15, 7f) //
				);

		kickRight.add(new MovementPhase("kickRight2", 30) //
				.add(INaoConstants.RHipPitch, 0, 2f) //
				.add(INaoConstants.RFootPitch, 0, 2f) //
				);
		kickRight.add(new MovementPhase("kickRight3", 20) //
				.add(INaoConstants.LHipPitch, 15, 2f) //
				.add(INaoConstants.LKneePitch, -30, 4f) //
				.add(INaoConstants.LFootPitch, 15, 2f) //
				.add(INaoConstants.RHipPitch, 15, 2f) //
				.add(INaoConstants.RKneePitch, -30, 4f) //
				.add(INaoConstants.RFootPitch, 15, 2f) //
				.add(INaoConstants.LHipRoll, 0, 1f) //
				.add(INaoConstants.LFootRoll, 0, 1f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.RFootRoll, 0, 1f) //
				);

		kickLeft = kickRight.getLeftVersion();

		kickRightFast = new Movement("kickRightStraightFast");
		kickRightFast.add(new MovementPhase("getOnOneLegFast1", 3)
				.add(INaoConstants.RFootRoll, 8, 2f) //
				.add(INaoConstants.RHipRoll, -8, 2f) //
				.add(INaoConstants.LFootRoll, 8, 2f) //
				.add(INaoConstants.LHipRoll, -8, 2f) //

				.add(INaoConstants.LHipPitch, 0, 4.5f) //
				.add(INaoConstants.LKneePitch, 0, 7f) //
				.add(INaoConstants.LFootPitch, 0, 2.5f) //

				.add(INaoConstants.RHipPitch, 0, 3.5f) //
				.add(INaoConstants.RKneePitch, 0, 7f) //
				.add(INaoConstants.RFootPitch, 0, 3.5f) //
				);
		kickRightFast.add(new MovementPhase("getOnOneLegFast2", 4)
				.add(INaoConstants.RFootRoll, 10, 1.5f) //
				.add(INaoConstants.RHipRoll, -10, 1.5f) //
				.add(INaoConstants.LFootRoll, 10, 1.5f) //
				.add(INaoConstants.LHipRoll, -10, 1.5f) //

				.add(INaoConstants.LHipPitch, 0, 3.5f) //
				.add(INaoConstants.LKneePitch, 0, 7f) //
				.add(INaoConstants.LFootPitch, 0, 3.5f) //
				);
		kickRightFast.add(new MovementPhase("moveLegBackward1", 14)
				.add(INaoConstants.LHipPitch, 5, 1f) //

				.add(INaoConstants.RHipPitch, 25, 2f) //
				.add(INaoConstants.RKneePitch, -40, 4f) //
				.add(INaoConstants.RFootPitch, 15, 4f) //
				);
		kickRightFast.add(new MovementPhase("kickFast", 10)
				.add(INaoConstants.LHipPitch, -15, 7f) //
				.add(INaoConstants.LKneePitch, 0, 2f) //
				.add(INaoConstants.LFootPitch, 5, 2f) //

				.add(INaoConstants.RHipPitch, 55, 7f) //
				.add(INaoConstants.RKneePitch, 0, 7f) //
				.add(INaoConstants.RFootPitch, -15, 7f) //
				);
		kickRightFast.add(new MovementPhase("moveToNormal1", 10)
				.add(INaoConstants.LHipPitch, 0, 2f) //
				.add(INaoConstants.LKneePitch, 0, 4f) //
				.add(INaoConstants.LFootPitch, 0, 2f) //

				.add(INaoConstants.RHipPitch, 20, 2f) //
				.add(INaoConstants.RKneePitch, -40, 4f) //
				.add(INaoConstants.RFootPitch, 20, 4f) //
				);
		kickRightFast.add(new MovementPhase("moveToNormal2", 20)
				.add(INaoConstants.RFootRoll, 0, 1f) //
				.add(INaoConstants.RHipRoll, 0, 1f) //
				.add(INaoConstants.LFootRoll, 0, 1f) //
				.add(INaoConstants.LHipRoll, 0, 1f) //
				.add(INaoConstants.RHipPitch, 15, 2f) //
				.add(INaoConstants.RKneePitch, -30, 4f) //
				.add(INaoConstants.RFootPitch, 15, 2f) //
				.add(INaoConstants.LHipPitch, 15, 2f) //
				.add(INaoConstants.LKneePitch, -30, 4f) //
				.add(INaoConstants.LFootPitch, 15, 2f) //
				);

		kickLeftFast = kickRightFast.getLeftVersion();

		currentMovement = standStill;
	}

	/**
	 * Default constructor
	 * 
	 * @param worldModel - the world model
	 * @param agentModel - the agent model
	 */
	public KickStraight(IGlobalMap worldModel, IAgentModel agentModel)
	{
		this(IBehaviorConstants.SHOOT_STRAIGHT, worldModel, agentModel,
				KICK_DIRECTION, BALL_FACING_DIRECTION, BEHIND_BALL_DISTANCE,
				MAX_KICK_DISTANCE, SHOOT_POLY);
	}

	@Override
	public void perform()
	{
		if (currentMovement.isFinished()) {
			if (currentMovement == standStill) {
				IThisPlayer thisPlayer = worldModel.getThisPlayer();
				Vector3D ballPos = worldModel.getBall().getPosition();
				if (thisPlayer.isInsidePolygonXY(ballPos, internalShootPoly)) {
					// ball is kickable by both feet
					boolean kickRight = true;
					if (thisPlayer.positionIsLeft(ballPos)) {
						kickRight = false;
					}
					IPlayer opponentAtBall = worldModel.getOpponentAtBall();
					boolean shootFast = false;
					if (opponentAtBall != null) {
						shootFast = opponentAtBall.getDistanceToXY(worldModel
								.getBall()) < 1
								&& Math.abs(worldModel.getThisPlayer()
										.getBodyDirectionTo(opponentAtBall.getPosition())
										.degrees()) < 100;
					}

					if (shootFast) {
						if (kickRight) {
							switchTo(kickRightFast);
						} else {
							switchTo(kickLeftFast);
						}

					} else {
						if (kickRight) {
							switchTo(getOnLeftLeg);
						} else {
							switchTo(getOnRightLeg);
						}
					}

				} else {
					isFinished = true;
					switchTo(standStill);
					return;
				}
			} else if (currentMovement == getOnLeftLeg) {
				switchTo(kickRight);
			} else if (currentMovement == getOnRightLeg) {
				switchTo(kickLeft);
			} else if (currentMovement == kickRight || currentMovement == kickLeft
					|| currentMovement == kickRightFast
					|| currentMovement == kickLeftFast) {
				switchTo(getReady);
			} else {
				isFinished = true;
				switchTo(standStill);
				return;
			}
		}

		super.perform();
	}

	@Override
	public void init()
	{
		super.init();
		switchTo(standStill);
	}
}
