/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.behavior.movement;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehaviorConstants;
import magma.agent.behavior.movement.Movement;
import magma.agent.behavior.movement.MovementPhase;
import magma.agent.worldmodel.IGlobalMap;
import magma.robots.nao.INaoConstants;
import magma.util.geometry.Angle;
import magma.util.geometry.Geometry;

import org.apache.commons.math.geometry.Vector3D;

/**
 * This behavior kicks somehow to the left with the right foot.
 * 
 * @author Stefan Glaser
 * 
 */
public class SideKickRight extends Kick
{
	private static final Angle KICK_DIRECTION = Angle.deg(90);

	private static final Angle BALL_FACING_DIRECTION = Angle.RAD0;

	private static final double BEHIND_BALL_DISTANCE = 0.15;

	private static final double MAX_KICK_DISTANCE = 3.5;

	/** length of the kickable area */
	private static final double MAX_X_KICK = 0.18;

	/** width of the kickable area */
	private static final double MAX_Y_KICK = 0.04;

	private static final Vector3D[] SHOOT_POLY = Geometry
			.getPolygon(new double[][] { { 0, MAX_Y_KICK },
					{ MAX_X_KICK, MAX_Y_KICK }, { MAX_X_KICK, -MAX_Y_KICK },
					{ 0, -MAX_Y_KICK } });

	/** movement to kick with right leg */
	protected Movement kickSide;

	/**
	 * Default constructor.
	 * 
	 * @param worldModel - the world model
	 * @param agentModel - the agent model
	 */
	public SideKickRight(IGlobalMap worldModel, IAgentModel agentModel)
	{
		this(IBehaviorConstants.SHOOT_SIDE_RIGHT, worldModel, agentModel,
				KICK_DIRECTION, BALL_FACING_DIRECTION, BEHIND_BALL_DISTANCE,
				MAX_KICK_DISTANCE, SHOOT_POLY);
	}

	public SideKickRight(String name, IGlobalMap worldModel,
			IAgentModel agentModel, Angle kickDirection,
			Angle ballFacingDirection, double maxBallDistance,
			double maxKickDistance, Vector3D[] shootPoly)
	{
		super(name, worldModel, agentModel, kickDirection, ballFacingDirection,
				maxBallDistance, maxKickDistance, shootPoly);

		getOnLeftLeg = new Movement("kickRightStraight");
		getOnLeftLeg.add(new MovementPhase("getOnLeftLeg1", 10)
				.add(INaoConstants.LHipRoll, -15, 1f) //
				.add(INaoConstants.LFootRoll, 15, 1f) //
				.add(INaoConstants.RHipRoll, -15, 1f) //
				.add(INaoConstants.RFootRoll, 15, 1f) //
				);
		getOnLeftLeg.add(new MovementPhase("getOnLeftLeg2", 15).add(
				INaoConstants.RHipRoll, -25, 1f) //
				);

		kickSide = new Movement("kickRightSide");
		kickSide.add(new MovementPhase("sideKick0", 15)
				.add(INaoConstants.LHipYawPitch, 0, .5f) //
				.add(INaoConstants.LHipPitch, 15, .5f) //
				.add(INaoConstants.LHipRoll, -17, 1f) //
				.add(INaoConstants.LKneePitch, -30, 1f) //
				.add(INaoConstants.LFootPitch, 15, .5f) //
				.add(INaoConstants.LFootRoll, 15, 1f) //
				.add(INaoConstants.RHipYawPitch, -20, 1.5f) //
				.add(INaoConstants.RHipRoll, -30, 1f) //
				.add(INaoConstants.RFootRoll, 20, 1f) //
				);
		kickSide.add(new MovementPhase("sideKick1", 10)
				.add(INaoConstants.RHipYawPitch, -30, 3f) //
				.add(INaoConstants.RHipPitch, 25, 1f) //
				.add(INaoConstants.RKneePitch, 0, 2.5f) //
				.add(INaoConstants.RFootPitch, -45, 4f) //
				.add(INaoConstants.RFootRoll, -25, 3f) //
				);
		kickSide.add(new MovementPhase("sideKick2", 10)
				.add(INaoConstants.LHipYawPitch, 0, .5f) //
				.add(INaoConstants.LHipPitch, 15, .5f) //
				.add(INaoConstants.LHipRoll, -17, 1f) //
				.add(INaoConstants.LKneePitch, -30, 1f) //
				.add(INaoConstants.LFootPitch, 15, .5f) //
				.add(INaoConstants.LFootRoll, 15, 1f) //
				.add(INaoConstants.RHipYawPitch, -1, 7f) //
				.add(INaoConstants.RHipPitch, 45, 2.0f) //
				.add(INaoConstants.RHipRoll, 10, 7f) //
				.add(INaoConstants.RKneePitch, 0, 7f) //
				.add(INaoConstants.RFootPitch, -25, 7f) //
				.add(INaoConstants.RFootRoll, 25, 1.5f) //
				);
		kickSide.add(new MovementPhase("sideKick3", 10)
				.add(INaoConstants.LHipYawPitch, 0, .5f) //
				.add(INaoConstants.LHipPitch, 15, .5f) //
				.add(INaoConstants.LHipRoll, -17, .5f) //
				.add(INaoConstants.LKneePitch, -30, 1f) //
				.add(INaoConstants.LFootPitch, 15, .5f) //
				.add(INaoConstants.LFootRoll, 15, .5f) //
				.add(INaoConstants.RHipYawPitch, 0, .5f) //
				.add(INaoConstants.RHipPitch, 35, 1f) //
				.add(INaoConstants.RHipRoll, -10, 2f) //
				.add(INaoConstants.RKneePitch, -30, 1f) //
				.add(INaoConstants.RFootPitch, -25, .5f) //
				.add(INaoConstants.RFootRoll, 0, 2.5f) //
				);
		kickSide.add(new MovementPhase("sideKick4", 30)
				.add(INaoConstants.LHipYawPitch, 0, .5f) //
				.add(INaoConstants.LHipPitch, 15, .5f) //
				.add(INaoConstants.LHipRoll, 0, .5f) //
				.add(INaoConstants.LKneePitch, -30, 1f) //
				.add(INaoConstants.LFootPitch, 15, .5f) //
				.add(INaoConstants.LFootRoll, 0, .5f) //
				.add(INaoConstants.RHipYawPitch, 0, .5f) //
				.add(INaoConstants.RHipPitch, 15, .5f) //
				.add(INaoConstants.RHipRoll, 0, .5f) //
				.add(INaoConstants.RKneePitch, -30, 1f) //
				.add(INaoConstants.RFootPitch, 15, 1.35f) //
				.add(INaoConstants.RFootRoll, 0, .5f) //
				);
		currentMovement = standStill;
	}

	@Override
	public void perform()
	{
		if (currentMovement.isFinished()) {
			if (currentMovement == standStill) {
				switchTo(getOnLeftLeg);
			} else if (currentMovement == getOnLeftLeg) {
				switchTo(kickSide);
			} else if (currentMovement == kickSide) {
				switchTo(getReady);
			} else {
				isFinished = true;
				switchTo(standStill);
				return;
			}
		}

		super.perform();
	}
}
