/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.robots.nao.general;

import java.util.Map;

import magma.agent.agentmodel.IAgentModel;
import magma.agent.behavior.IBehavior;
import magma.agent.behavior.basic.BehaviorParameter;
import magma.agent.behavior.complex.Attack;
import magma.agent.behavior.complex.DribbleBall;
import magma.agent.behavior.complex.GetInScorePosition;
import magma.agent.behavior.complex.GoaliePositioning;
import magma.agent.behavior.complex.Keep;
import magma.agent.behavior.complex.MoveWithBall;
import magma.agent.behavior.complex.PassivePositioning;
import magma.agent.behavior.complex.RunToPosition;
import magma.agent.behavior.complex.ShootToGoal;
import magma.agent.behavior.complex.WalkForward;
import magma.agent.belief.IBelief;
import magma.agent.belief.impl.BallInScorableArea;
import magma.agent.belief.impl.BallIsKeepable;
import magma.agent.belief.impl.ResourceBelieve;
import magma.agent.general.ComponentFactory;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.worldmodel.IGlobalMap;
import magma.robots.nao.INaoConstants;
import magma.robots.nao.behavior.basic.Cry;
import magma.robots.nao.behavior.basic.Dance;
import magma.robots.nao.behavior.complex.Walk;
import magma.robots.nao.behavior.dynamic.FocusBall;
import magma.robots.nao.behavior.dynamic.walk.Balance;
import magma.robots.nao.behavior.dynamic.walk.WalkBehaviorParameter;
import magma.robots.nao.behavior.movement.KeeperLeft;
import magma.robots.nao.behavior.movement.KeeperRight;
import magma.robots.nao.behavior.movement.KickBack;
import magma.robots.nao.behavior.movement.KickBackPull;
import magma.robots.nao.behavior.movement.KickStraight;
import magma.robots.nao.behavior.movement.SideKickLeft;
import magma.robots.nao.behavior.movement.SideKickRight;
import magma.robots.nao.belief.impl.SoccerPosition;
import magma.robots.nao.meta.agent.NaoAgentMetaModel;

public class NAOComponentFactory extends ComponentFactory
{
	/**
	 * Create a agent meta model
	 * 
	 * @return the agent meta model
	 */
	@Override
	public IAgentMetaModel createAgentMetaModel()
	{
		return new NaoAgentMetaModel();
	}

	@Override
	protected void createSpecificBeliefs(IAgentModel agentModel,
			IGlobalMap worldModel, Map<String, IBelief> beliefs)
	{
		IBelief belief = new ResourceBelieve("resourceBody", 1);
		beliefs.put(belief.getName(), belief);

		belief = new BallIsKeepable(agentModel, worldModel,
				INaoConstants.PLAYER_HEIGHT);
		beliefs.put(belief.getName(), belief);

		belief = new SoccerPosition(agentModel, worldModel);
		beliefs.put(belief.getName(), belief);

		belief = new BallInScorableArea(agentModel, worldModel,
				INaoConstants.KICKABLE_DISTANCE);
		beliefs.put(belief.getName(), belief);
	}

	@Override
	protected String getBehaviorFilesBasePath()
	{
		return "behaviors/nao/";
	}

	@Override
	protected void createSpecificBehaviors(IAgentModel agentModel,
			IGlobalMap worldModel, BehaviorParameter params,
			Map<String, IBehavior> behaviors)
	{

		// General behaviors
		IBehavior behavior = new FocusBall(worldModel, agentModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		// Movement behaviors
		// behavior = new RunToBall(worldModel, agentModel, params, behaviors);
		// behaviors.put(behavior.getName(), behavior);

		behavior = new RunToPosition(worldModel, agentModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new DribbleBall(worldModel, agentModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new Balance(worldModel, agentModel,
				new WalkBehaviorParameter());
		behaviors.put(behavior.getName(), behavior);

		behavior = new Walk(worldModel, agentModel, new WalkBehaviorParameter(),
				behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new WalkForward(worldModel, agentModel,
				new WalkBehaviorParameter(), behaviors);
		behaviors.put(behavior.getName(), behavior);

		// Positioning behaviors
		behavior = new GetInScorePosition(worldModel, agentModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new MoveWithBall(worldModel, agentModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new PassivePositioning(worldModel, agentModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new GoaliePositioning(worldModel, agentModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		// Kick behaviors
		behavior = new ShootToGoal(worldModel, agentModel, behaviors);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KickStraight(worldModel, agentModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KickBack(worldModel, agentModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KickBackPull(worldModel, agentModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new SideKickLeft(worldModel, agentModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new SideKickRight(worldModel, agentModel);
		behaviors.put(behavior.getName(), behavior);

		// Keeper behavior
		behavior = new KeeperLeft(worldModel, agentModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new KeeperRight(worldModel, agentModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new Keep(worldModel, agentModel, behaviors,
				INaoConstants.PLAYER_HEIGHT);
		behaviors.put(behavior.getName(), behavior);

		// Game result reflection behaviors
		behavior = new Cry(worldModel, agentModel);
		behaviors.put(behavior.getName(), behavior);

		behavior = new Dance(worldModel, agentModel);
		behaviors.put(behavior.getName(), behavior);

		// Other behaviors
		behavior = new Attack(worldModel, agentModel, behaviors);
		behaviors.put(behavior.getName(), behavior);
	}
}
