/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util;

import java.security.InvalidParameterException;
import java.util.Collection;
import java.util.List;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

import magma.agent.worldmodel.localizer.IReferenceLine;
import magma.agent.worldmodel.localizer.IReferencePoint;
import magma.util.geometry.MatrixUtil;
import magma.util.geometry.SingularValueDecomposition;

import org.apache.commons.math.geometry.Vector3D;

/**
 * This class provides several util functions for localization.
 * 
 * @author Stefan Glaser
 * 
 */
public class LocalizerUtil
{
	/**
	 * Calculates a normal vector to the plane defined by the point set. To do
	 * so, at least 3 points on the plane has to be passed.<br>
	 * This method performs an SVD of the covariance matrix to get the smallest
	 * eigenvector (which is already a normal vector of the least square fitting
	 * plane). After the calculation of the eigenvector it is ensured, that the
	 * returned normal vector always points towards the center of the coordinate
	 * system. (this means for example, if this method is used to compute the
	 * normal of a soccer pitch by a bunch of points on ground level, the
	 * resulting normal points always to the sky, since we can't be below the
	 * soccer pitch)
	 * 
	 * @param points the points forming a plane
	 * @return a normal vector of the plane, always pointing towards the center
	 *         of the coordinate system
	 */
	public static Vector3d calculatePlaneNormal(List<Vector3D> points)
	{
		// Check for minimum 3 points
		if (points.size() < 3) {
			return null;
		}

		Vector3D centroid;
		double[][] cov = new double[points.size()][3];
		double x = 0, y = 0, z = 0;

		// Calculate centroid
		for (Vector3D vec : points) {
			x += vec.getX();
			y += vec.getY();
			z += vec.getZ();
		}
		centroid = new Vector3D(x / points.size(), y / points.size(), z
				/ points.size());

		// Calculate co-variance matrix
		int i = 0;
		for (Vector3D vec : points) {
			cov[i][0] = vec.getX() - centroid.getX();
			cov[i][1] = vec.getY() - centroid.getY();
			cov[i][2] = vec.getZ() - centroid.getZ();
			i++;
		}

		// Compute SVD
		SingularValueDecomposition svd = new SingularValueDecomposition(cov,
				points.size(), 3);
		double[][] V = svd.getV();
		// Take the eigenvector to the smalles eigenvalue (which should in our
		// case be the third one - so take the third column of V)
		Vector3d res = new Vector3d(V[0][2], V[1][2], V[2][2]);

		// Calculate direction to centroid
		Vector3d centroid3d = new Vector3d(centroid.getX(), centroid.getY(),
				centroid.getZ());
		centroid3d.normalize();

		// Check if the normal points towards the bottom - and if so, let it point
		// towards the sky
		if (centroid3d.dot(res) > 0) {
			res.negate();
		}

		return res;
	}

	/**
	 * This method calculates the position from a set of reference points and the
	 * given orientation. The position is calculated by using the given
	 * orientation to transform the local position of a reference point into the
	 * global coordinate system. After this transformation, we can simply
	 * subtract this vector from the known position to retrieve our position.
	 * With more than one reference point, the resulting positions to the
	 * different reference points are simply averaged.
	 * 
	 * @param refPoints - a collection of reference points for position
	 *        calculation
	 * @param orientation - the rotation matrix to transform local positions into
	 *        global positions
	 * @return the position
	 */
	public static Vector3D calculatePosition(
			Collection<IReferencePoint> refPoints, Matrix3d orientation)
	{
		return calculatePosition(refPoints, null, orientation);
	}

	/**
	 * This method calculates the position from a set of reference points and
	 * reference lines and the given orientation. The position is calculated by
	 * using the given orientation to transform the local position of a reference
	 * point / reference line-point into the global coordinate system. After this
	 * transformation, we can simply subtract this vector from the known position
	 * to retrieve our position. With more than one reference point, the
	 * resulting positions to the different reference points are simply averaged.
	 * 
	 * @param refPoints - a collection of reference points for position
	 *        calculation
	 * @param refLines - a collection of reference lines for position calculation
	 * @param orientation - the rotation matrix to transform local positions into
	 *        global positions
	 * @return the position
	 */
	public static Vector3D calculatePosition(
			Collection<IReferencePoint> refPoints,
			Collection<IReferenceLine> refLines, Matrix3d orientation)
	{
		// localized x-, y-, z-Position
		double x = 0;
		double y = 0;
		double z = 0;
		int noOfUsedRefPoints = 0;

		Vector3d myPos;
		Vector3D localPosition;
		Vector3D knownPosition;

		// iterate through all reference points
		if (refPoints != null) {
			for (IReferencePoint flag : refPoints) {
				myPos = calculatePosition(flag.getLocalPosition(),
						flag.getKnownPosition(), orientation);

				// add calculated position
				x += myPos.getX();
				y += myPos.getY();
				z += myPos.getZ();
				noOfUsedRefPoints++;
			}
		}

		// iterate through all reference lines
		// TODO: Check if line is not seen completely and therefore not matching
		// both known positions and adjust the known position of the not matching
		// reference point
		if (refLines != null) {
			for (IReferenceLine line : refLines) {
				localPosition = line.getLocalPosition1();
				if (localPosition != null) {
					knownPosition = line.getAdjustedKnownPosition1();
					myPos = calculatePosition(localPosition, knownPosition,
							orientation);

					// add calculated position
					x += myPos.getX();
					y += myPos.getY();
					z += myPos.getZ();
					noOfUsedRefPoints++;
				}

				localPosition = line.getLocalPosition1();
				if (localPosition != null) {
					knownPosition = line.getAdjustedKnownPosition2();
					myPos = calculatePosition(localPosition, knownPosition,
							orientation);

					// add calculated position
					x += myPos.getX();
					y += myPos.getY();
					z += myPos.getZ();
					noOfUsedRefPoints++;
				}
			}
		}

		if (noOfUsedRefPoints == 0) {
			return null;
		}

		// average results
		x /= noOfUsedRefPoints;
		y /= noOfUsedRefPoints;
		z /= noOfUsedRefPoints;

		return new Vector3D(x, y, z);
	}

	/**
	 * This method calculates the position based on a single local-to-known
	 * position combination and the given orientation.
	 * 
	 * @param localPosition - the position in the global body system
	 * @param knownPosition - the known position in the global system
	 * @param orientation - the transformation from the local to the global
	 *        system
	 * @return resulting position
	 */
	private static Vector3d calculatePosition(Vector3D localPosition,
			Vector3D knownPosition, Matrix3d orientation)
	{
		// transform view-system to torso-system
		Vector3d flagVector = new Vector3d(-localPosition.getY(),
				localPosition.getX(), localPosition.getZ());

		// calculate absolute Flag-Position
		flagVector = MatrixUtil.mulMatrixVector(orientation, flagVector);

		// transform gyro-system to view-system
		flagVector = new Vector3d(flagVector.getY(), -flagVector.getX(),
				flagVector.getZ());

		// create a flag-vector
		Vector3d myPos = new Vector3d(knownPosition.getX(), knownPosition.getY(),
				knownPosition.getZ());

		// myPos = flagPositionVecotr - flagViewVector
		myPos.sub(flagVector);

		return myPos;
	}

	/**
	 * An implementation of the umeyama point set registration method (known from
	 * image processing). The result is an 3D rotation matrix, which aligns the
	 * given moving point set best to the fixed point set, in the sense of least
	 * square distances.
	 * 
	 * @param movingVec - the array of moving points to register
	 * @param fixedVec - the array of corresponding fixed points
	 * @return the 3D rotation matrix, which aligns the given moving point set
	 *         best to the fixed point set, in the sense of least square
	 *         distances
	 */
	public static Matrix3d umeyama(Vector3d[] movingVec, Vector3d[] fixedVec)
	{
		if (movingVec.length != fixedVec.length) {
			throw new InvalidParameterException();
		}

		Matrix3d cov = new Matrix3d();
		Matrix3d temp;

		for (int i = 0; i < movingVec.length; i++) {
			temp = new Matrix3d();
			temp.m00 = fixedVec[i].x * movingVec[i].x;
			temp.m01 = fixedVec[i].x * movingVec[i].y;
			temp.m02 = fixedVec[i].x * movingVec[i].z;

			temp.m10 = fixedVec[i].y * movingVec[i].x;
			temp.m11 = fixedVec[i].y * movingVec[i].y;
			temp.m12 = fixedVec[i].y * movingVec[i].z;

			temp.m20 = fixedVec[i].z * movingVec[i].x;
			temp.m21 = fixedVec[i].z * movingVec[i].y;
			temp.m22 = fixedVec[i].z * movingVec[i].z;

			cov.add(temp);
		}

		SingularValueDecomposition svd = new SingularValueDecomposition(cov);
		Matrix3d U = svd.getUMatrix();
		Matrix3d S = new Matrix3d();
		S.setIdentity();
		Matrix3d V = svd.getVMatrix();

		if (U.determinant() * V.determinant() == -1) {
			S.m22 = -1;
		}
		V.transpose();

		U.mul(S);
		U.mul(V);

		return U;
	}
}
