/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.util.command;

import java.util.ArrayList;
import java.util.List;

/**
 * List containing commands for undo functionality
 * 
 * @author Klaus Dorer
 */
public class UndoList
{
	/** the list holding the undo commands */
	private final List<ICommand> commands;

	/** index of the current active command */
	private int undoIndex;

	/** the maximal number of entries that should be stored */
	private final int maxCommands;

	/**
	 * Instantiates an UndoList object and initializes the maximum number of
	 * entries
	 * 
	 * @param max Maximum number of entries
	 */
	public UndoList(int max)
	{
		commands = new ArrayList<ICommand>();
		maxCommands = max;
		undoIndex = -1;
	}

	/**
	 * Adds the passed command to the list of commands. If the maximal amount is
	 * exceeded the oldest command is removed.
	 * @param newCommand the command to add
	 * @return True if successfull, False if not
	 */
	public boolean addCommand(ICommand newCommand)
	{
		// make sure that all commands after the undo index are removed
		for (int i = commands.size() - 1; i > undoIndex; i--) {
			commands.remove(i);
		}

		// check if the new command can be merged with the last
		if (commands.size() > 0
				&& commands.get(commands.size() - 1).isMergeable(newCommand)) {
			commands.get(commands.size() - 1).merge(newCommand);
			return true;
		}

		boolean result = commands.add(newCommand);
		undoIndex++;

		// make sure that list does not grow too big
		if (commands.size() > maxCommands) {
			commands.remove(0);
			undoIndex--;
		}

		return result;
	}

	/**
	 * Returns the command at the undo position and moves the pointer for the
	 * next undo command to the previous command
	 * @return the command at the undo position, null if no more undo available
	 */
	public ICommand getUndoCommand()
	{
		if (undoIndex >= 0) {
			return commands.get(undoIndex--);
		}

		return null;
	}

	/**
	 * Moves the pointer for the next undo command to the next in the list and
	 * returns it
	 * @return the command at the redo position, null if no more undo available
	 */
	public ICommand getRedoCommand()
	{
		if (undoIndex < commands.size() - 1) {
			return commands.get(++undoIndex);
		}

		return null;
	}

	/**
	 * @return the index of the next command that will be returned if undo is
	 *         called
	 */
	int getUndoIndex()
	{
		return undoIndex;
	}

	/**
	 * @return the number of commands stored in the list
	 */
	public int size()
	{
		return commands.size();
	}
}
