/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.function;

import magma.util.geometry.Geometry;

/**
 * Represents a piecewise linear function with any number of pieces
 * 
 * @author Klaus Dorer
 */
public class PiecewiseLinearFunction extends SupportPointFunction
{
	/** serial ID for serialization */
	private static final long serialVersionUID = 1L;

	/** name identifying a this kind of function */
	public static final String LINEAR_NAME = "linear";

	/**
	 * Instantiates a new PiecewiseLinearFunction object
	 * 
	 * @param xValues Array containing x values
	 * @param yValues Array containing y values
	 */
	public PiecewiseLinearFunction(double[] xValues, double[] yValues)
	{
		super(xValues, yValues, 0.0);
	}

	/**
	 * Creates the function from a sin line in a motor function file
	 * 
	 * @param csvLineContents the values that were comma separated in the file
	 */
	public PiecewiseLinearFunction(String[] csvLineContents)
	{
		super(csvLineContents);
	}

	/**
	 * Copy Constructor
	 * 
	 * @param source Source object to copy
	 */
	private PiecewiseLinearFunction(PiecewiseLinearFunction source)
	{
		super(source);
	}

	/**
	 * Instantiates the object from a maximum x value and a phase
	 * 
	 * @param maxX maximum x value
	 * @param phase sinus phase
	 */
	public PiecewiseLinearFunction(int maxX, double phase)
	{
		super(maxX, phase);
	}

	@Override
	protected String getName()
	{
		return LINEAR_NAME;
	}

	/**
	 * Returns the sin value at the passed x coordinate. Parameters are used as
	 * written in constructor.
	 * @param t the coordinate at which to return the functions value
	 * @return the sin value at the passed x coordinate
	 */
	@Override
	public double value(double t)
	{
		int x0Index = 0;
		int x1Index = 1;

		if (t < xValues[x0Index]) {
			throw new IllegalArgumentException(
					"Too small x value for linear function! x: " + t + " minX: "
							+ xValues[x0Index]);
		}

		// find suitable values
		while (t > xValues[x1Index]) {
			x0Index = x1Index;
			x1Index++;
			if (x1Index >= xValues.length) {
				throw new IllegalArgumentException(
						"Too big x value for linear function! x: " + t + " maxX: "
								+ xValues[x0Index]);
			}
		}

		// linear interpolation
		return Geometry.linearInterpolation(xValues[x0Index], yValues[x0Index],
				xValues[x1Index], yValues[x1Index], t);
	}

	/**
	 * @see magma.util.function.Function#copy()
	 */
	@Override
	public Function copy()
	{
		return new PiecewiseLinearFunction(this);
	}
}