/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.function;

/**
 * Represents a sin function with amplitude, period, phase and offset
 * 
 * @author Klaus Dorer
 */
public class SinFunction implements Function
{
	/** serial ID for serialization */
	private static final long serialVersionUID = 1L;

	/** name identifying a sine function */
	public static final String SIN_NAME = "sinus";

	/**
	 * parameters needed for the sin 0: period, 1: amplitude, 2: phase, 3: offset
	 */
	protected final double[] params;

	/**
	 * Creates the function from a sin line in a motor function file
	 * 
	 * @param csvLineContents the values that were comma separated in the file
	 * @param period Sinus period
	 */
	public SinFunction(String[] csvLineContents, double period)
	{
		params = new double[4];
		params[0] = period;
		params[1] = Double.valueOf(csvLineContents[2]);
		params[2] = Double.valueOf(csvLineContents[3]);
		params[3] = Double.valueOf(csvLineContents[4]);
	}

	/**
	 * Constructor
	 * 
	 * @param period Period
	 * @param amplitude Amplitude
	 * @param phase Phase shift
	 * @param offset Offset in Y-direction
	 */
	public SinFunction(double period, double amplitude, double phase,
			double offset)
	{
		params = new double[4];

		params[0] = period;
		params[1] = amplitude;
		params[2] = phase;
		params[3] = offset;
	}

	/**
	 * Copy constructor
	 * 
	 * @param source The source object to copy
	 */
	public SinFunction(SinFunction source)
	{
		params = new double[source.params.length];
		System.arraycopy(source.params, 0, params, 0, source.params.length);
	}

	/**
	 * Get Function name
	 * 
	 * @return String
	 */
	protected String getName()
	{
		return SIN_NAME;
	}

	@Override
	public double value(double t)
	{
		return params[1] * Math.sin(t / params[0] * 2 * Math.PI + params[2])
				+ params[3];
	}

	/**
	 * Convert object into a CSV string representation
	 * 
	 * @return String
	 */
	@Override
	public String toCSVString()
	{
		StringBuilder result = new StringBuilder();
		result.append(getName()).append(",");
		for (int i = 1; i < params.length; i++) {
			result.append(params[i]).append(",");
		}
		return result.toString();
	}

	@Override
	public double getMaxX()
	{
		return params[0];
	}

	@Override
	public double getMinX()
	{
		return 0.0;
	}

	@Override
	public Function copy()
	{
		return new SinFunction(this);
	}

	@Override
	public double getPhase()
	{
		return 0;
	}
}