/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.function;

import org.apache.commons.math.ArgumentOutsideDomainException;
import org.apache.commons.math.analysis.PolynomialSplineFunction;
import org.apache.commons.math.analysis.SplineInterpolator;

/**
 * Represents a spline function. Adapter to the PolynomialSplineFunction class.
 * 
 * @author Klaus Dorer
 */
public class SplineFunction extends SupportPointFunction
{
	/** serial ID for serialization */
	private static final long serialVersionUID = 1L;

	/** name identifying a spline */
	public static final String SPLINE_NAME = "spline";

	private PolynomialSplineFunction spline;

	private double xMin;

	private double xMax;

	/**
	 * Creates a Spline Function. Adapter to the PolynomialSplineFunction class.
	 * The supporting points passed have to make sure that x0 is the minimum x
	 * value needed later and xn is the highest x value.
	 * @param xValues the x coordinates of the supporting points
	 * @param yValues the y coordinates of the supporting points
	 */
	public SplineFunction(double xValues[], double yValues[])
	{
		this(xValues, yValues, 0.0);
	}

	/**
	 * Creates the function from a spline line in a motor function file
	 * @param csvLineContents the values that were comma separated in the file
	 */
	public SplineFunction(String[] csvLineContents)
	{
		super(csvLineContents);
		initialize();
	}

	@Override
	protected String getName()
	{
		return "spline";
	}

	/**
	 * Creates a Spline Function. Adapter to the PolynomialSplineFunction class.
	 * The supporting points passed have to make sure that x0 is the minimum x
	 * value needed later and xn is the highest x value.
	 * @param xValues the x coordinates of the supporting points
	 * @param yValues the y coordinates of the supporting points
	 * @param phase phase subtracted from x. Has to be within xMin and xMax
	 */
	private SplineFunction(double xValues[], double yValues[], double phase)
	{
		super(xValues, yValues, phase);
		initialize();
	}

	/**
	 * Copy constructor
	 * 
	 * @param source The source object to copy
	 */
	private SplineFunction(SplineFunction source)
	{
		super(source);

		xMin = source.xMin;
		xMax = source.xMax;
		initialize();
	}

	/**
	 * Creates a spline function from the previously set x and y values
	 */
	@Override
	public void initialize() throws IllegalArgumentException
	{
		SplineInterpolator interpolator = new SplineInterpolator();
		spline = (PolynomialSplineFunction) interpolator.interpolate(xValues,
				yValues);
		xMin = xValues[0];
		xMax = xValues[xValues.length - 1];
	}

	/**
	 * Returns the spline value at t - phase mapped to [xMin, xMax].
	 * @param t the coordinate at which to return the functions value
	 * @return the spline value at the calculated x coordinate
	 */
	@Override
	public double value(double t)
	{
		try {
			double x = t - phase;
			if (x < xMin) {
				x += xMax;
			} else if (x > xMax) {
				x -= xMax;
			}
			return spline.value(x);

		} catch (ArgumentOutsideDomainException e) {
			assert false : "Invalid spline value: " + t;
			e.printStackTrace();
			return 0.0;
		}
	}

	@Override
	public Function copy()
	{
		return new SplineFunction(this);
	}
}