/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.function;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;

/**
 * Represents a piecewise linear function with any number of pieces
 * @author Klaus Dorer
 */
public abstract class SupportPointFunction implements Function
{
	/**  */
	private static final long serialVersionUID = 1L;

	/**
	 * parameters needed for the start and end points
	 */
	protected double xValues[];

	protected double yValues[];

	/** phase shift of the function in x direction */
	protected final double phase;

	/**
	 * @param xValues x coordinates of support points
	 * @param yValues y coordinates of support points
	 * @param phase x phase shift
	 */
	public SupportPointFunction(double[] xValues, double[] yValues, double phase)
	{
		this.xValues = xValues;
		this.yValues = yValues;
		this.phase = phase;
	}

	/**
	 * Creates the function from a sin line in a motor function file
	 * @param csvLineContents the values that were comma separated in the file
	 */
	public SupportPointFunction(String[] csvLineContents)
	{
		int size = Integer.valueOf(csvLineContents[2]);
		xValues = new double[size];
		yValues = new double[size];

		int xpos = 3;
		// TODO: add phase to all piecewise linear function txt files
		phase = java.lang.Double.valueOf(csvLineContents[xpos++]);
		for (int i = 0; i < size; i++) {
			xValues[i] = Double.valueOf(csvLineContents[xpos++]);
			yValues[i] = Double.valueOf(csvLineContents[xpos++]);
		}
	}

	/**
	 * Copy constructor
	 * @param source the source to copy
	 */
	public SupportPointFunction(SupportPointFunction source)
	{
		xValues = source.xValues.clone();
		yValues = source.yValues.clone();
		phase = source.phase;
	}

	/**
	 * @param size number of support points to represent
	 * @param phase phase shift in x direction
	 */
	public SupportPointFunction(int size, double phase)
	{
		xValues = new double[size];
		yValues = new double[size];
		this.phase = phase;
	}

	/**
	 * @return x coordinates of all support points
	 */
	public double[] getXValues()
	{
		return xValues;
	}

	/**
	 * @return y coordinates of all support points
	 */
	public double[] getYValues()
	{
		return yValues;
	}

	@Override
	public String toCSVString()
	{
		DecimalFormatSymbols symbols = new DecimalFormatSymbols();
		symbols.setDecimalSeparator('.');
		DecimalFormat formatter = new DecimalFormat("##0.0#", symbols);
		StringBuilder result = new StringBuilder();
		result.append(getName()).append(",").append(xValues.length).append(",")
				.append(phase).append(",");
		for (int i = 0; i < xValues.length; i++) {
			result.append(formatter.format(xValues[i])).append(",");
			result.append(formatter.format(yValues[i])).append(",");
		}
		return result.toString();
	}

	/**
	 * @return name of this function
	 */
	protected abstract String getName();

	@Override
	public double getMaxX()
	{
		if (xValues.length == 0) {
			return 0.0;
		}
		return xValues[xValues.length - 1];
	}

	/**
	 * Checks if the function has a support point in the specified area
	 * @param right border (including) (joint coordinates)
	 * @param left border (including) (joint coordinates)
	 * @return true if at least one support point is inside the specified x area
	 */
	public boolean hasSupportPointInArea(double left, double right)
	{
		for (double xValue : xValues) {
			if (xValue >= left && xValue <= right) {
				return true;
			}
		}
		return false;
	}

	@Override
	public double getMinX()
	{
		if (xValues.length == 0) {
			return 0.0;
		}
		return xValues[0];
	}

	/**
	 * Get x-axis support value
	 * 
	 * @param index Entry index in the internal list
	 * @return X value
	 */
	public double getSupportX(int index)
	{
		return xValues[index];
	}

	/**
	 * Get y-axis support value
	 * 
	 * @param index Entry index in the internal list
	 * @return Y value
	 */
	public double getSupportY(int index)
	{
		return yValues[index];
	}

	/**
	 * @return the number of support points
	 */
	public int getSupportPointSize()
	{
		return xValues.length;
	}

	/**
	 * Changes the coordinates of a support point. Does not allow to be moved
	 * left or right of neighboring support points.
	 * @param index the index of the point to change.
	 * 
	 * @param x the new x coordinate
	 * @param y the new y coordinate
	 * @return the old coordinates of the point
	 */
	public double[] changePointCoordinates(int index, double x, double y)
	{
		double[] result = { xValues[index], yValues[index] };
		xValues[index] = x;
		yValues[index] = y;

		// do not allow x to be less than previous x
		if (index > 0 && x <= xValues[index - 1]) {
			xValues[index] = xValues[index - 1] + 0.1;
		}

		// do not allow x to be more than next x
		if (index < xValues.length - 1 && x >= xValues[index + 1]) {
			xValues[index] = xValues[index + 1] - 0.1;
		}
		initialize();
		return result;
	}

	/**
	 * Removes a support point of this function
	 * @param x the x coordinate of the new point
	 * @param y the y coordinate of the new point
	 * @return the index at which the point was added
	 */
	public int addPoint(double x, double y)
	{
		double[] newXValues = new double[xValues.length + 1];
		double[] newYValues = new double[yValues.length + 1];

		// copy values
		int addedIndex = -1;
		int newIndex = 0;
		for (int oldIndex = 0; oldIndex < xValues.length; oldIndex++, newIndex++) {
			if (addedIndex < 0 && x < xValues[oldIndex]) {
				// new point position reached
				newXValues[newIndex] = x;
				newYValues[newIndex] = y;
				addedIndex = newIndex;
				newIndex++;
			}
			newXValues[newIndex] = xValues[oldIndex];
			newYValues[newIndex] = yValues[oldIndex];
		}

		if (addedIndex < 0) {
			// new point added at end
			newXValues[newIndex] = x;
			newYValues[newIndex] = y;
			addedIndex = newIndex;
		}

		xValues = newXValues;
		yValues = newYValues;
		initialize();
		return addedIndex;
	}

	/**
	 * Removes a support point of this function
	 * @param index the index of the point to remove
	 * @return the coordinates of the removed point
	 */
	public double[] removePoint(int index)
	{
		double[] newXValues = new double[xValues.length - 1];
		double[] newYValues = new double[yValues.length - 1];

		double[] result = { xValues[index], yValues[index] };

		if (index > 0) {
			// copy left part
			System.arraycopy(xValues, 0, newXValues, 0, index);
			System.arraycopy(yValues, 0, newYValues, 0, index);
		}

		if (index < xValues.length - 1) {
			// copy right part
			System.arraycopy(xValues, index + 1, newXValues, index, xValues.length
					- index - 1);
			System.arraycopy(yValues, index + 1, newYValues, index, yValues.length
					- index - 1);
		}

		xValues = newXValues;
		yValues = newYValues;
		initialize();

		return result;
	}

	/**
	 * Called after changes to support points. Default does nothing.
	 */
	public void initialize()
	{
	}

	/**
	 * @return the phase
	 */
	@Override
	public double getPhase()
	{
		return phase;
	}
}