/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry;

import java.io.Serializable;

/**
 * Represents an angle. Objects are immutable.
 * 
 * @author Klaus Dorer
 */
public class Angle implements Serializable
{
	/** */
	private static final long serialVersionUID = 1L;

	/** Constant for -180 degrees in rad */
	public final static Angle RAD180;

	/** Constant for 360 degrees in rad */
	private final static double RAD360D;

	/** Constant for 180 degrees in rad */
	private final static double RAD180D;

	/** Constant for 90 degrees in rad */
	private final static double RAD90D;

	/** Constant for 0 degrees in rad */
	public final static Angle RAD0;

	/** the angle value (rad) this object represents */
	private double angle;

	static {
		/** Initialize constants */
		RAD360D = Math.toRadians(360);

		RAD180D = Math.toRadians(180);

		RAD90D = Math.toRadians(90);

		RAD180 = Angle.deg(-180);

		RAD0 = Angle.rad(0);
	}

	/**
	 * Creates a normalized angle [-RAD180 to RAD180[
	 * 
	 * @param value Angle in rad
	 * @return Angle object
	 */
	public static Angle rad(double value)
	{
		return new Angle(value);
	}

	/**
	 * Creates a normalized angle [-RAD180 to RAD180[
	 * 
	 * @param value Angle in degrees
	 * @return Angle object
	 */
	public static Angle deg(double value)
	{
		return new Angle(Math.toRadians(value));
	}

	private Angle(double angle)
	{
		this.angle = angle;
		normalize();
	}

	/**
	 * Retrieve the angle value in rad
	 * 
	 * @return Angle
	 */
	public double radians()
	{
		return angle;
	}

	/**
	 * Retrieve the angle value in degrees
	 * 
	 * @return Angle value
	 */
	public double degrees()
	{
		return Math.toDegrees(angle);
	}

	/**
	 * Adds another angleobject to this angle
	 * 
	 * @param other angle object
	 * @return a new Angle object representing the sum of this and the other
	 *         angle
	 */
	public Angle add(Angle other)
	{
		return new Angle(angle + other.angle);
	}

	/**
	 * Adds a constant to this angle
	 * 
	 * @param value Angle (in rad)
	 * @return a new Angle object representing the sum of this angle and the
	 *         constant
	 */
	public Angle add(double value)
	{
		return new Angle(angle + value);
	}

	/**
	 * Subtracts another angle object from this angle
	 * 
	 * @param other Angle object to subtract
	 * @return a new Angle object representing the difference between this and
	 *         the other angle
	 */
	public Angle subtract(Angle other)
	{
		return new Angle(angle - other.angle);
	}

	/**
	 * Subtracts a constant angle from this angle
	 * 
	 * @param value Angle in rad
	 * @return a new Angle object representing the difference between this angle
	 *         and the constant
	 */
	public Angle subtract(double value)
	{
		return new Angle(angle - value);
	}

	/**
	 * Normalize this angle
	 */
	private void normalize()
	{
		if (angle >= 0) {
			angle += RAD180D;
			angle = angle % RAD360D;
			angle -= RAD180D;
		} else {
			angle = -angle + RAD180D;
			angle = angle % RAD360D;
			angle = -(angle - RAD180D);
			if (angle >= RAD180D) {
				angle = -angle;
			}
		}
	}

	@Override
	public boolean equals(Object obj)
	{
		if (!(obj instanceof Angle)) {
			return false;
		}
		return ((Angle) obj).angle == this.angle;
	}

	@Override
	public String toString()
	{
		return "" + degrees();
	}

	/**
	 * An angle is "left" of another if it is bigger, but by less than 180
	 * degrees.
	 * 
	 * @param angle2 second angle object
	 * @return true if first angle is left of second
	 */
	public boolean isLeftOf(Angle angle2)
	{
		if (angle > RAD90D && angle2.angle < -RAD90D) {
			// value on -180 degree border
			return false;
		} else if (angle < -RAD90D && angle2.angle > RAD90D) {
			// value on +180 degree border
			return true;
		}

		double delta = angle2.subtract(angle).radians();
		return delta < 0 && delta > -RAD180D;

	}

	/**
	 * An angle is "right" of another if it is not "left" of it
	 * 
	 * @param angle2 second angle object
	 * @return true if first angle is right of second
	 */
	public boolean isRightOf(Angle angle2)
	{
		return !isLeftOf(angle2);
	}

	/**
	 * Calculate the average of two Angles
	 * 
	 * @param angle1 first angle to use
	 * @param angle2 second angle to use
	 * @return the average of the two passed angles
	 */
	public static Angle average(Angle angle1, Angle angle2)
	{
		Angle[] angles = { angle1, angle2 };
		return average(angles);
	}

	/**
	 * Calculate the average of a number of Angles
	 * 
	 * @param angles Angles array
	 * @return Average
	 */
	public static Angle average(Angle[] angles)
	{
		double average = angles[0].radians();

		for (int i = 1; i < angles.length; i++) {
			double angleToUse = angles[i].radians();

			if (average > RAD90D && angleToUse < -RAD90D) {
				// value on -180 degree border
				angleToUse += RAD360D;
			} else if (average < -RAD90D && angleToUse > RAD90D) {
				// value on +180 degree border
				angleToUse -= RAD360D;
			}
			average = (i * average + angleToUse) / (i + 1);
		}

		return Angle.rad(average);
	}
}
