/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry;

import java.io.Serializable;

import javax.vecmath.Vector3d;

/**
 * Holds parameters needed to perform Denavit/Hartenberg Transformations
 * @author Ingo Schindler
 * 
 */
public class DenavitHartenbergParameters implements Serializable
{
	private static final long serialVersionUID = 1090727267472880029L;

	private final double alpha;

	private final double beta;

	private double theta;

	private final Vector3d distance;

	/**
	 * Constructor
	 * 
	 * @param theta Theta value
	 * @param alpha Alpha value
	 * @param beta Beta value
	 * @param distance Distance
	 */
	public DenavitHartenbergParameters(double theta, double alpha, double beta,
			Vector3d distance)
	{
		this.theta = theta;
		this.alpha = alpha;
		this.beta = beta;

		this.distance = distance;
	}

	/**
	 * Get Alpha value in radians
	 * 
	 * @return Value
	 */
	public double getAlpha()
	{
		return Math.toRadians(alpha);
	}

	/**
	 * Get Beta value in radians
	 * 
	 * @return Value
	 */
	public double getBeta()
	{
		return Math.toRadians(beta);
	}

	/**
	 * Get Theta value in radians
	 * 
	 * @return Value
	 */
	public double getTheta()
	{
		return Math.toRadians(theta);
	}

	/**
	 * Set Theta value
	 * 
	 * @param value New value
	 */
	public void setTheta(double value)
	{
		theta = value;
	}

	/**
	 * Get distance in 3-dimensional space
	 * 
	 * @return Distance
	 */
	public Vector3d getDistance()
	{
		return distance;
	}
}
