/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry;

import javax.vecmath.Point2d;
import javax.vecmath.Point3d;

import magma.util.FuzzyCompare;

import org.apache.commons.math.geometry.Vector3D;

/**
 * Utilities for geometric calculations
 * 
 * @author Klaus Dorer
 */
public class Geometry
{
	/**
	 * Checks if the passed 2D point is inside the passed polygon
	 * @param position the point to check (x,y) (it has to be in the same
	 *        coordinate system as the points of the polygon
	 * @param polygon corner points 2D of the polygon in clockwise direction
	 * @return true if the passed point is inside the polygon
	 */
	public static boolean isInsidePolygon(Vector3D position, Vector3D[] polygon)
	{
		int index0 = polygon.length - 1;
		int index1 = 0;
		double x2 = position.getX();
		double y2 = position.getY();
		for (index1 = 0; index1 < polygon.length;) {
			double x0 = polygon[index0].getX();
			double y0 = polygon[index0].getY();
			double x1 = polygon[index1].getX();
			double y1 = polygon[index1].getY();
			// if points are clockwise then the determinant has to be positive
			// ----1 | x0 y0 1 |
			// A = - | x1 y1 1 |
			// ----2 | x2 y2 1 |
			double result = 0.5 * (x1 * y2 - y1 * x2 - x0 * y2 + y0 * x2 + x0 * y1 - y0
					* x1);
			if (result > 0) {
				return false;
			}

			index0 = index1;
			index1++;
		}

		return true;
	}

	/**
	 * Converts the passed points to a Vector3D array representing a polygon
	 * @param points the corner points (x,y)
	 * @return a Vector3D array representing a polygon
	 */
	public static Vector3D[] getPolygon(double[][] points)
	{
		Vector3D[] polygon = new Vector3D[points.length];
		for (int i = 0; i < points.length; i++) {
			double[] point = points[i];
			polygon[i] = new Vector3D(point[0], point[1], 0);
		}
		return polygon;
	}

	/**
	 * calculates with the given coordinates the desired angles for Hip and Knee
	 * the Hip-Coordinates are always (0|0). When Leg is straight Foot is
	 * (0|-0.25) and Knee is (0|-0.14)
	 * 
	 * @param xKnee - x Position of the Knee
	 * @param zKnee - z Position of the Knee
	 * @param xFoot - x Position of the Foot (Ankle)
	 * @param zFoot - y Position of the Foot (Ankle)
	 * @return two-Dimensional Array first Value is Hip-Angle Second is
	 *         Knee-Angle
	 */
	public static double[] calculateAngles(Point3d kneePosition,
			Point3d anklePosition)
	{
		double xKnee = kneePosition.x;
		double zKnee = kneePosition.z;
		double xFoot = anklePosition.x;
		double zFoot = anklePosition.z;

		double[] ret = { 0, 0 };

		double l1 = 0.14, l2 = 0.11;

		zFoot = -((l1 + l2) - zFoot);

		double hip = Math.toDegrees(Math.atan(zKnee / xKnee));

		if (xKnee < 0) {
			hip = hip - 180;
		}
		hip = 90 + hip;

		double l = -Math.sqrt((xFoot * xFoot) + (zFoot * zFoot));

		double calc = ((l2 * l2 + l1 * l1 - l * l) / (2 * l2 * l1));

		double knee = 0;

		if (!FuzzyCompare.eq(zKnee / xKnee, zFoot / xFoot, 0.01)) {
			if (calc > 1) {
				knee = Math.toDegrees(Math.acos(calc - 2));
			} else if (calc < -1) {
				knee = Math.toDegrees(Math.acos(calc + 2));
			} else {
				knee = Math.toDegrees(Math.acos(calc));
			}
			knee = -(180 - knee);
		}

		ret[0] = hip;
		ret[1] = knee;

		return ret;
	}

	/**
	 * Calculates intersection Point(s) of two circles. Center of the first
	 * circle is (x1|z1) and its radius is r1. The center of the second circle is
	 * given by the parameters x2 and z2. The radius of the second circle is r2.
	 * If there is no intersection between the two circles a straight line from
	 * center to center is calculated and the intersection of this line and the
	 * first circle is returned.
	 * 
	 * @param p1 X and Z coordinate of the first circle
	 * @param r1 Radius of the first circle
	 * @param p2 X and Z coordinate of the second circle
	 * @param r2 Radius of the second circle
	 * @return Array with two elements for the coordinate of the two
	 *         intersection-points
	 */
	public static Point2d[] intersectionPoint(Point2d p1, double r1, Point2d p2,
			double r2)
	{
		Point2d[] ret = new Point2d[2];
		// double[] ret = { 0, 0, 0, 0 };

		double xc1, xc2, yc1, yc2, a1, a2, b, c, d, e, f, g;
		double ret_y0, ret_y1, ret_x0, ret_x1;

		xc1 = p1.getX();
		yc1 = p1.getY();

		xc2 = p2.getX();
		yc2 = p2.getY();

		if (xc2 != 0) {
			a1 = r1 * r1 - xc1 * xc1 - yc1 * yc1;
			a2 = r2 * r2 - xc2 * xc2 - yc2 * yc2;
			b = (a1 - a2) / 2;
			c = b / (xc2 - xc1);
			d = (yc2 - yc1) / (xc2 - xc1);
			e = d * d + 1;
			f = 2 * (d * xc1 - yc1 - c * d);
			g = c * c - 2 * c * xc1 - a1;
			if (f * f - 4 * e * g < 0) {
				double dx = xc1 - xc2;
				double dy = yc1 - yc2;

				double dr = Math.sqrt(dx * dx + dy * dy);

				double D = xc1 * yc2 - xc2 * yc1;

				double det = r1 * r1 * dr * dr - D * D;

				ret_x1 = (D * dy - (dy < 0 ? -1 : 1) * dx * Math.sqrt(det))
						/ (dr * dr);

				ret_y1 = (-D * dx - Math.abs(dy) * Math.sqrt(det)) / (dr * dr);

				ret[0] = new Point2d(ret_x1, ret_y1);
				ret[1] = new Point2d(ret_x1, ret_y1);

			} else {

				ret_y0 = ((-f - Math.sqrt(f * f - 4 * e * g)) / (2 * e));
				ret_x0 = (c - d * ret_y0);

				ret_y1 = ((-f + Math.sqrt(f * f - 4 * e * g)) / (2 * e));
				ret_x1 = (c - d * ret_y1);

				ret[0] = new Point2d(ret_x1 > 0 ? ret_x1 : ret_x0,
						ret_x1 > 0 ? ret_y1 : ret_y0);
				ret[1] = new Point2d(ret_x1 > 0 ? ret_x0 : ret_x1,
						ret_x1 > 0 ? ret_y0 : ret_y1);
			}

		} else {
			c = -yc2;
			double a = 0.11;
			b = 0.14;

			double alpha = Math.toDegrees(Math.acos((b * b + c * c - a * a)
					/ (2 * b * c)));

			double h = b * Math.sin(Math.toRadians(alpha));

			ret[0] = new Point2d(h, -Math.sqrt(h * h + b * b - 2 * h * b
					* Math.cos(Math.toRadians(180 - 90 - alpha))));
			ret[1] = new Point2d(-h, -Math.sqrt(h * h + b * b - 2 * h * b
					* Math.cos(Math.toRadians(180 - 90 - alpha))));
		}

		return ret;
	}

	/**
	 * Returns a linear interpolation between the two passed points at the passed
	 * x coordinate. Depending on the ascending parameter the function grows
	 * (true) from 0 to 1 otherwise it decreases from 1 to 0. Outside the range
	 * specified by x0 and x1 the values are 0 (left, true or right, false) or 1
	 * respectively (left, false or right, true.
	 * @param x0 lower x value
	 * @param x1 higher x value
	 * @param ascending true if the lower x value represents 0 false if it
	 *        represents 1
	 * @param x the x value for which to get the fuzzy value
	 * @return a fuzzy value that is linearly interpolated between the two passed
	 *         values
	 */
	public static double getLinearFuzzyValue(double x0, double x1,
			boolean ascending, double x)
	{
		if (x0 > x1) {
			double temp = x0;
			x0 = x1;
			x1 = temp;
		}

		if (ascending) {
			if (x <= x0) {
				return 0.0;
			}
			if (x >= x1) {
				return 1.0;
			}
			return linearInterpolation(x0, 0, x1, 1, x);

		}

		if (x <= x0) {
			return 1.0;
		}

		if (x >= x1) {
			return 0.0;
		}

		return linearInterpolation(x0, 1, x1, 0, x);
	}

	/**
	 * Limits the passed value's absolute value by the limit passed. So if limit
	 * is 20 the returned value will be in the interval [-20, x, 20].
	 * @param value the value to limit
	 * @param limit positive upper limit of the interval
	 * @return the passed value's absolute value limited by the passed limit
	 */
	public static double limitAbs(double value, double limit)
	{
		if (value > limit) {
			return limit;
		}
		if (value < -limit) {
			return -limit;
		}
		return value;
	}

	/**
	 * Limits the passed value's absolute value by the limit passed. So if limit
	 * is 20 the returned value will be in the interval [-20, x, 20].
	 * @param value the value to limit
	 * @param limit positive upper limit of the interval
	 * @return the passed value's absolute value limited by the passed limit
	 */
	public static float limitAbs(float value, float limit)
	{
		if (value > limit) {
			return limit;
		}
		if (value < -limit) {
			return -limit;
		}
		return value;
	}

	/**
	 * Returns the linear interpolation (y value) of the passed points (x0,y0)
	 * and (x1,y1) at the passed coordinate x
	 * @param x0 x coordinate of the first point
	 * @param y0 y coordinate of the first point
	 * @param x1 x coordinate of the second point
	 * @param y1 y coordinate of the second point
	 * @param x x coordinate of the point of interest
	 * @return the y coordinate of position x as linear interpolation of the two
	 *         passed points
	 */
	public static double linearInterpolation(double x0, double y0, double x1,
			double y1, double x)
	{
		return y0 + (y1 - y0) / (x1 - x0) * (x - x0);
	}

	/**
	 * Method to rotate a vector in 2D by an angle about the origin.
	 * <p>
	 * <b><i>Note:</i> Rotation displaces the polygon</b>
	 * @param vectorToRotate - the vector that has to be rotated
	 * @param angleInDegrees - the angle of rotation in degrees
	 * @param clockwise - true if the vector has to be rotated clockwise, false
	 *        if anti-clockwise
	 * 
	 * @return Rotated vector
	 */
	public static Vector3D rotateVectorIn2D(Vector3D vectorToRotate,
			double angleInDegrees, boolean clockwise)
	{
		double originalX = vectorToRotate.getX();
		double originalY = vectorToRotate.getY();
		double newX, newY;

		if (clockwise) {
			newX = originalX * Math.cos(angleInDegrees) - originalY
					* Math.sin(angleInDegrees);
			newY = originalX * Math.sin(angleInDegrees) + originalY
					* Math.cos(angleInDegrees);
		} else {
			newX = originalX * Math.cos(angleInDegrees) + originalY
					* Math.sin(angleInDegrees);
			newY = -originalX * Math.sin(angleInDegrees) + originalY
					* Math.cos(angleInDegrees);
		}
		return new Vector3D(newX, newY);
	}

	/**
	 * Method to calculate the vector at a <code>distance</code> from
	 * <code>vector</code> at an <code>angle</code> to the norm of the passed
	 * vector
	 * @param vector - the vector from which to find the displaced and rotated
	 *        vector
	 * @param distance - the distance from the <code>vector</code>
	 * @param angle - Angle object
	 * @return Vector3D (the vector from Apache)
	 */
	public static Vector3D getVectorAtDistanceAndAngle(Vector3D vector,
			double distance, Angle angle)
	{
		double newX, newY;
		newX = vector.getX() + (distance * Math.cos(angle.radians()));
		newY = vector.getY() + (distance * Math.sin(angle.radians()));

		return new Vector3D(newX, newY, 0.0);
	}

	/**
	 * Calculates the area of the triangle passed. Uses co-ordinate geometry.
	 * @param t - triangle
	 * @return double
	 */
	public static double areaOfTriangle(Vector3D[] t)
	{
		Vector3D p1 = t[0];
		Vector3D p2 = t[1];
		Vector3D p3 = t[2];

		// area of a triangle by co-ordinate geometry
		double area = (p1.getX() * (p2.getY() - p3.getY()))
				+ (p2.getX() * (p3.getY() - p1.getY()))
				+ (p3.getX() * (p1.getY() - p2.getY()));

		return Math.abs(area);
	}

	/**
	 * @param source source point
	 * @param destination destination point
	 * @param distanceFromSource distance from source point
	 * @return a point on the line between source and destination with distance
	 *         from source as specified
	 */
	public static PositionOrientation getPointInterpolation(Vector3D source,
			Vector3D destination, double distanceFromSource)
	{
		Vector3D srcDest = destination.subtract(source).normalize()
				.scalarMultiply(distanceFromSource);
		Angle directionSrcDest = Angle.rad(srcDest.getAlpha());
		srcDest = srcDest.add(source);
		return new PositionOrientation(srcDest, directionSrcDest);
	}
}
