/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry;

import javax.vecmath.Matrix3d;
import javax.vecmath.Matrix4d;
import javax.vecmath.Vector3d;

/**
 * Implements a set of useful matrix functions.<br>
 * <h5>Some general Notation:</h5>
 * <ul>
 * <li><b>R</b> -- Rotation matrix around one ore more arbitrary axis</li>
 * <li><b>Rx</b> -- Rotation matrix around the x-axis</li>
 * <li><b>Ry</b> -- Rotation matrix around the y-axis</li>
 * <li><b>Rz</b> -- Rotation matrix around the z-axis</li>
 * <li><b>phi</b> -- Some angle</li>
 * <li><b>phi_x</b> -- Some angle around the x-axis</li>
 * <li><b>phi_y</b> -- Some angle around the y-axis</li>
 * <li><b>phi_x</b> -- Some angle around the z-axis</li>
 * <li><b>vec, v</b> -- Some vectors</li>
 * <li><b>m</b> -- Some matrix</li>
 * <li>Moreover 0 is also often used to describe zero-vectors (as well for 0^T
 * (the transposed version of a zero-vector))</li>
 * </ul>
 * 
 * 
 * @author Stefan Glaser
 * 
 */
public class MatrixUtil
{
	public enum RotationOrder {
		XYZ, XZY, YXZ, YZX, ZXY, ZYX
	}

	/**
	 * Multiply a 3-dimensional vector with a 4-dimensional matrix.<br>
	 * In the end this results in the following operation:<br>
	 * <b>vec = m * v|1</b>
	 * @param m Matrix
	 * @param v Vector
	 * 
	 * @return Resulting vector
	 */
	public static Vector3d mulMatrixVector(Matrix4d m, Vector3d v)
	{
		Vector3d ret = new Vector3d();

		ret.x = v.x * m.m00 + v.y * m.m01 + v.z * m.m02 + 1 * m.m03;
		ret.y = v.x * m.m10 + v.y * m.m11 + v.z * m.m12 + 1 * m.m13;
		ret.z = v.x * m.m20 + v.y * m.m21 + v.z * m.m22 + 1 * m.m23;

		return ret;
	}

	/**
	 * Multiply a 3-dimensional vector with a 3-dimensional matrix<br>
	 * In the end this results in the following operation:<br>
	 * <b>vec = m * v</b>
	 * @param m Matrix
	 * @param v Vector
	 * 
	 * @return Resulting vector
	 */
	public static Vector3d mulMatrixVector(Matrix3d m, Vector3d v)
	{
		Vector3d ret = new Vector3d();

		ret.x = v.x * m.m00 + v.y * m.m01 + v.z * m.m02;
		ret.y = v.x * m.m10 + v.y * m.m11 + v.z * m.m12;
		ret.z = v.x * m.m20 + v.y * m.m21 + v.z * m.m22;

		return ret;
	}

	/**
	 * Check if the given 3-dimensional matrix is symmetric
	 * 
	 * @param matrix Input matrix
	 * @return True if symmetric, false if not
	 */
	public static boolean isSymmetric(Matrix3d matrix)
	{
		return matrix.m01 == matrix.m10 && matrix.m02 == matrix.m20
				&& matrix.m12 == matrix.m21;

	}

	/**
	 * Perform a Choleski decomposition on a 3-dimensional matrix. The input
	 * matrix has to be symmetric and positive definite.
	 * 
	 * @param matrix 3-dimensional input matrix
	 * @return Resulting matrix
	 * @throws RuntimeException in case the input matrix is not symmetric and
	 *         positive definite
	 */
	public static Matrix3d cholesky(Matrix3d matrix) throws RuntimeException
	{
		// Check symmetry
		if (!isSymmetric(matrix)) {
			throw new RuntimeException("Matrix is not symmetric");
		}

		Matrix3d res = new Matrix3d();

		res.m00 = Math.sqrt(matrix.m00);
		res.m01 = matrix.m01 / res.m00;
		res.m02 = matrix.m02 / res.m00;

		double temp = matrix.m11 - res.m01 * res.m01;

		if (temp <= 0) {
			throw new RuntimeException("Matrix not positive definite");
		}
		res.m11 = Math.sqrt(temp);
		res.m12 = (matrix.m12 - res.m01 * res.m02) / res.m11;

		temp = matrix.m22 - res.m02 * res.m02 - res.m12 * res.m12;
		if (temp <= 0) {
			throw new RuntimeException("Matrix not positive definite");
		}
		res.m22 = Math.sqrt(temp);

		return res;
	}

	/**
	 * Method to create a 3D rotation matrix out of three rotation angles.<br>
	 * This Method represents a forward transformation:<br>
	 * <b>R = Rz * Ry * Rx</b>
	 * 
	 * @param rotX - rotation on x axis (in rad)
	 * @param rotY - rotation on y axis (in rad)
	 * @param rotZ - rotation on z axis (in rad)
	 * @return the rotation matrix
	 */
	public static Matrix3d createRotationMatrix(double rotX, double rotY,
			double rotZ)
	{
		Matrix3d res = new Matrix3d(1, 0, 0, 0, 1, 0, 0, 0, 1);

		// The forward rotation is R = Rz * Ry * Rx

		// first: Rotate X
		rotateX(res, rotX);

		// second: Rotate Y
		rotateY(res, rotY);

		// third: Rotate Z
		rotateZ(res, rotZ);

		return res;
	}

	/**
	 * Method to create a <u>inverse</u> 3D rotation matrix out of three rotation
	 * angles.<br>
	 * This Method represents a backward (inverse) transformation:<br>
	 * <b>R^-1 = Rx^-1 * Ry^-1 * Rz^-1</b>
	 * 
	 * @param rotX - rotation on x axis (in rad)
	 * @param rotY - rotation on y axis (in rad)
	 * @param rotZ - rotation on z axis (in rad)
	 * @return the inverse rotation matrix
	 */
	public static Matrix3d createInverseRotationMatrix(double rotX, double rotY,
			double rotZ)
	{
		Matrix3d res = new Matrix3d(1, 0, 0, 0, 1, 0, 0, 0, 1);

		// The forward rotation is R = Rz * Ry * Rx
		// Therefore the inverse rotation is R^-1 = Rx^-1 * Ry^-1 * Rz^-1
		// (take the inverse of each matrix and switch order of multiplying)
		// Since R^-1(phi) = R(-phi) holds for all rotation matrixes, we just have
		// to negate all angles and switch the order of multiplying, to get the
		// inverse rotation matrix

		// Rotate Z
		rotateZ(res, -rotZ);

		// Rotate Y
		rotateY(res, -rotY);

		// Rotate X
		rotateX(res, -rotX);

		return res;
	}

	/**
	 * Method to create a 4D translation matrix based on the three shift values.
	 * 
	 * @param x - translation in x direction
	 * @param y - translation in y direction
	 * @param z - translation in z direction
	 * @return the translation matrix
	 */
	public static Matrix4d createTranslationMatrix(double x, double y, double z)
	{
		return new Matrix4d(1, 0, 0, x, 0, 1, 0, y, 0, 0, 1, z, 0, 0, 0, 1);
	}

	/**
	 * Method to create a 4D translation matrix based on the given vector.
	 * 
	 * @param trans - vector with shift values
	 * @return the translation matrix
	 */
	public static Matrix4d createTranslationMatrix(Vector3d trans)
	{
		return createTranslationMatrix(trans.x, trans.y, trans.z);
	}

	/**
	 * Method to create a 4D <u>inverse</u> translation matrix based on the three
	 * shift values.
	 * 
	 * @param x - translation in x direction
	 * @param y - translation in y direction
	 * @param z - translation in z direction
	 * @return the translation matrix
	 */
	public static Matrix4d createInverseTranslationMatrix(double x, double y,
			double z)
	{
		// Since the inverse of a shift operation is simply the backward movement,
		// we just have to negate all shift parameters.
		return createTranslationMatrix(-x, -y, -z);
	}

	/**
	 * Method to create a 4D <u>inverse</u> translation matrix based on the given
	 * vector.
	 * 
	 * @param trans - vector with shift values
	 * @return the translation matrix
	 */
	public static Matrix4d createInverseTranslationMatrix(Vector3d trans)
	{
		// Since the inverse of a shift operation is simply the backward movement,
		// we just have to negate all shift parameters.
		return createTranslationMatrix(-trans.x, -trans.y, -trans.z);
	}

	/**
	 * Rotate a 4-dimensional matrix around three axis
	 * 
	 * @param matrix The input matrix to rotate
	 * @param rotX Rotation on the x-axis (in rad)
	 * @param rotY Rotation on the y-axis (in rad)
	 * @param rotZ Rotation on the z-axis (in rad)
	 */
	public static void rotateMatrix(Matrix4d matrix, double rotX, double rotY,
			double rotZ)
	{
		// TODO: Think about the Gyro-Story (and with that the initial motivation
		// of this method) again...

		// Rotate Y
		rotateY(matrix, rotY);

		// Rotate X
		rotateX(matrix, rotX);

		// Rotate Z
		rotateZ(matrix, rotZ);
	}

	/**
	 * Rotate a 4-dimensional matrix around the X axis.<br>
	 * In the end this results in the following operation:<br>
	 * <b>matrix = matrix * Rx(rotX)</b>
	 * 
	 * @param matrix The input matrix to rotate
	 * @param rotX Rotation on the X-axis (in rad)
	 */
	public static void rotateX(Matrix4d matrix, double rotX)
	{
		// Rotate X
		Matrix4d rotMatrix = new Matrix4d();
		rotMatrix.rotX(rotX);

		matrix.mul(rotMatrix);
	}

	/**
	 * Rotate a 4-dimensional matrix around the Y axis<br>
	 * In the end this results in the following operation:<br>
	 * <b>matrix = matrix * Ry(rotY)</b>
	 * 
	 * @param matrix The input matrix to rotate
	 * @param rotY Rotation on the Y-axis (in rad)
	 */
	private static void rotateY(Matrix4d matrix, double rotY)
	{
		// Rotate Y
		Matrix4d rotMatrix = new Matrix4d();
		rotMatrix.rotY(rotY);

		matrix.mul(rotMatrix);
	}

	/**
	 * Rotate a 4-dimensional matrix around the Z axis.<br>
	 * In the end this results in the following operation:<br>
	 * <b>matrix = matrix * Rz(rotZ)</b>
	 * 
	 * @param matrix The input matrix to rotate
	 * @param rotZ Rotation on the Z-axis (in rad)
	 */
	public static void rotateZ(Matrix4d matrix, double rotZ)
	{
		// Rotate Z
		Matrix4d rotMatrix = new Matrix4d();
		rotMatrix.rotZ(rotZ);

		matrix.mul(rotMatrix);
	}

	/**
	 * Rotate a 3-dimensional matrix around three axis
	 * 
	 * @param matrix The input matrix to rotate
	 * @param rotX Rotation on the x-axis (in rad)
	 * @param rotY Rotation on the y-axis (in rad)
	 * @param rotZ Rotation on the z-axis (in rad)
	 */
	public static void rotateMatrix(Matrix3d matrix, double rotX, double rotY,
			double rotZ)
	{
		// TODO: Think about the Gyro-Story (and with that the initial motivation
		// of this method) again...

		// Rotate Y
		rotateY(matrix, rotY);

		// Rotate X
		rotateX(matrix, rotX);

		// Rotate Z
		rotateZ(matrix, rotZ);
	}

	/**
	 * Rotate a 3-dimensional matrix around the X axis.<br>
	 * In the end this results in the following operation:<br>
	 * <b>matrix = matrix * Rx(rotX)</b>
	 * 
	 * @param matrix The input matrix to rotate
	 * @param rotX Rotation on the X-axis (in rad)
	 */
	public static void rotateX(Matrix3d matrix, double rotX)
	{
		// Rotate X
		Matrix3d rotMatrix = new Matrix3d();
		rotMatrix.rotX(rotX);

		matrix.mul(rotMatrix);
	}

	/**
	 * Rotate a 3-dimensional matrix around the Y axis<br>
	 * In the end this results in the following operation:<br>
	 * <b>matrix = matrix * Ry(rotY)</b>
	 * 
	 * @param matrix The input matrix to rotate
	 * @param rotY Rotation on the Y-axis (in rad)
	 */
	public static void rotateY(Matrix3d matrix, double rotY)
	{
		// Rotate Y
		Matrix3d rotMatrix = new Matrix3d();
		rotMatrix.rotY(rotY);

		matrix.mul(rotMatrix);
	}

	/**
	 * Rotate a 3-dimensional matrix around the Z axis.<br>
	 * In the end this results in the following operation:<br>
	 * <b>matrix = matrix * Rz(rotZ)</b>
	 * 
	 * @param matrix The input matrix to rotate
	 * @param rotZ Rotation on the Z-axis (in rad)
	 */
	public static void rotateZ(Matrix3d matrix, double rotZ)
	{
		// Rotate Z
		Matrix3d rotMatrix = new Matrix3d();
		rotMatrix.rotZ(rotZ);

		matrix.mul(rotMatrix);
	}

	/**
	 * Returns a 3-dimensional matrix which is rotated around the z-Axis, by the
	 * given value. It represents an orientation in the 3-dimensional space.
	 * 
	 * @param zRotation Rotation around the z-axis in degrees
	 * @return An orientation matrix calculated in respective to the given
	 *         z-rotation @
	 */
	public static Matrix3d getOrientationMatrix(float zRotation)
	{
		Matrix3d rot = new Matrix3d(1, 0, 0, 0, 1, 0, 0, 0, 1);
		rot.rotZ(Math.toRadians(zRotation));

		Matrix3d m = new Matrix3d(1, 0, 0, 0, 1, 0, 0, 0, 1);

		m.mul(rot);
		return m;
	}

	/**
	 * Retrieve the euler angles to a 3D rotation described by an matrix.
	 * 
	 * @param m the matrix representing a rotation
	 * @param order the order of the rotations (e.g. XYZ, XZY, YXZ, etc.)
	 * @return an array containing the three angles around the three system-axis
	 */
	public static double[] getEulerAngles(Matrix3d m, RotationOrder order)
	{
		double[] res = new double[3];

		switch (order) {
		case XYZ:
			if (m.m02 > 0.998) { // singularity 1
				res[0] = Math.atan2(m.m21, m.m11);
				res[1] = Math.PI / 2;
				res[2] = 0;
			} else if (m.m02 < -0.998) { // singularity 2
				res[0] = Math.atan2(m.m21, m.m11);
				res[1] = -Math.PI / 2;
				res[2] = 0;
			} else {
				res[0] = Math.atan2(-m.m12, m.m22);
				res[1] = Math.asin(m.m02);
				res[2] = Math.atan2(-m.m01, m.m00);
			}
			break;
		case XZY:
			if (m.m01 > 0.998) { // singularity 1
				res[0] = Math.atan2(-m.m12, m.m22);
				res[1] = -Math.PI / 2;
				res[2] = 0;
			} else if (m.m01 < -0.998) { // singularity 2
				res[0] = Math.atan2(-m.m12, m.m22);
				res[1] = Math.PI / 2;
				res[2] = 0;
			} else {
				res[0] = Math.atan2(m.m21, m.m11);
				res[1] = Math.asin(-m.m01);
				res[2] = Math.atan2(m.m02, m.m00);
			}
			break;
		case YXZ:
			if (m.m12 > 0.998) { // singularity 1
				res[0] = Math.atan2(-m.m20, m.m00);
				res[1] = -Math.PI / 2;
				res[2] = 0;
			} else if (m.m12 < -0.998) { // singularity 2
				res[0] = Math.atan2(-m.m20, m.m00);
				res[1] = Math.PI / 2;
				res[2] = 0;
			} else {
				res[0] = Math.atan2(m.m02, m.m22);
				res[1] = Math.asin(-m.m12);
				res[2] = Math.atan2(m.m10, m.m11);
			}
			break;
		case YZX:
			if (m.m10 > 0.998) { // singularity 1
				res[0] = Math.atan2(m.m02, m.m22);
				res[1] = Math.PI / 2;
				res[2] = 0;
			} else if (m.m10 < -0.998) { // singularity 2
				res[0] = Math.atan2(m.m02, m.m22);
				res[1] = -Math.PI / 2;
				res[2] = 0;
			} else {
				res[0] = Math.atan2(-m.m20, m.m00);
				res[1] = Math.asin(m.m10);
				res[2] = Math.atan2(-m.m12, m.m11);
			}
			break;
		case ZXY:
			if (m.m21 > 0.998) { // singularity 1
				res[0] = Math.atan2(m.m10, m.m00);
				res[1] = Math.PI / 2;
				res[2] = 0;
			} else if (m.m21 < -0.998) { // singularity 2
				res[0] = Math.atan2(m.m10, m.m00);
				res[1] = -Math.PI / 2;
				res[2] = 0;
			} else {
				res[0] = Math.atan2(-m.m01, m.m11);
				res[1] = Math.asin(m.m21);
				res[2] = Math.atan2(-m.m20, m.m22);
			}
			break;
		case ZYX:
			if (m.m20 > 0.998) { // singularity 1
				res[0] = Math.atan2(-m.m01, m.m11);
				res[1] = -Math.PI / 2;
				res[2] = 0;
			} else if (m.m20 < -0.998) { // singularity 2
				res[0] = Math.atan2(-m.m01, m.m11);
				res[1] = Math.PI / 2;
				res[2] = 0;
			} else {
				res[0] = Math.atan2(m.m10, m.m00);
				res[1] = Math.asin(-m.m20);
				res[2] = Math.atan2(m.m21, m.m22);
			}
			break;
		}

		return res;
	}
}
