/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry;

import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

/**
 * Implements a set of useful vector functions (related to javax.vecmath).
 * 
 * @author Stefan Glaser
 * 
 */
class VectorUtil
{
	/**
	 * Build a vector from the given azimuthal coordinates
	 * @param distance scale factor
	 * @param alpha azimuth (&alpha;) around Z (0 is +X, &pi;/2 is +Y, &pi; is -X
	 *        and 3&pi;/2 is -Y)
	 * @param delta elevation (&delta;) above (XY) plane, from -&pi;/2 to +&pi;/2
	 */
	public static Vector3d toVector(double distance, double alpha, double delta)
	{
		Vector3d v = new Vector3d();
		double cosDelta = Math.cos(delta);

		v.x = Math.cos(alpha) * cosDelta * distance;
		v.y = Math.sin(alpha) * cosDelta * distance;
		v.z = Math.sin(delta) * distance;

		return v;
	}

	/**
	 * Build a point from the given azimuthal coordinates
	 * @param distance scale factor
	 * @param alpha azimuth (&alpha;) around Z (0 is +X, &pi;/2 is +Y, &pi; is -X
	 *        and 3&pi;/2 is -Y)
	 * @param delta elevation (&delta;) above (XY) plane, from -&pi;/2 to +&pi;/2
	 */
	public static Point3d toPoint(double distance, double alpha, double delta)
	{
		Point3d v = new Point3d();
		double cosDelta = Math.cos(delta);

		v.x = Math.cos(alpha) * cosDelta * distance;
		v.y = Math.sin(alpha) * cosDelta * distance;
		v.z = Math.sin(delta) * distance;

		return v;
	}

	/**
	 * Get the azimuth to the vector.
	 * @return azimuth (&alpha;) to the vector, between -&pi; and +&pi;
	 */
	public static double alpha(Vector3d vec)
	{
		return Math.atan2(vec.y, vec.x);
	}

	/**
	 * Get the azimuth of the two vectors.
	 * @return azimuth (&alpha;) of the two points, between -&pi; and +&pi;
	 */
	public static double alpha(Vector3d vec1, Vector3d vec2)
	{
		return Math.atan2(vec2.y - vec1.y, vec2.x - vec1.x);
	}

	/**
	 * Get the elevation to the vector.
	 * @return elevation (&delta;) to the vector, between -&pi;/2 and +&pi;/2
	 */
	public static double delta(Vector3d vec)
	{
		return Math.asin(vec.z / vec.length());
	}

	/**
	 * Get the azimuth of the two points.
	 * @return azimuth (&alpha;) of the two points, between -&pi; and +&pi;
	 */
	public static double distance(Vector3d vec1, Vector3d vec2)
	{
		return new Point3d(vec1).distance(new Point3d(vec2));
	}
}
