/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.statemachine;

/**
 * Represents a transition from one state of a state machine to another
 * 
 * @param <K> the enum used to identify states
 * @param <E> the enum used to identify events
 * 
 * @author Klaus Dorer
 */
public class StateTransition<K, E>
{
	/** ID of the source state */
	private final K source;

	/** ID of the destination state */
	private final K destination;

	/** Event ID that triggers this transition */
	private final E trigger;

	/**
	 * @param source id of the source state from which to transition
	 * @param destination id of the destination state to which to transition
	 * @param trigger the event id that triggers the transition
	 */
	public StateTransition(K source, K destination, E trigger)
	{
		super();
		this.source = source;
		this.destination = destination;
		this.trigger = trigger;
	}

	/**
	 * @return the source state ID
	 */
	public K getSource()
	{
		return source;
	}

	/**
	 * @return the destination state ID
	 */
	public K getDestination()
	{
		return destination;
	}

	/**
	 * Checks if this transition can be triggered. This is true if the
	 * corresponding event occurred and the guard condition is satisfied.
	 * @param eventID the event that triggered checking
	 * @param context the object that triggered the event
	 * @return true if this transition is allowed
	 */
	public boolean canTrigger(E eventID, Object context)
	{
		if (eventID != trigger) {
			// not the same event
			return false;
		}

		return checkGuardCondition(context);

	}

	/**
	 * Checks the guard condition of this transition. The default implementation
	 * always returns true implementing the situation where there is no guard
	 * condition
	 * @param context the context object
	 * @return true if the guard of this transition is fulfilled. The default
	 *         implementation always returns true
	 */
	protected boolean checkGuardCondition(Object context)
	{
		return true;
	}

	/**
	 * Performs the transition part of a state change (but not the state change
	 * itself)
	 * @param context the object that triggered the event
	 */
	public void transitionAction(Object context)
	{
	}
}
