/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.agentmodel.impl;

import static org.easymock.EasyMock.createMock;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.replay;
import static org.easymock.EasyMock.verify;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotSame;
import static org.junit.Assert.assertSame;

import java.util.ArrayList;
import java.util.List;

import javax.vecmath.Vector3d;

import magma.agent.action.IAction;
import magma.agent.agentmodel.IBodyModel;
import magma.agent.meta.agent.IAgentMetaModel;
import magma.agent.meta.agent.IBodyPartConfiguration;
import magma.agent.meta.agent.impl.AgentMetaModel;
import magma.agent.meta.agent.impl.BodyPartConfiguration;
import magma.agent.meta.agent.impl.HingeJointConfiguration;
import magma.agent.meta.agent.impl.SensorConfiguration;
import magma.agent.perception.IAccelerometerPerceptor;
import magma.agent.perception.IGyroPerceptor;
import magma.agent.perception.IHingeJointPerceptor;
import magma.agent.perception.IPerception;
import magma.robots.nao.INaoConstants;

import org.apache.commons.math.geometry.Vector3D;
import org.junit.Before;
import org.junit.Test;

/**
 * 
 * @author dorer
 */
public class BodyModelTest
{
	private BodyModel testee;

	/**
	 * @throws java.lang.Exception
	 */
	@Before
	public void setUp() throws Exception
	{
		IAgentMetaModel metaModel2 = new AgentMetaModel("sceneString",
				getTestBodyPartConfig(), null);
		testee = new BodyModel(metaModel2);
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#BodyModel(magma.agent.meta.agent.IAgentMetaModel)}
	 * .
	 */
	@Test
	public void testBodyModelIAgentMetaModel()
	{
		assertEquals("torso", testee.getTorso().getName());
		assertEquals(1, testee.getTorso().getNoOfChildren());
		assertSame(testee.getTorso().getJoint(),
				testee.getJoint(INaoConstants.Torso));
		assertEquals("TorsoGyro", testee.getSensor("TorsoGyro").getName());
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#BodyModel(magma.agent.agentmodel.impl.BodyModel)}
	 * .
	 */
	@Test
	public void testBodyModelBodyModel()
	{
		IBodyModel copy = new BodyModel(testee);
		assertNotSame(copy.getTorso(), testee.getTorso());
		assertEquals(1, copy.getTorso().getNoOfChildren());
		assertEquals(testee.getSensor("TorsoGyro"), copy.getSensor("TorsoGyro"));
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#getJoint(java.lang.String)} .
	 */
	@Test
	public void testGetJoint()
	{
		assertEquals(INaoConstants.NeckYaw, testee
				.getJoint(INaoConstants.NeckYaw).getName());
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#generateJointActions(magma.agent.action.IAction)}
	 * .
	 */
	@Test
	public void testGenerateJointActions()
	{
		IAction action = createMock(IAction.class);
		action.setHingeEffectorSpeed("he1", 0.0f);
		replay(action);
		testee.generateJointActions(action);
		verify(action);
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#updateFromPerception(magma.agent.perception.IPerception)}
	 * .
	 */
	@Test
	public void testUpdateFromPerception()
	{
		IPerception perception = createMock(IPerception.class);
		IHingeJointPerceptor hingePerceptor = createMock(IHingeJointPerceptor.class);
		expect(hingePerceptor.getAxis()).andReturn(20.0f);

		IGyroPerceptor gyroPerceptor = createMock(IGyroPerceptor.class);
		expect(gyroPerceptor.getGyro()).andReturn(new Vector3D(1, 2, 3));

		IAccelerometerPerceptor accPerceptor = createMock(IAccelerometerPerceptor.class);
		expect(accPerceptor.getAcceleration()).andReturn(new Vector3d(1, 2, 3));

		expect(perception.getHingeJointPerceptor("hj1"))
				.andReturn(hingePerceptor);
		expect(perception.getGyroRatePerceptor("torso")).andReturn(gyroPerceptor);
		expect(perception.getAccelerationPerceptor("torso")).andReturn(
				accPerceptor);
		replay(perception, hingePerceptor, gyroPerceptor, accPerceptor);

		testee.updateFromPerception(perception);

		HingeJoint hingeJoint = (HingeJoint) testee
				.getJoint(INaoConstants.NeckYaw);
		assertEquals(20.0f, hingeJoint.getAxis(), 0.0001);
		assertEquals(0.0f, hingeJoint.getNextAxisSpeed(), 0.0001);
		// TODO switch these three lines on once the miraculous factor 2 in
		// performAxisPosition is removed
		// hingeJoint.performAxisPosition(25);
		// assertEquals(25.0f, hingeJoint.getAxis(), 0.0001);
		// assertEquals(5.0f, hingeJoint.getNextAxisSpeed(), 0.0001);

		verify(perception);
	}

	private List<IBodyPartConfiguration> getTestBodyPartConfig()
	{
		List<IBodyPartConfiguration> configs = new ArrayList<IBodyPartConfiguration>();

		IBodyPartConfiguration currentConfig;

		currentConfig = new BodyPartConfiguration(INaoConstants.Torso, null,
				new Vector3d(0, 0, 0), 1.2171f, new Vector3d(0.1, 0.1, 0.18), null,
				new Vector3d(0, 0, 0),
				new SensorConfiguration("TorsoGyro", "torso"),
				new SensorConfiguration("TorsoAccel", "torso"), null);
		configs.add(currentConfig);

		currentConfig = new BodyPartConfiguration(INaoConstants.Neck,
				INaoConstants.Torso, new Vector3d(0, 0, 0.09), 0.05f, new Vector3d(
						0.03, 0.03, 0.08), new HingeJointConfiguration(
						INaoConstants.NeckYaw, "hj1", "he1", new Vector3d(0, 0, 1),
						-120, 120, INaoConstants.MAX_JOINT_SPEED), new Vector3d(0, 0,
						0), null, null, null);
		configs.add(currentConfig);

		return configs;
	}

	/**
	 * Test method for
	 * {@link magma.agent.agentmodel.impl.BodyModel#BodyModel(magma.agent.agentmodel.impl.BodyModel)}
	 * .
	 */
	@Test
	public void testUpdateJointSpeed()
	{
		BodyModel copy = new BodyModel(testee);

		HingeJoint hj1 = (HingeJoint) copy.getJoint(INaoConstants.NeckYaw);
		hj1.performAxisSpeed(3.0f);
		testee.updateJointsSpeed(copy);

		HingeJoint hj1Source = (HingeJoint) testee
				.getJoint(INaoConstants.NeckYaw);
		assertEquals(3.0f, hj1Source.getNextAxisSpeed(), 0.00001);
	}

}
