/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/

package magma.agent.agentmodel.impl;

import static org.easymock.classextension.EasyMock.createMock;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotSame;
import static org.junit.Assert.assertSame;

import javax.vecmath.Matrix3d;
import javax.vecmath.Vector3d;

import magma.agent.agentmodel.IJoint;
import magma.agent.agentmodel.ISensor;
import magma.util.geometry.MatrixUtil;

import org.easymock.classextension.EasyMock;
import org.junit.Before;
import org.junit.Test;

/**
 * @author stilnox
 * 
 */
public class BodyPartTest
{

	private IJoint joint2;

	private IJoint joint3;

	private IJoint joint4;

	private IJoint joint5;

	private BodyPart torso;

	private BodyPart shoulderR;

	private BodyPart lowerarm;

	private BodyPart hand;

	/**
	 * @throws java.lang.Exception
	 */
	@Before
	public void setUp() throws Exception
	{
	}

	@Test
	public void test1()
	{
		joint2 = createJointMockNoRotation("joint2Name");
		joint3 = createJointMockNoRotation("joint3Name");

		torso = new BodyPart("torso", null, null, new Vector3d(0, 0, 0),
				new Vector3d(0, 0, 0), 0f, null);

		BodyPart part1 = new BodyPart("part1", torso, joint2, new Vector3d(10, 0,
				0), new Vector3d(0, 0, 0), 0f, null);

		BodyPart part2 = new BodyPart("part2", part1, joint3, new Vector3d(0, 0,
				0), new Vector3d(10, 0, 0), 0f, null);

		assertEquals(new Vector3d(10, 0, 0), part2.getPosition());
	}

	@Test
	public void test2()
	{
		joint2 = createJointMockNoRotation("joint2Name");
		joint3 = createJointMockNoRotation("joint3Name");

		torso = new BodyPart("torso", null, null, new Vector3d(0, 0, 0),
				new Vector3d(0, 0, 0), 0f, null);

		BodyPart part1 = new BodyPart("part1", torso, joint2, new Vector3d(10, 0,
				0), new Vector3d(0, 0, 0), 0f, null);

		BodyPart part2 = new BodyPart("part2", part1, joint3, new Vector3d(0, 10,
				0), new Vector3d(0, 0, 0), 0f, null);

		assertEquals(new Vector3d(10, 10, 0), part2.getPosition());
	}

	@Test
	public void test3()
	{
		createJointMocks();
		createBodyParts();

		Vector3d handPos = hand.getPosition();

		assertEquals(0.108, handPos.x, 0.00001);
		assertEquals(0.195, handPos.y, 0.00001);
		assertEquals(0.084, handPos.z, 0.00001);
	}

	@Test
	public void test4()
	{
		createJointMocks();
		joint2 = createJointMock(
				MatrixUtil.createRotationMatrix(Math.toRadians(90), 0, 0),
				"joint2Name");
		createBodyParts();

		Vector3d position = lowerarm.getPosition();

		assertEquals(0.098, position.x, 0.0001);
		assertEquals(-0.009, position.y, 0.0001);
		assertEquals(0.215, position.z, 0.0001);
	}

	@Test
	public void testCopyConstructor()
	{
		createTorsoAndJoint2();
		joint3 = createMock(IJoint.class);
		EasyMock.expect(joint2.copy()).andReturn(joint3);
		EasyMock.replay(joint2);

		shoulderR = new BodyPart("shoulderR", torso, joint2, new Vector3d(1.0,
				2.0, 3.0), new Vector3d(0, 0, 0), 0.5f, null);

		BodyPart copyTorso = new BodyPart(torso, null);
		BodyPart copyShoulder = copyTorso.getChild("shoulderR");

		assertEquals(1, copyTorso.getNoOfChildren());
		assertSame(copyTorso, copyShoulder.getParent());
		assertNotSame(torso, copyTorso);
		assertNotSame(shoulderR, copyShoulder);
		assertNotSame(shoulderR.getJoint(), copyShoulder.getJoint());

		assertNotSame(torso.getTransformation(), copyTorso.getTransformation());
		assertEquals(torso.getTransformation(), copyTorso.getTransformation());
	}

	@Test
	public void testGetBodyPart()
	{
		createJointMocks();
		createBodyParts();

		BodyPart result = torso.getBodyPart("hand");
		assertSame(hand, result);
	}

	@Test
	public void testGetSensor()
	{
		createJointMocks();
		createBodyParts();

		ISensor result = torso.getSensorDeep("joint2Name");
		assertSame(joint2, result);
	}

	@Test
	public void testGetCenterOfMass()
	{
		createTorsoAndJoint2();
		EasyMock.replay(joint2);
		shoulderR = new BodyPart("shoulderR", torso, joint2, new Vector3d(1.0,
				2.0, 3.0), new Vector3d(0, 0, 0), 0.5f, null);

		Vector3d result = torso.getCenterOfMass();
		assertEquals(0.2f, result.x, 0.00001);
		assertEquals(0.4f, result.y, 0.00001);
		assertEquals(0.6f, result.z, 0.00001);
	}

	private void createJointMocks()
	{
		joint2 = createJointMockNoRotation("joint2Name");
		joint3 = createJointMockNoRotation("joint3Name");
		joint4 = createJointMockNoRotation("joint4Name");
		joint5 = createJointMockNoRotation("joint5Name");
	}

	private IJoint createJointMockNoRotation(String name)
	{
		return createJointMock(MatrixUtil.createRotationMatrix(0, 0, 0), name);
	}

	private IJoint createJointMock(Matrix3d rotation, String name)
	{
		IJoint joint = createMock(IJoint.class);
		EasyMock.expect(joint.getRotationMatrix()).andReturn(rotation);
		EasyMock.expect(joint.getName()).andStubReturn(name);
		EasyMock.replay(joint);
		return joint;
	}

	/**
	 * @return
	 */
	private void createBodyParts()
	{
		torso = new BodyPart("torso", null, null, new Vector3d(0, 0, 0),
				new Vector3d(0, 0, 0), 0f, null);

		shoulderR = new BodyPart("shoulderR", torso, joint2, new Vector3d(0.098,
				0, 0.075), new Vector3d(0, 0, 0), 0f, null);

		BodyPart upperarmR = new BodyPart("upperarmR", shoulderR, joint3,
				new Vector3d(0.01, 0.02, 0), new Vector3d(-0.01, -0.02, 0), 0f,
				null);

		BodyPart elbow = new BodyPart("elbow", upperarmR, joint4, new Vector3d(
				-0.01, 0.07, 0.009), new Vector3d(0, 0, 0), 0f, null);

		lowerarm = new BodyPart("lowerarm", elbow, joint5, new Vector3d(0, 0.05,
				0), new Vector3d(0, -0.05, 0), 0f, null);

		hand = new BodyPart("hand", lowerarm, null, new Vector3d(0.01, 0.055, 0),
				new Vector3d(0, 0, 0), 0f, null);
	}

	/**
	 * 
	 */
	private void createTorsoAndJoint2()
	{
		joint2 = createMock(IJoint.class);
		EasyMock.expect(joint2.getRotationMatrix()).andReturn(
				MatrixUtil.createRotationMatrix(0, 0, 0));
		EasyMock.expect(joint2.getName()).andStubReturn("joint2Name");

		torso = new BodyPart("torso", null, null, new Vector3d(0, 0, 0),
				new Vector3d(0, 0, 0), 2.0f, null);
	}

}
