/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.agent.belief.impl;

import static org.easymock.EasyMock.createMock;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.replay;
import static org.junit.Assert.assertEquals;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import magma.agent.worldmodel.IGlobalMap;
import magma.agent.worldmodel.IMoveableObject;
import magma.agent.worldmodel.IPlayer;
import magma.agent.worldmodel.IThisPlayer;

import org.apache.commons.math.geometry.Vector3D;
import org.junit.Before;
import org.junit.Test;

/**
 * @author Klaus Dorer
 * 
 */
public class TeammateCloserToBallTest
{
	private TeammateCloserToBall testee;

	private IGlobalMap worldModelMock;

	private IThisPlayer thisPlayerMock;

	private IPlayer playerMock;

	private IPlayer player2Mock;

	private IPlayer opponentMock;

	private IMoveableObject ballMock;

	private Vector3D goalPosition;

	private Vector3D ballPosition;

	/**
	 * @throws java.lang.Exception
	 */
	@Before
	public void setUp() throws Exception
	{
		ballMock = createMock("ball", IMoveableObject.class);
		playerMock = createMock("player", IPlayer.class);
		player2Mock = createMock("player2", IPlayer.class);
		opponentMock = createMock("opponent", IPlayer.class);
		thisPlayerMock = createMock("thisPlayer", IThisPlayer.class);
		worldModelMock = createMock(IGlobalMap.class);
		ballPosition = new Vector3D();
		goalPosition = new Vector3D();
		expect(ballMock.getPosition()).andStubReturn(ballPosition);
		expect(player2Mock.getPosition()).andStubReturn(new Vector3D(1, 0, 0));
		expect(opponentMock.getPosition()).andStubReturn(new Vector3D(2, 0, 0));
		expect(playerMock.getPosition()).andStubReturn(new Vector3D(4, 0, 0));
		expect(playerMock.getID()).andStubReturn(1);
		expect(player2Mock.getID()).andStubReturn(2);
		expect(thisPlayerMock.getID()).andStubReturn(3);
		expect(worldModelMock.getThisPlayer()).andStubReturn(thisPlayerMock);
		expect(worldModelMock.getBall()).andStubReturn(ballMock);
		expect(worldModelMock.getOpponentAtBall()).andStubReturn(opponentMock);
		expect(worldModelMock.getOtherGoalPosition()).andStubReturn(goalPosition);

		testee = new TeammateCloserToBall(null, worldModelMock);
	}

	/**
	 * Test method for
	 * {@link magma.agent.belief.impl.TeammateCloserToBall#getTruthValue()}.
	 */
	@Test
	public void testGetTruthValueNoTeammates()
	{
		expect(worldModelMock.getTeammatesAtBall()).andReturn(
				new ArrayList<IPlayer>(0));
		expect(worldModelMock.getGlobalTime()).andReturn(0f);
		replay(worldModelMock, thisPlayerMock);

		assertEquals(0.0, testee.getTruthValue(), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.agent.belief.impl.TeammateCloserToBall#getTruthValue()}.
	 */
	@Test
	public void testGetTruthValueFarerTeammate()
	{
		expect(playerMock.getDistanceToXY(ballPosition)).andReturn(1.0);
		expect(thisPlayerMock.getPosition()).andStubReturn(
				new Vector3D(1.5, 0, 0));
		expect(playerMock.isLaying()).andStubReturn(false);
		expect(thisPlayerMock.isLaying()).andStubReturn(false);
		expect(thisPlayerMock.getDistanceToXY(ballPosition)).andReturn(0.5);
		expect(thisPlayerMock.getDistanceToXY(goalPosition)).andStubReturn(5.0);
		expect(playerMock.getDistanceToXY(goalPosition)).andStubReturn(5.0);
		expect(ballMock.getDistanceToXY(goalPosition)).andStubReturn(4.0);
		expect(worldModelMock.getTeammatesAtBall()).andReturn(
				Collections.singletonList(playerMock));
		expect(worldModelMock.getGlobalTime()).andReturn(0f);
		replay(worldModelMock, thisPlayerMock, playerMock, opponentMock, ballMock);

		assertEquals(0.0, testee.getTruthValue(), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.agent.belief.impl.TeammateCloserToBall#getTruthValue()}.
	 */
	@Test
	public void testGetTruthValueCloserTeammate()
	{
		expect(player2Mock.getDistanceToXY(ballPosition)).andReturn(1.0);
		expect(thisPlayerMock.getPosition()).andStubReturn(new Vector3D(0, 0, 0));
		expect(thisPlayerMock.getDistanceToXY(ballPosition)).andReturn(1.5);
		expect(worldModelMock.getTeammatesAtBall()).andReturn(
				Collections.singletonList(player2Mock));
		expect(worldModelMock.getGlobalTime()).andReturn(0f);
		expect(player2Mock.isLaying()).andStubReturn(false);
		expect(thisPlayerMock.isLaying()).andStubReturn(false);
		expect(thisPlayerMock.getDistanceToXY(goalPosition)).andStubReturn(5.0);
		expect(player2Mock.getDistanceToXY(goalPosition)).andStubReturn(5.0);
		expect(ballMock.getDistanceToXY(goalPosition)).andStubReturn(4.0);
		replay(worldModelMock, thisPlayerMock, player2Mock, opponentMock,
				ballMock);

		assertEquals(1.0, testee.getTruthValue(), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.agent.belief.impl.TeammateCloserToBall#getTruthValue()}.
	 */
	@Test
	public void testGetTruthValueTeammateCloserToGoal()
	{
		expect(player2Mock.getDistanceToXY(ballPosition)).andReturn(0.5);
		expect(thisPlayerMock.getPosition()).andStubReturn(new Vector3D(0, 0, 0));
		expect(thisPlayerMock.getDistanceToXY(ballPosition)).andReturn(2.0);
		expect(worldModelMock.getTeammatesAtBall()).andReturn(
				Collections.singletonList(player2Mock));
		expect(worldModelMock.getGlobalTime()).andReturn(0f);
		expect(player2Mock.isLaying()).andStubReturn(false);
		expect(thisPlayerMock.isLaying()).andStubReturn(false);
		expect(thisPlayerMock.getDistanceToXY(goalPosition)).andStubReturn(5.0);
		expect(player2Mock.getDistanceToXY(goalPosition)).andStubReturn(3.0);
		expect(ballMock.getDistanceToXY(goalPosition)).andStubReturn(4.0);
		replay(worldModelMock, thisPlayerMock, player2Mock, opponentMock,
				ballMock);

		assertEquals(0.0, testee.getTruthValue(), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.agent.belief.impl.TeammateCloserToBall#getTruthValue()}.
	 */
	@Test
	public void testGetClosestPlayerOpponentBetweenOther()
	{
		expect(playerMock.getDistanceToXY(ballPosition)).andStubReturn(1.0);
		expect(player2Mock.getDistanceToXY(ballPosition)).andStubReturn(2.0);

		expect(thisPlayerMock.getPosition()).andStubReturn(
				new Vector3D(-0.5, 0, 0));
		expect(thisPlayerMock.getDistanceToXY(ballPosition)).andStubReturn(3.5);

		List<IPlayer> players = new ArrayList<IPlayer>();
		players.add(playerMock);
		players.add(player2Mock);
		expect(worldModelMock.getTeammatesAtBall()).andReturn(players);
		expect(worldModelMock.getGlobalTime()).andReturn(0f);
		expect(playerMock.isLaying()).andStubReturn(false);
		expect(player2Mock.isLaying()).andStubReturn(false);
		expect(thisPlayerMock.isLaying()).andStubReturn(false);
		expect(thisPlayerMock.getDistanceToXY(goalPosition)).andStubReturn(5.0);
		expect(playerMock.getDistanceToXY(goalPosition)).andStubReturn(5.0);
		expect(player2Mock.getDistanceToXY(goalPosition)).andStubReturn(5.0);
		expect(ballMock.getDistanceToXY(goalPosition)).andStubReturn(4.0);
		replay(worldModelMock, thisPlayerMock, playerMock, player2Mock,
				opponentMock, ballMock);

		assertEquals(player2Mock, testee.getClosestPlayer());
	}

}
