/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.function;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;
import magma.util.FuzzyCompare;

import org.junit.Before;
import org.junit.Test;

/**
 * @author Klaus Dorer
 * 
 */
public class PiecewiseLinearFunctionTest
{
	private PiecewiseLinearFunction testee;

	/**
	 * Setup method
	 */
	@Before
	public void setUp()
	{
		String[] data = { "Joint", "linear", "4", "0", "0", "-2", "4", "2", "5",
				"2", "6", "0" };
		testee = new PiecewiseLinearFunction(data);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#PiecewiseLinearFunction(double[], double[])}
	 */
	@Test
	public void testConstructor()
	{
		double[] xvalues = { 0.0, 4.0, 5.0, 6.0 };
		double[] yvalues = { -2.0, 2.0, 2.0, 0.0 };

		PiecewiseLinearFunction f = new PiecewiseLinearFunction(xvalues, yvalues);

		compareDoubleArray(xvalues, f.getXValues(), 0.000001);
		compareDoubleArray(yvalues, f.getYValues(), 0.000001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#value(double)} .
	 */
	@Test(expected = IllegalArgumentException.class)
	public void testValueInvalidSmall()
	{
		testee.value(-1);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#value(double)} .
	 */
	@Test(expected = IllegalArgumentException.class)
	public void testValueInvalidBig()
	{
		testee.value(6.1);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#value(double)} .
	 */
	@Test
	public void testValueValid()
	{
		assertEquals(-2, testee.value(0), 0.0001);
		assertEquals(0, testee.value(6), 0.0001);
		assertEquals(1, testee.value(5.5), 0.0001);
		assertEquals(2, testee.value(4.5), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#getName()} .
	 */
	@Test
	public void testGetName()
	{
		assertEquals("linear", testee.getName());
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#addPoint(double, double)}
	 * .
	 */
	@Test
	public void testAddPointFirst()
	{
		testee.addPoint(-1, 7);
		assertEquals(5, testee.getSupportPointSize(), 0.0001);
		assertEquals(-1, testee.getSupportX(0), 0.0001);
		assertEquals(7, testee.getSupportY(0), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#addPoint(double, double)}
	 * .
	 */
	@Test
	public void testAddPointMiddle()
	{
		testee.addPoint(4.5, 7);
		assertEquals(5, testee.getSupportPointSize(), 0.0001);
		assertEquals(4.5, testee.getSupportX(2), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#addPoint(double, double)}
	 * .
	 */
	@Test
	public void testAddPointLast()
	{
		testee.addPoint(8, 7);
		assertEquals(5, testee.getSupportPointSize(), 0.0001);
		assertEquals(8, testee.getSupportX(4), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#removePoint(int)} .
	 */
	@Test
	public void testHasSupportPointInArea()
	{
		assertEquals(false, testee.hasSupportPointInArea(1.0, 2.0));
		assertEquals(true, testee.hasSupportPointInArea(3.9, 4.1));
		assertEquals(true, testee.hasSupportPointInArea(3.0, 7));
		assertEquals(true, testee.hasSupportPointInArea(-1.0, 1.0));
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#removePoint(int)} .
	 */
	@Test
	public void testRemovePointFirst()
	{
		testee.removePoint(0);
		assertEquals(3, testee.getSupportPointSize(), 0.0001);
		assertEquals(4, testee.getSupportX(0), 0.0001);
		assertEquals(2, testee.getSupportY(0), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#removePoint(int)} .
	 */
	@Test
	public void testRemovePointMiddle()
	{
		testee.removePoint(2);
		assertEquals(3, testee.getSupportPointSize(), 0.0001);
		assertEquals(6, testee.getSupportX(2), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#removePoint(int)} .
	 */
	@Test
	public void testRemovePointLast()
	{
		testee.removePoint(3);
		assertEquals(3, testee.getSupportPointSize(), 0.0001);
		assertEquals(5, testee.getSupportX(2), 0.0001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#getXValues()} .
	 */
	@Test
	public void testGetXValues()
	{
		double[] xvalues = testee.getXValues();

		assertEquals(xvalues[0], 0.0, 0.00000001);
		assertEquals(xvalues[1], 4.0, 0.00000001);
		assertEquals(xvalues[2], 5.0, 0.00000001);
		assertEquals(xvalues[3], 6.0, 0.00000001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#getXValues()} .
	 */
	@Test
	public void testGetXValuesException()
	{
		double[] xvalues = testee.getXValues();

		try {
			assertEquals(xvalues[4], 6.0, 0.00000001);
			fail("getXValues() Yielded too many values");
		} catch (IndexOutOfBoundsException ioe) {
			// Expected
		}
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#getYValues()} .
	 */
	@Test
	public void testGetYValues()
	{
		double[] yvalues = testee.getYValues();

		assertEquals(yvalues[0], -2.0, 0.00000001);
		assertEquals(yvalues[1], 2.0, 0.00000001);
		assertEquals(yvalues[2], 2.0, 0.00000001);
		assertEquals(yvalues[3], 0.0, 0.00000001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#getYValues()} .
	 */
	@Test
	public void testGetYValuesException()
	{
		double[] yvalues = testee.getYValues();

		try {
			assertEquals(yvalues[4], 6.0, 0.00000001);
			fail("getYValues() Yielded too many values");
		} catch (IndexOutOfBoundsException ioe) {
			// Expected
		}
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#changePointCoordinates(int, double, double)}
	 */
	@Test
	public void testChangePointCoordinates()
	{
		double[] old = testee.changePointCoordinates(1, 3.0, -2.5);

		assertEquals(old[0], 4.0, 0.00001);
		assertEquals(old[1], 2.0, 0.00001);

		double[] xvalues = testee.getXValues();
		double[] yvalues = testee.getYValues();

		assertEquals(xvalues[1], 3.0, 0.000001);
		assertEquals(yvalues[1], -2.5, 0.000001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#changePointCoordinates(int, double, double)}
	 */
	@Test
	public void testChangePointCoordinatesLessThanPrevious()
	{
		/** Illegal: x coordinate is less than the previous point */
		double[] old = testee.changePointCoordinates(2, 3.0, -2.5);

		assertEquals(old[0], 5.0, 0.00001);
		assertEquals(old[1], 2.0, 0.00001);

		/**
		 * New x value is calculated to be 0.1 more than that of the previous
		 * point
		 */
		double[] xvalues = testee.getXValues();
		double[] yvalues = testee.getYValues();

		assertEquals(xvalues[2], 4.1, 0.000001);
		assertEquals(yvalues[2], -2.5, 0.000001);
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#changePointCoordinates(int, double, double)}
	 */
	@Test
	public void testChangePointCoordinatesMoreThanNext()
	{
		/** Illegal: x coordinate is higher than the next point */
		double[] old = testee.changePointCoordinates(2, 7.0, 2.5);

		assertEquals(old[0], 5.0, 0.00001);
		assertEquals(old[1], 2.0, 0.00001);

		/**
		 * New x value is calculated to be 0.1 less than that of the previous
		 * point
		 */
		double[] xvalues = testee.getXValues();
		double[] yvalues = testee.getYValues();

		assertEquals(xvalues[2], 5.9, 0.000001);
		assertEquals(yvalues[2], 2.5, 0.000001);
	}

	/**
	 * Test method for {@link magma.util.function.PiecewiseLinearFunction#copy()}
	 */
	@Test
	public void testCopy()
	{
		PiecewiseLinearFunction copy = (PiecewiseLinearFunction) testee.copy();

		assertEquals(testee.getMaxX(), copy.getMaxX(), 0.0000001);
		assertEquals(testee.getMinX(), copy.getMinX(), 0.0000001);
		assertEquals(testee.getName(), copy.getName());
		assertEquals(testee.getPhase(), copy.getPhase(), 0.0000001);
		assertEquals(testee.getSupportPointSize(), copy.getSupportPointSize(),
				0.0000001);

		compareDoubleArray(testee.getXValues(), copy.getXValues(), 0.0000001);
		compareDoubleArray(testee.getYValues(), copy.getYValues(), 0.0000001);
	}

	private void compareDoubleArray(double[] array1, double[] array2,
			double range)
	{
		if (array1.length != array2.length)
			fail("Array size not equal: " + array1.length + " != " + array2.length);

		for (int i = 0; i < array1.length; i++) {
			if (!FuzzyCompare.eq(array1[i], array2[i], range))
				fail("Array index " + i + ": " + array1[i] + " != " + array2[i]);
		}
	}

	/**
	 * Test method for
	 * {@link magma.util.function.PiecewiseLinearFunction#toCSVString()}
	 */
	@Test
	public void testToCSVString()
	{
		assertEquals("linear,4,0.0,0.0,-2.0,4.0,2.0,5.0,2.0,6.0,0.0,",
				testee.toCSVString());
	}
}
