/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import org.junit.Test;

/**
 * @author Klaus Dorer
 * 
 */
public class AngleTest
{
	/**
	 * Test method for {@link magma.util.geometry.Angle#degrees()}.
	 */
	@Test
	public void testNormalize()
	{
		assertEquals(-90.0, Angle.deg(270).degrees(), 0.0001);
		assertEquals(-170.0, Angle.deg(190).degrees(), 0.0001);
		assertEquals(170.0, Angle.deg(-190).degrees(), 0.0001);
		assertEquals(-170.0, Angle.deg(-170).degrees(), 0.0001);
		assertEquals(-180.0, Angle.deg(-180).degrees(), 0.0001);
		assertEquals(-180.0, Angle.deg(180).degrees(), 0.0001);
		assertEquals(10.0, Angle.deg(370).degrees(), 0.0001);
		assertEquals(-10.0, Angle.deg(-370).degrees(), 0.0001);
	}

	/**
	 * Test method for {@link magma.util.geometry.Angle#add(Angle)}.
	 */
	@Test
	public void testAdd()
	{
		assertEquals(5.0, Angle.deg(2).add(Angle.deg(3)).degrees(), 0.0001);
		assertEquals(-160.0, Angle.deg(100).add(Angle.deg(100)).degrees(), 0.0001);
		assertEquals(14.422, Angle.deg(2).add(6.5).degrees(), 0.001);
	}

	/**
	 * Test method for {@link magma.util.geometry.Angle#subtract(Angle)}.
	 */
	@Test
	public void testSubtract()
	{
		assertEquals(-1.0, Angle.deg(2).subtract(Angle.deg(3)).degrees(), 0.0001);
		assertEquals(160.0, Angle.deg(-100).subtract(Angle.deg(100)).degrees(),
				0.0001);
	}

	/**
	 * Test method for {@link magma.util.geometry.Angle#isLeftOf(Angle)}.
	 */
	@Test
	public void testIsLeftOf()
	{
		assertTrue(Angle.deg(10).isLeftOf(Angle.deg(-10)));
		assertFalse(Angle.deg(90).isLeftOf(Angle.deg(-100)));
		assertTrue(Angle.deg(-100).isLeftOf(Angle.deg(100)));
		assertFalse(Angle.deg(100).isLeftOf(Angle.deg(-100)));
	}

	/**
	 * Test method for {@link magma.util.geometry.Angle#isRightOf(Angle)}
	 */
	@Test
	public void testIsRightOf()
	{
		assertTrue(Angle.deg(-10).isRightOf(Angle.deg(10)));
		assertFalse(Angle.deg(-100).isRightOf(Angle.deg(90)));
		assertTrue(Angle.deg(100).isRightOf(Angle.deg(-100)));
		assertFalse(Angle.deg(-100).isRightOf(Angle.deg(81)));
	}

	/**
	 * Test method for {@link magma.util.geometry.Angle#average(Angle[])}.
	 */
	@Test
	public void testAveragePositive()
	{
		// Simple case: All angles positve
		Angle[] angles = new Angle[3];
		angles[0] = Angle.deg(9.9);
		angles[1] = Angle.deg(20.0);
		angles[2] = Angle.deg(30.1);

		assertEquals(Angle.average(angles).degrees(), 20.00, 0.0001);
	}

	/**
	 * Test method for {@link magma.util.geometry.Angle#average(Angle[])}.
	 */
	@Test
	public void testAverageNegative()
	{
		// Simple case: All angles negative
		Angle[] angles = new Angle[3];
		angles[0] = Angle.deg(-9.9);
		angles[1] = Angle.deg(-20.0);
		angles[2] = Angle.deg(-30.1);

		assertEquals(Angle.average(angles).degrees(), -20.00, 0.0001);
	}

	/**
	 * Test method for {@link magma.util.geometry.Angle#average(Angle[])}.
	 */
	@Test
	public void testAverageMixed()
	{
		// Harder case: Mixed signs
		Angle[] angles = new Angle[3];
		angles[0] = Angle.deg(-12.5);
		angles[1] = Angle.deg(10.0);
		angles[2] = Angle.deg(41.3);

		assertEquals(Angle.average(angles).degrees(), 12.9333, 0.0001);
	}

	/**
	 * Test method for {@link magma.util.geometry.Angle#average(Angle[])}.
	 */
	/**
	 * 
	 */
	@Test
	public void testAverageCornerCases()
	{
		// Corner case: More than 90 degrees in both directions
		Angle[] angles = new Angle[3];
		angles[0] = Angle.deg(95.0);
		angles[1] = Angle.deg(-100.0);
		angles[2] = Angle.deg(41.3);

		assertEquals(Angle.average(angles).degrees(), 132.1, 0.0001);

		// Corner case: More than 90 degrees in both directions
		angles = new Angle[3];
		angles[0] = Angle.deg(-95.0);
		angles[1] = Angle.deg(100.0);
		angles[2] = Angle.deg(41.2);

		assertEquals(Angle.average(angles).degrees(), -104.6, 0.0001);

	}

	/**
	 * Test method for {@link magma.util.geometry.Angle#equals(Object)}.
	 */
	@Test
	public void testEquals()
	{
		assertTrue(Angle.deg(102.3).equals(Angle.deg(102.3)));
		assertTrue(Angle.deg(-12.4).equals(Angle.deg(-12.4)));

		assertFalse(Angle.deg(-10.1).equals(Angle.deg(-8.97)));
		assertFalse(Angle.deg(-10.1).equals(Angle.deg(10.3)));
		assertFalse(Angle.deg(24.46).equals(Angle.deg(-1.02)));
		assertFalse(Angle.deg(24.46).equals(Angle.deg(5.0001)));

		assertFalse(Angle.deg(-10.1).equals("12"));
	}

	/**
	 * Test method for {@link magma.util.geometry.Angle#toString()}.
	 */
	@Test
	public void testToString()
	{
		assertEquals("1.1233999999999902", Angle.deg(1.1234).toString());
	}
}
