/*******************************************************************************
 * Copyright 2008, 2011 Hochschule Offenburg
 * Klaus Dorer, Mathias Ehret, Stefan Glaser, Thomas Huber, Fabian Korak,
 * Simon Raffeiner, Srinivasa Ragavan, Thomas Rinklin,
 * Joachim Schilling, Ingo Schindler, Rajit Shahi, Bjoern Weiler
 *
 * This file is part of magmaOffenburg.
 *
 * magmaOffenburg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * magmaOffenburg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with magmaOffenburg. If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package magma.util.geometry;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import javax.vecmath.Matrix3d;
import javax.vecmath.Matrix4d;
import javax.vecmath.Vector3d;

import magma.util.geometry.MatrixUtil.RotationOrder;

import org.junit.Test;

/**
 * @author Stefan Glaser
 * 
 */
public class MatrixUtilTest
{
	private Matrix4d matrix_4d;

	private Matrix3d matrix_3d;

	private Vector3d vector_3d;

	@Test
	public void testMulMatrix4dVector3d()
	{
		vector_3d = new Vector3d(1, 1, 1);
		matrix_4d = new Matrix4d(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);

		vector_3d = MatrixUtil.mulMatrixVector(matrix_4d, vector_3d);

		assertEquals("Mul vector with unity matrix failed of x-Value check", 1,
				vector_3d.x, 0.001);
		assertEquals("Mul vector with unity matrix failed of x-Value check", 1,
				vector_3d.y, 0.001);
		assertEquals("Mul vector with unity matrix failed of x-Value check", 1,
				vector_3d.z, 0.001);

		// Check real situation
		matrix_4d = new Matrix4d(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14,
				15, 16);
		vector_3d = new Vector3d(1, 2, 3);

		vector_3d = MatrixUtil.mulMatrixVector(matrix_4d, vector_3d);

		// x-component = 1*1 + 2*2 + 3*3 + 4*1 (last term because of homogeneous
		// coordinates)
		assertEquals(
				"Mul vector with arbitrary matrix failed because of x-Value check",
				(1 + 2 * 2 + 3 * 3 + 4), vector_3d.x, 0.001);

		// y-component = 5*1 + 6*2 + 7*3 + 8*1 (last term because of homogeneous
		// coordinates)
		assertEquals(
				"Mul vector with arbitrary matrix failed because of x-Value check",
				(5 + 6 * 2 + 7 * 3 + 8), vector_3d.y, 0.001);

		// z-component = 9*1 + 10*2 + 11*3 + 12*1 (last term because of
		// homogeneous coordinates)
		assertEquals(
				"Mul vector with arbitrary matrix failed because of x-Value check",
				(9 + 10 * 2 + 11 * 3 + 12), vector_3d.z, 0.001);
	}

	@Test
	public void testMulMatrix3dVector3d()
	{
		vector_3d = new Vector3d(1, 1, 1);
		matrix_3d = new Matrix3d(1, 0, 0, 0, 1, 0, 0, 0, 1);

		vector_3d = MatrixUtil.mulMatrixVector(matrix_3d, vector_3d);

		assertEquals("Mul vector with unity matrix failed of x-Value check", 1,
				vector_3d.x, 0.001);
		assertEquals("Mul vector with unity matrix failed of x-Value check", 1,
				vector_3d.y, 0.001);
		assertEquals("Mul vector with unity matrix failed of x-Value check", 1,
				vector_3d.z, 0.001);

		// Check real situation
		matrix_3d = new Matrix3d(1, 2, 3, 4, 5, 6, 7, 8, 9);
		vector_3d = new Vector3d(1, 2, 3);

		vector_3d = MatrixUtil.mulMatrixVector(matrix_3d, vector_3d);

		// x-component = 1*1 + 2*2 + 3*3
		assertEquals(
				"Mul vector with arbitrary matrix failed because of x-Value check",
				(1 + 2 * 2 + 3 * 3), vector_3d.x, 0.001);

		// y-component = 4*1 + 5*2 + 6*3
		assertEquals(
				"Mul vector with arbitrary matrix failed because of x-Value check",
				(4 + 5 * 2 + 6 * 3), vector_3d.y, 0.001);

		// z-component = 7*1 + 8*2 + 9*3
		assertEquals(
				"Mul vector with arbitrary matrix failed because of x-Value check",
				(7 + 8 * 2 + 9 * 3), vector_3d.z, 0.001);
	}

	@Test
	public void testIsSymmetric()
	{
		// Test on false
		matrix_3d = new Matrix3d(//
				1, 1, 0, //
				1, 1, 1, //
				0, 0, 0);

		assertFalse(MatrixUtil.isSymmetric(matrix_3d));

		// Test on true
		matrix_3d = new Matrix3d(//
				1, 6, 7, //
				6, 1, 2.6, //
				7, 2.6, 0);

		assertTrue(MatrixUtil.isSymmetric(matrix_3d));

		// Test on each relevant component once
		matrix_3d = new Matrix3d(//
				1, 0, 1, //
				1, 1, 1, //
				1, 1, 1);

		assertFalse(MatrixUtil.isSymmetric(matrix_3d));

		matrix_3d = new Matrix3d(//
				1, 1, 1, //
				1, 1, 1, //
				0, 1, 1);

		assertFalse(MatrixUtil.isSymmetric(matrix_3d));

		matrix_3d = new Matrix3d(//
				1, 1, 1, //
				1, 1, 0, //
				1, 1, 1);

		assertFalse(MatrixUtil.isSymmetric(matrix_3d));
	}

	@Test
	public void testCreateTranslationMatrix()
	{
		vector_3d = new Vector3d(1, 1, 1);
		matrix_4d = new Matrix4d(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);

		matrix_4d = MatrixUtil.createTranslationMatrix(0, 0, 0);
		vector_3d = MatrixUtil.mulMatrixVector(matrix_4d, vector_3d);

		assertEquals(
				"Creation of translation matrix with 0-values failed because of x-Value check",
				1, vector_3d.x, 0.001);
		assertEquals(
				"Creation of translation matrix with 0-values failed because of x-Value check",
				1, vector_3d.y, 0.001);
		assertEquals(
				"Creation of translation matrix with 0-values failed because of x-Value check",
				1, vector_3d.z, 0.001);

		// Check real situation
		vector_3d = new Vector3d(1, 2, 3);
		matrix_4d = MatrixUtil.createTranslationMatrix(10, 20, 30);
		vector_3d = MatrixUtil.mulMatrixVector(matrix_4d, vector_3d);

		// The x-value should be shifted about 10 to 11
		assertEquals(
				"Creation of translation matrix with arbitrary values failed because of x-Value check",
				11, vector_3d.x, 0.001);

		// The y-value should be shifted about 20 to 22
		assertEquals(
				"Creation of translation matrix with arbitrary values failed because of y-Value check",
				22, vector_3d.y, 0.001);

		// The z-value should be shifted about 30 to 33
		assertEquals(
				"Creation of translation matrix with arbitrary values failed because of z-Value check",
				33, vector_3d.z, 0.001);
	}

	@Test
	public void testCreateInverseTranslationMatrix()
	{
		vector_3d = new Vector3d(1, 1, 1);
		matrix_4d = new Matrix4d(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);

		matrix_4d = MatrixUtil.createInverseTranslationMatrix(0, 0, 0);
		vector_3d = MatrixUtil.mulMatrixVector(matrix_4d, vector_3d);

		assertEquals(
				"Creation of inverse translation matrix with 0-values failed because of x-Value check",
				1, vector_3d.x, 0.001);
		assertEquals(
				"Creation of inverse translation matrix with 0-values failed because of x-Value check",
				1, vector_3d.y, 0.001);
		assertEquals(
				"Creation of inverse translation matrix with 0-values failed because of x-Value check",
				1, vector_3d.z, 0.001);

		// Check real situation
		vector_3d = new Vector3d(1, 2, 3);
		matrix_4d = MatrixUtil.createInverseTranslationMatrix(10, 20, 30);
		vector_3d = MatrixUtil.mulMatrixVector(matrix_4d, vector_3d);

		// The x-value should be shifted about 10 to 11
		assertEquals(
				"Creation of inverse translation matrix with arbitrary values failed because of x-Value check",
				-9, vector_3d.x, 0.001);

		// The y-value should be shifted about 20 to 22
		assertEquals(
				"Creation of inverse translation matrix with arbitrary values failed because of y-Value check",
				-18, vector_3d.y, 0.001);

		// The z-value should be shifted about 30 to 33
		assertEquals(
				"Creation of inverse translation matrix with arbitrary values failed because of z-Value check",
				-27, vector_3d.z, 0.001);
	}

	@Test
	public void testCreateRotationMatrix()
	{
		vector_3d = new Vector3d(1, 1, 1);

		matrix_3d = MatrixUtil.createRotationMatrix(0, 0, 0);
		vector_3d = MatrixUtil.mulMatrixVector(matrix_3d, vector_3d);

		assertEquals(
				"Creation of rotation matrix with 0-values failed because of x-Value check",
				1, vector_3d.x, 0.001);
		assertEquals(
				"Creation of rotation matrix with 0-values failed because of y-Value check",
				1, vector_3d.y, 0.001);
		assertEquals(
				"Creation of rotation matrix with 0-values failed because of z-Value check",
				1, vector_3d.z, 0.001);

		// Check real situation (rotate 90 deg around z)
		Vector3d tempVec_3d = new Vector3d(1, 2, 3);
		matrix_3d = MatrixUtil.createRotationMatrix(0, 0, Math.toRadians(90));
		vector_3d = MatrixUtil.mulMatrixVector(matrix_3d, tempVec_3d);

		// The x-value should be the negative of the old y-value
		assertEquals(
				"Creation of rotation matrix with z-rotation failed because of x-Value check",
				-tempVec_3d.y, vector_3d.x, 0.001);

		// The y-value should be the old x-value
		assertEquals(
				"Creation of rotation matrix with z-rotation failed because of y-Value check",
				tempVec_3d.x, vector_3d.y, 0.001);

		// The z-value should stay the same
		assertEquals(
				"Creation of rotation matrix with z-rotation failed because of z-Value check",
				tempVec_3d.z, vector_3d.z, 0.001);

		// Check real hard situation (rotate 90 deg around x and z)
		tempVec_3d = new Vector3d(1, 2, 3);
		matrix_3d = MatrixUtil.createRotationMatrix(Math.toRadians(90), 0,
				Math.toRadians(90));
		vector_3d = MatrixUtil.mulMatrixVector(matrix_3d, tempVec_3d);

		// The x-value should be the negative of the old y-value
		assertEquals(
				"Creation of rotation matrix with x- and z-rotation failed because of x-Value check",
				-tempVec_3d.y, vector_3d.x, 0.001);

		// The y-value should be the negative of the old z-value
		assertEquals(
				"Creation of rotation matrix with x- and z-rotation failed because of y-Value check",
				-tempVec_3d.z, vector_3d.y, 0.001);

		// The z-value should be the old x-value
		assertEquals(
				"Creation of rotation matrix with x- and z-rotation failed because of z-Value check",
				tempVec_3d.x, vector_3d.z, 0.001);
	}

	@Test
	public void testCreateInverseRotationMatrix()
	{
		vector_3d = new Vector3d(1, 1, 1);

		matrix_3d = MatrixUtil.createInverseRotationMatrix(0, 0, 0);
		vector_3d = MatrixUtil.mulMatrixVector(matrix_3d, vector_3d);

		assertEquals(
				"Creation of inverse rotation matrix with 0-values failed because of x-Value check",
				1, vector_3d.x, 0.001);
		assertEquals(
				"Creation of inverse rotation matrix with 0-values failed because of y-Value check",
				1, vector_3d.y, 0.001);
		assertEquals(
				"Creation of inverse rotation matrix with 0-values failed because of z-Value check",
				1, vector_3d.z, 0.001);

		// Check real situation (rotate 90 deg around z)
		Vector3d tempVec_3d = new Vector3d(1, 2, 3);
		matrix_3d = MatrixUtil.createInverseRotationMatrix(0, 0,
				Math.toRadians(90));
		vector_3d = MatrixUtil.mulMatrixVector(matrix_3d, tempVec_3d);

		// The x-value should be the old y-value
		assertEquals(
				"Creation of inverse rotation matrix with z-rotation failed because of x-Value check",
				tempVec_3d.y, vector_3d.x, 0.001);

		// The y-value should be negative of the old x-value
		assertEquals(
				"Creation of inverse rotation matrix with z-rotation failed because of y-Value check",
				-tempVec_3d.x, vector_3d.y, 0.001);

		// The z-value should stay the same
		assertEquals(
				"Creation of inverse rotation matrix with z-rotation failed because of z-Value check",
				tempVec_3d.z, vector_3d.z, 0.001);

		// Check real hard situation (rotate 90 deg around x and z)
		tempVec_3d = new Vector3d(1, 2, 3);
		matrix_3d = MatrixUtil.createInverseRotationMatrix(Math.toRadians(90), 0,
				Math.toRadians(90));
		vector_3d = MatrixUtil.mulMatrixVector(matrix_3d, tempVec_3d);

		// The x-value should be the old z-value
		assertEquals(
				"Creation of inverse rotation matrix with x- and z-rotation failed because of x-Value check",
				tempVec_3d.z, vector_3d.x, 0.001);

		// The y-value should be the negative of the old x-value
		assertEquals(
				"Creation of inverse rotation matrix with x- and z-rotation failed because of y-Value check",
				-tempVec_3d.x, vector_3d.y, 0.001);

		// The z-value should be the negative of the old y-value
		assertEquals(
				"Creation of inverse rotation matrix with x- and z-rotation failed because of z-Value check",
				-tempVec_3d.y, vector_3d.z, 0.001);
	}

	@Test
	public void testGetEulerAngles()
	{
		double rot1 = Math.toRadians(26);
		double rot2 = Math.toRadians(-64);
		double rot3 = Math.toRadians(78);

		double singularityRot1 = Math.toRadians(26);
		double singularityRot2 = Math.toRadians(90);
		double singularityRot3 = Math.toRadians(78);

		matrix_3d = new Matrix3d();

		// ==================================================
		// Test general rotations with XYZ rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateX(matrix_3d, rot1);
		MatrixUtil.rotateY(matrix_3d, rot2);
		MatrixUtil.rotateZ(matrix_3d, rot3);

		double[] eulerAnles = MatrixUtil.getEulerAngles(matrix_3d,
				RotationOrder.XYZ);

		assertEquals("wrong euler angle 1 by XYZ rotation order", rot1,
				eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by XYZ rotation order", rot2,
				eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by XYZ rotation order", rot3,
				eulerAnles[2], 0.001);

		// Test first singularity rotations with XYZ rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateX(matrix_3d, singularityRot1);
		MatrixUtil.rotateY(matrix_3d, singularityRot2);
		MatrixUtil.rotateZ(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.XYZ);

		assertEquals("wrong euler angle 1 by singular1 XYZ rotation order",
				singularityRot1 + singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by singular1 XYZ rotation order",
				singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 singular1 by XYZ rotation order", 0,
				eulerAnles[2], 0.001);

		// Test second singularity rotations with XYZ rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateX(matrix_3d, singularityRot1);
		MatrixUtil.rotateY(matrix_3d, -singularityRot2);
		MatrixUtil.rotateZ(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.XYZ);

		assertEquals("wrong euler angle 1 by singular2 XYZ rotation order",
				singularityRot1 - singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by XYZ singular2 rotation order",
				-singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by XYZ singular2 rotation order", 0,
				eulerAnles[2], 0.001);
		// ==================================================

		// ==================================================
		// Test general rotations with XZY rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateX(matrix_3d, rot1);
		MatrixUtil.rotateZ(matrix_3d, rot2);
		MatrixUtil.rotateY(matrix_3d, rot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.XZY);

		assertEquals("wrong euler angle 1 by XZY rotation order", rot1,
				eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by XZY rotation order", rot2,
				eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by XZY rotation order", rot3,
				eulerAnles[2], 0.001);

		// Test first singularity rotations with XZY rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateX(matrix_3d, singularityRot1);
		MatrixUtil.rotateZ(matrix_3d, singularityRot2);
		MatrixUtil.rotateY(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.XZY);

		assertEquals("wrong euler angle 1 by singular1 XZY rotation order",
				singularityRot1 - singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by singular1 XZY rotation order",
				singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 singular1 by XZY rotation order", 0,
				eulerAnles[2], 0.001);

		// Test second singularity rotations with XZY rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateX(matrix_3d, singularityRot1);
		MatrixUtil.rotateZ(matrix_3d, -singularityRot2);
		MatrixUtil.rotateY(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.XZY);

		assertEquals("wrong euler angle 1 by singular2 XZY rotation order",
				singularityRot1 + singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by XZY singular2 rotation order",
				-singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by XZY singular2 rotation order", 0,
				eulerAnles[2], 0.001);
		// ==================================================

		// ==================================================
		// Test general rotations with YXZ rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateY(matrix_3d, rot1);
		MatrixUtil.rotateX(matrix_3d, rot2);
		MatrixUtil.rotateZ(matrix_3d, rot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.YXZ);

		assertEquals("wrong euler angle 1 by YXZ rotation order", rot1,
				eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by YXZ rotation order", rot2,
				eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by YXZ rotation order", rot3,
				eulerAnles[2], 0.001);

		// Test first singularity rotations with YXZ rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateY(matrix_3d, singularityRot1);
		MatrixUtil.rotateX(matrix_3d, singularityRot2);
		MatrixUtil.rotateZ(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.YXZ);

		assertEquals("wrong euler angle 1 by singular1 YXZ rotation order",
				singularityRot1 - singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by singular1 YXZ rotation order",
				singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by singular1 YXZ rotation order", 0,
				eulerAnles[2], 0.001);

		// Test second singularity rotations with YXZ rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateY(matrix_3d, singularityRot1);
		MatrixUtil.rotateX(matrix_3d, -singularityRot2);
		MatrixUtil.rotateZ(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.YXZ);

		assertEquals("wrong euler angle 1 by singular2 YXZ rotation order",
				singularityRot1 + singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by singular2 YXZ rotation order",
				-singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by singular2 YXZ rotation order", 0,
				eulerAnles[2], 0.001);
		// ==================================================

		// ==================================================
		// Test general rotation with YZX rotation order
		// woohoo
		matrix_3d.setIdentity();
		MatrixUtil.rotateY(matrix_3d, rot1);
		MatrixUtil.rotateZ(matrix_3d, rot2);
		MatrixUtil.rotateX(matrix_3d, rot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.YZX);

		assertEquals("wrong euler angle 1 by YZX rotation order", rot1,
				eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by YZX rotation order", rot2,
				eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by YZX rotation order", rot3,
				eulerAnles[2], 0.001);

		// Test first singularity rotations with YZX rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateY(matrix_3d, singularityRot1);
		MatrixUtil.rotateZ(matrix_3d, singularityRot2);
		MatrixUtil.rotateX(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.YZX);

		assertEquals("wrong euler angle 1 by singular1 YZX rotation order",
				singularityRot1 + singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by singular1 YZX rotation order",
				singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by singular1 YZX rotation order", 0,
				eulerAnles[2], 0.001);

		// Test second singularity rotations with YZX rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateY(matrix_3d, singularityRot1);
		MatrixUtil.rotateZ(matrix_3d, -singularityRot2);
		MatrixUtil.rotateX(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.YZX);

		assertEquals("wrong euler angle 1 by singular2 YZX rotation order",
				singularityRot1 - singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by singular2 YZX rotation order",
				-singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by singular2 YZX rotation order", 0,
				eulerAnles[2], 0.001);
		// ==================================================

		// ==================================================
		// Test general rotation with ZXY rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateZ(matrix_3d, rot1);
		MatrixUtil.rotateX(matrix_3d, rot2);
		MatrixUtil.rotateY(matrix_3d, rot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.ZXY);

		assertEquals("wrong euler angle 1 by ZXY rotation order", rot1,
				eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by ZXY rotation order", rot2,
				eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by ZXY rotation order", rot3,
				eulerAnles[2], 0.001);

		// Test first singularity rotations with ZXY rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateZ(matrix_3d, singularityRot1);
		MatrixUtil.rotateX(matrix_3d, singularityRot2);
		MatrixUtil.rotateY(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.ZXY);

		assertEquals("wrong euler angle 1 by singular1 ZXY rotation order",
				singularityRot1 + singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by singular1 ZXY rotation order",
				singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by singular1 ZXY rotation order", 0,
				eulerAnles[2], 0.001);

		// Test second singularity rotations with ZXY rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateZ(matrix_3d, singularityRot1);
		MatrixUtil.rotateX(matrix_3d, -singularityRot2);
		MatrixUtil.rotateY(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.ZXY);

		assertEquals("wrong euler angle 1 by singular2 ZXY rotation order",
				singularityRot1 - singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by singular2 ZXY rotation order",
				-singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by singular2 ZXY rotation order", 0,
				eulerAnles[2], 0.001);
		// ==================================================

		// ==================================================
		// Test general rotations with ZYX rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateZ(matrix_3d, rot1);
		MatrixUtil.rotateY(matrix_3d, rot2);
		MatrixUtil.rotateX(matrix_3d, rot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.ZYX);

		assertEquals("wrong euler angle 1 by ZYX rotation order", rot1,
				eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by ZYX rotation order", rot2,
				eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by ZYX rotation order", rot3,
				eulerAnles[2], 0.001);

		// Test first singularity rotations with ZYX rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateZ(matrix_3d, singularityRot1);
		MatrixUtil.rotateY(matrix_3d, singularityRot2);
		MatrixUtil.rotateX(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.ZYX);

		assertEquals("wrong euler angle 1 by singular1 ZYX rotation order",
				singularityRot1 - singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by singular1 ZYX rotation order",
				singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by singular1 ZYX rotation order", 0,
				eulerAnles[2], 0.001);

		// Test second singularity rotations with ZYX rotation order
		matrix_3d.setIdentity();
		MatrixUtil.rotateZ(matrix_3d, singularityRot1);
		MatrixUtil.rotateY(matrix_3d, -singularityRot2);
		MatrixUtil.rotateX(matrix_3d, singularityRot3);

		eulerAnles = MatrixUtil.getEulerAngles(matrix_3d, RotationOrder.ZYX);

		assertEquals("wrong euler angle 1 by singular2 ZYX rotation order",
				singularityRot1 + singularityRot3, eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by singular2 ZYX rotation order",
				-singularityRot2, eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by singular2 ZYX rotation order", 0,
				eulerAnles[2], 0.001);
		// ==================================================
	}

	@Test
	public void testGetEulerAnglesDifference()
	{
		double rot11 = Math.toRadians(67);
		double rot12 = Math.toRadians(12);
		double rot13 = Math.toRadians(-53);

		double rot21 = Math.toRadians(26);
		double rot22 = Math.toRadians(-64);
		double rot23 = Math.toRadians(78);

		// current orientation
		matrix_3d = new Matrix3d();
		matrix_3d.setIdentity();
		MatrixUtil.rotateX(matrix_3d, rot11);
		MatrixUtil.rotateY(matrix_3d, rot12);
		MatrixUtil.rotateZ(matrix_3d, rot13);

		// target orientation (difference)
		Matrix3d targetM_3d = new Matrix3d();
		targetM_3d.setIdentity();
		MatrixUtil.rotateX(targetM_3d, rot21);
		MatrixUtil.rotateY(targetM_3d, rot22);
		MatrixUtil.rotateZ(targetM_3d, rot23);

		// inverse of current orientation
		Matrix3d invMatrix_3d = new Matrix3d(matrix_3d);
		invMatrix_3d.transpose();

		// create the target matrix
		targetM_3d.mul(matrix_3d, targetM_3d);

		// multiply target matrix with inverse of current orientation
		targetM_3d.mul(invMatrix_3d, targetM_3d);

		// retrieve the euler angle difference
		double[] eulerAnles = MatrixUtil.getEulerAngles(targetM_3d,
				RotationOrder.XYZ);

		assertEquals("wrong euler angle 1 by rotation difference", rot21,
				eulerAnles[0], 0.001);
		assertEquals("wrong euler angle 2 by rotation difference", rot22,
				eulerAnles[1], 0.001);
		assertEquals("wrong euler angle 3 by rotation difference", rot23,
				eulerAnles[2], 0.001);

	}
}
